"""File Wrapper module.

The class FileWrapper defined here writes to a temporary file, which is 
moved to the real file on close.  Optionally, it can also keep a back-up of 
the originial output file, check whether the new file is different from the 
old one, and be chatty (verbose).  For files opened for reading or 
appending, no behavioral changes compared to ordinary files occur.

Of course, there are numerous ways in which the semantics of a program may 
be subtly changed by using this class instead of a standard file object, 
(e.g., when other references to the file are made while it is still open 
for writing), so user beware!

"""

import os
import sys
import string

class FileWrapper:
	
	def __init__(self, filename, mode="r", bufsize=-1):
		self._fp = None
		self._flags = ""
		if mode[:1] == "w":
			self._open_write(filename, mode, bufsize)
		else:
			self._fp = open(filename, mode, bufsize)
		self._add_methods()
	
	def _add_methods(self):
		fp = self._fp
		for m in fp.__methods__:
			if m == 'close' and self._flags:
				continue
			setattr(self, m, getattr(fp, m))
	
	def _open_write(self, filename, mode, bufsize):
		head, tail = os.path.split(filename)
		if not os.path.isabs(head):
			head = os.path.join(os.getcwd(), head)
		self._file = os.path.join(head, tail)
		self._temp = os.path.join(head, '@'+tail)
		self._back = os.path.join(head, tail+'~')
		self._flags = mode
		mode = ""
		for c in self._flags:
			if c not in string.uppercase:
				mode = mode+c
		self._fp = open(self._temp, mode, bufsize)
	
	def __del__(self):
		self.close()
	
	def close(self, abort=None):
		if self._fp:
			fp = self._fp
			self._fp = None
			fp.close()
			if self._flags:
				self._close_actions(abort)
				self._flags = ""
	
	def _close_actions(self, abort=None):
		if abort:
			self._report(ABORTREPORT)
			_remove(self._temp)
			return
		if 'C' in self._flags:
			try:
				if not _compare(self._file, self._temp):
					_remove(self._temp)
					self._report(EQUALREPORT)
					return
			except (IOError, os.error):
				self._report(EQUALERROR)
		if 'B' in self._flags:
			if os.path.exists(self._file):
				try:
					_rename(self._file, self._back)
					self._report(BACKUPREPORT)
				except os.error:
					self._report(BACKUPERROR)
		_rename(self._temp, self._file)
		self._report(RENAMEREPORT)
	
	def _report(self, message):
		if 'V' in self._flags:
			sys.stderr.write(message % self.__dict__ + "\n")

ABORTREPORT = "Write aborted, remove temp file %(_temp)s"
EQUALREPORT = "No changes, remove temp file %(_temp)s"
BACKUPREPORT = "Made backup on %(_back)s"
RENAMEREPORT = "Moved temp file to %(_file)s"
EQUALERROR = "Comparison or rename of %(_temp)s failed"
BACKUPERROR = "Backup of %(_file)s failed"

def _compare(f1, f2):
	"""Helper to compare two files, return 0 if they are equal."""
	
	BUFSIZE = 8192
	fp1 = open(f1)
	try:
		fp2 = open(f2)
		try:
			while 1:
				b1 = fp1.read(BUFSIZE)
				b2 = fp2.read(BUFSIZE)
				if not b1 and not b2: return 0
				c = cmp(b1, b2)
				if c: return c
		finally:
			fp2.close()
	finally:
		fp1.close()

def _rename(f1, f2):
	"""Helper function to rename a file.
	
	When the rename system call fails, the destination is removed and
	we try again.  This is needed for platforms (Mac, Win) where rename
	doesn't atomically replace the target.
	
	"""
	try:
		os.rename(f1, f2)
	except os.error:
		_remove(f2)
		os.rename(f1, f2)

def _remove(f):
	"""Helper function to remove a file silently.
	
	(I.e. no exception is raised if the file doesn't exist
	or cannot be removed for another reason.)
	
	"""
	try:
		os.unlink(f)
	except os.error:
		pass

def _test():
	file = "AAPJE"
	
	f = FileWrapper(file, "wBCV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.close()
	
	f = FileWrapper(file, "wV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.close()
	
	f = FileWrapper(file, "wBV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.close()
	
	f = FileWrapper(file, "wCV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.close()
	
	f = FileWrapper(file, "wBCV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.close()
	
	f = FileWrapper(file, "wBCV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.write("line 3\n")
	f.close()
	
	f = FileWrapper(file, "wBV")
	f.write("line 1\n")
	f.write("line 2\n")
	f.close()
	
	raw_input("OK?")

if __name__ == '__main__':
	_test()
