"""A very simple bookmark utility.

It parses the grail/netscape bookmarks file and creates dictionaries
for entries.  

Here are keys I've decided on:
url: the url we load
description: the text string supplied by the title and/or user
add_date: ??, from NS
last_visit: time of last check (time.time()), (from NS)
last_modified: time of last modification (time.time())
digest: md5 checksum of response body (last non-error)
error: error code for last_visit
last_success: date of last successful load
kss: name of service station used for last visit
latency: avg. time to perform request
tries: # of tries from this site
timing: list of (kss, [time, time, time, ...])
"""

import regex
import string
import marshal
import copy
import httplib
import md5
import time
import types
import operator
import sys

rx_entry = regex.compile('<A HREF="\([^"]*\)"\(.*\)>\(.*\)</A>')
rx_attrs = regex.compile('\([A-Z_]+\)="\([0-9]+\)')

def parse(io):
    bookmarks = {}
    while 1:
        line = io.readline()
        if line == '':
            break
        if rx_entry.search(line) == -1:
            continue
        entry = {}
        url, attrs, title = rx_entry.group(1, 2, 3)
        entry['url'] = url
        entry['title'] = title
        for attr in string.split(attrs):
            if rx_attrs.search(attr) >= 0:
                key, val = rx_attrs.group(1, 2)
                entry[string.lower(key)] = string.atoi(val)
        bookmarks[url] = entry
    return bookmarks

class Bookmarks:

    def __init__(self, raw='', file=''):
        self.entries = {}
        if raw:
            self.parse(open(raw))
        if file:
            self.load(open(file))

    def __getstate__(self):
        return marshal.dumps(self.entries)

    def __setstate__(self, state):
        self.entries = marshal.loads(state)
            
    def parse(self, io):
        while 1:
            line = io.readline()
            if line == '':
                break
            if rx_entry.search(line) == -1:
                continue
            entry = {}
            url, attrs, title = rx_entry.group(1, 2, 3)
            entry['url'] = url
            entry['title'] = title
            for attr in string.split(attrs):
                if rx_attrs.search(attr) >= 0:
                    key, val = rx_attrs.group(1, 2)
                    key = string.lower(key)
                    val = string.atoi(val)
                    entry[key] = val
            self.entries[url] = entry

    def dump(self, io):
        io.write(marshal.dumps(self.entries))

    def load(self, io):
        self.entries = marshal.loads(io.read())

    def html(self, io=sys.stdout):
        fmt = "<DT><A HREF=\"%s\" ADD_DATE=\"%s\" LAST_VISIT=\"%s\" LAST_MODIFIED=\"%s\">%s</A>\n"
        for url, entry in self.entries.items():
            io.write(fmt % (url, entry['add_date'], entry['last_visit'],
                            entry['last_modified'], entry['title']))  
        

def ims_check(entry):
    """Use an If-Modified-Since request to check the page"""

def dig_check(entry):
    """Use a digest of the context to check the page"""

def http_req(entry, headers={}):
    req = httplib.Request(entry['url'], headers=headers)
    cn = httplib.Connection(req)
    b = time.time()
    cn.open()
    cn.send(req)
    resp = cn.receive()
    cn.close()
    e = time.time()
    return resp, e - b
    
def make_dig(resp):
    """Compute a digest for an entry."""
    return md5.new(resp.body).digest()

def choose(entries):
    """Choose the fastest kss based on timing info"""
    # what kind of errors would be appropriate?
    best_kss = ''
    best_time = 0.0
    print entries[0]['url']
    for entry in entries:
        for kss, times in entry['timing']:
            loads = filter(lambda x:type(x)==types.FloatType, times)
            if not loads:
                continue
            avg = reduce(operator.add, loads) / len(loads)
            print "\t", kss, avg
            if avg < best_time or best_time == 0.0:
                best_time = avg
                best_kss = kss
    return best_kss, best_time
    
def merge(*bookmarks):
    """Merge two new entries and select kss."""
    new = Bookmarks()
    urls = {}
    for b in bookmarks:
        for url in b.entries.keys():
            urls[url] = 1

    for url in urls.keys():
        entries = filter(lambda x:x,
                         map(lambda d, u=url: d.has_key(u) and d[u],
                             map(lambda b:b.entries, bookmarks)))
        best_kss, best_time = choose(entries)
        entry = copy.copy(entries[0])
        del entry['timing']
        entry['kss'] = best_kss
        entry['latency'] = best_time
        entry['tries'] = 1
        new.entries[url] = entry
    return new
