"""Base class for KP auctioneer.  Interfaces in types/MarketAPI.isl."""

# Start at medium-grained stepping - between items, but not between bids.
# To increase/decrease grain, enter a '+'/'-' at the step prompt.  Single
# step for final wrapup (when lot is consumed) is level 1.
SINGLESTEP = 2

import auctiontools
verbose=auctiontools.verbose

from iotools import get_io_mechanism

import random, time

# XxX AUCTION_BELLVAR_TEMPLATE = 'kos/plugins/market/conditions/%s Bell'
BIDINCR = 5


class Auctioneer:
    """Auction conductor.

    The auction proceedings start when all expected bidders have registered."""
    # XxX Currently we start the auction contingent to arrival of
    #     particular set of bidders.  Should instead start at appointed
    #     time, signaling a bit in advance with condition var...
    # XxX Authentication - make the credentials real structures, and get
    #     pgp validation going.
    # XxX We simulate open bidding using random order poll - *bogus*, but
    #     at least we can avoid ties without *appearing* to give
    #     preference...

    bidincr = BIDINCR

    def __init__(self, kos, name, private_key, expected_bidders, items):
        self.kos = kos
        self.name = name
        self.private_key = private_key
        self.expected_bidders = expected_bidders
        self.lot = auctiontools.AuctionLot(items)
        self.lot_surrogate = None
        self.bidders = {}               # who: bidder-object

        self.io = get_io_mechanism(kos, height=25)

    def open_shop(self):
        """Start serving, so clients can register, etc."""

        # Register ourself for service:
        self.kos.bind_service(self.name, self, "MarketAPI.BasicAuctioneer")

        # Register lot so we can pass surrogate to clients:
        self.lot_id = `time.time()`
        self.kos.bind_service(self.lot_id,
                              auctiontools.BidderAuctionLot(self.lot),
                              self.lot.IF_NAME)
        self.lot_surrogate = self.kos.lookup_service(self.lot.IF_NAME,
                                                     self.lot_id).Open()
        self.note("Opening shop, awaiting %s\n", self.expected_bidders)

        self.kos.run()                  # Serve registrations.
        self.note("Ready to run auction...  ")
        self.check_singlestep()
        self.note("\n")
        self.run_auction()
        time.sleep(1)                   # Allow bidders output first.
        self.wrapup()

    def Register(self, who, bidder_obj):
        """Add a bidder to the roster."""
        # XxX Should not accept any new registrations once the auction
        # starts!  (Right now we *restart* the auction if a bidder
        # reenters, but that's going to go away.)
        self.bidders[who] = bidder_obj
        self.note(" %s registered...\n", who)
        for n in self.expected_bidders:
            if not self.bidders.has_key(n):
                return
        # Done with registrations.
        self.note("Expected bidders have all arrived.\n")
        self.kos.stop()

    #def When(self):
    #   """Return time in seconds until the auction session."""
    #   return 0

    def Lot(self):
        """Return interface to current auction lot."""
        return self.lot_surrogate

    def CurrentBid(self):
        """Return the description of the current bid, including item info."""
        # XxX *** item name, description, details, current_bid, start_bid
        curr = self.lot.current()
        if curr:
            return (self.curr.desc(), self.lot.currbid(),
                    curr.openingprice or curr.price)
        else:
            kdu = self.kos.exception("MarketAPI.DataUnavailable")
            raise kdu, "No current item."

    #def warning_bell(self, secs):
    #   """Trigger condition warns interested parties of impending auction."""
    #   self.var = self.kos.create_condition(self.name " Bell")

    def run_auction(self):
        # Guido's suggestions (10/1):
        #  - follow through on top-down - make .
        tabled = settled = 0

        while self.lot.next():
            item = self.lot.current()

            self.note("Auctioning %s:\n %s\n", item.brief, item.details)
            self.check_singlestep(level=2)

            self.lot.set_currbid(0)

            who, price = self.auction_item(item)

            if who:
                self.award_item(item, who, price)
            else:
                self.table_item(item, price)

    def wrapup(self):
        settled = self.lot.report('settled')
        tabled = self.lot.report('tabled')
        self.note('     -=-             -=-             -=-\n')
        self.note("Wrapping up the auction, settled %d, tabled %d.\n",
                  len(settled), len(tabled))
        self.note("Who\t  Price\tWhat\n---\t  -----\t----\n")
        for i in self.lot.report('done'):
            w, p, r = self.lot.disposition(i)
            if not w:
                w, p = '    *', ' *'
            self.note("%-9s %-7s %s\n", w, p, i.desc()['brief'])
        self.note('     -=-             -=-             -=-\n')
            
        self.kos.stop()
        self.check_singlestep(msg="CR to conclude.\n")
        self.io.close()

    def award_item(self, item, who, price):
        """Award item to high bidder and notify all of settlement."""
        self.note("Awarding %s to %s @ %d.\n\n",
                  item.brief, who, price)
        make_receipt = auctiontools.make_receipt
        receipt = make_receipt(item.desc(), who, price, self.private_key)
        for w, b in self.bidders.items():
            # Inform all of settlement:
            b.Settled(item.desc(), who, price, len(self.lot.report('pending')))
            if w == who:
                b.Award(receipt)
        self.lot.awarded(receipt)

    def table_item(self, item, price):
        """Award item to high bidder and notify all of settlement."""
        self.note("Tabling %s, no sale.\n\n", item.brief)
        for who, b in self.bidders.items():
            # Inform all of tabling: 
            b.Tabled(item.desc(), len(self.lot.report('pending')))
        item = self.lot.tabled(price)

    def auction_item(self, item):
        """Conduct auction of one item, returning who got it for how much.

        ... or (None, low-ask) if no sale.

        Increment bid levels depending on response.  We use random order
        polling, so noone gets overt preference.  (XxX However, there's no
        regulation of how long they take to decide, so at the least it's
        not robust).  Highest bidder above ask sets new ask, and is
        eligible for award."""

        # Open bidding at initial price.
        # Decrease until we get a bid or we hit bottom. 
        # If we hit bottom (minprice), table the item.
        # If we get a bid, increase price until no more nibbles, then
        # settle with high bidder.

        # (XxX The self.notes are places to plug in monitoring reports.)

        pendbid = currbid = None        # (price, who)
        ask = item.openingprice or item.price
        while 1:
            # Gather bids in this cycle:
            # XxX Make this asynchronous.
            self.note("%d", ask)
            self.check_singlestep(msg=" ? ", level=3)

            currhigh = (pendbid and pendbid[1]) or ""
            currbid = self.solicit_bids(ask, item.desc(), currhigh)

            # Handle the new bid (or lack thereof):

            if currbid:
                # New bid - go for greater heights.
                pendbid = self.currbid = currbid
                self.lot.set_currbid(currbid[0])
                currbid = None
                ask = pendbid[0] + self.bidincr
                self.note(" -> %s,%s",
                          pendbid[1], ((SINGLESTEP>2 and "\n") or " "))
                continue

            elif pendbid:
                # Last high bid is the peak - found our buyer.
                self.note(" - <no takers>; %s is high bidder.\n", pendbid[1])
                return pendbid[1], pendbid[0]

            else:
                # Whoops - musta started too high.
                if hasattr(item, 'minprice') and (ask > item.minprice):
                    # Decrement towards minprice and try for more.
                    ask = ask - self.bidincr
                    if ask <= item.minprice:
                        ask = item.minprice
                    self.note(" - no, ")
                    continue

                else:
                    # No nibbles, and we've hit bottom - give up on this one.
                    self.note(" - no.  No takers...\n")
                    return None, ask

    def solicit_bids(self, ask, desc, currhigh):
        """Solicit bids on current item, breaking when we get one."""
        for who, bobj in auctiontools.random_order(self.bidders.items()):
            # Cycle through bidders, retaining highest bid at or above ask.
            # XxX debug:
            gotprice = bobj.SolicitBid(ask, desc, currhigh)
            if gotprice:
                return ask, who
        return None
        

    def note(self, msg, *args):
        self.io.write(msg % args)

    def check_singlestep(self,
                         msg="Continue? ('-' pause less, '+' pause more) ",
                         level=1):
        global SINGLESTEP
        if SINGLESTEP >= level:
            self.io.write(msg)
            got = self.io.readline()
            if got[0:1] == '-':
                SINGLESTEP = SINGLESTEP - 1
            if got[0:1] == '+':
                SINGLESTEP = SINGLESTEP + 1
