"""Base class for KP auction bidders.  Interfaces in types/MarketAPI.isl."""

import auctiontools
verbose=auctiontools.verbose

import nstools

from iotools import get_io_mechanism

import random, time

RegistrationFailure = "RegistrationFailure"

class Bidder:
    """Participate in an auction.

    To use, from your derived class:

     - invoke the .__init__() method,
     - invoke the .setfactors() method, to set the bidding factors,
     - then invoke the .register() method, to register with the auctioneer
       and to start serving the bidding interface."""

    OPPOSE = []         # Names of bidders we particularly bid against.
    IF_NAME = "MarketAPI.AuctionBidder"

    def __init__(self, kos, name, auction_name, funds):
        self.kos = kos
        self.name = name
        self.auction_name = auction_name
        self.funds = funds

        self.last_to = None
        self.highbidder = None
        self.io = get_io_mechanism(kos, use_dev='tty', height=5)
        self.io.write("(%s initting)\n" % name)
        self.lot = None
        self.current_item = None
        self.auction = self.get_auction()

    def setfactors(self, bandwidth, chunk, funds,
                       bang_factor=3,
                       chunk_factor=5,
                       bandwidth_factor=2,
                       urgency_factor=2,
                       opposition_factor=6):
        if funds:
            bang = float(bandwidth) / funds
        else:
            bang = 0
        self.factors = {'bandwidth': float(bandwidth),
                        'satisfied': 0.0,
                        'chunk': float(chunk),
                        'bang': float(bang),
                        'bang_factor': float(bang_factor),
                        'chunk_factor': float(chunk_factor),
                        'bandwidth_factor': float(bandwidth_factor),
                        'urgency_factor': float(urgency_factor),
                        'opposition_factor': float(opposition_factor)}
                        

    def run(self):
        """Register myself for bidding, and start running services."""
        # Fancy dance - register ourself as an "anonymous" (actually,
        # unique gensym) connector, then fetch a surrogate, pass it to the
        # auctioneer, and start serving it.
        if not self.auction:
            self.io.write("%s bailing out - failed to find auction!\n"
                          % self.name)
            return
        self.lot = self.auction.Lot()
        self.lotsize = self.pendsize = len(self.lot.Report('lot'))
        # XxX Use namespace id for uniqueness?
        myid = `time.time()`            # Unique "random" number
        self.svcid = self.kos.bind_service(myid, self, self.IF_NAME)
        surrogate = self.kos.lookup_service(self.IF_NAME, myid)
        self.auction.Register(self.name, surrogate.Open())
        self.kos.run()
        self.wrapup()

    def get_auction(self, timeout=30):
        """Obtain auction interface, if it's available within timeout."""
        while timeout:
            try:
                descr = self.kos.lookup_service("MarketAPI.BasicAuctioneer",
                                                self.auction_name)
                return descr.Open()
            except nstools.BadPathError:
                timeout = timeout - 1
                time.sleep(1)
                if not (timeout % 7):
                    msg = ("(%s: awaiting auctioneer, %d secs remain)\n"
                           % (self.name, timeout))
                    self.io.write(msg)
        return None

    def wrapup(self):
        """Cleanup in preparation for conclusion."""
        # XxX *** Report to home, check funds fuel, what else?
        self.io.write("(%s wrapup)\n" % self.name)

    ### Session Bidding Public Interface Methods ###

    def NextItem(self, item, remaining):
        """Method by which a new round is conveyed."""
        self.current_item = item
        self.pendsize = remaining


    def SolicitBid(self, asking, what, currhigh):
        """Bid being sought - return 1 to bid or 0 to pass."""
        # We use 'contrast' factors to express the satisfaction of our
        # criteria.  The emphasis of each is expressed in the final
        # equation.  See self.setfactors() for weight settings.

        contrast = auctiontools.contrast

        if asking == 0:
            return asking

        if currhigh == self.name:
            # We have the current high bid - don't bid against ourself!
            return 0

        if self.funds < asking:
            # Insufficient funds.
            return 0

        details = auctiontools.parse_details(what['details'])

        factors = self.factors
        bangf, bwf = factors['bang_factor'], factors['bandwidth_factor']
        chf, urf = factors['chunk_factor'], factors['urgency_factor']
        opf = factors['opposition_factor']

        b, s = factors['bandwidth'], factors['satisfied']
        if s >= b:
            # We've satisfied our bandwidth quota - pass.
            return 0

        # Better as asking bang-per-buck exceeds what is sought:
        bpb = bangf * contrast(float(details['bandwidth'])/asking,
                               factors['bang'],
                               exp=2)

        # Better as asking chunk is to closer to sought chunk -
        # (Goes negative when the difference exceeds a quarter.)
        chunking = chf * (0.5 - abs(contrast(details['chunk'],
                                             factors['chunk'],
                                             exp=2)))
        verbose("%s chunking\n\twant %.1f, is %.1f, contrast %.2f,"
                " chf %.2f, ing %f",
                self.name,
                factors['chunk'], details['chunk'], 
                contrast(details['chunk'], factors['chunk'], 0.5),
                chf, chunking,
                level=2)

        # Reduce bandwidth factor by the proportion we've satisfied our
        # quota:
        bwf = max(bwf * (b - s)/b, 0)
        # And now compute the entire thing:
        bw = bwf * contrast(details['bandwidth'],
                            factors['bandwidth'], 
                            exp=0.5)

        # Spiteful opposition:-)
        if currhigh in self.OPPOSE:
            # Stir things up a bit, but not regularly enough to get the bid
            # - at least, most of the time...
            opp = opf * random.choice([0, 1])
        else:
            opp = 0

        # Increasing urgency as remaining prospects diminish:
        urgency = urf * contrast(self.lotsize, self.pendsize)

        # Normalize to between 0 and 1 by dividing out sum of factors:
        # (Not useful until we're comparing between items
        composite = ((bpb + chunking + bw + urgency + opp)
                     / (bangf + chf + bwf + urf + opf))

        verbose("(%s bidding breakdown,\n\tFor %s:\n"
                "\tbang-per-buck %.1f, chunk %.1f, bandwidth %.1f,\n"
                "\turgency %.1f, sum %.1f, norm %.1f)",
                self.name, details,
                bpb, chunking, bw, urgency,
                bpb + chunking + bw + urgency, composite, level=3)

        return composite >= 0

    def Settled(self, item, to, price, remaining):
        """Item was sold, and 'remaining' number of items are remaining."""
        self.last_to = to
        if not remaining:
            self.kos.stop()

    def Tabled(self, item, remaining):
        """Item was not sold, with 'remaining' number of items remaining."""
        if not remaining:
            self.kos.stop()

    def Award(self, receipt):
        """Called when we get the high bid.  (Yippee.)

        Derived classes can wrap this in their own bookkeeping method."""
        self.funds = self.funds - receipt['payment']
        parse_details=auctiontools.parse_details
        bw = parse_details(receipt['desc']['details'])['bandwidth']
        self.factors['satisfied'] = self.factors['satisfied'] - bw
