"""Incidental tools for KP-based auction.

See types/MarketAPI for type definitions."""

VERBOSE=0

import random, time

class AuctionLot:
    """Track the disposition of items in the auction lot.

    'begin() method must first be called to prime the pump."""

    IF_NAME = "MarketAPI.AuctionLot"

    def __init__(self, lot_items):
        self.__lot = lot_items[:]
        self.reset()
        self.__pending = lot_items[:]
        self.__processed = []                   # Ordered history - [item, ...]
        # Qualified history - {item: (who, price, receipt), ...}:
        self.__processed_details = {}
    def next(self):
        """Shift the state, moving current to hist and ending to current."""
        if self.__current:
            # Ditch this item - shoulda been __done'd.
            self.__done(who=None, price=0, receipt=None)
        if self.__pending:
            self.__current = self.__pending[0]
            del self.__pending[0]
        else:
            # We're dry.
            self.__current = None
        return self.__current
    def reset(self):
        """Position the first item on the block (as current)."""
        self.__current = None
        self.__pending = self.__lot[:]
        # We don't reset _processed.
        self.__currbid = None
    def set_currbid(self, bid):
        self.__currbid = bid
    def current(self):
        return self.__current
    def currbid(self):
        return self.__currbid
    def tabled(self, lowask):
        """Register the current item as unsold, and shift to next pending."""
        self.__done(who=None, price=lowask, receipt=None)
    def awarded(self, receipt):
        """Register sale of current item."""
        who = receipt['buyercred']      # XxX Change w/receipt cred gets real.
        price = receipt['payment']
        self.__done(who=who, price=price, receipt=receipt)
    def __done(self, who, price, receipt):
        """Register item disposition and return next item."""
        self.__processed.append(self.__current)
        self.__processed_details[self.__current] = (who, price, receipt)
        self.__current = None
    def report(self, qual):
        """Return list of items in lot meeting some qualification:

         'lot' - all items,
         'done' - all already-auctioned items,
         'pending' - all items not yet auctioned,
         'settled' - all done items that were sold,
         'tabled' - all done items that were not sold."""
        if qual == 'lot':
            return self.__lot
        elif qual == 'pending':
            return self.__pending
        elif qual == 'done':
            return self.__processed
        elif qual == 'settled':
            # All those for which we have a 'who':
            return filter(lambda x, d=self.__processed_details: d[x][0],
                          self.__processed)
        elif qual == 'tabled':
            # All those for which we lack a 'who':
            return filter(lambda x, d=self.__processed_details: not d[x][0],
                          self.__processed)
        else:
            print "unimplemented qualifier..."
            kve = self.kos.exception("KOS.ValueError")
            raise kve, "Unimplemented qualifier '%s'" % qual
    def disposition(self, item):
        """Return settlement details - who, price, receipt - for item.

        Or None if it's not been settled."""
        if not self.__processed_details.has_key(item):
            return None
        return self.__processed_details[item]

class BidderAuctionLot:
    """Bidder view of auction lot, publishing items sanitized of price info."""
    def __init__(self, lot):
        self.__lot = lot

    def CurrentItem(self):
        """Report current item's description, sans pricing info."""
        return self.__lot.current.desc()

    def Report(self, qual):
        """Return list of items descriptions.

        See AuctionLot.report for qualifiers."""
        return map(lambda x: x.desc(), self.__lot.report(qual))

class Item:
    def __init__(self, name, owner, brief, details, price,
                 supplier=None, minprice=None, openingprice=None):
        self.name = name
        self.owner = owner
        self.brief = brief
        self.details = details
        self.price = price
        self.supplier = supplier
        if minprice != None:
            self.minprice = minprice
        if openingprice != None:
            self.openingprice = openingprice
    def desc(self):
        """Dictify the description part for transmission."""
        return {'name': self.name, 'owner': self.owner,
                'brief': self.brief, 'details': self.details}

def random_order(seq):
    """Return random-order copy of sequence of arbitrary elements."""
    got, copy = [], seq[:]
    while copy:
        this = random.randint(0, len(copy) - 1)
        got.append(copy[this])
        del copy[this]
    return got

def parse_details(details):
    """Parse bandwidth-auction details string into 'name': numeric-val dict."""
    # Details, eg: "bandwidth: 150, chunk: 30"
    import string
    got = {}
    for item in string.splitfields(details, ','):
        try:
            [name, val] = map(string.strip, string.splitfields(item, ':'))
            got[name] = string.atof(val)
        except ValueError:
            verbose("(Bad details line: '%s')", details, level=0)
    return got

def contrast(a, b, exp=1):
    """Rate how much bigger A is than B, on a scale of -1 to 1.

    1 means A is *infinitely* bigger than B, -1 means the opposite.
    
    The curve is weighted according to an optional exponent, exp:
    
     - exp=1 (default) for linear behavior.     _
     - 0 < exp < 1 to weight towards extremes (/  decreasing slope from 0).
     - 1 < exp to weight towards center (_/ increasing slope from 0).

    Negative exponents are disallowed - they don't converge."""
    # The desired effect may be achievable based on math.atan()
    
    if exp < 0:
        raise ValueError, "Negative exponent diverges."
    # Zeros
    a, b = float(a), float(b)
    if a < b:
        if b == 0: return a and -1
        sign = -1
        fract = 1 - a/b
    else:
        if a == 0: return b and 1
        sign = 1
        fract = 1 - b/a
    return sign * pow(fract, exp)

def make_receipt(itemdesc, buyercred, payment, private_key):
    """Return timestamped, validated receipt."""
    got = {'desc': itemdesc,
           'buyercred': buyercred,
           'payment': payment,
           'date': time.ctime(time.time())}
    got['validation'] = pgp_sign(got.values(), private_key)
    return got

def pgp_sign(strlist, private_key):
    """Return a pgp signature for a sequence of strings and a key."""
    # XxX Not yet implemented.
    return "(PGP signature not yet implemented)"

def verbose(msg, *args, **keywords):
    """Emit message when global VERBOSE setting exceeds level (default 1).

    Optional keyword arg, level, changes threshold."""
    level = ((not keywords.has_key('level') and 1) or keywords['level'])
    if VERBOSE >= level:
        import sys
        sys.stderr.write("auctiontools: " + (msg % args) + "\n")
