"""CGI bridge between a web server and content-providing KPs.

The bridge is run as a CGI script by the web server and resolves remaining
path components to access either a file provided by a KP or a binding handle
for a ContentProvider object that is used to call on the KP to dynamically
generate the content in much the same way as a CGI script.

The bridge consults the content registry maintained by the plugin.  If a file
is available which corresponds to the URL, the content-type is guessed and
the file is returned as the content of the response.  If no file is available
and a ContentProvider interface has been registered by the KP, the
ContentProvider object is used to request the content from the KP.  If no
file is available and there is not ContentProvider, an HTTP 404 response is
returned.
"""

import os
import string
import sys

# Make sure we can get the the shared code for this application:
basedir = os.path.normpath(sys.path[0])
sys.path[0] = basedir
sys.path.insert(1, os.path.normpath(
    os.path.join(basedir, os.pardir, "common")))
sys.path.insert(2, os.path.normpath(
    os.path.join(basedir, os.pardir, os.pardir, os.pardir)))

import statefile
import traceback


# include here any env. variables that expose system information:
IGNORE_ENV_VARS = ["HTTP_HOST", "SCRIPT_NAME", "SCRIPT_FILENAME", "PATH",
                   "DOCUMENT_ROOT", "PATH_TRANSLATED", "REQUEST_URI"]

DEFAULT_CONTENT_TYPE = "text/plain"
DEFAULT_INDEX_NAMES = ["index.html", "index.htm", "default.htm"]


def main():
    try:
        real_main()
    except:
        traceback.print_exc(file=sys.stdout)

def real_main():
    env = {}
    for k, v in os.environ.items():
        if k not in IGNORE_ENV_VARS:
            env[k] = v
    url = os.environ.get("PATH_INFO", '')
    while url and url[0] == os.sep:
        url = url[1:]
    try:
        state = statefile.load()
    except IOError, err:
        http_error(404, "No registered content providers")
        return
    parts = string.split(url, os.sep)
    provider = parts[0]
    if not provider:
        list_providers(state)
        return
    try:
        provider_info = state[provider]
    except KeyError:
        http_error(404, "Unknown content provider: " + provider)
        return
    if len(parts) == 1:
        print "Location: http://%s%s/" \
              % (os.environ["SERVER_NAME"], os.environ["REQUEST_URI"])
        return
    basedir, sbh = provider_info
    if basedir:
        filename = find_file(basedir, url)
        if os.path.isfile(filename):
            use_file(filename)
            return
    if sbh:
        use_provider(sbh, url, env)
        return
    http_error(404, "No such resource -- no dynamic content")


def find_file(basedir, url):
    """Attempt to locate a file to return, handling directory indexes."""
    partial = string.join(string.split(url, os.sep)[1:], os.sep)
    filename = os.path.join(basedir, partial)
    if os.path.isdir(filename):
        for indexname in DEFAULT_INDEX_NAMES:
            fn = os.path.join(filename, indexname)
            if os.path.isfile(fn):
                return fn
    return filename


def use_file(filename):
    """Return the content of a file as the HTTP response."""
    try:
        fp = open(filename)
    except IOError:
        http_error(403, "Could not open file")
        return
    import mimetypes
    ctype, encoding = mimetypes.guess_type(filename)
    if not ctype:
        ctype = DEFAULT_CONTENT_TYPE
    print "Content-Type:", ctype
    if encoding:
        print "Content-Encoding:", encoding
    print
    while 1:
        buf = fp.read(10240)
        if not buf:
            break
        sys.stdout.write(buf)
    sys.stdout.flush()


def use_provider(sbh, url, env):
    """Allow a ContentProvider to respond to the request."""
    import environment
    import ilu
    import redirector
    from koe.interfaces.stubs import ContentProviderAPI
    fp = redirector.Redirector(sys.stdin, sys.stdout)
    cp = ilu.ObjectOfSBH(ContentProviderAPI.ContentProvider, sbh)
    env = environment.Environment(env)
    cp.GetContent(url, env, fp)


def list_providers(state):
    # there is no PATH_INFO in the environment
    print 'Content-Type: text/html'
    print
    providers = state.keys()
    print '''\
<html><head><title>Content Providers</title></head>
<body><h1>Content Providers</h1>
'''
    if providers:
        print '<ul>'
        for pname in providers:
            print '<li><a href="%s/">%s</a>' % (pname, pname)
        print '</ul></body></html>'
    else:
        print 'No dynamic content providers are currently registered.'


def http_error(errno, msg, explanation=None, ctype="text/html"):
    """Format and return a very simple error message."""
    print "Status:", errno, msg
    print "Content-Type:", ctype
    print
    if explanation:
        print explanation
    else:
        print ("<html><head><title>%s</title></head>"
               "\n<body><h1>%s</h1>"
               "\n<p>Error %d generated in KOE content bridge."
               "\n</body></html>" % (msg, msg, errno))
