"""Plugin implementation of the ContentRegistrarAPI."""

import sys
import os
import errno
import ilu
from koe.common import pathhack
from koe.common import pluginlib
from koe.common import Suitcase
from koe.interfaces.stubs import ContentRegistrarAPI
from koe.interfaces.stubs import ContentProviderAPI

# XXX: Yuk!  We should packagize this stuff...
sys.path.insert(0, os.path.join(os.pardir, 'common'))

import statefile

DEFAULT_PROVIDER_DIR = os.path.normpath(
    os.path.join(pathhack.kos_root(),
                 'demos',
                 'provider',
                 'var',
                 'providers'))



def main():
    kos, ns = pluginlib.register('Registrar')
    registrar = Registrar()
    svcid = kos.bind_service('Registrar', registrar,
                             'ContentRegistrarAPI.ContentRegistrar')
    try:
        kos.run()
    finally:
        kos.unbind_service(svcid)



class Registrar:
    def __init__(self):
        # __providers is the data structure that forms the interface between
        # the plugin and the CGI bridge.  it is what gets saved by the plugin
        # as a pickle, and read by the bridge.  it's format is as follows:
        #
        # key => (dir, sbh)
        #
        # where `key' is the provider name registered in RegisterProvider;
        # `dir' is the absolute directory in the filesystem where the uploaded 
        # static content for the provider is stored. `sbh' is the ILU string
        # binding handle for the KP's ContentProviderAPI.ContentProvider
        # object which will be contacted for dynamic content.
        #
        # if dir is '', then there is no static content
        # if sbh is '', then there is no dynamic content
        self.__providers = {}
        self.__byhash = {}

    def __save(self):
        statefile.save(self.__providers)

    def __getname(self, ident):
        try:
            return self.__byhash[ident]
        except KeyError:
            raise ContentRegistrar.BadIdentifierError

    def RegisterProvider(self, name):
        if self.__providers.has_key(name):
            raise ContentRegistrarAPI.NonUniqueNameError, hash(name)
        self.__providers[name] = ('', '')         # no static or dynamic yet
        ident = hash(name)
        self.__byhash[ident] = name
        self.__save()
        return ident

    def UnRegisterProvider(self, ident):
        name = self.__getname(ident)
        del self.__byhash[ident]
        del self.__providers[name]
        self.__save()

    def RegisterLiveContent(self, ident, provider):
        name = self.__getname(ident)
        dir, sbh = self.__providers[name]
        self.__providers[name] = (dir, provider.IluSBH())
        self.__save()

    def UnRegisterLiveContent(self, ident):
        name = self.__getname(ident)
        dir, sbh = self.__providers[name]
        self.__providers[name] = (dir, '')        # no dynamic
        self.__save()

    def UploadStaticContent(self, ident, repr):
        name = self.__getname(ident)
        dir, sbh = self.__providers[name]
        # create the subdirectory
        omask = os.umask(002)
        try:
            try:
                pdir = os.path.join(DEFAULT_PROVIDER_DIR, name)
                makedirs(pdir)
            except os.error, (code, msg):
                if code == errno.EEXIST:
                    pass
                else:
                    raise
            # suitcase is a NamedPart.  Sadly, a KP cannot pass a
            # SuitcaseFilesystem through an ILU interface, so it must pass the 
            # KPRepr's NamedPart (which it actually rips out of a
            # SuitcaseFilesystem).  But to make our lives easier here, we
            # create yet another SuitcaseFilesystem object to manage the copy
            # from the suitcase to the real on-disk filesystem.
            sfs = Suitcase.SuitcaseFilesystem(repr)
            sfspath = Suitcase.SuitcaseFilesystemPath(sfs)
            todisk(pdir, sfs, sfspath)
        finally:
            os.umask(omask)
        self.__providers[name] = (pdir, sbh)
        self.__save()

    def UnloadStaticContent(self, ident):
        name = self.__getname(ident)
        dir, sbh = self.__providers[name]
        pdir = os.path.join(DEFAULT_PROVIDER_DIR, name)
        wipe(pdir)
        self.__providers[name] = ('', sbh)        # no static content
        self.__save()



def todisk(pdir, sfs, sfspath):
    for f in sfs.listdir():
        diskf = os.path.join(pdir, f)
        if sfspath.isdir(f):
            try:
                os.mkdir(diskf)
            except os.error, (code, msg):
                if code == errno.EEXIST:
                    pass
                else:
                    raise
            try:
                here = sfs.getcwd()
                sfs.chdir(f)
                todisk(diskf, sfs, sfspath)       # recurse
            finally:
                sfs.chdir(here)
        else:
            src = sfs.open(f)
            dst = open(diskf, 'w')
            dst.write(src.read())
            src.flush()
            src.close()
            dst.flush()
            dst.close()


# This definition is very handy, but is only in Python 1.5.2.
try:
    makedirs = os.makedirs
except AttributeError:
    # Ripped from Python 1.5.2
    def makedirs(name, mode=0777):
        """makedirs(path [, mode=0777]) -> None

        Super-mkdir; create a leaf directory and all intermediate ones.
        Works like mkdir, except that any intermediate path segment (not
        just the rightmost) will be created if it does not exist.  This is
        recursive.

        """
        head, tail = os.path.split(name)
        if head and tail and not os.path.exists(head):
            makedirs(head, mode)
        os.mkdir(name, mode)


# useful for debugging, unnecessary in Python 1.5.2
def rmdir(file):
    try:
        os.rmdir(file)
    except os.error, (code, msg):
        msg = msg + ': ' + file
        raise os.error(code, msg)

def unlink(file):
    try:
        os.unlink(file)
    except os.error, (code, msg):
        msg = msg + ': ' + file
        raise os.error(code, msg)

def _wipe(dir):
    for f in os.listdir(dir):
        file = os.path.join(dir, f)
        if os.path.isdir(file):
                _wipe(file)                       # recurse
                rmdir(file)
        else:
            unlink(file)

def wipe(dir):
    """Recursively wipe out a directory.

    Yes, this is very dangerous and should be used *very* carefully!  For
    my protection, wipe will only remove directories under
    DEFAULT_PROVIDER_DIR.
    """
    pdir = DEFAULT_PROVIDER_DIR
    if os.path.commonprefix([pdir, dir]) <> pdir:
        raise ValueError, dir
    _wipe(dir)
    rmdir(dir)


main()
