"""A Rights Management demo KP.

This demo is intended to be run interactively, but it can run
non-interactively too (e.g. on autopilot), primarily for
debugging/regression testing purposes.

The following options are recognized:

    --autopilot
    -a            -- run non-interactively.  Default is to run interactively.

    --repository kssname
    -r kssname
                  -- the name of the kernel running the image
                     repository.  Default is `imagerepo'.

    --sleeptime secs
    -s secs
                  -- sleep secs number of seconds after arriving at
                     the new location.  Default is zero seconds
                     (i.e. no sleeping between migrations).

    --timeout secs
    -t
                  -- amount of time to wait for KP synchronization.
                     Default is 150 seconds.  A negative timeout says
                     to never timeout (KP hangers on must be killed
                     differently).

Use the submit.py command line to pass these options to the KP:

    % submit.py rightsdemo.py -- --autopilot --sleeptime 5 --timeout 200

"""

import sys
import nstools
import string
import marshal
import iotools
import time
import getopt

# states
STARTING   = 'starting'                 # haven't gone anywhere yet
CURIOUS    = 'curious'                  # getting list of images
WHICHIMG   = 'whichimg'                 # asking which image to retrieve
RETRIEVING = 'retrieving'               # getting the requested image
RETRIEVED  = 'retrieved'                # image has been retrieved
WAITING    = 'waiting'                  # waiting for rendezvous
CLONED     = 'cloned'                   # cloned to make payment
GETKEY     = 'getkey'                   # get key from repository
RENDEZVOUS = 'rendezvous'               # rendezvous'ing with player
DONE       = 'done'                     # done for good

# stand-in for un-import-able module (really just their exceptions)
PlayerAPI = None
_TIMEOUT = 150
_SLEEP = 0



# This should really be a method on the KOS object.  Returns either
# the requested open'd service, or None if the timeout was reached.
# Timeout is in seconds.

def wait_for_service(kos, svcapi, svcname, timeout=120):
    # Currently, the KOS doesn't support registering interest in the
    # arrival of a service (the correct paradigm), so instead we check
    # continuously inside a busy loop.
    while timeout > 0:
        try:
            return kos.lookup_service(svcapi, svcname).Open()
        except nstools.BadPathError:
            # the service is not yet available
            timeout = timeout - 1
            time.sleep(1)
    return None

def usage(status):
    print __doc__ % globals()
    sys.exit(status)



class KP:
    __state = STARTING
    __iofp = None
    __iofp_owner_p = None
    __kos = None
    __myhdl = None
    __imagehost = 'imagerepo'
    __interactive = 1
    __sleeptime = _SLEEP
    __timeout = _TIMEOUT

    def __main__(self, kos):
        func = getattr(self, self.__state)
        if not self.__kos:
            self.__kos = kos
            self.__myhdl = kos.get_kphandle()
##          self.__debug('My handle: %s\n' % self.__myhdl)
        self.__whereami()
        time.sleep(self.__sleeptime)
        func()

    def __close(self, msg=''):
        if msg:
            self.__write(msg)
        if self.__iofp and self.__iofp_owner_p:
            self.__iofp.close()
        sys.exit()

    def __debug(self, msg='---'):
        if msg[-1] == '\n': msg = msg[:-1]
        text = '%s: %s' % (self.__state, msg)
        print text

    def __open_iofp(self):
        if not self.__interactive:
            use_dev='tty'
        else:
            use_dev=None
        # make a connection to an I/O window
        self.__iofp = iotools.get_io_mechanism(self.__kos,
                                               width=60,
                                               height=20,
                                               use_dev=use_dev)
        self.__iofp_owner_p = 1

    def __write(self, msg=''):
        try:
            self.__iofp.write(msg)
        # TBD: this is highly bogus since KPs should never ever see
        # ILU errors.  However we aren't doing the right amount of
        # bastionizing of the connectors so they can filter through,
        # like when the SimpleIO window gets closed...
        except:
            self.__open_iofp()
            self.__iofp.write(msg)

    def __whereami(self):
        self.__hostname = self.__kos.get_kos_name()
##      self.__debug('I am now at: %s\n' % self.__hostname)

    def __prompt(self, prompt, list):
        self.__write(prompt)
        answer = None
        while not answer:
            count = 1
            for item in list:
                self.__write('    %d: %s\n' % (count, item))
                count = count + 1
            try:
                raw = string.strip(self.__iofp.readline())
                if raw == '':
                    answer = 0
                else:
                    answer = string.atoi(raw)
            except ValueError:
                answer = -1
            if answer == 0:
                self.__write('You fell through the secret escape hatch!\n')
                self.__debug('You fell through the secret escape hatch!\n')
                self.__close()
                # never returns
            if answer < 1 or answer > len(list):
                self.__write('Not a good answer: %s\n' % raw)
                answer = None
        return list[answer-1]

    def __ask_noyes(self, question):
        noyes = {'No':0, 'Yes':1}
        return noyes[self.__prompt(question, noyes.keys())]

    def __singlestep(self):
        self.__write("Hit `Enter' to continue...\n")
        return self.__iofp.readline()

    def __parseargs(self):
        try:
            opts, args = getopt.getopt(
                sys.argv[1:],
                'as:t:r:h',
                ['autopilot', 'sleeptime=', 'timeout=', 'repository=', 'help'])
        except getopt.error, msg:
            print msg
            usage(1)

        for opt, val in opts:
            if opt in ('-h', '--help'):
                usage(0)
            elif opt in ('-s', '--sleeptime'):
                try:
                    self.__sleeptime = string.atof(val)
                except ValueError:
                    self.__debug('--sleeptime invalid argument, ignoring.')
            elif opt in ('-a', '--autopilot'):
                self.__interactive = 0
            elif opt in ('-t', '--timeout'):
                try:
                    self.__timeout = string.atof(val)
                except ValueError:
                    self.__debug('--timeout invalid argument, ignoring.')
            elif opt in ('-r', '--repository'):
                self.__imagehost = val

        if not self.__interactive:
            self.__debug('running on autopilot...\n')

    #
    # state machine
    #

    def starting(self):
        self.__parseargs()
        # remember my home base
        self.__homebase = self.__hostname
        #
        # figure out which kernel has the image repository, so get
        # list of kernels in the world, and search out the Image
        # Repository.  but only do it if in interactive mode
        #
        if self.__interactive:
            kpns = self.__kos.get_namespace()
            kosns = nstools.Lookup(kpns, 'world/kos').Open('Namespace.Context')
            koslist = kosns.List()
            koslist.sort()
            self.__imagehost = self.__prompt(
                'Which one has the image database?\n',
                koslist)
        self.__write('Migrating to image repository: %s\n' % self.__imagehost)
        self.__state = CURIOUS
        try:
            self.__kos.migrate(self.__imagehost)
        except self.__kos.exception('KOS.CommunicationError'), msg:
            self.__write('%s.  Bye bye!\n' % msg)

    def curious(self):
        # get the Image Repository plugin
        try:
            self.__imagerepo = self.__kos.lookup_service(
                'RAPK.Repository', 'BassImages').Open()
        except nstools.BadPathError:
            self.__write('No Bass Image Repository found...\n')
            self.__debug('No Bass Image Repository found...\n')
            self.__close()
        # retrieve the list of image handles so we can query the user
        self.__images = self.__imagerepo.Scan()
        self.__state = WHICHIMG
        self.__kos.migrate(self.__homebase)

    def whichimg(self):
        # find out which one the user wants and go get it...
        if self.__interactive:
            self.__imagehdl = self.__prompt(
                'which would you like to retrieve?\n',
                self.__images)
        else:
            self.__write('The Image Repository contained these handles:\n')
            for imghdl in self.__images:
                self.__write('    %s\n' % imghdl)
            import random
            self.__imagehdl = random.choice(self.__images)
        self.__write('Retrieving image with handle: %s\n' % self.__imagehdl)
        self.__state = RETRIEVING
        self.__kos.migrate(self.__imagehost)

    def retrieving(self):
        #
        # get the digital object
        #
        do = self.__imagerepo.GetDO(self.__imagehdl)
        #
        # find out what disseminations are valid
        #
        dometa = do.GetMetadata()
        dtypes = string.split(dometa.GetValue('Disseminations'))
        self.__write('%d dissemination types found: %s\n' %
                     (len(dtypes), dtypes))
        #
        # get the thumbnail if available
        #
        if 'thumbnail' in dtypes:
            td = do.GetDissemination('thumbnail')
            tc = td.GetTermsAndConditions()
            required = string.split(tc.GetValue('RequiredTerms'))
            if required:
                self.__write('thumbnail required Ts&Cs: %s\n' % required)
                self.__thumbnail = ''
            else:
                self.__write('retrieving thumbnail data\n')
                size = string.atoi(dometa.GetValue('ThumbnailSize'))
                self.__thumbnail = td.Read(size)
                td.Close()
                self.__write('Got all %d bytes of the thumbnail\n' % size)
        if 'hires' in dtypes:
            td = do.GetDissemination('hires')
            self.__tc = td.GetTermsAndConditions()
            # we know the high res image will have Ts & Cs we need to
            # negotiate, and we happen to know the order in which
            # these must be negotiated.
            self.__tc.SetValue('authentity', self.__myhdl)
            self.__ticket = self.__tc.GetValue('ticket')
##          self.__debug('my ticket: %s\n' % `self.__ticket`)
##          self.__debug('stealth ticket: %s\n' %
##                       `self.__tc.GetValue('ticket')`)
            # read the (encrypted) hires data
            size = string.atoi(dometa.GetValue('HiresSize'))
            self.__hires = td.Read(size)
            td.Close()
            self.__price = string.atoi(dometa.GetValue('HiresPrice'))
            self.__write('Got all %d bytes of the high resolution\n' % size)
        self.__state = RETRIEVED
        self.__kos.migrate(self.__homebase)

    def retrieved(self):
        global PlayerAPI
        #
        # We need to get these exceptions for later, and we aren't
        # allowed to import the PlayerAPI module.  Make it look
        # like it though... :-)
        #
        PlayerAPI = self.__kos.all_exceptions('PlayerAPI')
        #
        # Now that the thumbnail image data has been retrieved, we
        # want to display it to the user and see if they want to pay
        # for the high resolution image.  To prove the hires is
        # encrypted, we'll show it too.
        #
        factory = self.__kos.lookup_service(
            'PlayerAPI.PlayerFactory', 'PlayerFactory').Open()
        #
        # play the thumbnail
        #
        player1 = factory.GetNewPlayer()
        player1.SetImageName(self.__imagehdl + ' [thumbnail]')
        player1.SetImageData(self.__thumbnail)
        try:
            player1.Play()
        except PlayerAPI.ImageError:
            self.__write('The thumbnail could not be played\n')
        #
        # play the high resolution image
        #
        player2 = factory.GetNewPlayer()
        player2.SetImageName(self.__imagehdl + ' [hi resolution]')
        player2.SetImageData(self.__hires)
        try:
            player2.Play()
        except PlayerAPI.ImageError, msg:
##          self.__debug('Parent Play() failed: %s\n' % msg)
            self.__write('The hi resolution could not be played\n')
        #
        # ask the user if s/he wants to pay the cost to get the hi res image
        #
        if self.__interactive and not self.__ask_noyes(
            'Purchase the high resolution image for $%d?\n' % self.__price
            ):
            self.__close()
            # never returns
        else:
            # TBD: should we dig this out of the suitcase?  I'm
            # thinking we put a representation of your wallet account
            # in the suitcase, and then we can check the price against
            # that and do the debit if we have enough money.  Might be
            # interesting to hook an ecash system up here.  For now...
            pass
        #
        self.__debit = self.__price
        #
        # clone myself, and send the clone out to get the key.
        #
        state = self.__state
        self.__state = CLONED
        new_kpid = self.__kos.clone()
        #
        # We are in the parent (uncloned)
        #
        player = wait_for_service(self.__kos,
                                  'PlayerAPI.DecryptPlayer',
                                  'TollPlayer',
                                  timeout=self.__timeout)
        if not player:
            # the service was never available.  boo hoo!
            self.__write('Never found the TollPlayer service... oh well!\n')
            self.__debug('Never found the TollPlayer service... oh well!\n')
            return
        player.SetImageName(self.__imagehdl + ' [hi resolution]')
        player.SetImageData(self.__hires)
        # timing dependent.  this can fail if the clone hasn't yet set
        # the key.  no matter, just keep trying until it succeeds,
        # with a time out
        timeout = self.__timeout
        while timeout > 0:
            try:
                player.Play()
                break
            except PlayerAPI.AuthorizationFailed, msg:
                # We beat the key carrying KP back to the homebase.
                # Let's just keep waiting for it to show up.
##              self.__debug('timeout: %d' % timeout)
                timeout = timeout - 1
                time.sleep(2)
            except PlayerAPI.ImageError, msg:
                self.__write('The hi resolution could not be played\n')
                break
            except:
                # TBD: this is highly bogus since KPs should never
                # ever see ILU errors.  However we aren't doing the
                # right amount of bastionizing of the connectors so
                # they can filter through, like when the SimpleIO
                # window gets closed...
                break
        else:
            self.__write('Decryption rendezvous failed!\n')
        player.Close()
        self.__close()

    def cloned(self):
        # carry these...
        #
        # self.__iofp
        # self.__interactive
        self.__iofp_owner_p = None
        # self.__state
        # self.__kos
        # self.__homebase
        # self.__myhdl
        # self.__imagehost
        # self.__tc
        # self.__ticket
        # self.__debit
        #
        # leave these
        #
        del self.__hostname
        del self.__imagerepo
        del self.__images
        del self.__imagehdl
        del self.__thumbnail
        del self.__hires
        del self.__price
        #
        self.__write('Spending $%d to get the key.\n' % self.__debit)
        self.__write('Fetching key from: %s\n' % self.__imagehost)
        self.__state = GETKEY
        if self.__interactive:
            self.__singlestep()
        self.__kos.migrate(self.__imagehost)

    def getkey(self):
        global Mapping
        Mapping = self.__kos.all_exceptions('Mapping')
        # meet the terms and conditions
##      self.__debug('meeting terms and conditions for high res image\n')
        self.__tc.SetValue('ticket',      self.__ticket)
        self.__tc.SetValue('receipt',     `self.__debit`)
        self.__tc.SetValue('destination', self.__homebase)
        #
        # we should now be able to extract the decryption key
        #
##      self.__key = self.__tc.GetValue('key')
        try:
            self.__key = self.__tc.GetValue('key')
        except Mapping.KeyError, msg:
            self.__debug("Couldn't get my key!\n")
            self.__write("Couldn't get my key!\n")
            return
##      self.__debug('my key: %s\n' % `self.__key`)
##      self.__debug('stealth key: %s\n' % `self.__tc.GetValue('key')`)
        self.__state = RENDEZVOUS
        self.__kos.migrate(self.__homebase)

    def rendezvous(self):
        if self.__interactive:
            status = self.__singlestep()
            if status == '':
                # the readline returned an EOF because the talk window
                # was closed.  In this case none of the following will
                # work.
                self.__debug('Rendezvous timed out, waiting for input')
                return
        global PlayerAPI
        #
        # We need to get these exceptions for later, and we aren't
        # allowed to import the PlayerAPI module.  Make it look
        # like it though... :-)
        #
        PlayerAPI = self.__kos.all_exceptions('PlayerAPI')
        player = wait_for_service(self.__kos,
                                  'PlayerAPI.DecryptPlayer',
                                  'TollPlayer',
                                  timeout=self.__timeout)
        if not player:
            # the service was never available.  boo hoo!
            self.__write('Never found the TollPlayer service... oh well!\n')
            self.__debug('Never found the TollPlayer service... oh well!\n')
            return
        #
        # uncomment this line to test rendezvous logic in retrieved()
        # method above
        #
##      time.sleep(30)
        player.SetImageKey(self.__key)
##      self.__debug('Clone has set the key... exiting\n')
