import pagegraph
import link
import exclude

import urllib
import urlparse
import string
import UserList
import re

import random

import operator

VERBOSE = 0

class Queue(UserList.UserList):
    
    def getNext(self):
        next = self.data[0]
        del self.data[0]
        return next

class Response:
    """Cache the results of reads on an HTTP response"""
    def __init__(self, addinfourl):
        self.addinfourl = addinfourl
        self.__data = None
    def __getattr__(self, attr):
        return getattr(self, self.__dict__['addinfourl'])
    def read(self):
        if self.__data is None:
            self.__data = self.addinfourl.read()
        return self.__data

# helper for managing calls to superclass constructors
def superInit(superclass, self, *args, **kw):
    name = '_%s__init' % superclass.__name__  # mangle
    try:
        flag = getattr(self, name)
    except AttributeError:
        setattr(self, name, None)
        flag = None
    if flag is None:
        setattr(self, name, 1)
        apply(superclass.__init__, (self,) + args, kw)

class Crawler:

    def __init__(self, root):
        root = link.canonicalizeURL(root)
        self.initGraph(root)
        self.initURLopen()
        self.extractorClass = link.LinkExtractor
        self.summarizers = {}
        self.queue = Queue()
        self.queue.append(root)
        self.initRobotLimit()

    def initGraph(self, root):
        self.graph = pagegraph.Graph(root)

    def initURLopen(self):
        self.urlopen = urllib.URLopener().open

    def initRobotLimit(self):
        self.robotPolicy = exclude.Policy()
        sch, host, x, x, x, x = urlparse.urlparse(self.graph.rootUrl)
        url = urlparse.urlunparse((sch, host, '/robots.txt', '', '', ''))
        try:
            lines = self.urlopen(url).readlines()
        except IOError:
            self.robotPolicy.allows = lambda x: 1
            return
        else:
            self.robotPolicy.parse(lines, '') # no user-agent

    def crawl(self):
        while self.queue:
            url = self.queue.getNext()
            page = self.graph.get(url)

            if not self.robotPolicy.allows(url):
                if VERBOSE:
                    print "excluded by robots.txt", url
                continue

            if page.alreadyVisited():
                continue

            if VERBOSE:
                print url

            try:
                f = self.urlopen(url)
            except IOError, err:
                # collection the HTTP message and continue
##                print url, err
                msg = err[3]
                loc = msg.getheader('location')
                if loc:
                    self.queue.append(loc)
                    self.graph.addRedirect(url, loc)
                else:
                    page.addInfo(msg)
                continue
            else:
                page.addInfo(f.info())

            # avoid multiple reads by adding attr to f
            wrap = Response(f)
            self.summarize(page, wrap)
            self.extractLinks(page, wrap)
            
            f.close()

            if VERBOSE:
                print "queue size %4d\tgraph size %4d" % (len(self.queue),
                                                          len(self.graph))

    def addSummarizer(self, summarizer, types):
        for type in types:
            l = self.summarizers.get(type, None)
            if l is None:
                self.summarizers[type] = l = []
            l.append(summarizer)

    def summarize(self, page, resp):
        type = page.info.gettype()
        sums = self.summarizers.get(type, [])
        for summarizer in sums:
            summarizer(page, resp, self.graph)

    def extractLinks(self, page, resp):
        if page.info.gettype() != 'text/html':
            return
        
        body = resp.read()

        newlinks = []
        ext = self.extractorClass(page.url, body)
        for href in ext.extractLocalLinks():
            self.graph.addLocalEdge(page, href)
            newlinks.append(href)

        self.updateQueue(newlinks)

        for href in ext.extractExternalLinks():
            self.graph.addExternalEdge(page, href)

    def updateQueue(self, links):
        self.queue = self.queue + links
        
class DiameterLimitCrawler(Crawler):
    
    def __init__(self, root, diameter=None, width=None):
        superInit(Crawler, self, root)
        self.diameter = diameter
        self.width = width

    def extractLinks(self, page, resp):
        if VERBOSE:
            print page.url, self.graph.getDepth(page)
        if self.diameter and self.graph.getDepth(page) > self.diameter:
            if VERBOSE:
                print "will not crawl", page.url, ": too far away"
            return
        Crawler.extractLinks(self, page, resp)
            
    def updateQueue(self, links):
        if self.width and len(links) > self.width:
            if VERBOSE:
                print "skipping %d of %d links" % (len(links)-self.width, len(links))
            while len(links) > self.width:
                i = random.randint(0, len(links) - 1)
                del links[i]
        self.queue = self.queue + links

class StatsCollector:
    def __init__(self, crawler):
        import zlib
        self.__rawData = []
        crawler.addSummarizer(self.savePageSize, ('text/html',))

    def savePageSize(self, page, resp, graph):
        self.__rawData.append(resp.read())

    def getSize(self):
        return reduce(operator.add, map(len, self.__rawData))

    def getSizeZ(self):
        return len(zlib.compress(string.join(self.__rawData, '')))

class TitleGrabber:
    rx = re.compile('<TITLE>(.*)</TITLE>', re.I | re.S)
    def __init__(self, crawler):
        crawler.addSummarizer(self.grabTitle, ('text/html',))

    def grabTitle(self, page, resp, graph):
        body = resp.read()
        mo = self.rx.search(body)
        if mo:
            page.title = mo.group(1)

class TheCrawler(DiameterLimitCrawler):
    def __init__(self, root, diameter=None, width=None):
        superInit(DiameterLimitCrawler, self, root, diameter, width)
        # add summarizers
        # XXX how should a client access them
        StatsCollector(self)
        TitleGrabber(self)

if __name__ == "__main__":
    import sys
    import getopt
    import cPickle, pickle
    import zlib

    DIAMETER = None
    WIDTH = None
    VERBOSE = 0
    PROFILE = 0
    OUTPUT = None

    opts, args = getopt.getopt(sys.argv[1:], 'd:w:vpo:')
    for k, v in opts:
        if k == '-d':
            DIAMETER = string.atoi(v)
        elif k == '-w':
            WIDTH = string.atoi(v)
        elif k == '-v':
            VERBOSE = 1
        elif k == '-p':
            PROFILE = 1
        elif k == '-o':
            OUTPUT = v
            
    base = args[0]
    if DIAMETER or WIDTH:
        crawler = TheCrawler(base, DIAMETER, WIDTH)
    else:
        crawler = TheCrawler(base)

    if PROFILE:
        import profile
        import os
        dest = 'crawler.%d' % os.getpid()
        profile.run('crawler.crawl()', dest)
        print "profile stats in", dest
    else:
        crawler.crawl()

    p = pickle.dumps(crawler.graph, 1)

    if OUTPUT:
        f = open(OUTPUT, 'w')
        f.write(p)
        f.close()
        if VERBOSE:
            print "graph saved to", OUTPUT
    
    z = zlib.compress(p)

    print "graph size", len(crawler.graph)
    print "ext. data size", len(crawler.graph.extData)
    print "pickle size %6d (%6d .gz)" % (len(p), len(z))

    ps = crawler.getSize()
    psz = crawler.getSizeZ()
    print "page size   %6d (%6d .gz)" % (ps, psz)

    print "combined    %6d (%6d .gz)" % (ps + len(p), psz + len(z))

    print "reduction factor"
    print "pickle vs. pages\traw %6.2f\tgz %6.2f" % (float(ps) / len(p),
                                                     float(psz) / len(z))
