"""A graph of Web pages."""

import string
import random
import UserDict
import mimetools
import copy
import types

import link

# XXX need to canonicalize URLs

class Message(mimetools.Message):
    # redefine message so that we can pickle it
    def __init__(self, msg):
        self.dict = {}
        for k, v in msg.items():
            self.dict[k] = v
        self._reMimeIt()

    def _reMimeIt(self):
        self.encodingheader = self.getheader('content-transfer-encoding')
        self.typeheader = self.getheader('content-type')
        self.parsetype()
        self.parseplist()

    def __getstate__(self):
        return self.dict

    def __setstate__(self, state):
        self.dict = state
        self._reMimeIt()

# XXX would like a function to walk the tree (even though we have a
 # cyclic graph)

# XXX need to turn a graph of URLs into a hierarchy based on paths

def createHierarchy(urls):
    """Organize list of URLs into directories by looking at the path"""
    import os, urlparse

    unparse = {}
    paths = []
    for url in urls:
        path = urlparse.urlparse(url)[2]
        unparse[path] = url
        paths.append(path)
    paths.sort()
    dirs = {}
    for path in paths:
        base, ext = os.path.split(path)
        dir = dirs.get(base, None)
        if dir is None:
            dir = []
        dir.append(unparse[path])
        dirs[base] = dir
    return dirs

class Graph(UserDict.UserDict):
    def __init__(self, root):
        self.rootUrl = root
        self.data = {}
        # only provide dict-like access to the local graph
        self.extData = {}
        self.depth = {root:0}

    def get(self, url):
        """Returns an existing page or creates a new ones"""
        page = self.data.get(url, None)
        if page is None:
            page = self.data[url] = self._newPage(url)
        return page

    def _newPage(self, url):
        return Page(url)

    def getRootURL(self):
        return self.rootUrl

    def getRoot(self):
        return self.get(self.rootUrl)
        
    def addLocalEdge(self, page, url):
        """Add link from page to url.
        
        Creates the child page if it doesn't already exist.
        """
        if page.url == url:
            return
        child = self.get(url)
        page.children[child] = page.children.get(child, 0) + 1
        if self.depth.has_key(page.url):
            cur = self.depth.get(url, None)
            if cur is None:
                self.depth[url] = self.depth[page.url] + 1
            else:
                new = self.depth[page.url] + 1
                if new < cur:
                    self.depth[url] = new
        child.parents[page] = child.parents.get(page, 0) + 1

    def getDepth(self, urlOrPage):
        if type(urlOrPage) != types.StringType:
            url = urlOrPage.url
        else:
            url = urlOrPage
        depth = self.depth.get(url, None)
        if not depth is None:
            return depth
        page = self.data[url]
        min = None
        minParent = None
        for parent in page.parents.keys():
            if min is None:
                minParent = parent
                min = self.getDepth(parent)
            else:
                d = self.getDepth(parent)
                if d < min:
                    min = d
                    minParent = parent
        self.depth[url] = min
        return min

    def addRedirect(self, source, dest):
        page = self.data[source]
        page.url = dest

    def getExt(self, url):
        page = self.extData.get(url, None)
        if page is None:
            page = self.extData[url] = self._newPage(url)
        return page

    def addExternalEdge(self, page, url):
        """Add link from page to url.
        
        Creates the child page if it doesn't already exist.
        """
        child = self.getExt(url)
        page.extChildren[child] = page.extChildren.get(child, 0) + 1
        child.extParents[page] = child.extParents.get(page, 0) + 1

class GraphNoExternalLinks(Graph):
    """A serialization wrapper that doesn't save external links"""
    def __init__(self, root=None, graph=None):
        if graph:
            # we are wrapping another graph after the fact
            self.graph = graph         # do I need this for any reason?
            self.rootUrl = graph.root
            self.data = graph.data
            self.extData = {}
            self.depth = graph.depth
        elif root:
            Graph.__init__(self, root)
        else:
            raise ValueError, "must supply root URL or graph to constructor"

    def _newPage(self, url):
        return PageNoExternalLinks(url)

    def __getstate__(self):
        # throw away external links, recreate self.depth on demand
        return self.rootUrl, self.data

    def __setstate__(self, (rootUrl, data)):
        self.rootUrl = rootUrl
        self.data = data
        self.extData = {}
        self.depth = {self.rootUrl:0}
        self.graph = None

class Page:
    """A node in the graph"""
    
    def __init__(self, url):
        # metadata
        self.url = url
        self.info = None
        # local edges
        self.children = {}
        self.parents = {}
        # external edges
        self.extChildren = {}
        self.extParents = {}

    def __getinitargs__(self):
        return (self.url,)

    def __repr__(self):
        return '<%s>' % self.url

    def alreadyVisited(self):
        if self.info:
            return 1
        else:
            return 0

    def addInfo(self, info):
        self.info = Message(info)

class PageNoExternalLinks(Page):
    def __getstate__(self):
        return self.url, self.info, self.children, self.parents
    
    def __setstate__(self, (url, info, children, parents)):
        self.url = url
        self.info = info
        self.children = children
        self.parents = parents
