"""Demonstration of a parallel programming style for Knowbot Programs.

Performs a standard merge sort that uses cloning during the divide
phase to spread the computation across multiple systems. Each KP
sorts a list of integers, located in the suitcase in the file
'unsorted', and places the results in a file 'sorted'. During each
divide phase, if each subproblem is large enough, a clone is created
to carry out the computation.

When the KP starts, it sets some global state that limits the
number of clones created based on the constant KPS_PER_KOS and the
number of KOSes listed in the namespace. (When a clone is created, it
is sent to a random KOS; so there may be more than KPS_PER_KOS at one
time.)

The global state is a threshold for the minimum size problem that a KP
will perform without creating a clone. When the number of elements to
sort is less than the cutoff_size, the entire problem is performed by
the current KP. If it is greater than the cutoff_size, two clones are
created to a sort half the elements.

The cutoff_size is based on a programmer defined limit on the maximum
number of clones in the world at one time. Realizing that after n
levels of recursion in merge sort we have 2^n - 1 clones, we simply
need to choose an value for 'n' that guarantees we won't have too many
clones. Instead of storing the maximum recursion depth and the current
depth, we store the size of the subproblem at that level. We
can easily determine the size of the current problem, which can be
compared to the cutoff_size.
"""

import nstools
import marshal
import whrandom
import ilu
 
KPS_PER_KOS = 2

class KP:

    def __init__(self):
        self.parent = None
        self.data = []
        self.results = []
        self.cutoff_size = None
        self.state = 'main_sorting'

    def __main__(self, kos):
        "Get the function named by the state attribute and call it."
        main = getattr(self, self.state)
        main(kos)

    # two different main methods depending on the KP's state
    # (could actually have a third for setting global state)

    def main_sorting(self, kos):
        "Sort data in suitcase and return results to parent."
        self.load_unsorted_data(kos)
        if self.cutoff_size is None:
            self.set_global_state(kos)

        # need good seed. data to sort is only good if numbers aren't
        # repeated often
        #rand.srand(self.data[0])

        if self.len_data > self.cutoff_size:
            # send clones
            me = kos.get_kphandle()
            where = kos.get_kos_name()
            recv = Receiver(kos)
            kos.bind_service(me, recv, "Sort.Merger")
            childinfo = (where, me)
            self.create_clones(kos, childinfo)
            kos.run()       # returns after both children have returned
            self.results = self.merge(recv.one, recv.two)
        else:
            # just perform the computation locally
            # self.results = self.merge_sort(self.data)
            self.data.sort()
            self.results = self.data
        self.store_data(kos)
        self.return_to_parent(kos)

    def main_merging(self, kos):
        "Return sorted data to the parent KP."
        self.load_sorted_data(kos)
        merger = kos.lookup_service("Sort.Merger", self.parent[1]).Open()
        try:
            merger.Receive(self.data)
        except (ilu.IluProtocolError, ilu.IluGeneralError):
            raise RuntimeError, "failed to deliver data to %s" % `merger`

    def load_unsorted_data(self, kos):
        sfs = kos.get_suitcase()
        f = sfs.open('unsorted')
        self.data = marshal.loads(f.read())
        self.len_data = len(self.data)
        f.close()
        sfs.unlink('unsorted')   # don't need it anymore

    def load_sorted_data(self, kos):
        sfs = kos.get_suitcase()
        f = sfs.open('sorted')
        self.data = marshal.loads(f.read())
        self.len_data = len(self.data)
        f.close()
        sfs.unlink('sorted')     # don't need it anymore

    def store_data(self, kos):
        sfs = kos.get_suitcase()
        f = sfs.open('sorted', 'w')
        f.write(marshal.dumps(self.results))
        f.close()
        if self.parent is None:
            # don't kill this so that we can print it
            return
        self.results = []       # we're doing this instead of pickling

    def set_global_state(self, kos):
        "Decide when to create clones."
        import math
        def log2(x, base=math.log10(2), log=math.log10):
            return log(x) / base

        hosts = nstools.WorldOpen('kos').List()
        max_kps = len(hosts) * KPS_PER_KOS + 1
        max_depth = int(log2(max_kps))
        self.cutoff_size = int(self.len_data / int(math.pow(2, max_depth)))
        if self.cutoff_size < 256:
            # don't be completely silly
            self.cutoff_size = 256

    def pick_rand_hosts(self, kos):
        """Pick two hosts to migrate the subproblems to.

        Tries to avoid sending clones to parent host or current host,
        and tries to send them to different hosts."""

        # looks like a place for rater.d

        kernels = nstools.WorldOpen('kos').List()
        this_kernel = kos.get_kos_name()
        try:
            kernels.remove(this_kernel)
        except ValueError:
            pass
        if not kernels:
            return this_kernel, this_kernel

        first = whrandom.choice(kernels)
        if len(kernels) > 1:
            kernels.remove(first)
            second = whrandom.choice(kernels)
            return first, second
        return first, first

    def create_clones(self, kos, parent):
        "Create clones to solve the subproblems."
        original_parent = self.parent
        original_data = self.data
        self.parent = parent
        self.data = []

        first_dest, second_dest = self.pick_rand_hosts(kos)

        # here's the equivalent of the merge sort divide phase
        middle = self.len_data / 2
        self.write_clone_data(kos, original_data[:middle])
        try:
            kos.clone(first_dest)
        except (ilu.IluProtocolError, ilu.IluGeneralError):
            raise RuntimeError, "Failed to clone to %s" % first_dest
        
        self.write_clone_data(kos, original_data[middle:])
        try:
            kos.clone(second_dest)
        except (ilu.IluProtocolError, ilu.IluGeneralError):
            raise RuntimeError, "Failed to clone to %s" % second_dest
            kos.clone()

        self.data = original_data
        self.parent = original_parent
        
    def write_clone_data(self, kos, data):
        sfs = kos.get_suitcase()
        f = sfs.open('unsorted', 'w')
        f.write(marshal.dumps(data))
        f.close()

    def return_to_parent(self, kos):
        if self.parent is None:
            # we're the last one, so return home
            print self.results
        else:
            self.state = 'main_merging'
            parent_host = self.parent[0]
            kos.migrate(parent_host)
            return
            # perhaps we ought to look for exceptions here
            try:
                kos.migrate(parent_host)
            except ilu.IluProtocolError:
                raise RuntimeError, "failed migration to %s" % parent_host

    # what was this KP supposed to do after all this setup???
    # oh, yeah! it was going to sort stuff

    def merge_sort(self, list):
        if len(list) > 1:
            middle = len(list) / 2
            left = self.merge_sort(list[:middle])
            right = self.merge_sort(list[middle:])
            merged = self.merge(left, right)
            return merged
        else:
            return list

    def merge(self, sorted1, sorted2):
        # HACK HACK HACK
        dest = sorted1 + sorted2
        dest.sort()
        return dest
        
        # two counters
        dest = []
        c1 = c2 = 0
        l1 = len(sorted1)
        l2 = len(sorted2)
 
        while c1 < l1 and c2 < l2:
            if sorted1[c1] > sorted2[c2]:
                dest.append(sorted2[c2])
                c2 = c2 + 1
            else:
                dest.append(sorted1[c1])
                c1 = c1 + 1
        if c1 < l1:
            dest = dest + sorted1[c1:]
        elif c2 < l2:
            dest = dest + sorted2[c2:]
 
        return dest

class Receiver:
    """Receives the results of clone KPs solving the subproblems.

    When both KPs have returned and delivered their sorted data sets
    to the Receiver, it calls kos.stop(). The call returns control to
    the KP inside its main function, where it can merge the two
    subsets together and return to its parent."""

    def __init__(self, kos):
        self.one = []
        self.two = []
        self.receipts = 0
        self.done = kos.stop

    def Receive(self, data):
        if self.receipts == 0:
            self.one = data
            self.receipts = 1
        elif self.receipts == 1:
            self.two = data
            self.receipts = 2
            self.done()

    
