"""Kernel service that coordinates connector interchange."""

from koe.interfaces.stubs import KOS, KOS__skel
from koe.interfaces.stubs import KOSKernel
from koe.interfaces.stubs import Namespace
from koe.common import nstools, NamespaceImpl

class ConnectionManager:
    """Kernel service that coordinates use of connectors.

    - Maintains a portion of the namespace for connectors.

    - Is responsible for cleanup of connectors left over from processes
    that suffered non-standard exits."""

    def __init__(self):
        self.db = {}            # owner: {instance: (type, name, context)}
        self.init_services_ns()

    def init_services_ns(self):
        """Establish the categorical contexts for connector bindings."""

        # Install the 'services' and 'services-by-type' contexts.
        #
        # For example:
        #
        # <some_kos>
        #      [...]
        #      services
        #	   simple-client-server-demo <Seller1.Seller connector>
        #	   simpleio <SimpleIOAPI.SimpleIOFactory connector>
        #	   test_seller <Seller1.Seller connector>
        #      services-by-type
        #	   Seller1.Seller
        #	       simple-client-server-demo <Seller1.Seller connector>
        #	       test_seller <Seller1.Seller connector>
        #	   SimpleIOAPI.SimpleIOFactory
        #	       simpleio <SimpleIOAPI.SimpleIOFactory connector>

	self.services_ns = nstools.Context(NamespaceImpl.Context())
	self._nsroot.Bind('services', self.services_ns.GetDescriptor())
	self.services_by_type_ns = nstools.Context(NamespaceImpl.Context())
	self._nsroot.Bind('services-by-type',
			  self.services_by_type_ns.GetDescriptor())

    def Wrapup(self):
        """Shutdown connector management."""

        # Clear out the connector entries and framework from namespace.
        self._nsroot.Unbind('services')
        self._nsroot.Unbind('services-by-type')

    def RegisterConnector(self, requester, instance, type, name):
        """Record connectors, creating type, name slots in namespace as needed.
        """
        # Create type slot, if necessary, for use by client:
        context = None
        try:
            self.services_by_type_ns.Lookup1(type)
        except nstools.BadPathError:
            context = NamespaceImpl.Context()
            self.services_by_type_ns.Bind(type, context.GetDescriptor())

        if self.db.has_key(requester):
            self.db[requester][instance] = (type, name, context)
        else:
            self.db[requester] = {instance: (type, name, context)}
        
    def UnregisterConnector(self, requester, instance):
        """Remove registration for connector instance owned by requester."""
        try:
            del self.db[requester][instance]
        except KeyError:
            raise KOSKernel.NotFound, ("%s does not have connecter %s"
                                       % (requester, instance))
            
    def DeleteOwnersConnectors(self, owner):
        """Cleanup all service entries for specified owner.

        This is cleanup for untidy exit of owner's supervisor."""

        if not self.db.has_key(owner):
            return
        for t, n, c in self.db[owner].values():
            try:
                self.services_by_type_ns.Lookup1(t).Open().Unbind(n)
            except (nstools.BadPathError, Namespace.BindError):
                continue
            try:
                self.services_ns.Unbind(n)
            except Namespace.BindError:
                continue
