"""KOS kernel true object that tracks a resident KP's characteristics.

Characteristics includes kpID, metadata, and run state."""

import sys

import ilu
from koe.interfaces.stubs import KOS
from koe.interfaces.stubs import KOSKernel, KOSKernel__skel
from koe.interfaces.stubs import ReportingAPI, ReportingAPI__skel
from koe.interfaces.stubs import Namespace
from koe.common import nstools
from koe.common import NamespaceImpl
    
from ProcessEntryImpl import ProcessEntry

TRANSFER_CHUNK_SIZE = 10240



class KPEntry(KOSKernel__skel.KPEntry, ProcessEntry):
    """KOS kernel communications link with a KP subprocess."""

    proctype = "KP"
    runstate = ''

    def __init__(self, kernel, kosns,
                 repr,
                 reporting_sbh,
                 reporting_id,
                 kos_name = '',
                 visualizer = None):

        # from the namespace for our kernel, derive the namespace
        # context for KPs and for the worldroot
        self._kosns = kosns
        self._kpsns = kosns.Lookup1('kps').Open(typeobj='Namespace.Context')
        if not self._kpsns:
            # XXX Shouldn't happen.
            raise nstools.BadPathError, "Could not get `kps' namespace"
        self._worldns_d = kosns.Lookup1('world')
        self._visualizer = visualizer

        ProcessEntry.__init__(self, kernel)

        self.repr = repr                # Encapsulated KP representation
        self.reporting_sbh = reporting_sbh
        self.reporting_id = reporting_id        # kpid viz. reporting station
        self.kos_name = kos_name

        self._ns = nstools.Context(NamespaceImpl.Context())
        self._ns.Bind('kos', self._kosns.GetDescriptor())
        self._ns.Bind('world', self._worldns_d)
        self._nsd = self._ns.GetDescriptor()
        self._kpsns.Rebind(self.reporting_id, self._nsd)
        # XXX If this fails, something else is wrong!

        # XXX There's something else the matter here as well: we
        # assume the reporting_id is unique.  Since the client has
        # control over the reporting_id, this assumption could be
        # wrong.  It would be better if the kpid (which is guaranteed
        # to be unique and never reused) were used as the name, but
        # then we would have to have an interface to ask for the
        # reporting-id (e.g. for the visualiser).  Sigh.  Later...

    def SetPID(self, pid):
        """Register underlying-os process id.  This signifies process start."""
        # Called on the kernel side, this is not part of the remote interface.
        ProcessEntry.SetPID(self, pid)

    # Handshakes between kernel and supervisor

    def Hello(self, link):
        "Initial handshake between supervisor and kernel."
        if self.runstate:
            raise KOS.CommunicationError, \
                  "Inappropriate Hello call, runstate=%s" % `self.runstate`
        self.EstablishProcessLink(link)
        self.runstate = 'running'
        return (self.kos_name,
                self.kpid,
                self.repr,
                self._visualizer,
                self._nsd)

    def Goodbye(self, new_kpid, new_kos_name):
        "This KP has left us via migration."
        self.byebye()

    def Terminated(self, reason, details):
        "This KP has terminated."
        self.byebye()

    # Utilities

    def byebye(self):
        if self.runstate != 'running':
            print "Inappropriate byebye() call, runstate=%s, repid=%s" % (
                `self.runstate`, `self.reporting_id`)
            return
        self.process_died()

    def process_died(self):
        self.runstate = 'exiting'
        # XXX We should only unbind if this is bound to *our*
        # namespace.  When we are migrating to ourselves, it is
        # already bound to the namespace of the new incarnation.
        # Note that there's no race condition: if we migrated to
        # ourselves, the change is guaranteed to have happened, and if
        # we don't, it is guaranteed not to happen, ever :-).
        it = self._kpsns.Lookup1(self.reporting_id)
        self.kernel.DeleteOwnersConnectors(self.kpid)
        if it.reference() == self._nsd.reference():
            self._kpsns.Unbind(self.reporting_id)

    def reporting_station(self):
        """Retrieve connection to the reporting station for this KP.

        Useful to the Receiver, perhaps others.
        """
        # XXX Note - we may want to make a shared cache of reporting
        # station surrogates in the class, keyed on SBH'es, to reduce tcp
        # connections when multiple KPs come from a common reporting stn.
        rs = ilu.ObjectOfSBH(ReportingAPI.ReportingStation, self.reporting_sbh)
        return rs
