"""Kernel service that receives and manages local KP processes.
"""

import sys, os, string
import ilu
import signal

from StringIO import StringIO
import KPEntryImpl, PluginEntryImpl

from koe.common import pathhack
from koe.common import nstools
from koe.common import access
from koe.common import KPReprImpl

from koe.interfaces.stubs import KOS, KOS__skel
from koe.interfaces.stubs import Mapping
from koe.interfaces.stubs import KPRepr


# profiling information
prof = 0
prof_dir = ''

# KP starter interface
SBHFD = 3                               # Fd to read sbh from
PIDFD = 4                               # Fd to write pid to
STSFD = 5                               # Fd to write status to

class Receiver(KOS__skel.Receiver):

    # Command to run Python scripts (e.g. kpstarter.py or bootstrap.py).
    # The first item must be the full path to the Python interpreter;
    # the remaining items can be additional arguments (e.g. '-v')
    BOOTSTRAP = [sys.executable]

    def __init__(self, name, kernel, visualizer):
        self.name = name
        self.visualizer = visualizer
        self.kernel = kernel

        self.kp_entries = {}
        self.ext_entries = {}
        self.kosns = None
        self.kpstarter_pid = None

    def bindnames(self, kosns):
        """Bind receiver names in the kernel's namespace context."""
        self.kosns = kosns
        # get a descriptor for our receiver and bind it in our namespace
        desc = nstools.Descriptor(typename='KOS.Receiver',
                                  reference=self.IluSBH())
        kosns.Bind('receiver', desc)

    def SubmitRepr(self, orig_repr):
        if not access.ilu_allow_access('Submit', exception=KOS.ResourceError):
            raise KOS.AuthorizationError, "No authorization to submit KP"
        self.check_duplicate_repid(orig_repr)
        repr = KPReprImpl.Representation()
        repr.loadfromrepr(orig_repr)
        return self.submit_repr(repr)

    def SubmitKP(self, type, trans_repr):
        if not access.ilu_allow_access('Submit', exception=KOS.ResourceError):
            raise KOS.AuthorizationError, "No authorization to submit KP"
        if type != KOS.AlphaFormat:
            raise KOS.EncodingError, "unknown submission type"
        repr = KPReprImpl.Representation()
        repr.loadfrommarshal(trans_repr)
        return self.submit_repr(repr)

    def allowed_access(self, type):
        """Examine the access file and the credentials of the caller
           to insure that we wish to allow a connection of this type."""
        try:
            fqdn = access.ilu_callers_fqdn()
        except:
            raise KOS.ResourceError, \
                  "Failed to access callers credentials"
        if not fqdn:
            raise KOS.ResourceError, \
                          "Address: %s not equal gethostbyname(%s)!" \
                          % (address, same_address)
        try:
            # obtain the appropriate LimitParser object
            perm = access.limitparser_by_type(type)
            perm.readfile()
        except:
            raise KOS.ResourceError, \
                  "Failed to obtain Limit%sParser object" % type

        if perm.is_allowed_from(fqdn):
            return 1
        return 0
            
    def check_duplicate_repid(self, repr):
        """Check for duplicate reporting ID.

        We don't want to admit KPs with a reporting ID that's
        identical to that of a KP we are already hosting, since this
        would screw up the namespace (since the reporting ID is used
        as the name for the KP's own namespace).

        So we raise KOS.MetadataError if we detect a duplicate.

        However, we have to avoid complaining when we're migrating
        from ourselves, in which case we will find a sure match!  We
        can check for this case by realizing that migrate() actually
        submits our *own* KP repr object, so we can compare the
        IlUSBH() (not assuming that ILU actually guarantees to give us
        the true object here -- it might though).

        Additional frustrating detail: we need to skip entries whose
        run state is exiting, since they may be dying earlier
        incarnations of the same KP that have migrated but whose
        process exit status hasn't been reported yet.

        Sigh...
        """
        try:
            reporting_id = repr.GetMetadata().GetValue('reporting-id')
        except Mapping.KeyError, msg:
            raise KOS.EncodingError, 'Missing metadata value for %s.' % msg
        for entry in self.kp_entries.values():
            if entry.runstate == 'exiting':
                continue
            if entry.reporting_id == reporting_id:
                if entry.repr.IluSBH() != repr.IluSBH():
                    raise KOS.MetadataError, \
                          "Duplicate reporting-id: %s" % `reporting_id`
    
    def KillKP(self, kpid, signal, password):
        # Since there's no table mapping a kpid to a KP entry,
        # search the table indexed by pid.
        for entry in self.kp_entries.values():
            # Allow callers to specify kpid or the reporting station id
            if entry.kpid == kpid or entry.reporting_id == kpid:
                break
        else:
            raise KOS.ResourceError, "no KP id %s" % `kpid`
        if not entry.pid:
            raise KOS.ResourceError, "no Unix process for KP id %s" % `kpid`
        
        # Consult access.conf and fake password validation
        if not access.ilu_allow_access('Kpkill', exception=KOS.ResourceError) \
           or not password:
            raise KOS.AuthorizationError, \
                  "No authorization to kill KP id %s" % `kpid`

        try:
            os.kill(entry.pid, signal)
        except os.error, msg:
            raise KOS.ResourceError, \
                  "OS error killing KP id %s: %s" % (`kpid`, str(msg))
        else:
            self.visualizer.KPTerminated(entry.reporting_id,
                                          "Killed by kernel.", `signal`,
                                          entry.repr)

    def submit_repr(self, repr):
        metadata = repr.GetMetadata()
        try:
            repid  = metadata.GetValue('reporting-id')
            repsbh = metadata.GetValue('reporting-station')
            # We require the following keys; if one isn't set we get KeyError:
            metadata.GetValue('knowbot-class')
            metadata.GetValue('main-module')
        except Mapping.KeyError, msg:
            raise KOS.MetadataError, 'Missing metadata value for %s.' % msg

        entry = KPEntryImpl.KPEntry(self.kernel,
                                    self.kosns,
                                    repr,
                                    repsbh,
                                    repid,
                                    kos_name=self.name,
                                    visualizer=self.visualizer)

        # Always report reception to visualizer and reporting station:
        try:
            self.visualizer.KPMigrationReceived(entry.reporting_id,
                                                entry.kpid, self.name)
        except (ilu.IluGeneralError, ilu.IluProtocolError):
            print "failed to contact visualizer for kp %s" % (entry.kpid)
        rs = entry.reporting_station()
        if not rs:
            print "Could not reach reporting station.  Oh well."
        else:
            # this call can fail. what should we do?????
            (repid, kpid) = (entry.reporting_id, entry.kpid)
            try:
                rs.KPMigrationReceived(repid, kpid, self.name)
            except (ilu.IluProtocolError, ilu.IluGeneralError):
                print "Could not notify reporting station.  Oh well."

        pid = self.bootstrap_kp(entry)

        # These statements *must* be in this order.
        # Bonus points if you understand why:
        entry.SetPID(pid)               # Notify entry of succesful fork
        self.kp_entries[pid] = entry

        return kpid

    def bootstrap_kp(self, entry):
        """Bootstrap the KP and return its process id.

        This version uses the kpstarter process.
        """
        if not self.kpstarter_pid:
            self.create_kpstarter()
        try:
            os.write(self.kpstarter_sbhpipe, entry.IluSBH() + '\n')
            line = os.read(self.kpstarter_pidpipe, 100)
        except os.error, msg:
            raise KOS.ResourceError, "KP starter pipe I/O error: " + str(msg)
        if not line:
            raise KOS.ResourceError, "KP starter pipe read got EOF"
        try:
            pid = string.atoi(line[:-1])
        except string.atoi_error:
            parts = string.split(line)
            if parts[0] == "ResourceError":
                raise KOS.ResourceError, \
                      "KP starter: " + string.join(parts[1:])
            raise KOS.ResourceError, "KP starter protocol error: " + `line`
        return pid

    def RegisterPlugin(self, pid):
        """Allocate an PluginEntry for a KOS plugin process.

        Returns the entry to the registering process."""
        # assume plugin entries are going to get created by kernel
        # before it is launched
        try:
            return self.ext_entries[pid].IluSBH()
        except KeyError:
            entry = self.create_plugin_entry(self.kernel, pid)
            return entry.IluSBH()

    def create_plugin_entry(self, kernel, pid, name=""):
        entry = PluginEntryImpl.PluginEntry(self.kernel, name)
        entry.SetPID(pid)
        self.ext_entries[pid] = entry
        return entry

    def reaper(self):
        """Reap deceased children without blocking."""

        while 1:
            try:
                pid, sts = os.waitpid(-1, os.WNOHANG)
            except os.error:
                break
            if not pid:
                break
            self.report_process_status(pid, sts)

    def report_process_status(self, pid, sts):
        """Report status on a terminated process."""

        if sts & 0xff:
            message = "killed by signal %d" % (sts & 0x7f)
            if sts & 0x80:
                message = message + " -- core dumped"
        elif sts & 0xff00:
            message = "exit status %d" % (sts >> 8)
        else:
            message = ""

        if self.kp_entries.has_key(pid):
            entry = self.kp_entries[pid]
            del self.kp_entries[pid]
            self.report_kp_demise(entry, pid, message)
        elif self.ext_entries.has_key(pid):
            entry = self.ext_entries[pid]
            del self.ext_entries[pid]
        elif pid == self.kpstarter_pid:
            self.kpstarter_pid = 0
            print "KP starter died, pid %d, %s" % (pid, message or "exit 0")
            # bootstrap_kp() will attempt to create a new one.
            # XXX Should we mark all current children as orphaned?
        else:
            # A child we don't know about has died.
            # This may have been forked by another module, or our
            # administration may be messed up.  We can't tell.
            # In any case, report this to the console, so someone
            # might be alerted that there's a problem (most likely,
            # they might correlate it to missing children reported
            # elsewhere).
            print "Unknown child died, pid %d, %s" % (pid, message or "exit 0")

    def report_kp_demise(self, entry, pid, message):
        """Report the demise of a KP process."""

        if message:
            print
            print "*** Supervisor process", pid, message

        if entry.runstate == 'exiting':
            # An expected death.
            return

        # Unexpected demise.
        if not message:
            message = "exited unexpected with status 0"
            print "*** Supervisor process", pid, message

        entry.process_died()

        # Get the reporting station and try to send the report:
        rs = entry.reporting_station()
        if not rs:
            print "Could not reach reporting station.  Oh well."
        else:
            message = "Supervisor process " + message
            repid = entry.reporting_id
            # This call can fail. What should we do?????
            rs.KPTerminated(repid, "ProcessFailed", message, entry.repr)

    def create_kpstarter(self, quiet=None):
        """Create the KP starter subprocess and setup pipes to it.

        If this raises an exception, the KOS kernel cannot function
        properly.  Potential exceptions are in the pipe() and fork()
        calls.  Exceptions in the execv() call in the child must be
        handled differently -- it's too late to raise an exception in
        the parent!

        Note that it is sort of okay (in a debug setting) to kill the
        KP starter process if it has no children -- reaper() will zap
        kpstarter_pid, and the next call to bootstrap_kp() will call
        create_kpstarter() again.  This may be useful to test a new
        version of the supervisor code.

        """

        # Set the KP starter command line
        args = self.BOOTSTRAP[:]
        args.append(os.path.join(pathhack.kos_root(),
                                 'supervisor/kpstarter.py'))
        if prof:
            args.append('-p')
            args.append(self.name) 
            if prof_dir:
                args.append('--profile-dir')
                args.append(prof_dir)
        program = args[0]
        if not os.path.isfile(program):
            raise RuntimeError, "No KP starter script: " + `program`
        args[0] = os.path.basename(program) # Looks nicer in ps listing

        # Create pipes
        # NB os.pipe() returns a 2-tuple: (read_end, write_end)
        sbhpipe = os.pipe()
        pidpipe = os.pipe()
        stspipe = os.pipe()

        # Fork a subprocess
        pid = os.fork()
        if pid:
            # Parent -- save pid, keep some pipe ends, close others
            self.kpstarter_pid = pid
            self.kpstarter_sbhpipe = sbhpipe[1]
            self.kpstarter_pidpipe = pidpipe[0]
            self.kpstarter_stspipe = stspipe[0]
            os.close(sbhpipe[0])
            os.close(pidpipe[1])
            os.close(stspipe[1])
            ilu.RegisterInputHandler(self.kpstarter_stspipe,
                                     self.stspipe_handler)
            if not quiet:
                print "KP starter created"
        else:
            # Child -- shuffle pipes around, exec KP starter script
            os.dup2(sbhpipe[0], SBHFD)
            os.dup2(pidpipe[1], PIDFD)
            os.dup2(stspipe[1], STSFD)
            map(os.close, sbhpipe+pidpipe+stspipe) # Close all six
            # Close all file descriptors except 0, 1, 2, and
            # {SBH,PID,STS}FD.
            # XXX Assume file descriptors are below 100.
            # XXX Assume standard Unix streams are 0, 1, 2.
            fdlist = range(3, 100)
            for i in SBHFD, PIDFD, STSFD:
                fdlist.remove(i)
            for i in fdlist:
                try:
                    os.close(i)
                except os.error:
                    pass
            try:
                os.execv(program, args)
            except os.error, msg:
                try:
                    print "execv", program, "failed:", msg
                except:
                    pass
                os._exit(1)
            except:
                os._exit(127)

    def stspipe_handler(self):
        BUFSIZE = 100
        fd = self.kpstarter_stspipe
        data = os.read(fd, BUFSIZE)
        if not data:
            ilu.UnregisterInputHandler(self.kpstarter_stspipe)
            self.kpstarter_stspipe = -1
            os.close(fd)
            print "KP starter sts pipe was closed"
            self.reaper()
            return
        while data[-1] != '\n':
            newdata = os.read(fd, BUFSIZE)
            if not newdata:
                break
            data = data + newdata
        lines = string.split(data[:-1], '\n')
        for line in lines:
            try:
                i = string.index(line, ':')
                pid = string.atoi(line[:i])
                sts = string.atoi(line[i+1:])
            except (string.atoi_error, string.index_error):
                print "KP starter sts pipe protocol error:", `line`
            else:
                self.report_process_status(pid, sts)

    def shutdown(self):
        for pid in self.ext_entries.keys():
            try:
                os.kill(pid, signal.SIGTERM)
            except os.error, msg:
                print "kernel shutdown: failed to kill pid", pid, ":", msg
