"""Supervisor-side handling of triggers.

Implements the class TriggerManager which is intended to be used as a
mixin for the Kernel class.
"""

import sys
import random
import ilu
from koe.interfaces.stubs import KOSKernel
from koe.common import nstools, NamespaceImpl



class TriggerManager:
    """Manages triggers and callbacks."""

    def __init__(self):
        self.events = {}
        self.triggers = {}
        # self.receiver defined by Kernel class
        # Create the conditions namespace for clients to use:
        self.condition_ns = nstools.Context(NamespaceImpl.Context())
        self._nsroot.Bind('conditions', self.condition_ns.GetDescriptor())

    def notify(self, var, broadcast=0):
        "Queue and process a broadcast or signal event."

        # Here's the rub: All the notifications for one Broadcast should
        # happen before any of the notifications for the next
        # Broadcast. But the signal/broadcast call goes through the ILU
        # runtime, which implements a kind of multi-threading using a
        # recursive call inside the IluMainLoop.
        #
        # Solution: We *can* guarantee that any section of code
        # that does not involve calls through ILU will be executed
        # atomically; i.e. the ILU runtime is not truly
        # multithreaded. Thus, we can just assure that the test-and-set
        # at the begin of this routine has no ILU calls; then it will be
        # executed atomically.

        try:
            self.events[var].append(broadcast)
        except KeyError:
            raise KOSKernel.NotFound, `var`
        if len(self.events[var]) > 1:
            return                   # already handling a signal/broadcast

        # Begin processing events if this is the first event in the queue.
        while self.events[var]:
            broadcast = self.events[var][0]      # get event type
            if not self.triggers[var]:
                # no one to deliver event to
                procs = []
            if broadcast:
                # Get list of processes to notify. Can't use the
                # trigger list itself, because it may be updated by
                # registrations that arrive while this broadcast is
                # being handled.
                procs = self.triggers[var]
                self.triggers[var] = []
            else:
                # Choose an arbitrary process to receive the signal.
                one = random.choice(self.triggers[var])
                self.triggers[var].remove(one)
                procs = [one]

            for pid in procs:
                try:
                    # Get communication link to process.
                    proclink = self.processlinks[pid]
                except KeyError:
                    print "Tried to fire trigger on unknown pid %s" % pid
                else:
                    try:
                        if broadcast:
                            proclink.Broadcast(var)
                        else:
                            proclink.Signal(var)
                    except (ilu.IluGeneralError, ilu.IluProtocolError), msg:
                        print "ILU error firing trigger on pid %s:" % pid
                        print str(sys.exc_type)+':', sys.exc_value
        
            # Done handling event. Delete it from queue.
            del self.events[var][0]

    def CreateCondition(self):
        """assign a new variable id"""
        var = random.randint(0, 32767)
        while self.events.has_key(var):
            var = random.randint(0, 32767)

        self.triggers[var] = []
        self.events[var] = []

        return var

    def Signal(self, var):
        self.notify(var, 0)

    def Broadcast(self, var):
        self.notify(var, 1)

    def RegisterTrigger(self, var, pid):
        """Register a new trigger"""
        try:
            self.triggers[var].append(pid)
        except KeyError:
            raise KOSKernel.NotFound, str(var)

    def UnregisterTrigger(self, var, pid):
        """Unregister an existing trigger."""
        try:
            self.triggers[var].remove(pid)
        except ValueError:
            # pid wasn't in list
            pass
        except KeyError:
            raise KOSKernel.NotFound, str(var)

