#! /usr/bin/env python

# This documentation string also serves as a usage message
"""Run the KOS kernel process.

Usage:  kernel.py [-n name] [-v] [-h]

Where:

    --name name
    -n name      -- Set the name of the KOS (default: local hostname).

    --verbose
    -v           -- turn on verbose output in the KPEntryImpl module.

    --quiet
    -q           -- be extra quiet about plugin startup

    --profile
    -p           -- turn on profiling

    --profile-dir dir
                 -- send profiling output to the named directory

    --help
    -h           -- print a help message.

"""

# path hackery, no longer requires $PYTHONPATH to be set properly
import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../"))

import os
import sys
import time
import getopt
import string

import pluginutil
# XXX need this until Python 1.5.2 is required
from koe.pythonlib import ConfigParser
import ilu
import Receiver
import CnManager
import TriggerMgr
import AdministratorImpl

from koe.common import pathhack
from koe.common import qsplit
from koe.common import ilutools
from koe.common import nstools
from koe.common import NamespaceImpl
from koe.interfaces.stubs import Namespace

DuplicateNameError = 'kernel.DuplicateNameError'
quiet = None


class Kernel(CnManager.ConnectionManager, TriggerMgr.TriggerManager):

    def __init__(self, name):
        self.name = name
        self._administrator = AdministratorImpl.Administrator(self)
        reporter = self._administrator.get_reporter()
        self.receiver = Receiver.Receiver(self.name, self, reporter)
        # create a place to hold a kpid->processlink mapping
        self.processlinks = {}
        # bind all namespaces
        self.bindnames()

        CnManager.ConnectionManager.__init__(self)
        TriggerMgr.TriggerManager.__init__(self)

        # launch plugins & bind their names
        self.launch_plugins()
        self.receiver.create_kpstarter(quiet)
        self.start_receiving()

    def bindnames(self):
        """Bind names in various namespaces.

        - bind our KOS kernel name into the /worldroot/kos context
        - create kernel root namespace for kps, etc.
        """

        # Get the worldroot namespace context
        nsworld = nstools.GetWorldroot()
        nskos = nstools.WorldOpen("kos")

        # Make sure no live kernel with the same name already exists
        try:
            other = nstools.WorldOpen("kos/%s/receiver" % self.name)
        except nstools.BadPathError:
            pass
        else:
            raise DuplicateNameError, "A live kernel %s exists" % `self.name`

        # Delete a binding with the same name if one exists -- it's dead
        try:
            nskos.Unbind(self.name)
        except Namespace.BindError, msg:
            pass
        else:
            if not quiet:
                print "Warning: replaced dead host %s" % `self.name`

        # Create our root namespace and create subdirectories
        self._nsroot = nsroot = nstools.Context(NamespaceImpl.Context())
        self._kpscontext = NamespaceImpl.Context()
        nsroot.Bind('kps', self._kpscontext.GetDescriptor())
        self._pluginscontext = NamespaceImpl.Context()
        nsroot.Bind('plugins', self._pluginscontext.GetDescriptor())

        # Bind the public interfaces; bind us into the world last
        self._administrator.bindnames(nsroot)
        self.receiver.bindnames(nsroot)
        nsroot.Bind('world', nsworld.GetDescriptor())
        nskos.Bind(self.name, nsroot.GetDescriptor())

    def launch_plugins(self):
        """Launch the plugins registered to run-at-boot."""
        try:
            opts = pluginutil.PluginOpts(kosname=self.name)
        except ConfigParser.ParsingError, e:
            print "Error parsing configuration files for %s" % self.name
            print e
        except ConfigParser.Error, e:
            print "Unexpected error loading configuration files for %s" % \
                  self.name
            print e

        if not quiet:
            print 'Launching plugins...'
        self.fix_python_path()
        for pluginname in opts.sections():
            try:
                bool = opts.getboolean(pluginname, 'run-at-boot')
            except ConfigParser.Error, e:
                print "Error in configuration of plugin %s for %s" % \
                      (pluginname, self.name)
                print e
                print "Failure: plugin %s not launched" % pluginname
            except ValueError, e:
                print "Error in configuration of plugin %s for %s" % \
                      (pluginname, self.name)
                print "Invalid boolean value for 'run-at-boot': %s" % e
                print "Failure: plugin %s not launched" % pluginname
            if bool:
                pid = os.fork()
                if pid:
                    self.receiver.create_plugin_entry(self, pid, pluginname)
                    # probably want to share pipe between parent &
                    # child, and pass the sbh of the entry across the
                    # pipe. would need to change pluginlib.
                else:
                    try:
                        x = opts.get(pluginname, 'executor')
                        d = opts.get(pluginname, 'dir')
                        f = opts.get(pluginname, 'file')
                        a = opts.get(pluginname, 'args')
                    except ConfigParser.Error, e:
                        # It doesn't matter which opt fails.  We ain't
                        # gonna launch this plugin.
                        print
                        print "Error in configuration of plugin %s for %s" % \
                              (pluginname, self.name)
                        print e
                        print "Failure: plugin %s not launched" % pluginname
                        os._exit(1)
                    a0 = os.path.basename(x)
                    if f:
                        command = (a0, f) + tuple(qsplit.qsplit_nq(a))
                    else:
                        command = (a0, ) + tuple(qsplit.qsplit_nq(a))
                    print pluginname, "\n", d, command
                    try:
                        os.chdir(d)
                    except os.error:
                        print ("Failure: %s directory not found: %s"
                               % (pluginname, d))
                        os._exit(1)
                    if not quiet:
                        print '    plugin: %s\n\t%s' % (pluginname, command)
                    # Close all file descriptors except 0, 1, 2.
                    # Need this to avoid hanging clients when the kernel
                    # dies but a plugin still lives.
                    # XXX Assume file descriptors are below 100.
                    # XXX Assume standard Unix streams are 0, 1, 2.
                    for i in range(3, 100):
                        try:
                            os.close(i)
                        except os.error:
                            pass
                    os.execvp(x, command)

    def fix_python_path(self):
        env = os.environ
        key = 'PYTHONPATH'
        try:
            rawpath = env[key]
        except KeyError:
            path = []
        else:
            path = string.split(rawpath, os.pathsep)
        root = pathhack.kos_root()
        if root not in path:
            path.insert(0, root)
            env[key] = string.join(path, os.pathsep)

    def start_receiving(self):
        try:
            if not quiet:
                print "Kernel serving as '%s'..." % self.name
            ilutools.RunInterruptableMainLoop()
        finally:
            if not quiet:
                print "Kernel exiting..."
            self.receiver.shutdown()
            self._administrator.withdraw_kos()
            CnManager.ConnectionManager.Wrapup(self)
            try:
                kos_ns = nstools.WorldOpen("kos", "Namespace.Context")
            except nstools.BadPathError:
                return
            try:
                kos_ns.Unbind(self.name)
            except (ilu.IluGeneralError, Namespace.BindError):
                pass

    def shutdown(self):
        ilutools.ExitMainLoop()

            

if __name__ == "__main__":
    global quiet
    from koe.common import kptools
    name = kptools.default_kernel()

    usage = 0
    error = 0
    args = None
    try:
        optlist, args = getopt.getopt(sys.argv[1:], 'hn:vpq',
                                      ['help', 'name=',
                                       'verbose', 'quiet',
                                       'profile-dir=', 'profile'])
    except getopt.error, msg:
        print msg
        error = 1
        usage = 1
        optlist = []

    if args:
        error = 1
        usage = 1
        optlist = []

    for opt, val in optlist:
        if opt in ('-n', '--name'):
            name = val
        elif opt in ('-h', '--help'):
            usage = 1
        elif opt in ('-v', '--verbose'):
            import KPEntryImpl
            KPEntryImpl.verbose = 1
        elif opt in ('-p', '--profile'):
            Receiver.prof = 1
        elif opt == '--profile-dir':
            Receiver.prof_dir = val
        elif opt in ('-q', '--quiet'):
            quiet = 1

    if usage:
        print __doc__
        sys.exit(error)

    # XXX: This is a horrible kludge; the story goes like this: ILU maintains
    # a file descriptor budget such that when connections are idle, their fds
    # can be hijacked for other ongoing connections.  That's all fine and
    # dandy except that sometime between ILU 2.0a9 and 2.0a12, something
    # changed (read: broke) such that at some random times RPC method calls
    # will just fail with an IluProtocolError: lost connection.
    #
    # I have no idea why this happens, nor why this stuff all worked fine in
    # 2.0a9 but is broken with 2.0a12.  Dan Larner (of the ILU team at Xerox)
    # says in a message:
    #
    # 1. Threaded Python 1.5 with ILUalpha 12 still has strange problems.
    # Python 1.4 is fine.
    #
    # 2. Forking - When you fork, you end up with multiple copies sharing
    # things like file descriptors.  What goes on in one process (e.g.
    # closing a connection) affects others - in general forking can result in
    # all sort of wierd behavior due to inconsistencies in the shared
    # resources.  This is probably the main cause of your problems.
    #
    # I'm not sure I buy this, since my test program was very careful not to
    # import any ILU specific stuff until it was in the leaves of a forking
    # tree.  Still, who knows?  In any case, if what Dan says is true then
    # we're going to have to significantly re-organize our code, which is way
    # to much work at this time.
    #
    # The upshot is that the simple workaround is to just crank the fd budget
    # up to some ungodly number, and 2000 seems about as good as anything.  It
    # is high enough such that other system resources (e.g. memory, swap
    # space) would probably be exhausted before ILU's fd budget.
    #
    ilu.SetFDBudget(2000)
    try:
        kernel = Kernel(name)
    except nstools.NoWorldrootError, msg:
        print "No worldroot:", msg
        sys.exit(1)
    except DuplicateNameError, msg:
        print "Duplicate name:", msg
        sys.exit(1)
