"""Client side resource management (connectors, conditions, triggers, ...).

Used, with some slight differences, by both KPs (supervisor) and plugins.
"""

import sys, imp, types
import os
import string
import whrandom

import ilu
from ilutools import RunMainLoop, ExitMainLoop
if sys.modules.has_key('Tkinter'):      # use other main loop if Tk is used
    # XXX I'm not so sure that this works, since there's no ExitMainLoop
    from ilu_tk import RunMainLoop

import nstools
import NamespaceImpl
import traceback
import types

from koe.interfaces.stubs import KOSKernel
from koe.interfaces.stubs import KOSKernel__skel
from koe.interfaces.stubs import Namespace



class ConnectionBroker:
    """Provide KP connector lookup and service registration services."""

    # List of names a KP can access
    PUBLICS = (
        'run',
        'stop',
        'bind_service',
        'unbind_service',
        'stop_all_services',
        'lookup_service',
        'list_services',
        'list_service_types',
        'create_condition',
        'signal',
        'broadcast',
        'register_trigger',
        'unregister_trigger',
        'stop_all_triggers',
        'all_exceptions',
        'exception',
        )

    def __init__(self, pentry):
        self._pentry = pentry
        self.ns = None
        # regular connector management
        self.services_by_type_ns = None
        self.services_ns = None
        self.connector_db = {}  # descriptor: (name, type, impl, true, cid)
        # Conditions/triggers management
        self.condition_ns = None
        self.condition_db = {}  # descriptor: name
        self.triggers = {}      # maps from ids to callbacks
        # mainloop management
        self.server = None

    def handle_cnbroker_termination(self):
        """Destroy (presumably deactivate) the server and deregister bindings.
        """
        self.stop_all_services()
        self.stop_all_triggers()
        self.stop_all_conditions()
        self.server = None

    # set_ns, get_ns -- utilities for the Supervisor

    def set_ns(self, ns):
        """Set the namespace after initialization.

        As a side effect, the condition and services namespaces are
        inserted in it."""

        self.ns = ns
        by_name = 'kos/services'
        by_type ='kos/services-by-type'
        self.services_ns = nstools.Lookup(ns, by_name).Open()
        self.services_by_type_ns = nstools.Lookup(ns, by_type).Open()
        self.condition_ns = nstools.Lookup(ns, 'kos/conditions').Open()

    def get_ns(self):
        return self.ns

    # starting and stopping

    # This is a little complicated because ILU doesn't like you to
    # call ExitMainLoop() outside the mainloop (i.e., before
    # RunMainLoop()).  However, we really want this -- e.g. when you
    # start a service and expect a specific call to terminate the
    # service, that specific call could arrive before you enter your
    # main loop when you make additional ILU calls before starting the
    # main loop.  So, we keep two flags: in_main_loop and
    # stop_main_loop.  Both are instance variables that shadow class
    # variables that provide the initial values.  The meaning of
    # in_main_loop is easy enough: it is set just before entering the
    # main loop and reset on leaving.  stop_main_loop is set by
    # stop(), and reset by run() when it exits.  When run() finds
    # stop_main_loop true on entry, it resets it and exits
    # immediately.

    in_main_loop = 0
    stop_main_loop = 0

    def run(self):
        """Enter the ILU main loop, unless stop() has already be called."""
        if self.in_main_loop:
            raise RuntimeError, "kos.run(): recursive main loop not supported"
        if self.stop_main_loop:
            self.stop_main_loop = 0
            return
        self.in_main_loop = 1
        try:
            RunMainLoop()
        finally:
            self.in_main_loop = 0
            self.stop_main_loop = 0

    def stop(self):
        """Stop the ILU main loop (the current or the next one)."""
        self.stop_main_loop = 1
        if self.in_main_loop:
            ExitMainLoop()

    # interface to services

    def stop_all_services(self):
        """Unregister services and stop the mainloop."""
        # Unbind everything we own.
        for descriptor in self.connector_db.keys():
            try:
                self.unbind_service(descriptor)
            except KOSKernel.NotFound, msg:
                print ("Warning: %s: '%s' '%s'" %
                       (msg,
                        self.connector_db[descriptor][0],
                        self.connector_db[descriptor][1]))

    def bind_service(self, name, impl, type):
        """Register implementation object vs connector type with kernel.

        Returns a namespace descriptor for the implementation, which is
        also used to identify the service to .unbind_service().

        '.run()' activates the bound services."""

        if not self.server:
            # Instantiating from the skeleton class in combine would create
            # the server, if not already done, so lets do it explicitly:
            self.server = ilu.CreateServer()

        true = self.combine(impl, type)
        cid = true.IluSBH()
        descriptor = nstools.Descriptor(reference = cid,
                                        typename = type)

        self._pentry.RegisterConnector(cid, type, name)

        self.services_ns.Bind(name, descriptor)
        # By-type-ns categories are created by RegisterConnector when needed.
        type_ns = self.services_by_type_ns.Lookup1(type).Open()
        type_ns.Bind(name, descriptor)

        self.connector_db[descriptor] = (name, type, impl, true, cid)

        return descriptor

    def unbind_service(self, descriptor):
        """Unregister service from kernel and namespace."""
        try:
            (name, type, impl, true, cid) = self.connector_db[descriptor]
        except KeyError:
            raise ValueError, "Service not found: %d" % `descriptor`

        # XXX Any of the ILU calls in this method could fail.
        #
        # There is an inherent race condition during shutdown here.  When
        # koehalt.py is run, kernels are shutdown by calling the Shutdown()
        # method on their Administrator interface.  This will exit the
        # kernel's mainloop and eventually call the Receiver's shutdown()
        # method.
        #
        # During Receiver.shutdown() a SIGTERM signal is sent to all the
        # subprocesses it is managing.  This in turn exits the plugin's
        # mainloop, passing control back to the plugin usercode.  The next
        # thing plugin usercode typically does is unbind the service, getting
        # us here.
        #
        # However, it is possible that by the time we get here, the kernel
        # which contains the process entry object will have already exited.
        # In this case any of the true objects being referenced below may be
        # invalid.  Depending on which of the calls below fail (and it always
        # changes), a BadPathError, BindError or IluGeneralError may result.
        #
        # The shutdown procedure needs to be redesigned to avoid such race
        # conditions.  Preferrably, instead of a signal from kernel to
        # plugins, an explicit Shutdown call would be added (much like the
        # Administrator API).  Until then, we'll just ignore any errors that
        # can occur.  It shouldn't matter much because the processes are all
        # existing anyway.
        try:
            self.services_by_type_ns.Lookup1(type).Open().Unbind(name)
            self.services_ns.Unbind(name)
            self._pentry.UnregisterConnector(cid)
        except (nstools.BadPathError,
                Namespace.BindError,
                ilu.IluGeneralError), e:
            print 'Unbinding call failed:', e
            pass
        del self.connector_db[descriptor]

    def lookup_service(self, type, name):
        """Return descriptor for specified registered service."""

##        return nstools.Lookup(self.services_by_type_ns,
##                              "%s/%s" % (type, name))
        try:
            type_ns = nstools.Lookup(self.services_by_type_ns, type).Open()
        except nstools.BadPathError:
            raise nstools.BadPathError, "No services of type %s" % type
        try:
            return nstools.Lookup(type_ns, name)
        except nstools.BadPathError:
            raise nstools.BadPathError, "No service of type %s named %s" % (type, name)

    def list_services(self, type):
        """Return name list of services of specific type."""

        try:
            return nstools.Lookup(self.services_by_type_ns,
                                  type).Open().List()
        except nstools.BadPathError:
            return []

    def list_service_types(self):
        """Return name list of types of registered services."""
        
        return self.services_by_type_ns.List()

    # Triggers methods.
    # the trigger methods below call on the same methods in the
    # ProcessEntry. the ProcessEntry is an ILU surrogate for the true
    # object in the kernel.

    # API methods first

    def create_condition(self, name):
        """Create a new conditional variable in this process.
        
        Bind the name in the conditions namespace; return the
        variable.
        
        """

        var = self._pentry.CreateCondition()
        condition = Condition(var)
        desc = nstools.Descriptor(reference=condition.IluSBH(),
                                  typename="KOSKernel.Condition")
        self.condition_ns.Bind(name, desc)
        self.condition_db[var] = name
        return condition

    def signal(self, cvar):
        var = cvar.GetID()
        self._pentry.Signal(var)

    def broadcast(self, cvar):
        var = cvar.GetID()
        self._pentry.Broadcast(var)

    def register_trigger(self, cvar, callback):
        """Register the function callback on a condition."""
        if not isinstance(cvar, Condition):
            raise TypeError, "first argument must be a condition variable"
        var = cvar.GetID()
        try:
            self.triggers[var].append(callback)
        except KeyError:
            self.triggers[var] = [callback]
            self._pentry.RegisterTrigger(var)

    def unregister_trigger(self, cvar, callback):
        """Unregister the function callback on a condition."""
        if not isinstance(cvar, Condition):
            raise TypeError, "first argument must be a condition variable"
        var = cvar.GetID()
        try:
            self.triggers[var].remove(callback)
        except (KeyError, ValueError):
            pass
        else:
            if not self.triggers[var]:
                del self.triggers[var]
                self._pentry.UnregisterTrigger(var)

    # Methods called by the KOS kernel
    # These methods implement the ProcessLink interface

    def Signal(self, var):
        self.receive_signal(var)
        
    def Broadcast(self, var):
        self.receive_broadcast(var)

    # internal methods for triggers

    def receive_signal(self, var):
        try:
            callbacklist = self.triggers[var]
        except KeyError:
            return
        callback = whrandom.choice(callbacklist)
        callbacklist.remove(callback)
        if callbacklist:
            # the kernel is going to forget your interest,
            # but if you've got more than one trigger
            # you need to re-register
            self._pentry.RegisterTrigger(val)
        else:
            # no more callbacks -- delete the list
            del self.triggers[var]
        try:
            callback()
        except:
            print "Condition variable callback raised exception"
            traceback.print_exc()

    def receive_broadcast(self, var):
        try:
            callbacklist = self.triggers[var]
        except KeyError:
            return
        del self.triggers[var]
        for callback in callbacklist:
            try:
                callback()
            except:
                print "Condition variable callback raised exception"
                traceback.print_exc()

    def stop_all_triggers(self):
        """Unregister triggers."""
        variables = self.triggers.keys()
        self.triggers = {}
        for var in variables:
            self._pentry.UnregisterTrigger(var)

    def stop_all_conditions(self):
        """Unregister conditions."""
        for n in self.condition_db.values():
            self.condition_ns.Unbind(n)

    # exceptions

    def all_exceptions(self, type_module):
        """Return module containing exceptions from specified type module.

        type_module
            Name of type module as string."""

        mod = get_type_mod(type_module)
        got = imp.new_module(type_module)
        for k in dir(mod):
            if k and k[0] != "_":
                v = getattr(mod, k)
                if type(v) == types.StringType:
                    setattr(got, k, v)
        return got
        
    def exception(self, typenm):
        """Return specified exception-object value.

        typenm
            Name of exception, as 'module.exception'."""

        modnm, objnm = split_type_name(typenm)
        mod = get_type_mod(modnm)
        try:
            got = getattr(mod, objnm)
            if type(got) == types.StringType:
                return got
        except AttributeError:
            pass
        # Fall here if exception is not found or value is not string.
        raise ValueError, "Exception not found: %s" % `typenm`

    def combine(self, impl, skel_type):
        """Return instance combining implementation instance w/ skeleton type.
        The skeleton type can be either the name of the type or the actual
        class object for the type."""

        if type(skel_type) == types.ClassType:
            skel_class = skel_type
        else:
            skel_class = get_type_class(skel_type)
        skel_obj = skel_class()
        # Populate the skeleton object with method attrs from the impl obj:
        #
        # BAW note: Ken's original implementation scanned the skel
        # class for its attributes, but did not take into account
        # walking the base classes.  In fact, it's quite problematic
        # to do this from the skel class because the ILU class chain
        # does not use Python inheritance on the skel side (even
        # though it does on the surrogate side).
        #
        # I've changed the delegate (method call forwarding) search
        # algorithm so that it scans the implementation classes and
        # follows their inheritance chain.  This can mean some methods
        # that aren't part of the skel get put in there anyway, but
        # that shouldn't matter since the skel classes only have API
        # methods, which do not start with underscore (and we filter
        # out underscores here).
        #
        # TBD: ask the ILU folks to add a reference to the inheritance
        # chain in the skel classes, either true inheritance, or an
        # attribute pointing to the `base' class.
        def do_class(skelobj, implobj, class_, recurse):
            # depth-first
            for baseclass in class_.__bases__:
                recurse(skelobj, implobj, baseclass, recurse)
            for name in dir(class_):
                if name[:1] == '_': continue
                try:
                    attr = getattr(implobj, name)
                    setattr(skelobj, name, attr)
                except AttributeError:
                    pass
        do_class(skel_obj, impl, impl.__class__, do_class)
        return skel_obj


class Condition(KOSKernel__skel.Condition):

    def __init__(self, id):
        self.id = id

    def GetID(self):
        return self.id


def split_type_name(name):
    """Return type name split into module and object components."""
    try:
        [mod, obj] = string.splitfields(name, ".")
        return mod, obj
    except ValueError:
        raise ValueError, "Invalid type name: %s." % `name`
    

def get_type_mod(name, skel=0):
    """Return module object for specified connector type."""
    modname = name
    if skel:
        modname = name + "__skel"
    try:
        # all interfaces live in the koe.interfaces.stubs package (relative to
        # $KOSROOT), so you need to first get a hold of the package, then dig
        # the module out of there
        pkg = __import__('koe.interfaces.stubs',
                         globals(), locals(),
                         [modname])
        return getattr(pkg, modname)
    except (ImportError, AttributeError):
        skelmsg = (skel and " skeleton") or ""
        raise ImportError, ("Connector type%s %s not found."
                            % (skelmsg, `name`))
    

def get_type_class(typenm):
    """Return class object corresponding to connector type."""
    modnm, objnm = split_type_name(typenm)
    mod = get_type_mod(modnm, skel=1)
    try:
        skel_class = getattr(mod, objnm)
    except AttributeError:
        raise AttributeError, "%s: %s not defined in %s" % (typenm, objnm,
                                                            modnm)
    return skel_class
