"""Knowbot Program MIME Parser version 0.1

Parses a MIME document, enforcing the format specified in 'MIME
Transport of Knowbot Programs, version 0.1'.

Classes:

        KPMimeParser
            High level interface to KP MIME format

Exceptions:

        KPMimeError
            An error occurred in the parsing of KP MIME format for the
            given version (currently 0.1).

"""

from MimeParser import MimeParser
from types import *
import rfc822
import string

KPMimeError = 'KPMimeParser.KPMimeError'



def _parse_external_body(headers, body):
    atype = headers.getparam('access-type')
    url = headers.getparam('url')
    if string.lower(atype) <> 'url' or not url:
        raise KPMimeError, 'invalid inclusion by reference'
    return MimeParser(body).index() + (url,)



class KPMimeParser:
    """MIME Transport of Knowbot Programs 0.1.

    This class parses the data in a file-like object, digging out KP
    information, enforcing the format as described in the above
    document.

    Public Methods:

    KPMimeParser(fp)  -- pass in a file-like object
    metadata()        -- returns the metadata information
    code()            -- returns the code information
    state()           -- returns the state information
    suitcase()        -- returns the suitcase information

    """
    def __init__(self, fp):
        self._mime = MimeParser(fp)
        self._metadata = {}
        self._code = {}
        self._state = {}
        self._suitcase = {}
        self._parse()

    def metadata(self):
        """Returns a dictionary of the metadata headers.

        This dictionary contains a mapping of the MIME headers found
        in the bodies of the MIME multipart/knowbot-metadata parts.
        The keys are the metadata name and the values are the string
        value of the header.

        Note that all metadata parts are normalized into a single
        part.  Metadata parts cannot be included by reference.  They
        must have a Content-Type of message/rfc822.  Extraneous
        headers are ignored.

        """
        return self._metadata

    def code(self):
        """Returns a dictionary of the KP program code parts.

        The keys of the dictionary are the Content-Name: values found
        in the various parts, which are required for each program code
        part, and which must be unique within the
        multipart/knowbot-code part.

        The value of the dictionary is a tuple of two elements.  The
        first element is the mimetools.Message() object for that part.
        The second element is the contents of the part, as a string.

        Note that for code included as human readable text, the main
        type of the part (as found in the content-type entry in the
        first dictionary) will be text.  For binary data, it whole
        content-type will be application/knowbot-code-object.

        TBD: parts included by reference are not yet support.

        """
        return self._code

    def state(self):
        """Returns a dictionary of the KP program state parts.

        The keys of the dictionary are the Content-Name: values found
        in the various parts, which are required for each program code
        part, and which must be unique within the
        multipart/knowbot-state part.

        The value of the dictionary is a tuple of two elements.  The
        first element is the mimetools.Message() object for that part.
        The second element is the contents of the part, as a string.

        Note that for code included as human readable text, the main
        type of the part (as found in the content-type entry in the
        first dictionary) will be text.  For binary data, it whole
        content-type will be application/knowbot-state-object.

        TBD: parts included by reference are not yet support.

        """
        return self._state

    def suitcase(self):
        """Returns a dictionary of the KP suitcase sections.

        Structure is same as for state(), but used for different
        purposes.

        """
        return self._suitcase

    def _parse(self):
        ctype, msgobj, body = self._mime.index()
        # verify that the top-level package has the right
        # content-type: header and version information.
        if ctype <> 'multipart/knowbot':
            raise KPMimeError, 'multipart/knowbot expected'
        if msgobj.getparam('version') <> '0.1':
            raise KPMimeError, 'version 0.1 expected'
        if type(body) <> type([]):
            raise KPMimeError, 'list of subparts expected'
        # Now we need to scan through all the subparts, collating
        # information as we go.  We expect to see one of the four
        # following multiparts: metadata, code, source, suitcase.
        # These parts are further broken down.
        for part in body:
            ctype, msgobj, body = part.index()
            if ctype == 'multipart/knowbot-metadata':
                self._parse_metadata(body)
            elif ctype == 'multipart/knowbot-code':
                self._parse_code(body)
            elif ctype == 'multipart/knowbot-state':
                self._parse_state(body)
            elif ctype == 'multipart/knowbot-suitcase':
                self._parse_suitcase(body)
            else:
                raise KPMimeError, 'undefined subpart: %s' % ctype

    def _parse_metadata(self, body):
        for part in body:
            ctype, msgobj, body = part.index()
            if ctype <> 'message/rfc822':
                raise KPMimeError, 'Bad metadata content-type: %s' % ctype
            # Parse all headers found in this part.  Any that start
            # with KPMD- are metadata proper.  All others go in the
            # `extras' dictionary.
            mdmsg = rfc822.Message(body)
            for k in map(string.lower, mdmsg.keys()):
                if self._metadata.has_key(k):
                    raise KPMimeError, 'Duplicate metadata key: %s' % k
                else:
                    self._metadata[k] = mdmsg[k]

    def _parse_code(self, body):
        for part in body:
            ctype, msgobj, body = part.index()
            cname = msgobj.getheader('content-name')
            if not cname:
                raise KPMimeError, 'Unnamed code part.'
            if self._code.has_key(cname):
                raise KPMimeError, 'Duplicate code part name.'
            # get the data
            pos = body.tell()
            s = body.read()
            body.seek(pos)
            #
            # TBD: We currently have to undo the special encoding used
            # by the KPMimeWriter.KPMimeWriter class for
            # knowbot-code-objects.  This encoding is used so the
            # marshaled string is not binary data, unsuitable for CVS.
            #
            if ctype == 'application/knowbot-code-object':
                s = eval(s)
            #
            # END TBD
            #
            self._code[cname] = msgobj, s

    def _parse_state(self, body):
        for part in body:
            ctype, msgobj, body = part.index()
            cname = msgobj.getheader('content-name')
            if not cname:
                raise KPMimeError, 'Unnamed state part.'
            if self._state.has_key(cname):
                raise KPMimeError, 'Duplicate state part name.'
            # get the data
            pos = body.tell()
            s = body.read()
            body.seek(pos)
            self._state[cname] = msgobj, s

    def _parse_suitcase(self, body):
        for part in body:
            ctype, msgobj, body = part.index()
            cname = msgobj.getheader('content-name')
            if not cname:
                raise KPMimeError, 'Unnamed suitcase part.'
            if self._suitcase.has_key(cname):
                raise KPMimeError, 'Duplicate suitcase part name.'
            # get the data
            pos = body.tell()
            s = body.read()
            body.seek(pos)
            self._suitcase[cname] = msgobj, s


if __name__ == '__main__':
    import TestKPMimer
    TestKPMimer.unmime_test()
