"""Knowbot Program MIME Writer version 0.1

Writes a MIME document conformant to 'MIME Transport of Knowbot
Programs, version 0.1'.

Classes:

        KPMimeWriter
            High level interface to KP MIME format.

Exceptions:

        KPMimeError
            An error occurred in the parsing of KP MIME format for the
            given version (currently 0.1).

"""

from MimeWriter import MimeWriter
from types import *
import rfc822
import string


KPMimeError = 'KPMimeWriter.KPMimeError'


CLOSED=-2
OUT=-1
PKG=0
MD=1
CODE=2
STATE=3
SUITCASE=4
COMPLEXMD=5
SRC=6
STATE_SRC=7
SUITCASE_FILE=8

def cap(k, prefix=''):
    # crufty stuff we don't strictly need
    return prefix + string.joinfields(
        map(lambda s: string.upper(s[0]) + s[1:],
            string.splitfields(k, '-')),
        '-')


class KPMimeWriter:
    """Writes a KP to MIME format.

    Methods:

    __init__(fp)
    currentpart()
    start_kp()
        start_metadata()
            metadata(metadata={})
        end_metadata()
        start_code()
            code_object(modulename, repr)
            start_code_source(modulename, url='')
            end_code_source()
        end_code()
        start_state()
            state_pickle(modulename, repr)
            start_state_source(modulename, url='')
            end_state_source()
        end_state()
        start_suitcase()
            start_suitcase_file(filename, url='')
            end_suitcase_file()
        end_suitcase()
    end_kp()

    Note that both code objects and state objects can be included
    inline or by reference.  The REPR argument to code_object() and
    state_pickle() are different depending on how the objects are
    included.  If the object is to be included inline, REPR must be
    the marshalled or pickled object.  If the object is to be included
    by reference REPR must be a tuple of the form (URL LEN), where URL
    is location of the remote object, and LEN is its length in bytes.

    TBD: Inclusion by reference is not yet supported.
    XXX: suitcase should also support this if it is supported at all.

    """

    def __init__(self, fp):
        self._inpart = OUT
        self._part = [MimeWriter(fp)]

    def _assert(self, part):
        if self._inpart <> part:
            raise KPMimeError, 'out of order'

    def currentpart(self): return self._part[-1]

    def start_kp(self):
        self._assert(OUT)
        part = self._part[-1]
        fp = part.startmultipartbody('knowbot',
                                     plist=[('version', '0.1')],
                                     prefix=0)
        fp.write('This is a multi-part Knowbot Program in MIME format.\n')
        self._inpart = PKG

    def end_kp(self):
        self._assert(PKG)
        self._part[-1].lastpart()
        del self._part[-1]
        self._inpart = CLOSED

    def start_metadata(self):
        self._assert(PKG)
        part = self._part[-1].nextpart()
        self._part.append(part)
        fp = part.startmultipartbody('knowbot-metadata')
        self._inpart = MD

    def end_metadata(self):
        self._assert(MD)
        self._part[-1].lastpart()
        del self._part[-1]
        self._inpart = PKG

    def metadata(self, metadata={}):
        self._assert(MD)
        part = self._part[-1].nextpart()
        self._part.append(part)
        w = MimeWriter(part.startbody('message/rfc822'))
        for k in metadata.keys():
            w.addheader(cap(k), metadata[k])
        w.flushheaders()
        del self._part[-1]

    def start_code(self):
        self._assert(PKG)
        part = self._part[-1].nextpart()
        self._part.append(part)
        fp = part.startmultipartbody('knowbot-code')
        self._inpart = CODE

    def end_code(self):
        self._assert(CODE)
        self._part[-1].lastpart()
        del self._part[-1]
        self._inpart = PKG

    def start_code_source(self, modulename, url=''):
        self._assert(CODE)
        part = self._part[-1].nextpart()
        self._part.append(part)
#       if url:
            # included by reference
#           w = MimeWriter(part.startbody('message/external-body',
#                                         [('access-type', 'url'),
#                                          ('url', url)]))
        part.addheader('Content-Name', modulename)
        self._inpart = SRC

    def end_code_source(self):
        self._assert(SRC)
        del self._part[-1]
        self._inpart = CODE

    def code_object(self, modulename, repr):
        self._assert(CODE)
        part = self._part[-1].nextpart()
        part.addheader('Content-Name', modulename)
        #
        # TBD: write the repr() of the object repr is binary data and
        # CVS diff doesn't handle that too well.
        #
        repr = `repr`
        #
        # END TBD
        #
        objlen = len(repr)
        part.addheader('Content-Length', `objlen`)
        fp = part.startbody('application/knowbot-code-object')
        fp.write(repr)

    def start_state(self):
        self._assert(PKG)
        part = self._part[-1].nextpart()
        self._part.append(part)
        fp = part.startmultipartbody('knowbot-state')
        self._inpart = STATE

    def end_state(self):
        self._assert(STATE)
        self._part[-1].lastpart()
        del self._part[-1]
        self._inpart = PKG

    def start_state_source(self, modulename, url=''):
        self._assert(STATE)
        part = self._part[-1].nextpart()
        self._part.append(part)
        part.addheader('Content-Name', modulename)
        self._inpart = STATE_SRC

    def end_state_source(self):
        self._assert(STATE_SRC)
        del self._part[-1]
        self._inpart = STATE

    def state_pickle(self, modulename, repr):
        self._assert(STATE)
        part = self._part[-1].nextpart()
        part.addheader('Content-Name', modulename)
        objlen = len(repr)
        part.addheader('Content-Length', `objlen`)
        fp = part.startbody('application/knowbot-state-object')
        fp.write(repr)

    def start_suitcase(self):
        self._assert(PKG)
        part = self._part[-1].nextpart()
        self._part.append(part)
        fp = part.startmultipartbody('knowbot-suitcase')
        self._inpart = SUITCASE

    def end_suitcase(self):
        self._assert(SUITCASE)
        self._part[-1].lastpart()
        del self._part[-1]
        self._inpart = PKG

    def start_suitcase_file(self, filename, url=''):
        self._assert(SUITCASE)
        part = self._part[-1].nextpart()
        self._part.append(part)
        part.addheader('Content-Name', filename)
        self._inpart = SUITCASE_FILE

    def end_suitcase_file(self):
        self._assert(SUITCASE_FILE)
        del self._part[-1]
        self._inpart = SUITCASE


if __name__ == '__main__':
    import TestKPMimer
    TestKPMimer.mime_test()
