"""KP representation as shared by kernels and supervisors.
"""

import string
from StringIO import StringIO
import MappingImpl
from koe.interfaces.stubs import KPRepr
from koe.interfaces.stubs import KPRepr__skel


class NamedPart(KPRepr__skel.NamedPart):

    def __init__(self):
        self.parts = {}

    def GetNames(self):
        return self.parts.keys()

    def GetPart(self, name):
        try:
            return self.parts[name]
        except KeyError:
            raise KPRepr.PartNameError, name

    def DeletePart(self, name):
        if self.parts.has_key(name):
            del self.parts[name]
        else:
            raise KPRepr.PartNameError, name

    def SetPart(self, name, header, data):
        # clobber already existing data, if any
        self.parts[name] = header, data

    def GetInfo(self, name):
        try:
            return (self.parts[name])[0]
        except KeyError:
            raise KPRepr.PartNameError, name

    def GetData(self, name):
        try:
            return (self.parts[name])[1]
        except KeyError:
            raise KPRepr.PartNameError, name

    def SetInfo(self, name, info):
        try:
            hdrs, body = self.parts[name]
            self.parts[name] = info, body
        except KeyError:
            raise KPRepr.PartNameError, name
            
    def SetData(self, name, buffer):
        try:
            hdrs, body = self.parts[name]
            self.parts[name] = hdrs, buffer
        except KeyError:
            raise KPRepr.PartNameError, name

    # true object only methods:
    
    def __getstate__(self):
        def rep((hdr, body)):
            return hdr.name, hdr.type, body
        return map(rep, self.parts.values())

    def __setstate__(self, guts):
        for name, type, body in guts:
            self.parts[name] = KPRepr.ContentHeader(name, type), body


class Representation(KPRepr__skel.Representation):
    def __init__(self, fp=None):
        # initialize
        self._metadata = MappingImpl.StringMap()
        self._code = NamedPart()
        self._state = NamedPart()
        self._suitcase = NamedPart()
        self._parser = None
        # parse
        if fp:
            self.loadfromfile(fp)

    # true object only interface:
    def loadfromfile(self, fp):
        try:
            import KPMimeParser
            self._parser = p = KPMimeParser.KPMimeParser(fp)
        except KPMimeParser.KPMimeError, errmsg:
            raise KPRepr.KPMimeError, errmsg

        # metadata
        self._metadata = MappingImpl.StringMap(p.metadata())

        def fill_part(part, src_dict):
            for cname, (msgobj, data) in src_dict.items():
                cprops = {}
                for pname in map(string.lower, msgobj.getparamnames()):
                    cprops[pname] = msgobj.getparam(pname)
                part.SetPart(cname, 
                             KPRepr.ContentHeader(cname, msgobj.gettype()),
                             data)

        fill_part(self._code, p.code())
        fill_part(self._state, p.state())
        fill_part(self._suitcase, p.suitcase())

    def GetMetadata(self):
        return self._metadata

    def GetCode(self):
        return self._code

    def GetState(self):
        return self._state

    def GetSuitcase(self):
        return self._suitcase

    def Marshal(self):
        import marshal
        return marshal.dumps(self.__getstate__())

    def Pickle(self):
        import pickle
        return pickle.dumps(self)

    def FreezeDry(self):
        strfp = StringIO()
        write(strfp, self)
        return strfp.getvalue()

    def get_traceback(self):
        try:
            return self._state.GetData('traceback')
        except KPRepr.PartNameError:
            return None

    # true object only interfaces

    def parser(self): return self._parser

    def loadfromrepr(self, repr):
        self.loadfrommarshal(repr.Marshal())

    def loadfrommarshal(self, s):
        import marshal
        self.__setstate__(marshal.loads(s))

    def __getstate__(self):
        return (self._metadata.get_dict(),
                self._code.__getstate__(),
                self._state.__getstate__(),
                self._suitcase.__getstate__())

    def __setstate__(self, (metadata, code, state, suitcase)):
        self._metadata = MappingImpl.StringMap(metadata)
        self._code.__setstate__(code)
        self._state.__setstate__(state)
        self._suitcase.__setstate__(suitcase)



def parse(fp):
    """Return Representation based on input from file.
    """
    return Representation(fp)


# XXX This should be a true-object-only method, like loadfromfile!!!
def write(fp, kprepr):
    """Write external MIME representation based on Representation object.
    """
    from KPMimeWriter import KPMimeWriter
    w = KPMimeWriter(fp)
    w.start_kp()
    #
    # metadata
    #
    w.start_metadata()
    mdmap = kprepr.GetMetadata()
    md = {}
    import sys
    for k, v in mdmap.GetItems():
        md[k] = v
    w.metadata(md)
    w.end_metadata()
    #
    # code
    #
    w.start_code()
    part = kprepr.GetCode()
    for name in part.GetNames():
        info = part.GetInfo(name)
        data = part.GetData(name)
        ctype = info['type']
        if ctype == 'application/knowbot-code-object':
            w.code_object(name, data)
        else:
            w.start_code_source(name)
            bodyfp = w.currentpart().startbody(ctype)
            bodyfp.write(data)
            w.end_code_source()
    w.end_code()
    #
    # state
    #
    w.start_state()
    part = kprepr.GetState()
    for name in part.GetNames():
        info = part.GetInfo(name)
        data = part.GetData(name)
        ctype = info['type']
        if ctype == 'application/knowbot-state-object':
            w.state_pickle(name, data)
        else:
            w.start_state_source(name)
            bodyfp = w.currentpart().startbody(ctype)
            bodyfp.write(data)
            w.end_state_source()
    w.end_state()
    #
    # suitcase
    #
    w.start_suitcase()
    part = kprepr.GetSuitcase()
    for name in part.GetNames():
        info = part.GetInfo(name)
        data = part.GetData(name)
        ctype = info['type']
        w.start_suitcase_file(name)
        bodyfp = w.currentpart().startbody(ctype)
        bodyfp.write(data)
        w.end_suitcase_file()
    w.end_suitcase()
    # done!
    w.end_kp()
                
