"""Implementation of a basic Namespace server.

See interfaces/Namespace.isl for details.

Classes defined here include:

Context -- your basic Context object.

PredefinedContext -- backwards compatibility hack.

PersistentContext -- a Context object that saves and restores its
    state in a persistent mapping object (such as a shelve or a dbm
    file).

DatabaseWrapper -- a wrapper around a dbm-like object that maps keys
    using a prefix.

"""

import sys
import string
from koe.interfaces.stubs import Namespace, Namespace__skel


class Context(Namespace__skel.Context):
    """Your basic Context object."""

    _SEPARATOR = '/'
    _FQ_CLASSNAME = 'Namespace.Context'

    def __init__(self):
        """Constructor.  Initializes an empty context."""
        self._bindings = {}

    def _make_descriptor(self, typename='', reference=''):
        """Convenience routine for ensuring type correctness."""
        if not typename:
            typename = self._FQ_CLASSNAME
        if not reference:
            reference = self.IluSBH()
        return {'typename': typename, 'reference': reference}

    #
    # ISL Interface
    #

    def Lookup1(self, name):
        """Look up a single name.

        The name should not contain any pathname separators, nor be
        empty.

        """

        if not name:
            raise Namespace.ResolutionError, "Can't resolve empty path"
        if self._SEPARATOR in name:
            raise Namespace.ResolutionError, \
                  "Can't resolve multi-name path %s" % `name`
        if self._bindings.has_key(name):
            return self._bindings[name]
        raise Namespace.ResolutionError, 'No binding for %s' % `name`

    def List(self):
        """Return an unsorted list of names of all current bindings."""
        return self._bindings.keys()

    def Bind(self, name, desc):
        """Bind a new name to a descriptor.

        The name should not contain any pathname separators, nor be
        empty, and should not currently be bound.

        """
        
        self.__bind(name, desc, 0)

    def Rebind(self, name, desc):
        """Replace or create a name binding.

        The name should not contain any pathname separators, nor be
        empty; it may or may not be bound currently.

        """
        self.__bind(name, desc, 1)

    def __bind(self, name, desc, rebind_ok):
        """Common code for Bind and Rebind."""
        
        # XXX: These cases used to raise ResolutionError but that was
        # not described in the ISL file.  In any event, all these
        # error conditions should be separated out, so really neither
        # is correct.  To be done in a future release.
        if not name:
            raise Namespace.BindError, "Can't bind empty path"
        if self._SEPARATOR in name:
            raise Namespace.BindError, \
                  "Can't bind multi-name path %s" % `name`

        if not rebind_ok:
            if self._bindings.has_key(name):
                raise Namespace.BindError, 'Name already bound: %s' % `name`

        # Sanity check on descriptor
        try:
            typename = desc['typename']
            reference = desc['reference']
        except KeyError:
            raise Namespace.BindError, 'Bad Descriptor: %s' % `desc`
        newdesc = self._make_descriptor(typename, reference)
        self._bindings[name] = newdesc

    def Unbind(self, name):
        """Remove a name binding.

        The name should not contain any pathname separators, nor be
        empty; it should currently be bound.

        """
        
        # XXX: These cases used to raise ResolutionError but that was
        # not described in the ISL file.  In any event, all these
        # error conditions should be separated out, so really neither
        # is correct.  To be done in a future release.
        if not name:
            raise Namespace.BindError, "Can't unbind empty path"
        if self._SEPARATOR in name:
            raise Namespace.BindError, \
                  "Can't unbind multi-name path %s" % `name`
        if not self._bindings.has_key(name):
            raise Namespace.BindError, 'Name not bound: %s' % `name`
        del self._bindings[name]

    def GetDescriptor(self):
        """Get the descriptor for this context."""
        return self._make_descriptor()


class PredefinedContext(Context):

    """Backwards compatibility hack."""

    def __init__(self, dict):
        Context.__init__(self)
        for key, value in dict.items():
            self.Bind(key, value)


class PersistentContext(Context):

    """Context object that implements persistency.

    You must pass it a persistent dictionary (e.g. a shelf).

    """

    def __init__(self, db):
        self._db = db
        Context.__init__(self)
        for key in self._db.keys():
            value = self._db[key]
            self.Bind(key, value)

    # Internal versions of Bind/Unbind that don't have persistent side effects

    _bind = Context.Bind
    _rebind = Context.Rebind
    _unbind = Context.Unbind

    # Override Bind/Unbind to record the effect in the persistent database

    def Bind(self, name, desc):
        self._bind(name, desc)
        self._db[name] = desc

    def Rebind(self, name, desc):
        self._rebind(name, desc)
        self._db[name] = desc

    def Unbind(self, name):
        self._unbind(name)
        del self._db[name]


class DatabaseWrapper:

    """Wrapper around a dbm-like object that maps keys using a prefix.

    E.g. if prefix == "foo/", then a key of "bar" passed to this
    module translates to a key of "foo/bar" passed to the underlying
    database.  The keys() method etc. are suitably modified as well.

    """

    def __init__(self, db, prefix):
        if prefix and prefix[-1] != '/':
            raise ValueError, "prefix must end in '/'"
        self._db = db
        self._prefix = prefix

    def __getitem__(self, key):
        return self._db[self._prefix + key]

    def __setitem__(self, key, value):
        self._db[self._prefix + key] = value

    def __delitem__(self, key):
        del self._db[self._prefix + key]

    def __len__(self, key):
        return len(self.keys())

    def has_key(self, key):
        return self._db.has_key(self._prefix + key)

    def keys(self):
        rawkeys = self._db.keys()
        prefix = self._prefix
        n = len(prefix)
        keys = []
        for key in rawkeys:
            if key[:n] == prefix:
                keys.append(key[n:])
        return keys

    def items(self):
        rawitems = self._db.items()
        prefix = self._prefix
        n = len(prefix)
        items = []
        for (key, value) in rawitems:
            if key[:n] == prefix:
                items.append((key[n:], value))
        return items

    def values(self):
        rawitems = self._db.items()
        prefix = self._prefix
        n = len(prefix)
        values = []
        for (key, value) in rawitems:
            if key[:n] == prefix:
                values.append(value)
        return values

    def close(self):
        self._db.close()
