"""Base KOS.ReportingStation implementations.
"""
import os
from types import *
import ilu
from koe.interfaces.stubs import ReportingAPI, ReportingAPI__skel



class NullRS(ReportingAPI__skel.ReportingStation):
    """Minimal reporting station which can be served if necessary.

    ``Real'' reporting stations should use this as a base class rather
    than inheriting directly from the ILU base class.  This allows
    subclasses to define only those methods in which they are
    interested.

    """
    def __init__(self):
        self.__next = {}
        self.__pid = os.getpid()

    #
    # API
    #

    def AllocateID(self, prefix):
        """Generate a reporting-id unique to this reporting station."""
        if not self.__next.has_key(prefix):
            prefix = "%05d-%s" % (self.__pid, prefix)
            if not self.__next.has_key(prefix):
                self.__next[prefix] = 0
        self.__next[prefix] = next = self.__next[prefix] + 1
        id = "%s.%d" % (prefix, next)
        self.__next[id] = 0
        return id

    def KPTerminated(self, id, reason, details, kprepr):
        pass
        
    def KPMigrationReceived(self, id, kpidTo, migratingTo):
        pass


class DebuggingRS(NullRS):
    """A reporting station that prints exception stack traces.

    A derived class can override print_exception() to customize what
    gets printed.
    """

    #
    # API
    #

    def KPTerminated(self, id, reason, details, kprepr):
        if reason:
            from koe.interfaces.stubs import KPRepr
            state = kprepr.GetState()
            try:
                trace = state.GetData("traceback")
            except KPRepr.PartNameError:
                trace = None
            self.print_exception(id, reason, details, trace)
        else:
            NullRS.KPTerminated(self, id, reason, details, kprepr)

    def print_exception(self, id, reason, details, trace):
        print "---------- Exception in", `id`, "----------"
        if trace:
            print "Traceback (innermost last):"
            print trace,
            if trace[-1] != '\n': print
        print "%s: %s" % (reason, details)
        print "-------------------------------------------"



class MulticastRS(NullRS):
    """Multi-way reporting station dispatcher.
    """
    def __init__(self, sbhOrRSlist):
        """Create a reporting-station exploder object.

        sbhlist
            Sequence of binding handles or reporting-station objects to be
            included on the initial list.
        """
        NullRS.__init__(self)
        self.__errors = []
        self.__stations = {}
        map(self.add_station, sbhOrRSlist)

    def add_station(self, sbhOrRS, del_cb=None):
        """Add a reporting station to the list."""
        if type(sbhOrRS) == StringType:
            try:
                sbh = sbhOrRS
                station = ilu.ObjectOfSBH(ReportingAPI.ReportingStation, sbh)
            except (ilu.IluGeneralError, ilu.IluProtocolError):
                t, v = sys.exc_type, sys.exc_value
                self.__errors.append((sbh, t, v))
                return
        else:
            station = sbhOrRS
            sbh = station.IluSBH()
        self.__stations[sbh] = (station, del_cb)

    def remove_station(self, sbh):
        """Remove a station from the set of stations we contact.

        If a callback was registered with the station, it will be called.
        """
        if self.__stations.has_key(sbh):
            station, del_cb = self.__stations[sbh]
            del self.__stations[sbh]
            if callable(del_cb):
                del_cb()

    def get_station_count(self):
        return len(self.__stations)

    def get_error_count(self):
        return len(self.__errors)

    def get_error(self, index):
        """Return information about a particular error, by index."""
        return self.__errors[index]

    # API for derived classes

    def _apply(self, method, args):
        results = []
        for sbh, (rs, del_cb) in self.__stations.items():
            try:
                v = apply(getattr(rs, method), args)
            except (ilu.IluGeneralError, ilu.IluProtocolError):
                self.remove_station(sbh)
            else:
                results.append(v)
        results = filter(None, results)
        if results:
            return results[0]
        return None

    #
    # ISL interface
    #

    def KPTerminated(self, *args):
        self._apply("KPTerminated", args)

    def KPMigrationReceived(self, *args):
        self._apply("KPMigrationReceived", args)
