"""A segment of a file seen as a file-like object.

Classes:

SubFile - the only thing here.

"""


class SubFile:

    """A segment of a file seen as a file-like object, open for reading.

    Methods:

    __init__()
    __repr__()
    isatty()
    tell()
    seek(pos[, whence])
    readline()
    read([nbytes])
    readlines()

    Note the absence of a close() method.

    To use this, pass the constructor a file(-like) object, and a
    start and end position relative to that file.  The sequence of
    bytes in that file indicated by those positions can be accessed
    through the SubFile, mapping position 0 of the SubFile to the
    start position in the underlying file.  The underlying file must
    be seekable.

    The initial position of the SubFile is 0.

    The default end is the end of the underlying file.

    The default start is the current position of the underlying file.

    SubFiles may be used recursively (i.e., the underlying object may
    be a SubFile).  One day, this may even be optimized.

    """

    def __init__(self, fp, start=None, end=None):
        pos = fp.tell()
        # Test seekability of the underlying file
        fp.seek(pos)
        if start is None:
            start = pos
        if end is None:
            end = -1
        self._fp = fp
        self._start = start
        self._end = end
        self._pos = 0

    def __repr__(self):
        return "SubFile(%s, %s, %s)" % (self._fp, self._start, self._end)

    def isatty(self):
        return 0

    def tell(self):
        return self._pos

    def seek(self, pos, whence=0):
        if whence == 1:
            pos = pos + self._pos
        elif whence == 2:
            if self._end < 0:
                raise IOError, "EOF relative seek not supported"
            pos = pos + (self._end - self._start)
        self._pos = max(0, pos)

    def readline(self):
        self._fp.seek(self._start + self._pos)
        data = self._fp.readline()
        return self._fixupread(data)

    def read(self, n=-1):
        self._fp.seek(self._start + self._pos)
        if n < 0:
            if self._end >= 0:
                n = max(0, self._end - self._start - self._pos)
            data = self._fp.read(n)
        else:
            data = self._fp.read(n)
        return self._fixupread(data)

    def readlines(self):
        list = []
        while 1:
            line = self.readline()
            if not line:
                break
            list.append(line)
        return list

    def _fixupread(self, data):
        self._pos = self._pos + len(data)
        if self._end >= 0:
            excess = self._start + self._pos - self._end
            if excess > 0:
                data = data[:-excess]
                self._pos = self._pos - excess
        return data


def _test():
    f = open("/etc/passwd")
    f.readline()
    f.readline()
    start = f.tell()
    for i in range(10): f.readline()
    end = f.tell()-1
    g = SubFile(f, start, end)
    g.seek(0)
    for line in g.readlines():
        print `line`
    print "-"*10
    g.seek(0)
    while 1:
        print g.tell(),
        line = g.readline()
        print `line`
        if not line: break
    print "-"*10
    g.seek(0)
    print g.read()
    print g.tell()
    print "-"*10
    g.seek(0)
    print g.read(100000)
    print g.tell()
    g.seek(100000)
    print g.tell()
    print g.read()
    print g.tell()


if __name__ == '__main__':
    _test()
