"""A POSIX-style interface to the suitcase.

Includes three classes:
- SuitcaseFilesystem
- SuitcaseFilesystemPath
- SuitcaseFile

The interface implemented by each class are somewhat incomplete. The
current practice is to add new operations as they are needed.
"""

from koe.interfaces.stubs import KPRepr
import errno
import KPReprImpl
import string
import posixpath
import StringIO
import regex
from types import ListType, StringType
try:
    import gzip
except ImportError:
    gzip = None
    Nogzip = "No module found for gzip"

error = IOError

MimeFile = 'application/suitcase-file'
MimeDir = 'application/suitcase-directory'

def FileHeader(name, type=MimeFile):
    return KPRepr.ContentHeader(name, type)

def DirHeader(name, type=MimeDir):
    return KPRepr.ContentHeader(name, type)

class SuitcaseFilesystem:

    """The SuitcaseFilesystem is a POSIX-like interface to the suitcase.

    This object should allow a KP to treat its suitcase just like a
    standard file system, that it can read and write to, do listdir, walk,
    etc.

    The methods of this object all read to and write from the suitcase
    representation where each file and directory is a separate
    entry in the NamedPart. Each directory contains the full path
    names of all the files in it. Each file contains its own contents.
    """

    def __init__(self, repr):
        self.repr = repr
        self.path = SuitcaseFilesystemPath(self)
        self.cwd = '/'
        self.newfiles = []
        # some metadata caches
        self.type_cache = [('', '')] * 13
        self.dir_cache = {}
        self.new_dirs = []
        if gzip:
            gzip._py_open = self.open

    def __getinitargs__(self):
        return (self.repr,)

    def _flush(self):
        # writeback any changes in dir_cache
        # assume that this doesn't get too big during the lifetime of
        #    your average kp
        for name in self.new_dirs:
            self.repr.SetPart(name, DirHeader(name), 
                              string.joinfields(self.dir_cache[name], '\012'))
            del self.dir_cache[name]
        for name, data in self.dir_cache.items():
            self.repr.SetData(name, string.joinfields(data, '\012'))
        self.new_dirs = []

    # externally visible interface

    def mkdir(self, name):
        name = self._normalize_path(name)
        parent, new = self.path.split(name)

        type = self._repr_type(parent)
        if type != MimeDir:
            if type:
                raise error, (errno.EEXIST, 'File exists')
            else:
                raise error, (errno.ENOENT, 'No such file or directory')
        self._repr_mkdir(name)
        if name <> '/':
            self.__dir_addfile(parent, name)

    def open(self, name, mode = 'r'):
        name = self._normalize_path(name)
        if mode[0] == 'r':
            type = self._repr_type(name)
            if type != MimeFile:
                if type:
                    raise error, (errno.EISDIR, 'Is a directory')
                else:
                    raise error, (errno.ENOENT, 'No such file or directory')
            f = SuitcaseFile(self, name)
            return f
        elif mode[0] in 'aw':
            # expected common case: opening new file
            parent, filename = self.path.split(name)
            if self._repr_isdir(parent):
                if name in self.__dir_contents(parent):
                    type = self._repr_type(name)
                    if type == MimeDir:
                        raise error, (errno.EISDIR, 'Is a directory')
                else:
                    self._repr_mkfile(name)
                    self.__dir_addfile(parent, name)
            else:
                type = self._repr_type(parent)
                if type:
                    raise error, (errno.ENOTDIR, 'Not a directory')
                else:
                    raise error, (errno.ENOENT, 'No such file or directory')
            if mode[0] == 'w':
                f = SuitcaseFile(self, name, 'w')
            else:
                f = AppendingSuitcaseFile(self, name, mode)
            return f
        else:
            raise error, (0, 'unknown mode')

    def open_gz(self, filename, mode="r", compressionlevel=9):
        if gzip:
            return gzip.GzipFile(filename, mode, compressionlevel,
                                 self.open(filename))  
        else:
            raise Nogzip

    def chdir(self, name):
        name = self._normalize_path(name)
        type = self._repr_type(name)
        if type != MimeDir:
            raise error, (errno.ENOENT, 'No such file or directory')
        self.cwd = name

    def getcwd(self):
        return self.cwd

    def listdir(self, name=None, rx=regex.compile('/\([^/]+\)$')):
        if not name:
            name = self.cwd
        name = self._normalize_path(name)
        if not self._repr_isdir(name):
            raise error, (errno.ENOTDIR, 'Not a directory')
        results = []
        for file in self.__dir_contents(name):
            if file:
                rx.search(file)         # what if this fails
                results.append(rx.group(1))
        return results
        
    def unlink(self, path):
        name = self._normalize_path(path)
        if self._repr_isdir(name):
            # not sure if this is the correct error to raise
            raise error, (errno.EPERM, 'Not owner')

        parent, filename = self.path.split(name)
        self.__dir_delfile(parent, name)
        self._repr_delete(name)

    remove = unlink

    def rmdir(self, path):
        name = self._normalize_path(path)
        type = self._repr_type(name)
        if type != MimeDir:
            if type:
                raise error, (errno.ENOTDIR, 'Not a directory')
            else:
                raise error, (errno.ENOENT, 'No such file or directory')
        if self.__dir_contents(name):
            raise error, (errno.EEXIST, 'File exists')
        parent, dir = self.path.split(name)
        self.__dir_delfile(parent, name)
        self._repr_delete(name)

    # helper routines

    sep = '/'

    def _normalize_path(self, path, 
                        neednorm = regex.compile('\(/\\|^\).\.?\(/\|$\)')):
        # remove trailing slashes
        if path[-1:] == '/' and not path == '/':
            path = path[:-1]
        # if it is absolute, we're done
        #if not self.path.isabs(path):
        if not path[0] == '/':
            if self.cwd == '/':
                path = '/' + path
            else:
                path = self.cwd + '/' + path

        # check for '.' or '..'
        if neednorm.search(path) != -1:
            path = self.path.normpath(path)

        return path

    # helper routines for managing directories

    def __dir_contents(self, name):
        try:
            return self.dir_cache[name]
        except KeyError:
            try:
                buf = self.repr.GetData(name)
            except KPRepr.PartNameError:
                buf = ''
        if buf:
            self.dir_cache[name] = string.splitfields(buf, '\012')
        else:
            self.dir_cache[name] = []
        return self.dir_cache[name]

    def __dir_addfile(self, dir, name):
        try:
            self.dir_cache[dir].append(name)
        except KeyError:
            self.__dir_contents(dir)
            self.dir_cache[dir].append(name)

    def __dir_delfile(self, dir, name):
        try:
            self.dir_cache[dir].remove(name)
        except KeyError:
            self.__dir_contents(dir)
            self.dir_cache[dir].remove(name)

    # helper routines for talking KP repr object

    def _repr_mkdir(self, name):
        self.dir_cache[name] = []
        self.new_dirs.append(name)

    def _repr_mkfile(self, name):
        self.newfiles.append(name)
        self.type_cache[hash(name) % 13] = (name, MimeFile)

    def _repr_isdir(self, name):
        return self._repr_type(name) == MimeDir

    def _repr_isfile(self, name):
        return self._repr_type(name) == MimeFile

    def _repr_type(self, name):
        if self.dir_cache.has_key(name):
            return MimeDir
        hidx = hash(name) % 13
        hname, hval = self.type_cache[hidx]
        if hname == name:
            return hval
        try:
            hdr = self.repr.GetInfo(name)
            self.type_cache[hidx] = (name, hdr.type)
            return hdr.type
        except KPRepr.PartNameError:
            if name in self.newfiles:
                return MimeFile
        if name == '/':
            # we're lazy about creating this
            self._repr_mkdir('/')
            return MimeDir
        return ''

    def _repr_delete(self, name):
        try:
            self.repr.DeletePart(name)
            hkey = hash(name) % 13
            if self.type_cache[hkey] == name:
                self.type_cache[hkey] = ('', '')
            if self.dir_cache.has_key(name):
                del self.dir_cache[name]
        except KPRepr.PartNameError:
            pass

class SuitcaseFile(StringIO.StringIO):

    """A file that read/writes its content from/to the suitcase."""
    # Just a few changes to StringIO.

    def __init__(self, suitcase, name, mode='r'):
        if mode == 'r':
            StringIO.StringIO.__init__(self, suitcase.repr.GetData(name))
        elif mode == 'a':
            try:
                StringIO.StringIO.__init__(self, suitcase.repr.GetData(name))
            except KPRepr.PartNameError:
                StringIO.StringIO.__init__(self)
        else:
            StringIO.StringIO.__init__(self)
        self.suitcase = suitcase
        self.name = name
        self.flushed = 0

    def close(self):
        self.flush()
        StringIO.StringIO.close(self)

    def flush(self):
        self.suitcase.repr.SetPart(self.name, FileHeader(self.name), 
                                   self.getvalue())
        if not self.flushed:
            try:
                self.suitcase.newfiles.remove(self.name)
            except ValueError:
                pass
            self.flushed = 1


class AppendingSuitcaseFile(SuitcaseFile):
    def __init__(self, suitcase, name, mode='a'):
        SuitcaseFile.__init__(self, suitcase, name, 'a')
        self.__readable = '+' in mode
        if mode == "a":
            # seek to end-of-file
            self.seek(0, 2)

    def write(self, data):
        if data:
            self.buflist.append(data)
            self.seek(self.tell() + len(data))

    def read(self, length=-1):
        if self.__readable:
            return SuitcaseFile.read(self, length)
        raise IOError, (errno.EBADF, 'Bad file number')

    def readline(self, length=None):
        if self.__readable:
            return SuitcaseFile.readline(self, length)
        raise IOError, (errno.EBADF, 'Bad file number')

    # readlines() is implemented in terms of readline(),
    # so don't do anything extra for it.


class SuitcaseFilesystemPath:
    """Implements posixpath style functions."""

    def __init__(self, sfs):
        self.sfs = sfs

    def __getinitargs__(self):
        return (self.sfs,)

    # special case the root
    # this is clumsy, but the fix that requires fewest changes

    def exists(self, p):
        name = self.sfs._normalize_path(p)
        test = self.sfs._repr_type(name)
        if test:
            return 1
        if p == '/':
            return 1
        return 0

    def isdir(self, p):
        name = self.sfs._normalize_path(p)
        return self.sfs._repr_isdir(name)

    def isfile(self, path):
        name = self.sfs._normalize_path(path)
        t = self.sfs._repr_type(name)
        if t == MimeFile:
            return 1
        if name in self.sfs.newfiles:
            return 1
        return 0

    # use posixpath whenever possible
    # TBD: should code be copied in the interest of speed

    def normcase(self, s):
        return posixpath.normcase(s)

    def isabs(self, p):
        return posixpath.isabs(p)

    def split(self, p):
        return posixpath.split(p)

    def join(self, *paths):
        return apply(posixpath.join, paths)

    def splitext(self, p):
        return posixpath.splitext(p)

    def splitdrive(self, p):
        return posixpath.splitdrive(p)
    
    def basename(self, p):
        return posixpath.basename(p)

    def dirname(self, p):
        return posixpath.dirname(p)

    def commonprefix(self, m):
        return posixpath.commonprefix(m)

    def samefile(self, f1, f2):
        return self.normpath(f1) == self.normpath(f2)

    def sameopenfile(self, f1, f2):
        return f1 == f2

    def normpath(self, path):
        return posixpath.normpath(path)

    def walk(self, root, func, arg):
        queue = [root]
        
        while queue:
            dir = queue[0]
            del queue[0]
            if self.isdir(dir):
                files = self.sfs.listdir(dir)
                func(arg, dir, files)
                for file in files:
                    path = self.join(dir, file)
                    if self.isdir(path):
                        queue.append(path)
