#! /usr/bin/env python
#
# Implement an infrastructure suitable for parsing of NCSA httpd style
# configuration files
#
# Classes:
#
#    Abstract superclasses:
#       BaseConfParser - Base class for parsing all 'groups'
#       BaseKey - Base class for all 'container' objects
#
#    Container classes:
#       Limit - container class for all <Limit> groups
#
#    User classes:
#       LimitParser - Parser class for all <Limit> groups
#
#    Admin classes:
#       Test - encapsulation of all the routines needed for testing
#
# Examples:
#
#      # For use with the KP Receiver
#      try:
#          p = access.LimitParser()
#      except OpenError:
#          raise MyIOException
#      except ParseError:
#          raise MyBadFileException
# 
#      if p.is_allowed_from('submit', 'host_we_wish_to_check.unknown.domain'):
#           print 'OK'
#      else:
#           print 'Access Denied'
#
#      # For use with the HTTP plugin (specify an alternate access.conf file)
#      p = access.LimitParser(filename='plugins/http/access.conf')
#      if p.is_allowed_to('get', 'my_restriced_web_server.mydomain'):
#           print 'OK'
#      else:
#           print 'Access Denied'
#
# Author: Roger E. Masse
# rmasse@CNRI.Reston.Va.US

# XXX: should convert all regex to re
import regex
import re

import string
import os
import string
import sys
import socket

ParseError = "Parse Error"
OpenError = "Open Error"
ResourceError = "Resource Error"

VALIDTYPES = {'limit': [{'tag' : 'get','port' : '80'},
                        {'tag' : 'post', 'port' : '80'},
                        {'tag' : 'submit', 'port' : 'all'},
                        {'tag' : 'kpkill', 'port' : 'all'},
                        {'tag' : 'worldrootpeer', 'port' : '7438'},
                        {'tag' : 'worldrootshutdown', 'port' : 'all'},
                        {'tag' : 'visushutdown', 'port' : 'all'},
                        {'tag' : 'kosshutdown', 'port' : 'all'}]}

GROUPS = {}
BEGINTAGS = {}
ENDTAGS = {}
for name in VALIDTYPES.keys():
    prog = regex.compile('</' + name + '>')
    # end tags - just the key as the key
    ENDTAGS[name] = prog
    for tag_info in VALIDTYPES[name]:
        # a keyword delimiter is a space since '\s' is not supported :(
        prog = regex.compile('<' + name + ' ' + tag_info['tag'] + '>')
        # Begins - use the key and the tag as the 'longkey'
        BEGINTAGS[name + '/' + tag_info['tag']] = prog

class BaseConfParser:
    """Basic framework for parsing HTTPD style configuration files.          
    """
    
    def __init__(self, filename, parse_method, container_class, strong_types=0,
                 debug=0):
        """We expect a container_class to be instansiatable.  It will hold
           all the group information found.  

         i.e  <Limit GET>
              </Limit>       ...has a name of 'limit'
                                and a tag of 'get'.  (Limit
                                could also have a tag of post, submit, etc)

                                 
           Non zero 'strong_types'  will force the precompilation and strong
           type checking of tags from the global VALIDTYPES data.  Otherwise,
           all tags are parsed and saved.
           
           if 'strong_types' is non zero, all supported group name / tag-port
           combinations must be appear in the VALIDTYPES global or parsing
           errors will most likely occur.

           This method raises OpenError upon encountering an IOError
           opening the configuration file.  I calls readfile() which
           may raise ParseError upon encountering syntax errors.
        """
        
        self.debug = debug
        self.strong_types = strong_types
        self.filename = filename
        self.parse_method = parse_method
        self.container_class = container_class
        self.stack = []
        self.any_tag = regex.compile('<.*>')
        self.end_tag = regex.compile('</.*>')

        self.checkfile()

    def checkfile(self):
        """Check to see if self.filename has been modified since
           last time we parsed it (if we had at all) and call
           readfile if necessary.
        """

        stat = os.stat(self.filename)
        _size_mtime = (stat[6], stat[8])
        try:
            groups = GROUPS[self.filename]
            if groups['_size_mtime'] == _size_mtime:
                if self.debug:
                    print '\t\t**Using cached information for:\n\t\t  File: ',\
                          self.filename
                return
        except KeyError:
            pass

        if self.debug:
            print '\t\t**(Re)Parsing\n\t\t  File: ', self.filename
            
        GROUPS[self.filename] = {'_size_mtime' : _size_mtime}    
        self.readfile()

    def readfile(self):
        """ parse through the file parsing groups and ordering them on a stack.

            This method raises ParseError upon encountering unexpected syntax.

            readfile has the side effect of closing self.f
        """
        try:
            self.f = open(self.filename, 'r')
        except IOError:
            raise OpenError, "\n%s\n%s" % (self.filename, sys.exc_value[1])
        self.lineno = 0
        while 1:
            line = self.f.readline()
            self.lineno = self.lineno + 1
            if not line:
                self.f.close()
                return
            # Tag line
            if self.any_tag.match(line) <> -1:

                # found a group tag - insure case insensitive comparisons
                line = string.lower(line)
                if self.strong_types:
                    if not self.check_strong_begin(line):
                        self.check_strong_end(line)
                else:
                    if self.end_tag.match(line) <> -1:
                        name = string.split(string.split(line, '<')[1], '>')[0]
                        self.group_end(string.strip(name)[1:])
                    else:
                        name_tag = string.split(string.split(line, '<')[1],
                                                '>')[0]
                        name_tag = string.strip(self.rm_extra_space(name_tag))
                        both = string.split(name_tag, ' ')
                        self.group_begin(both[0], both[1])
                    
            # parse the line saving results in the container at the top
            elif self.stack:
                self.parse_method(line, self.stack[0])
                
    def check_strong_begin(self, line):
        """push a new container object on the stack and return
           1 if a valid begin tag is found, else return 0"""
        
        for key in BEGINTAGS.keys():
            if BEGINTAGS[key].match(line) <> -1:
                both = string.split(key, '/')
                name = both[0]
                tag = both[1]
                try:
                    tags = VALIDTYPES[name]
                except KeyError:
                    raise ParseError, \
                    "Bad group: '%s'\nFile: %s\nLine: %d, Valid groups: %s" \
                    % (name, self.filename,
                       self.lineno, self.valid_start_tags())
                for tag_info in tags:
                    if tag_info['tag'] == tag:
                        self.group_begin(name, tag)
                        return 1
                raise ParseError, \
           "Bad tag: '%s', Group: '%s'\nFile: %s\nLine: %d, Valid groups: %s" \
                      % (tag, name, self.filename,
                         self.lineno, self.valid_start_tags())
        return 0

    def group_begin(self, name, tag):
        """Push a new container on the stack for name and tag.
           Assume all name/tag combinations are valid"""

        if not name or not tag:
            raise ParseError, "Valid groups must start with <name tag>"
        container = self.container_class(name, tag)
        self.push(container, self.stack)

    def check_strong_end(self, line):
        """Pop the container object off the stack, add it to groups"""
        
        for name in ENDTAGS.keys():
            if ENDTAGS[name].match(line) <> -1:
                self.group_end(name)

    def group_end(self, name):
        """Attempt to close a 'name' block by popping the stack"""

        try:
            if name <> self.stack[0].name:
                raise ParseError, "Block imbalance.\nFile: %s\nLine: %d" \
                      % (self.filename, self.lineno)
        except IndexError:
            if self.strong_types:
                error = \
                 "Group tags imbalance.\nFile:\n%s\nLine: %d, Valid tags: %s" \
                        % (self.filename, self.lineno, self.valid_start_tags())
            else:
                error = "Group tags imbalance. File:\n%s\nLine: %d" \
                        % (self.filename, self.lineno)
                        
            raise ParseError, error

        group = self.pop(self.stack)
        key = group.name + '/' + group.tag
        try:
            GROUPS[self.filename][key]
        except KeyError:
            GROUPS[self.filename][key] = group
        else:
            print 'Multiple <%s %s> groups in file:\n%s\nLine: %d, Ignoring.' \
                  % (group.name, group.tag, self.lineno, self.filename)
        

    def rm_extra_space(self, str, delimiter=' '):
        """Remove ALL spaces in a string, replacing a group of
           contigious spaces with a single delimiter.
        """

        parts = string.split(str, ' ')
        i = 0
        try:
            while 1:
                if not parts[i]:
                    del(parts[i])
                else:
                    i = i + 1
        except IndexError:
            return string.join(parts, delimiter)
  
    def valid_start_tags(self):
        """return a pretty printed string of all the valid start tags"""
        str = '\n\n'
        for name in VALIDTYPES.keys():
            for tag_info in self.VALIDTYPES[name]:
                str = str + '\t\t<' + name + ' ' + tag_info['tag'] + '>\n'
        return str

    def push(self, item, stack):
        """Insert item at the beginning of the list"""
        stack.insert(0, item)

    def pop(self, stack):
        """Get the first item off the list and return it. Delete item"""
        item = stack[0]
        del(stack[0])
        return item
                

class LimitParser(BaseConfParser):
    """Provide the implementation to fill a Generic Limit object.
    
    All keywords supported in this class fall inside
    a <Limit /something/> block and are case insensitve.  The following
    directives are currently implemented:

                order deny,allow
                order allow, deny
                deny from       all
                deny from .cnri.reston.va.us, all
                allow from host.mumble.edu
                allow to host.mumble.edu
                allow to        host.mumble.edu, 8001
                allow to all
                deny to badhost.evil.empire.com
                deny to all
                deny to badhost.evil.empire.com,81

    Future Note:
    
    Currently all 'allow from' and deny from' are parsed and saved.
    However, in the case of KPs, since they can do their own 'source
    routing', and the only semi-trustworthy origin information for
    a connecting host is via "ILU caller Identity" for the host that
    is *currently* connecting to us.  Since there is no guarentee that
    that host didn't allow a random connection, in order to more safely
    use the 'from' clauses, KP's will have to provide some form of
    authentication or origin which is not yet implemented.  For the
    moment, the alternative is to insure that *every* KOS in your
    Intranet KOE enforces the same set of 'allow_from' and 'deny_from'
    rules so that weak hosts cannot be used as a springboard.

    'deny to' and 'allow to' currently restrict or allow specific
    requests *to* those sights to be processed (in the case of the
    http plugin), as apposed to *from* hosts that are connecting to us
    (in the case of using these classes to validate KP submission).
    """
    def __init__(self, filename=None, debug=0, strong_types=0):

        if not filename:
            import pathhack
            filename = pathhack.kos_root() + '/config/access.conf'
            
        self.prog_order = regex.compile('order')
        self.prog_deny_from = regex.compile('deny from')
        self.prog_allow_from = regex.compile('allow from')
        self.prog_deny_to = regex.compile('deny to')
        self.prog_allow_to = regex.compile('allow to')

        self.valid_where = ('to', 'from')
        
        # last call the base classes initialzer
        BaseConfParser.__init__(self, filename, self.group_line, Limit,
                                debug=debug, strong_types=strong_types)
                
    def group_line(self, line, container):
        """handle a line of input for the limit group"""

        # Make <Limit> handling case and tab insensitive 
        line = string.strip(string.expandtabs(string.lower(line), 1))

        # nothing done for comments
        if '#' in line:
            return
        if self.debug:
            print '\t\t**In group: <%s %s> parsing line:\n\t\t%s' \
                  % (container.name, container.tag, line)
        if self.prog_order.match(line) <> -1:
            tag = string.strip(string.split(line, 'order ')[1])
            if container.order:
                raise ParseError, \
           "Multiply defined order clause. Group: <%s %s>\nFile: %s\nLine: %d"\
                      % (self.container.name, self.container.tag,
                         self.filename, self.lineno)
            tag = self.rm_extra_space(tag, delimiter='')
            valid_orders =  ('allow,deny', 'deny,allow')
            if tag not in valid_orders:
                raise ParseError, \
      "Bad order clause, Group: <%s %s>\nFile: %s\nLine: %d, Must be in: %s" \
                  % (self.container.name, self.container.tag,
                     self.filename, self.lineno, str(self.valid_where))
            container.order = tag
        elif self.prog_deny_from.match(line) <> -1:
            self.add_host('deny from ', line,
                          container.deny_from, container.tag)

        elif self.prog_allow_from.match(line) <> -1:
            self.add_host('allow from ', line,
                          container.allow_from, container.tag)
            
        elif self.prog_deny_to.match(line) <> -1:
            self.add_host('deny to ', line, container.deny_to, container.tag)

        elif self.prog_allow_to.match(line) <> -1:
            self.add_host('allow to ', line, container.allow_to, container.tag)

    def add_host(self, splitter, line, list, tag):

        host = string.split(line, splitter)[1]
        host_port = self.get_host_port(host, tag)
        if self.debug:
            print "\t\t**Adding: %s\n\t\t  to %s group's: %slist" \
                  % (str(host_port), tag, splitter)
        list.append(host_port)

    def get_host_port(self, host, tag):
        """ Return a tuple containing host and port parsed
            from a string containing host and (an optional)
            port separated by a comma.  If no port is specified,
            return the default port as the port
        """
        host_port = string.split(host, ',')
        if len(host_port) == 1:
            return string.strip(host), self.default_port(tag)
        if len(host_port) == 2:
            host = string.strip(host_port[0])
            port = string.strip(host_port[1])
            return host, port
        raise ParseError, "Bad host/port combination: '%s'" % host

    def in_list(self, group_list, list_name, host, port):
        """Return TRUE if host/port are in group_list"""

        if not group_list:
            if self.debug:
                print '\t\t**Empty list: ', list_name
            return 0
        if self.debug:
            print '\t\t**Looking for %s/%s in %s' % (host, port, list_name)
            print '\t\t**Group: %s' \
                  % string.join(string.split(str(group_list),
                                             '),'),'),\n\t\t\t ')

        for known_host, known_port in group_list:

            if known_host == 'all':
                if known_port == 'all' or known_port == port:
                    if self.debug:
                        print "\t\t**Found an 'all'"
                    return 1
            
            n = len(known_host)
            if string.lower(host[-n:]) == string.lower(known_host):
                if string.lower(known_port) == 'all' or known_port == port:
                    if self.debug:
                        print '\t\t**Found'
                    return 1
        if self.debug:
            print '\t\t**Not Found'
        return 0

    def default_port(self, tag):
        """Find the default port in the list of valid groups for <Limit /tag/>
        """
        
        for tag_info in VALIDTYPES['limit']:
            if tag_info['tag'] == tag:
                return tag_info['port']
        return 'all'

    def valid_tag(self, tag):
        """Return 1 if the tag is a valid tag"""
        for tag_info in VALIDTYPES['limit']:
            if tag_info['tag'] == tag:
                return 1
        return 0
            
    def is_allowed(self, tag, where, host, port=None):
        """Implement the list walking for host/port entries"""

        # Fill in the default port corresponding to tag if port is None
        if not port:
            port = self.default_port(tag)
            
        tag = string.lower(tag)
        if self.strong_types and not self.valid_tag(tag):
            print "'%s' is an invalid Limit tag." % tag
            return 0
        
        try:
            grp_name = 'limit/' + tag
            self.checkfile()
            grp = GROUPS[self.filename][grp_name]
        except KeyError:
            # If we do not find a group corresponding to tag, access is denied.
            if self.debug:
                print "\t\t** no group: %s in:\n%s" \
                      % (grp_name, str(GROUPS[self.filename]))
            return 0

        where = string.lower(where)
        if where not in self.valid_where:
            raise UsageError, "'where' must be in: %s" % str(self.valid_where)

        deny_name = 'deny_' + where
        allow_name = 'allow_' + where
        allow_list = getattr(grp, allow_name)
        deny_list = getattr(grp, deny_name)

        if self.debug:
            print '\t\t**<Limit %s> group, order is: %s' % (tag, grp.order)
        # Believe it or not, but according to the NCSA docs and the
        # Apache sources, "order allow,deny" means that the deny list
        # overrides the allow list, and vice versa.
        # The default also differs depending on the order directive:
        # for "order allow,deny" the default is to deny while for
        # the other order the default is to allow (!).
        if grp.order == 'deny,allow':
            if self.in_list(allow_list, allow_name, host, port):
                return 1
            if self.in_list(deny_list, deny_name, host, port):
                return 0
            default = 1
        else: # implied order is 'allow,deny' if not specified
            if self.in_list(deny_list, deny_name, host, port):
                return 0
            if self.in_list(allow_list, allow_name, host, port):
                return 1
            default = 0
        if self.debug:
            print '\t\t**%s/%s not found, defaulting to' % (host, port),
            print default and 'allow' or 'deny'
        return default
  
    def is_allowed_from(self, tag, host, port=None):
        """return true if we're allowing submissions from host
        """
        return self.is_allowed(tag, 'from', host, port=port)
    
    def is_allowed_to(self, tag, host, port=None):
        """return true if we should allowed to connect to host
        """
        return self.is_allowed(tag, 'to', host, port=port)
        
class BaseKey:
    """Base class for all containers"""
    def __init__(self, name, tag):
        self.name = name
        self.tag = tag

class Limit(BaseKey):
    """container class for the <Limit> groups"""
    def __init__(self, name, tag):

        self.order = None
        self.deny_from = []          
        self.allow_from = []    
        self.deny_to = []    
        self.allow_to = [] 
        
        # Last call the base class init
        BaseKey.__init__(self, name, tag)


def get_fqdn(addr, rx=regex.compile('\.')):
    # The only time we won't get a fully qualified name should be when
    # the name is in our own domain. We'd like to just append our
    # domain, but there isn't any easy way to get it (and be sure
    # we're not getting an NIS domain). Instead, we'll just look for
    # the first fully qualified name and use that.
    name, aliases, nums = socket.gethostbyaddr(addr)
    if rx.search(name) != -1:
        return name
    for name in aliases:
        if rx.search(name) != -1:
            return name
    return None
    

# Regular expression describing the string returned by ILU's generate
# ConnectionIdentity
iluIDcre = re.compile(
    r'\w+\s+'                                     # protocol, e.g. `sunrpcrm'
    r'over\s+'                                    # `over'
    r'\w+\s+'                                     # transport, e.g. `TCP'
    r'(?P<ip>(\d{1,3}\.){3}\d{1,3})\s+'           # IP address
    r'\d+'                                        # ???
    )


def ilu_callers_fqdn(cre=iluIDcre):
    """Utility to get and verify (to some degree) an ILU
    callers fully qualified domain name.

    XXX: This security really isn't that great because of IPv4 spoofability.

    This is meant to be used from an ILU true server.
    """
    
    import ilu
    caller = ilu.CallerIdentity()

    # Try to discover the IP address of the caller, which will get sent to
    # get_fqdn() to find the fully qualified domain name of the caller's
    # host.  XXX: Lots of spoofability here!  But let's assume for now that
    # this gives us authentic information
    #
    # We use the most generic mechanism for discovering the IP address of the
    # caller.  We get the ConnectionIdentity, and parse the resultant string,
    # assuming that that format doesn't change (ILU doesn't specify what the
    # format is, but at least for W3NG and SunRPC transport, the format is the 
    # same, as discovered experimentally).

    identity = caller.lookupIdentity('ConnectionIdentity')
    mo = None
    if type(identity) == type(''):
        mo = cre.match(identity)
    if not mo:
        # Couldn't authenticate
        return None
    address = mo.group('ip')
    fqdn = get_fqdn(address)
    # be paranoid
    if not fqdn or address <> socket.gethostbyname(fqdn):
        return None
    return fqdn

def ilu_port_from_sbh(sbh):
    """return the port number as a string from the sbh.
       Note: This may be non-portable between ILU versions
       and/or trasports :( ...this one's for sunrpc
    """
    port = string.split(sbh, '_')[-1]
    for char in port:
        if not char in string.digits:
            return None
    return port

def ilu_allow_access(tag, exception=ResourceError, where='from', fqdn=None):
    """Examine the access file and the credentials of the ILU caller
       to insure that we wish to allow a connection of this type.

       This utility is designed to be called from an ILU true server
       that wishes to identify the caller and determine if access
       is granted

       This utility will raise ResourceError upon encountering
       an error if exception is not specified.
    """
    if not fqdn:
        fqdn = ilu_callers_fqdn()
        if not fqdn:
            raise exception, "Can't get fully qualified domain name for caller"
##      if debug:
##          print 'Caller: ', fqdn
        
    # open and parse the configuration file
    try:
        perm = LimitParser()
    except OpenError:
        raise exception, sys.exc_value
    except ParseError:
        raise exception, sys.exc_value

    if perm.is_allowed(tag, where, fqdn, port=None):
        return 1
    return 0
            

class Test:

    def __init__(self, debug=0, strong_types=0):
        self.debug = debug
        self.strong_types = strong_types
        
    def test_host(self, host, testset, p):
        print '    Connections allowed from: %s/%s?' % (host, testset[1]),
        if self.debug:
            print
        if p.is_allowed_from(testset[0], host, port=testset[1]):
            print 'Yes'
        else:
            print 'No'
        print '    Connections allowed to: %s/%s?' % (host, testset[1]),
        if self.debug:
            print
        if p.is_allowed_to(testset[0], host, port=testset[1]):
            print 'Yes'
        else:
            print 'No'
        
    def test_group(self, p, testset):
        print '  Testing %s group' % testset[0]
        self.test_host('knowbot.cnri.reston.va.us', testset, p)
        self.test_host('badhost.evil.empire.com' , testset, p)
        self.test_host('host.mumble.edu', testset, p)
        self.test_host('host.mumble.edu', testset, p)
        self.test_host('badhost.evil.empire.com', testset, p)

    def run(self):
        import pathhack

        basepath = pathhack.kos_root()
        testfiles = ['plugins/http/access.conf', 'access.conf']
        testsets = [('get', None),
                    ('submit', None),
                    ('kpkill', None),
                    ('kosshutdown', None),
                    ('visushutdown', None),
                    ('worldrootpeer', None),
                    ('worldrootpeer', '7438'),
                    ('worldrootshutdown', None),
                   ]
        for filename in testfiles:
            fullfilename = os.path.join(basepath, filename)
            print '\nParsing: ', fullfilename
            p = LimitParser(fullfilename, debug=self.debug,
                            strong_types=self.strong_types)
            for testset in testsets:
                self.test_group(p, testset)

def main():

    import getopt

    opts, args = getopt.getopt(sys.argv[1:], 'ds')
    debug = strong_types = 0
    for o, a in opts:
        if o == '-d':
            debug = 1
            continue
        if o == '-s':
            strong_types = 1
            continue

    test = Test(debug=debug, strong_types=strong_types)
    test.run()
    

if __name__ == '__main__':
    main()
