"""ILU version abstraction utilities.

Functions in this module perform services for the caller which allow a
reduction in awareness of the details of ILU calls which have changed
between version of ILU.

When this module is imported, a new handler is installed for
exceptions not caught by implementation methods and not defined as a
valid ISL exception for the method.  The new handler prints a
traceback and the program identification for the program raising the
exception.  The default handler provided by ILU is saved as
`old_CaughtUnexpectedException'.  The new handler is installed as
`CaughtUnexpectedException' in modules `ilu' and `iluRt'.
"""

import ilu
import iluRt
import os
import sys
import traceback


if ilu.Version < '2.0':
    def LookupObject(objname, type):
        """Locate a published object by name.

        Version 1.8 of the ILU API is provided.
        """
        return ilu.LookupObject(objname, type)

    def RunMainLoop(handle=None):
        ilu.RunMainLoop()

    def ExitMainLoop(handle=None):
        ilu.ExitMainLoop()

else:
    def LookupObject(objname, type):
        """Locate a published object by name.

        Version 1.8 of the ILU API is provided.
        """
        import string
        [instHandle, serverID] = string.splitfields(objname, '@')
        return ilu.LookupObject(serverID, instHandle, type)

    def RunMainLoop(handle=None):
        global _loophandle
        if not handle:
            _loophandle = handle = ilu.CreateLoopHandle()
        ilu.RunMainLoop(handle)

    def ExitMainLoop(handle=None):
        global _loophandle
        if not handle:
            try:
                handle = _loophandle
            except NameError:
                raise RuntimeError, "ExitMainLoop called before RunMainLoop"
        ilu.ExitMainLoop(handle)


def RunInterruptableMainLoop():
    """Install a signal handler and a 1-second watchdog timer.

    The signal handler turns SIGINT and SIGTERM into calls to
    ExitMainLoop (but it won't interrupt incoming calls that are
    currently being handled).

    The watchdog timer does nothing but allows the signal handler to
    execute (ILU normally doesn't let the Python interpreter run when
    a signal arrives).

    This is intended for simple servers only -- there is no way to
    stop the watchdog or reset the signal handler.

    XXX Once ILU is fixed so the signal handler will run without help,
    we should remove the watchdog.

    """
    import signal
    signal.signal(signal.SIGINT, _SignalHandler)
    signal.signal(signal.SIGTERM, _SignalHandler)
    _StartWatchdog()
    RunMainLoop()

def _SignalHandler(*args):
    ExitMainLoop()

_watchdog_alarm = None

def _StartWatchdog():
    global _watchdog_alarm
    if not _watchdog_alarm:
        _watchdog_alarm = ilu.CreateAlarm()
    _Watchdog()

def _Watchdog():
    when = ilu.FineTime_Now() + ilu.FineTime(1)
    _watchdog_alarm.set(when, _Watchdog, ())


IluErrorTuple = (ilu.IluProtocolError, ilu.IluGeneralError)
