"""A toolkit for constructing and submitting KPs.

Implements a Receiver, and a KPConstructor that can be
used to instantiate KP Representations and submit them. 

Provides methods that simply standard procedures like assigning
metadata or filling the code part of the KP Repr.

Also allows for Franken-KP-stein like operations that instantiates the
pickled part of a terminated KP.   
"""

import ilu
import os
try:
    import cPickle
except ImportError:
    import pickle
    cPickle = pickle
import time
import traceback

import RS
import KPReprImpl
import nstools
import kptools
from koe.interfaces.stubs import KPRepr, KOS

NoMainMod = "No Main Module"

class Receiver:
    """Submits a KP to a KOS."""

    def __init__(self, name):
        self.__receiver = nstools.WorldOpen("kos/%s/receiver" % name)
        
    def submit(self, kp, reporter=RS.NullRS()):
        """Submits a KP and returns a unique ID for it."""
        # setup the metadata properly
        id = reporter.AllocateID(str(hash(kp))) # a cheap "unique" string
        rs = reporter.IluSBH()
        kp.set_metadata({
                          "reporting-station": rs,
                          "reporting-id": id
                          })
        # submit it
        self.__receiver.SubmitKP(kp.get_representation())
        return id

class KPConstructor:

    """Create a KP that can be given to a receiver.

    Given an instance of a Python object, this class can be used to
    turn it into a real KP (a real boy).
    """

    def __init__(self, instance=None, modules=[], metadata={}, suitcase=[]):
        self.__repr = kptools.Representation()
        self.__suitcase = None
        if metadata:
            self.set_metadata(metadata)
        if modules:
            self.set_code(modules)
        if instance:
            self.set_state(instance)
        if suitcase:
            self.set_suitcase(suitcase)

    def set_representation(self, repr):
        """Alternate way to initialize a KPConstructor."""
        # destroys any previous states
        self.__repr = repr

    def get_representation(self):
        """Get the Representation -- used by Receiver.submit"""
        return self.__repr

    def set_state(self, instance):
        state = self.__repr.GetState()
        inst_pickle = cPickle.dumps(instance)
        state.SetPart('state', 
                      KPRepr.ContentHeader('state',
                                           'application/knowbot-state-object'),
                      inst_pickle)
        self.set_metadata( {'knowbot-class': instance.__class__.__name__} )

    def set_code(self, modules):
        self.__repr.load_main_module(modules[0])
        for mod in modules[1:]:
            self.__repr.load_module(mod)

    def set_metadata(self, metadata):
        meta = self.__repr.GetMetadata()
        for k, v in metadata.items():
            meta.SetValue(k, v)

    def set_suitcase(self, files):
        if self.__suitcase is None:
            from Suitcase import SuitcaseFilesystem
            self.__suitcase = SuitcaseFilesystem(self.__repr.GetSuitcase())
        print files
        for path, buf in files:
            if path[0] != '/':
                path = '/' + path
            self.__suitcase._create_parent_parts(path)
            f = self.__suitcase.open(path, 'w')
            f.write(buf)
            f.close()

    def get_suitcase_fs(self):
        if self.__suitcase == None:
            from Suitcase import SuitcaseFilesystem
            self.__suitcase = SuitcaseFilesystem(self.__repr.GetSuitcase())
        return self.__suitcase
        

class KPManager(RS.NullRS):

    def __init__(self):
        RS.NullRS.__init__(self)
        self.recv = {}
        self.kps = {}

    def construct(self, instance, modules, metadata={}, files=[]):
        return KPConstructor(instance, modules, metadata, files)

    def send(self, kos_name, kp, callback):
        recv = self.get_receiver(kos_name)
        id = self.AllocateID(str(hash(kp))) # a cheap "unique" string
        rs = self.IluSBH()
        kp.set_metadata({"reporting-station": rs,
                         "reporting-id": id
                         })
        self.kps[id] = callback
        recv.SubmitKP(KOS.AlphaFormat,
                      kp.get_representation().Marshal())  

    unpack_times = None

    def KPTerminated(self, id, reason, details, kprepr):
        t1 = time.time()
        try:
            pickled_kp = kprepr.GetState().GetData('state')
        except ilu.IluGeneralError:
            traceback.print_exc()
            print "Detected in koe.common.kpapp.KPManager.KPTerminated()"
            instance = None
        else:
            instance = cPickle.loads(pickled_kp)
            if self.unpack_times is not None:
                self.unpack_times.append(time.time() - t1)
        try:
            callback = self.kps[id]
        except KeyError:
            print "Could not locate information about KP", id
            print "Detected in koe.common.kpapp.KPManager.KPTerminated()"
        else:
            del self.kps[id]
            callback(instance, id, reason, details, kprepr)

    def get_receiver(self, kos):
        if self.recv.has_key(kos):
            return self.recv[kos]
        recv = nstools.WorldOpen("kos/%s/receiver" % kos)
        self.recv[kos] = recv
        return recv
