"""Facilitator class for loading and augmenting a KP object.

KP creation using the class defined in this module is simplified over
using the KPReprImpl.Representation class by providing simpler ways to
add modules to the KP.  The constructor provides for the use of a
search path for module loading.  Modules are loaded into the KP by
name.  The main module can be distinguished at creation time or after
creation using load_main_module.

The module uses names as argument in several places. Sometimes these
names are module names and sometimes they are file names. Be careful
to use the right kind of name. For example, this module is named
'kptools' (this is the name you use to import it), but it's filename is
'kptools.py'.

Classes
-------

Representation
    Subclass of KPReprImpl.Representation with added functionality for
    easily adding modules and submitting the KP.

IOServer
    I/O server class useful to redirect a KP's standard I/O streams.

Constants
---------

DEFAULT_KP_CLASS
    Default class name for a KP.  If the class name is not specified when
    loading the main module for a KP, the loader methods will use this name.
    It will not be used if the main module is not specified.

Exceptions
----------

LoadingError
    This will be raised by the constructor for the `Representation' class if
    a named module could not be located.

Functions
---------

default_kernel()
    Return the default kernel name for KP launches


"""

import pathhack
import nstools
import os
import string
import sys
import KPReprImpl
import RS
import ilutools
import qsplit
from koe.interfaces.stubs import KOS
from koe.interfaces.stubs import Mapping
from koe.interfaces.stubs import KPRepr
from koe.interfaces.stubs import FileAPI, FileAPI__skel

LoadingError = "kptools.LoadingError"

DEFAULT_KP_CLASS = "KP"
DEFAULT_KP_PATH = ('.', os.path.join(pathhack.kos_root(), "local"))
LEGAL_EXTENSIONS = ('', '.py')



class Representation(KPReprImpl.Representation):
    """KP object with enhanced loading facilities.
    """

    def __init__(self, modules=(), main=None,
                 klass=DEFAULT_KP_CLASS, path=None):
        """Create a KP with specified modules.

        The module search path can be specified with the path
        argument. If it isn't specified, the KPPATH environment
        variable is used. If that doesn't exist either, the default is
        the current directory and the directory containing kptools.py.

        """
        KPReprImpl.Representation.__init__(self, None)
        if path:
            self._path = path
        elif os.environ.has_key("KPPATH"):
            self._path = string.splitfields(os.environ["KPPATH"], ":")
        else:
            self._path = DEFAULT_KP_PATH
        self._path = tuple(self._path)
        for name in modules:
            self.load_module(name)
        if main:
            self.load_main_module(main, klass)
        self.__sfs = None

    def load_module(self, modname):
        """Loads a single module by name.

        modname -- name of the module to load.

        Any module already loaded with the same name will be replaced
        if successful.

        Raise LoadingError if the module could not be found or if it
        has a bad extension.
        Raise IOError if the file is found but cannot be read.
        Raise SyntaxError if the module has a Python syntax error.

        """
        name, ext = os.path.splitext(modname)
        if string.lower(ext) not in LEGAL_EXTENSIONS:
            # why is this check done? the docs say you pass the name
            # of the module not the name of the file
            raise LoadingError, "bad extension in module " + `name+ext`
        filename = name + ".py"
        modname = os.path.basename(name)  # why is this needed?
        source = self._get_module_source(filename)
        # Compile the source, just to check for syntax errors
        # (Otherwise we get a lousy error from the KOS)
        try:
            compile(source, filename, "exec")
        except SyntaxError, value:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (errorfilename, lineno, offset, line) = value
            except:
                pass
            else:
                try:
                    value = SyntaxError(msg, (filename, lineno, offset, line))
                except:
                    value = msg, (filename, lineno, offset, line)
            raise SyntaxError, value
            
        self.add_module(modname, source)

    def _get_module_source(self, filename):
        """Locates & returns the source code for a single module by filename.

        filename -- file to load, must include `.py' extension.

        This method may be overridden by subclasses.  It provides path
        searching and file loading.  (Perhaps these two operations should be
        split into distrinct methods?)

        Raise IOError if a file is found but could not be read.
        Raise LoadingError if no file is found.
        """
        if os.path.isabs(filename) and os.path.exists(filename):
            return open(filename, 'r').read()
        else:
            path = self._path
            for dir in path:
                fname = os.path.join(dir, filename)
                if os.path.exists(fname):
                    return open(fname, 'r').read()
        raise LoadingError, "can't find module " + `filename`

    def load_main_module(self, name, klass=DEFAULT_KP_CLASS):
        """Loads the main module for a KP, setting metadata appropriately.

        name -- module name of the module to load.
        klass -- name of the KP class within the module.

        Return conditions are the same as for `load_module()'.  If module
        loading is successful, the metadata attributes for `main-module' and
        `knowbot-class' are set appropriately.

        """
        self.load_module(name)
        meta = self.GetMetadata()
        name = os.path.basename(name)      # why needed?
        name, ext = os.path.splitext(name) # why needed?
        meta.SetValue("main-module", name)
        meta.SetValue("knowbot-class", klass)

    def add_module(self, filename, source):
        """Add module from source string.

        name -- filename to use for the module
        source -- the actual source code

        """
        code = self.GetCode()
        code.SetPart(filename,
                     KPRepr.ContentHeader(filename, "text/plain"), source)

    def add_main_module(self, filename, source, klass=DEFAULT_KP_CLASS):
        """Add main module from source string.

        name -- filename to use for the main module
        source -- the actual source code
        klass -- name of the KP class within the module.

        The metadata attributes for `main-module' and `knowbot-class' are set
        appropriately.
        """
        self.add_module(filename, source)
        meta = self.GetMetadata()
        meta.SetValue("main-module", filename)
        meta.SetValue("knowbot-class", klass)

    def __load_suitcase_fs(self):
        if self.__sfs is None:
            from Suitcase import SuitcaseFilesystem
            self.__sfs = SuitcaseFilesystem(self.GetSuitcase())
        return self.__sfs

    def get_suitcase_fs(self):
        return self.__load_suitcase_fs()

    def add_file(self, path, string=None, file=None, srcpath=None):
        suitcase = self.__load_suitcase_fs()
        f = suitcase.open(path, 'w')
        if string:
            f.write(string)
        elif file:
            f.write(file.read())
        elif srcpath:
            fin = open(srcpath, 'r')
            f.write(fin.read())
            fin.close()
        f.close()

    def add_argv(self, args):
        """Assign the KP's sys.argv to the list args."""
        meta = self.GetMetadata()
        meta.SetValue('program-arguments',
                      qsplit.qjoin(args))

    def submit(self,
               kosname=None,
               receiver=None,
               reporter=None,
               kpname=None,
               repid=None,
               repsbh=None,
               ):
        """Submit this KP to a KOS, with a plethora of options.

        Must specify either kosname or receiver. All other arguments
        are optional.

        Returns the KP's reporting id and the KOS id assigned by the
        receiver. 

        kosname -- name for the KOS to submit to.

        receiver -- the receiver object
                    default: look up /kos/<kosname>/receiver

        reporter -- reporting station
                    default: new NullRS instance

        kpname -- name for the KP
                  default: metadata['main-module']; 'kp' if none

        repid -- Reporting ID for the KP
                 default: new ID from reporter

        repsbh -- SBH of reporting station 
                  default: reporter.IluSBH()

        """
        if (kosname and receiver) or not (kosname or receiver):
            raise ValueError, "exactly one of kosname and receiver required"
        if not receiver:
            receiver = nstools.WorldOpen("kos/%s/receiver" % kosname,
                                         typeobj="KOS.Receiver")
        if not reporter and (not repid or not repsbh):
            reporter = RS.NullRS()
        metadata = self.GetMetadata()
        if not kpname:
            try:
                kpname = metadata.GetValue('main-module')
            except Mapping.KeyError:
                kpname = "kp"
        id = repid or reporter.AllocateID(kpname)
        sbh = repsbh or reporter.IluSBH()
        metadata.SetValue('reporting-id', id)
        metadata.SetValue('reporting-station', sbh)
        return id, receiver.SubmitKP(KOS.AlphaFormat, self.Marshal())

    def Marshal(self):
        if self.__sfs:
            self.__sfs._flush()
        return KPReprImpl.Representation.Marshal(self)


class ReprWrapper:

    def __init__(self, repr, copy=1):
        if copy:
            self.__repr = KPReprImpl.Representation()
            self.__repr.loadfromrepr(repr)
        else:
            self.__repr = repr
        self.__sfs_fs = None

    def __load_suitcase_fs(self):
        if self.__sfs_fs is None:
            from Suitcase import SuitcaseFilesystem
            self.__sfs_fs = SuitcaseFilesystem(self.__repr.GetSuitcase())
        return self.__sfs_fs

    def __getattr__(self, attr):
        return getattr(self.__repr, attr)

    def get_suitcase_fs(self):
        return self.__sfs_fs

    def get_file(self, path, file=None, dstpath=None):
        suitcase = self.__load_suitcase_fs()
        f = suitcase.open(path, 'r')
        if file:
            file.write(f.read())
            f.close()
        elif dstpath:
            fout = open(dstpath, 'w')
            fout.write(f.read())
            fout.close()
            f.close()
        else:
            buf = f.read()
            f.close()
            return buf


class SimpleReporter(RS.NullRS):

    """A reporting station that supports callback on termination.

    To make effective use of callbacks, a KP needs to be submitted
    using the kpname optional argument (see Representation.submit
    above). Whenever a KP with that name terminates, the callback
    function will be called.

    example:
       rs = SimpleReporter()
       rs.register_callback('mine', my_callback)
       kp.submit('kos', kpname='mine')

    """

    def __init__(self):
        RS.NullRS.__init__(self)
        self.__callbacks = {}
        self.__next = {}
        self.__pid = os.getpid()
        self.__pending = {}
        self.__wait = 0

    def register_callback(self, name, callback):
        self.__callbacks[name] = callback

    def wait(self):
        if not self.__wait:
            self.__wait = 1
            self.start()

    def AllocateID(self, prefix):
        try:
            self.__next[prefix] = next = self.__next[prefix] + 1
        except KeyError:
            self.__next[prefix] = next = 0
        return "%s.%d.%d" % (prefix, self.__pid, next)

    def KPMigrationReceived(self, id, kpid, dest):
        self.__pending[id] = 1

    def KPTerminated(self, id, reason, details, repr):
        del self.__pending[id]
        i = string.find(id, '.')
        if i != -1:
            try:
                func = self.__callbacks[id[:i]]
            except KeyError:
                pass
            else:
                func(id, reason, details, ReprWrapper(repr))
        if self.__wait and not self.__pending:
            self.stop()
            self.__wait = 0

    # you can subclass these methods to use a different main loop

    def start(self):
        ilutools.RunInterruptableMainLoop()

    def stop(self):
        ilutools.ExitMainLoop()

class KPReporter(SimpleReporter):

    def __init__(self, kos):
        self.__kos = kos
        SimpleReporter.__init__(self)

    def start(self):
        self.__kos.run()

    def stop(self):
        self.__kos.stop()



class IOServer(FileAPI__skel.File):

    """A simple I/O server class.

    This implements most of the FileAPI.File interface (which looks a
    lot like Python's standard file objects).  The main use is to
    provide a simple way to connect the standard I/O stream of a KP to
    that of its submittor: the submittor instantiates an IOServer
    instance, and sticks its SBH in the metadata of the KP as
    stdin-server, stdout-server and stderr-server.  (In fact, there's
    a true-object-only method that does this for your.)

    Note: We map close() to flush() on the output stream.

    """

    # XXX Since the underlying methods raise IOError, not
    # XXX FileAPI.FileIOError, exception behavior is bogus.
    # XXX Fixing that would cause a lot of overhead, and since
    # XXX I/O errors are rarely observed on reads/writes,
    # XXX I don't think it's worth it.  YMMV.

    def __init__(self, r=None, w=None):
        self.r = r or sys.stdin
        self.w = w or sys.stdout
        self.close = self.w.flush
        self.flush = self.w.flush
        self.write = self.w.write
        self.readline = self.r.readline
        try:
            self.read = self.r.read
        except AttributeError:
            pass
        try:
            self.writelines = self.w.writelines
        except AttributeError:
            pass
        try:
            self.readlines = self.r.readlines
        except AttributeError:
            pass

    def linktokp(self, kprepr):
        # You can also do this yourself, e.g. if you don't want all
        # three streams connected to the same server
        sbh = self.IluSBH()
        metadata = kprepr.GetMetadata()
        metadata.SetValue('stdout-server', sbh)
        metadata.SetValue('stderr-server', sbh)
        metadata.SetValue('stdin-server', sbh)

    # The methods open(), seek() and tell() are not implemented.
    # After all, this is for standard I/O streams.

    # The following methods are only invoked when the underlying
    # r/w objects don't support them.  We assume that the basic
    # input method is readline(), not read().

    def read(self, n=-1):
        if n < 0:
            return string.join(self.readlines(), "")
        raise FileAPI.FileIOError, "read(n) not implemented"

    def readlines(self):
        lines = []
        line = self.readline()
        while line:
            lines.append(line)
            line = self.readline()
        return lines

    def writelines(self, data):
        filter(self.write, data)



def default_kernel():
    """Return the default kernel name for KP launches etc."""
    return pathhack.get_short_hostname()
