"""Namespace tools.

Exceptions:

    NoWorldrootError
    BadPathError

Classes:

    Descriptor
    Context

Functions:

    Lookup(context, path)
    GetWorldroot()
    WorldOpen(path, typeobj)

"""

import sys
import string
import urllib
import ilu
import pathhack
from types import *
from koe.interfaces.stubs import Namespace



# Exceptions

class NamespaceError:
    """Generic namespace exception.

    Public instance variables:
        msg  -- the error message

    """
    def __init__(self, msg=''):
        self.msg = msg

    def __repr__(self):
        return self.msg


class NoWorldrootError(NamespaceError):
    """Exception raised when no live Worldroot could be found.

    Derived from:
        NamespaceError

    Public instance variables:
        sbhfile -- the filename that was consulted to find the SBH of a
                   live Worldroot.

    """
    def __init__(self, sbhfile, msg=''):
        self.sbhfile = sbhfile
        NamespaceError.__init__(self, msg)


class BadPathError(NamespaceError):
    """Exception raised when a path could not be completely resolved.

    Derived from:
        NamespaceError

    Public instance variables:
        n/a

    """
    pass




class Context:
    """A wrapper around ILU Contexts (defined in Namespace.isl).

    Wraps around an ILU Context, but provides a more Python-esque
    interface. I.e. Descriptors are instances of the
    nstools.Descriptor class, not the Python dictionaries ILU RECORDs
    typically map to.

    These classes are only useful on the client (surrogate) side.

    """
    def __init__(self, context, parent=None, linkname=None):
        """Creates a context wrapper.

        Pass in the surrogate Namespace.Context object.  Optional
        `parent' is a context that created this context, and optional
        `linkname' is the name used in the parent context to reach
        this node.  If provided, these are used to automatically
        rebind the backend Namespace.Context object in case of
        failure.

        """
        self.__context = context
        self.__parent = parent
        self.__linkname = linkname

    _MAXREBIND = 5

    def __apply(self, func, *args):
        # We want to execute the function on the back-end context-object.
        # When connections are lost to the back-end object, we will try to
        # rebind it some number of times.  If the rebinding continues to fail, 
        # we will raise an exception, but this will true wedge the system
        rebindcnt = 0
        while 1:
            try:
                method = getattr(self.__context, func)
                return apply(method, args)
            except (ilu.IluGeneralError, ilu.IluProtocolError), msg:
                rebindcnt = rebindcnt + 1
                if rebindcnt > self._MAXREBIND:
                    raise BadPathError(msg)
                # Try rebinding the back-end context object.  If we can't
                # even attempt to rebind it, then just go around the loop
                # again
                if not self.__parent or not self.__linkname:
                    continue
                # Attempt to recontact the parent and find the back-end
                # Context object.  Then get it's method and re-apply, bumping
                # up the rebind counter
                contextwrapper = self.__parent.Lookup1(self.__linkname).Open()
                self.__context = contextwrapper.__context

    def Lookup1(self, component):
        try:
            desc = self.__apply('Lookup1', component)
        except Namespace.ResolutionError, msg:
            raise BadPathError(msg)
        return Descriptor(reference=desc['reference'],
                          typename=desc['typename'],
                          parent=self,
                          linkname=component)

    def List(self):
        return self.__apply('List')

    def Bind(self, name, desc):
        try:
            desc = desc.descriptor_of()
        except AttributeError:
            pass
        return self.__apply('Bind', name, desc)

    def Rebind(self, name, desc):
        try:
            desc = desc.descriptor_of()
        except AttributeError:
            pass
        return self.__apply('Rebind', name, desc)

    def Unbind(self, name):
        return self.__apply('Unbind', name)

    def GetDescriptor(self):
        desc = self.__apply('GetDescriptor')
        return Descriptor(reference=desc['reference'],
                          typename=desc['typename'])



class Descriptor:
    """A wrapper around ILU Descriptors (defined in Namespace.isl).

    An ILU Descriptor is defined as a Record which is mapped to a
    Python dictionary.  These dictionaries are not terribly convenient
    to use, hence this class, which also provides a functional
    interface to creating new descriptors, and opening descriptors, etc.

    """
    def __init__(self, reference, typename='Namespace.Context',
                 parent=None, linkname=None):
        """Create a new Descriptor object.

        External interface keyword arguments:

        - reference -- the SBH of the object (required)
        - typename  -- its type (optional, default 'Namespace.Context')

        Optional internal interface keyword arguments:
        
        - parent    -- the parent Context (wrapper) object that
                       created the Descriptor
        - linkname  -- the name of the link from the parent to this
                       Descriptor

        """
        self.__ref = reference
        self.__type = typename
        self.__parent = parent
        self.__linkname = linkname

    def __getitem__(self, name):
        if name == 'reference':
            return self.__ref
        elif name == 'typename':
            return self.__type
        else:
            raise KeyError, name

    def __repr__(self):
        return '<Descriptor of %s at %s>' % (self.__type, self.__ref)

    def reference(self): return self.__ref
    def typename(self): return self.__type
    def descriptor_of(self):
        return {'typename' : self.__type,
                'reference': self.__ref
                }

    def Open(self, typeobj=None):
        """Opens the descriptor, returning the surrogate object.

        Returns None if the descriptor is invalid or if the true
        object could not be reached.

        If optional `typeobj' is given, it is the type object that the
        descriptor must match.  If it doesn't match, then an None is
        returned.  If not provided, the type given by self.typename()
        is used. `typeobj' can be a class or a string.

        XXX and TBD: Returns None for failure, which is bogus.

        """

        # we can check the type this early if it is a string
        if typeobj and \
           type(typeobj) == StringType and \
           typeobj <> self.__type:
            return None

        # typename is a string of the form Module.Class
        [modname, classname] = string.splitfields(self.__type, '.')
        # now that the KOE software has been packagized, all interface stubs
        # reside in the koe.interfaces.stubs package (relative to $KOEROOT),
        # so what we're really trying to execute is something like:
        #
        # from koe.interfaces.stubs import modname
        try:
            # this returns the package, not the module
            package = __import__('koe.interfaces.stubs',
                                 globals(), locals(),
                                 [modname])
        except ImportError:
            return None

        # dig out the module attribute from the package, and then the class
        # attribute from the module
        try:
            module = getattr(package, modname)
            class_ = getattr(module, classname)
        except AttributeError:
            return None

        # check the actual type
        if typeobj and type(typeobj) == ClassType and class_ <> typeobj:
            return None

        # create the object from the SBH
        try:
            object = ilu.ObjectOfSBH(class_, self.__ref)
        except (TypeError, ilu.IluGeneralError):
            return None

        # should we wrap it in a context?  TBD: this is a kludge
        if self.__type == 'Namespace.Context':
            return Context(object, self.__parent, self.__linkname)
        return object


def Lookup(context, path):
    """Performs a complete lookup of the given path, starting at the
    given context.

    Raises `BadPathError' if the entire path is not resolvable.  The
    value of the exception is the portion of the path that could not
    be resolved.  Otherwise, returns the descriptor for the resolved
    object.

    """
    names = string.split(path, '/')
    for name in names[:-1]:
        desc = context.Lookup1(name)
        newcontext = desc.Open('Namespace.Context')
        if not newcontext:
            raise BadPathError(path)
        context = newcontext
    return context.Lookup1(names[-1])


_worldroot = None

def GetWorldroot():
    """Find a worldroot server.

    Return value is a Context object for the worldroot.

    If no worldroot was found, a NoWorldrootError exception is raised.

    """
    global _worldroot
    if _worldroot:
        return _worldroot
    sbhfile = pathhack.kos_root() + '/var/worldroot.sbh'
    try:
        fp = open(sbhfile, 'r')
        sbh = string.strip(fp.readline())
        fp.close()
    except IOError, msg:
        try:
            #
            # We might be an applet, try sbhfile as a URL
            import urlparse
            sbhfile = urlparse.urljoin(sys.path[-1], 'var/worldroot.sbh')
            fp = urllib.urlopen(sbhfile)
            lines = fp.readlines()
            # Handle potential Apache 1.4 urllib dilemma and rip out the sbh
            sbh = string.strip(lines[-1])
            fp.close()
        except IOError, msg:
            raise NoWorldrootError(
                sbhfile,
                "can't read worldroot SBH file %s: %s" % (`sbhfile`, str(msg))
                )
    _worldroot = Descriptor(reference=sbh).Open()
    if _worldroot is None:
        raise NoWorldrootError(sbhfile,
                               "can't open worldroot read from %s" % `sbhfile`)
    try:
        _worldroot.List()
    except (ilu.IluGeneralError, ilu.IluProtocolError):
        _worldroot = None
        raise NoWorldrootError(sbhfile, 'worldroot is unreachable')
    return _worldroot


def WorldOpen(path, typeobj=None):
    """Handy utility to open an object by a path relative to the worldroot.

    Return the opened object if successful.

    Raise NoWorldrootError when the worldroot can't be found.
    Raise BadPathError when the path lookup fails.

    """
    worldroot = GetWorldroot()
    desc = Lookup(worldroot, path)
    obj = desc.Open(typeobj)
    if not obj:
        raise BadPathError(
            "Open() of %s%s failed" % \
            (`path`, typeobj and " as %s" % `typeobj` or "")
            )
    return obj
