'''All path hacking for KOS.

In order to bootstrap this, the environment variable PYTHONPATH must include
the directory containing the pathhack module, *or* the script must contain a
line like this before importing pathhack:

import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../../"))

The number of "../" segments should reflect the depth of the script file
relative to $KOSROOT.

This module has knowledge of where it lives relative to $KOSROOT.  If its
location changes, it needs to be updated.  It contains a sophisticated warning
feature that tells you immediately when such an update is necessary :-)


Functions (see individual doc strings for more information):

    kos_root():
        Return the root of the KOS source tree.

    common_root():
        Return the root of the shared development tree.

    platform_name():
        Return a name for a directory containing architecture-specific
        files.

    get_short_hostname():
        Return a shortened version of the current host name (i.e.,
        with the domain name stripped off).

'''

import sys
import os
import string

_kosroot = None                         # kos root path cache
_commonroot = None                      # common root path
_path_is_set = None
_script_path = None                     # path to the main script



def kos_root():
    """Returns absolute path to the root directory of the KOS environment.
    """
    global _kosroot
    if _kosroot:
        return _kosroot

    if os.environ.has_key('KOSROOT'):
        _kosroot = os.environ['KOSROOT']
    else:
        try:
            file = __file__
        except NameError:
            # By mistake, Grail sets __filename__ instead of __file__ :-(
            file = __filename__
        pkgdir, file = os.path.split(file)
        parent, subdir = os.path.split(pkgdir)
        # Use old (Python 1.4) join semantics so pathhack works in an applet
        dir = os.path.join(os.getcwd(), parent)
        dir = os.path.join(dir, '..')
        _kosroot = os.path.normpath(dir)
    return _kosroot


def common_root():
    """Returns absolute path to the common root directory of the KOS.
    """
    global _commonroot
    if _commonroot:
        return _commonroot

    if os.environ.has_key('KOSCOMMONROOT'):
        _commonroot = os.environ['KOSCOMMONROOT']
    else:
        # XXX: yes, hardcoded.
        _commonroot = '/projects/koe/develop'
    return _commonroot


def platform_name():
    """Calculate our platform name, for use as a subdirectory of common/.

    Usually sys.platform will give us the right thing, however Python
    is broken in that sys.platform is "sunos5" for both Solaris/Sparc
    and Solaris/Intel (and presumably Solaris/PPC).  Do some
    additional hackery in those cases.  If these hacks fail (e.g. when
    imported in an applet), return sys.platform.

    XXX Should eventually do this for Linux and NT as well.

    """
    plat = sys.platform
    if plat != 'sunos5': return plat
    try:
        arch = os.uname()[4]
        if arch[:4] == 'sun4':
            arch = 'sparc'
    except (AttributeError, os.error):
        return plat
    else:
        return plat + '-' + arch
    

def get_short_hostname():
    """Return the shortened (with domain stripped), lowercased host name."""
    import socket
    name = socket.gethostname()
    return string.lower(string.split(name, '.')[0])
