"""Utilities by which plugin modules connect with the KOS kernel."""

import getopt
import os
import sys
import string

# CnBroker *must* be imported before any surrogate classes, since it imports
# ilu_tk.  Import order dependencies exist in ILU 2.0a12 between ilu_tk and
# surrogate classes of any object defined COLLECTIBLE.
import ilu
import ilutools                         # install exception handler
import CnBroker

import NamespaceImpl, nstools
from koe.interfaces.stubs import KOS
from koe.interfaces.stubs import KOSKernel

# XXX Hack alert!
# Patch CnBroker to make kos.run() interruptable.
# We don't want this to happen for the Supervisor so patch it here.
# This obviously depends on the fact that CnBroker has a line saying
# ``from ilutools import RunMainLoop''.
if not sys.modules.has_key('Tkinter'):
    CnBroker.RunMainLoop = ilutools.RunInterruptableMainLoop


COMMON_ARGS = "n:"
LONG_COMMON_ARGS = ['kernel=']



def register(plugin_name, kernel_name=None,
             command_args="", long_command_args=[]):
    """Register with kernel, and return plugin entry for communications."""

    long_args = LONG_COMMON_ARGS + long_command_args
    opts, args = getopt.getopt(sys.argv[1:], COMMON_ARGS + command_args,
                               long_args)
    for opt, val in opts:
        if opt in ("-n", "--kernel"):
            kernel_name = val
    if not kernel_name:
        import kptools
        kernel_name = kptools.default_kernel()
    kernel_name = string.strip(kernel_name)

    # Locate kernel via worldroot:
    try:
        wr = nstools.GetWorldroot()
    except nstools.NoWorldrootError, msg:
        error("can't get the worldroot: %s", str(msg))
    path = "kos/"+kernel_name
    try:
        kosns = nstools.WorldOpen(path)
    except nstools.BadPathError, msg:
        error("can't open KOS namespace %s"%`path`)

    # Get the plugins namespace and Bind this plugin name
    plugins_ns = kosns.Lookup1('plugins').Open()
    if not plugins_ns:
        error("kernel does not have a `plugins' namespace")
    # XXX The following test actually makes a lot of sense to have.
    # XXX Why is it commented out?
    # XxX I believe it was for developers convenience - restarting plugins
    # XxX without having to restart a kos.  It should be reenabled, or
    # XxX perhaps made an optional behavior...
##     if plugin_name in plugins_ns.List():
##      error("plugin %s already bound in kernel's `plugins' namespace" %
##            plugin_name)

    # Get the receiver through the namespace
    receiver = kosns.Lookup1('receiver').Open(typeobj='KOS.Receiver')
    if not receiver:
        error("kernel's receiver is not responding")

    this_plugin_ns = nstools.Context(NamespaceImpl.Context())
    this_plugin_ns.Bind('kos', kosns.GetDescriptor())
    this_plugin_ns.Bind('world', wr.GetDescriptor())
    plugins_ns.Rebind(plugin_name, this_plugin_ns.GetDescriptor())
    pentry_sbh = receiver.RegisterPlugin(os.getpid())
    pentry = ilu.ObjectOfSBH(KOSKernel.PluginEntry, pentry_sbh)
    broker = CnBroker.ConnectionBroker(pentry)
    broker.set_ns(this_plugin_ns)

    # XXX Hack to be able to kill a kernel with ^C.
    # (Catching KeyboardInterrupt doesn't work because
    # ILU doesn't propagate it out of its main loop.)
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    return broker, this_plugin_ns


def error(format, *args):
    """Print error message, and exit."""
    sys.stdout = sys.stderr
    print "Error:", format % args
    sys.exit(1)


def spawnserver(filename, args=(), flagoption="-f"):
    """Do the little dance needed to fork a child process with a server.

    This creates a pipe to which the child process must write its SBH.
    The file descriptor is passed to the child as a command line
    option, before any other options or arguments.  The option is
    normally named -f, but you can change this with the flagoption
    argument.

    Arguments:

    filename -- the executable file containing the server program
    args -- optional sequence of additional command line arguments
    flagoption -- name for the option to pass the pipe file
                  descriptor; default "-f"

    Return value:

    This returns a tuple (sbh, pid, pids) where sbh is the child
    server's SBH, pid is the child's process id, and pids is a list of
    recently deceased process ids as returned by reaper().  The latter
    two items can be safely ignored by applications that don't do
    their own child process management.

    Errors:

    If the fork succeeds but the exec in the child fails, this routine
    will return normally with sbh set to the empty string.

    For various reasons, os.error may be raised (can't create a pipe,
    can't fork).

    """
    pids = reaper()
    r, w = os.pipe()
    pid = os.fork()
    if pid:
        # Parent
        os.close(w)
        sbh = string.strip(os.read(r, 1024))
        os.close(r)
        return (sbh, pid, pids)
    else:
        # Child
        try:
            os.close(r)
            command = (filename, flagoption, `w`) + tuple(args)
            try:
                os.execvp(filename, command)
            except os.error, msg:
                sys.stderr.write(
                    "pluginlib.spawnserver: execv of %s failed: %s\n" %
                    (str(command), str(msg)))
        finally:
            os._exit(127)

def reaper():
    """Collect deceased children.  Return a list of their pids."""
    pids = []
    while 1:
        try:
            pid, sts = os.waitpid(-1, os.WNOHANG)
        except os.error:
            break
        if not pid:
            break
        pids.append(pid)
    return pids
