"""String utility to split a string into quoted components.

The main function (`qsplit()') was originally written by Guido and posted
to comp.lang.python in response to a question.  Documentation and some
associated utility functions have been added.
"""

import regex
import string
_qprog = regex.compile('''"[^"]*"\|\'[^\']*\'\|[^"\' \t]+''')
_jprog = regex.compile('["\'' + string.whitespace + ']')

def qsplit(s):
    """Splits a string into a list of whitespace-separated quoted strings.

    s
        String to split.

    Similar to `string.split' but allows component strings to be quoted.

    Example
    -------

      from qsplit import qsplit
      res = qsplit("part1 'part 2' '  another  '")

    `res' will have the value `["part1", "'part 2'", "'  another  '"]'.
    Note that the original quotation marks are included in the resulting
    components.
    """
    list = []
    i = 0
    while 1:
        i = _qprog.search(s, i)
        if i < 0:
            break
        it = _qprog.group(0)
        list.append(it)
        i = i + len(it)
    return list


def rem_quotes(list):
    """Remove outer quotation marks from list components.

    list
        The list of strings which may need quotation-mark removal.

    Returns a copy of `list' with matching outer quotation marks stripped
    from the string. All list components must be strings.  This should be
    used in conjunction with the result of the `qsplit()' function.
    """
    list = map(None, list)
    for i in range(len(list)):
        s = list[i]
        if len(s) >= 2 and s[0] in '\'"' and s[0] == s[-1]:
            list[i] = s[1:-1]
    return list


def qsplit_nq(s):
    """Like qsplit, but without the quotation marks.

    E.g. Splits a string into a list of whitespace-separated quoted
    strings, without the quotation marks.

    s
        String to split.

    Similar to `qsplit()' but removes the original quotation marks from the
    component string before returning the list.
    """
    return rem_quotes(qsplit(s))


def qjoin(list):
    """Joins a list of words, preserving whitespace.

    list
        List to join.

    If one of the words in list has embedded whitespace, wrap the word
    in quotes so that the corresponding qsplit will correctly split
    the string.
    """
    work = list[:]
    for i, word in map(None, range(len(work)), work):
        if _jprog.search(word) >= 0:
            quote = '"'
            if string.find(word, quote) >= 0:
                quote = "'"
            work[i] = '%s%s%s' % (quote, word, quote)
    return string.join(work)


#
#  end of qsplit.py
