"""Mostly standalone Rate.d client library.

This module speaks to the `koe' service of an AGS.  That service returns a
non-sense port number.  The interesting data is in the contact information,
which contains exactly the kernel name of the machine that is accepting the
RFS.

Note that this makes an assumption that there's only one KOE running on any
machine that's also running a Rate.d, because Rate.d can't mediate between
multiple kernels running on the same machine.

"""

import os
import string
import struct
import socket
import select



class RatedError(Exception):
    pass

class RatedTimeoutError(RatedError):
    pass

class RatedInvalidFDError(RatedError):
    pass



class RatedClient:
    def __init__(self):
        # sequence number for jobids
        self.__seqnum = 1
        # static jobid prefix
        addr = socket.gethostbyname(socket.gethostname())
        bytes = map(string.atoi, string.split(addr, '.'))
        self.__jobid_prefix = struct.pack(
            '4Bh', bytes[0], bytes[1], bytes[2], bytes[3],
            os.getpid())
        self.__responses = {}
        # rated socket.  create when needed
        self.__rsock = None

    # rate.d constants which should really be imported from ratedconfig.py out
    # of /projects/ags
    RATED_VERSION = 1
    REQUEST_FOR_SERVICE = 1
    KOE_SERVICE_ID = 333

    # Other constants, this ought to be configurable
    TIMEOUT = 2.0

    def requestService(self, addr, port):
        """Issue a request for service (RFS) to an AGS.

        addr is the broadcast address of the AGS network, e.g. XXX.XXX.XXX.0

        port is the external access port of the AGS broadcast address

        Returned is the kernel name of the accepting host, as a string.

        """
        #
        # send the request for service, select on the rated socket, and wait
        # for a response or a timeout
        #
        jobid = self.sendRequest(addr, port)
        return self.receiveResponse(jobid)

    def sendRequest(self, addr, port):
        fmt = '>bbh6shll'
        length = struct.calcsize(fmt)
        msg = struct.pack(fmt,
                          self.RATED_VERSION, self.REQUEST_FOR_SERVICE,
                          length, self.__jobid_prefix, self.__seqnum,
                          self.KOE_SERVICE_ID,
                          0)
        self.__seqnum = self.__seqnum + 1
        jobid = msg[4:12]
        # create the rated socket if not already available
        if not self.__rsock:
            self.__rsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.__rsock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.__fdlist = [self.__rsock.fileno()]
        #
        self.__rsock.sendto(msg, (addr, port))
        return jobid

    def receiveResponse(self, jobid):
        if self.__responses.has_key(jobid):
            return self.__responses[jobid]
        #
        # Wait for a response with a matching JobID
        while 1:
            r, w, x = select.select(self.__fdlist, [], [], self.TIMEOUT)
            if not r:
                raise RatedTimeoutError()

            # decode the rate.d response
            pkt, addr = self.__rsock.recvfrom(1024)
            # XXX: for now we basically ignore the jobid, except to match it
            # against the jobid of the message we sent.  Later we should be
            # able to handle asynchronous interleaved service requests, in
            # which case, we'll probably want to use the jobid as a key into a
            # dictionary, or some such
            vers, func, pktlen, r_jobid, r_addr, r_port, r_len = struct.unpack(
                '>bbh8slll', pkt[:24])
            kosname = pkt[24:24+r_len]
            if r_jobid == jobid:
                # We got a response for the request we sent out, so we're
                # done.  Otherwise we must have gotten an answer for a RFS
                # that we previous timed out on.  In that case, just discard
                # this response and go back to the top of the loop
                break
            self.__responses[r_jobid] = kosname

        # we basically ignore the returned address and port since these are
        # less important than the contact info, which will contain the name of
        # the kernel that the KP wants to use, either to migrate or clone to.
        # Dig out that info and return it
        return kosname
