"""Object wrapper with selective attribute coupling to original object.

Shadow objects provide the means for copying objects, with selected
attributes coupled to the corresponding attributes in the original.  The
shared attributes in the shadow are said to 'defer' to the original's
attributes, such that setting or getting a shared attribute value sets or
gets the current value of the attribute in the original object.

See the shadow class for more details, and the test routine for an example
of how this can be useful with the sys module."""

class shadow:

    """Object wrapper with selective attribute coupling to the original.
    
    Selected attributes will be consulted from and set in the original
    object.

    original
        An object to be shadowed.

    defers
        A list of attribute names to be coupled with those of the original
        object."""
    
    # '__originals' is situated outside the instance for protection in
    # rexec.  We use 'id(self)' as index instead of just 'self' to
    # circumvent an infinite getattr recursion on __hash__.

    def __init__(self, original, defers):
        global __originals
        try:
            __originals[id(self)] = (original, defers)
        except NameError:
            __originals = {id(self): (original, defers)}
        
    def __getattr__(self, attr):
        (original, defers) = __originals[id(self)]
        if attr in defers:
            return getattr(original, attr)
        else:
            raise AttributeError, attr

    def __setattr__(self, attr, val):
        (original, defers) = __originals[id(self)]
        if attr in defers:
            setattr(original, attr, val)
        else:
            self.__dict__[attr] = val

    def __repr__(self):
        original = __originals[id(self)][0]
        if hasattr(original, '__class__'):
            orig_name = "%s" % `original.__class__.__name__`
        elif hasattr(original, '__name__'):
            orig_name = "%s" % `original.__name__`
        else:
            orig_name = "unidentified"
        return ("<%s of %s %s at %s>"
                % (self.__class__.__name__,
                   orig_name,
                   type(original).__name__,
                   hex(id(self))[2:]))

    def __del__(self):
        del __originals[id(self)]

def test():
    """Test shadow attribute deferral."""
    import sys
    shadow_sys = shadow(sys, defers=['exc_type', 'exc_value',
                                     'last_type', 'last_value'])
    try:
        x
    except NameError:
        try:
            if sys.exc_type != shadow_sys.exc_type:
                raise SystemError, "Shadow attribute reference test failed!"
        except AttributeError:
                raise SystemError, "Shadow attribute reference missed!"

    shadow_sys.exc_type = 1
    if sys.exc_type != shadow_sys.exc_type:
        raise SystemError, "Shadow attribute assignment test failed!"

    print "Shadow tests successful."

if __name__ == "__main__":
    test()
