"""Modestly generic formatted text to index terms conversion

The getTokens function converts ASCII text, including all sorts of
punctuation and other crud, and returns a list of tokens suitable for
putting in a search engine.

The basic approach is to convert everything to lowercase and treat
everying other than letters, numbers, and hyphen as token separators.
Tokens must be at least 3 characters long and may not appear in the
list of stopwords.

The hyphen is treated specially.  A compound word, e.g.
the-hyphenated-word, is converted into four different tokens:
"the-hyphenated-word", "the", "hyphenated", and "word".

getTokens is implemented as an instance of TokenizerChain.  The
TokenizerChain allows you to chain together a text-to-token-list
function with an arbitrary set of filters. 
"""

import string
import re

import stop

goodchars = string.lowercase + string.digits + '-'
pattern = re.compile('[^%s]' % re.escape(goodchars))

def atleast3(s):
    return len(s) > 2

def isntStopword(s):
    return not stop.words.has_key(s)

def getRawTokens(buf, pattern=pattern):
    stripped = pattern.sub(' ', string.lower(buf))
    return string.split(stripped)

def getTokenKeys(tokens):
    multi = filter(lambda s:'-' in s, tokens)
    subparts = map(lambda s:string.split(s, '-'), multi)

    # deduplicate
    dict = {}
    for tok in tokens:
        dict[tok] = tok
    for list in subparts:
        for tok in list:
            dict[tok] = tok

    return dict.keys()

class TokenizerChain:
    def __init__(self, source, filters, predicates):
        self.source = source
        self.filters = filters
        self.predicates = predicates

    def __call__(self, buf):
        tokens = self.source(buf)
        for filt in self.filters:
            tokens = filt(tokens)
        for pred in self.predicates:
            tokens = filter(pred, tokens)
        return tokens

getTokens = TokenizerChain(getRawTokens,
                           [getTokenKeys],
                           [isntStopword, atleast3])
