(* Web content provider registration interface
 *
 * Used by KPs that wish to provide content services, this interface
 * allows the KPs to register, unregister, off-load, and unload static 
 * and dynamic pages.
 *)

INTERFACE ContentRegistrarAPI
    IMPORTS ContentProviderAPI, KPRepr END;

(* basic types *)

TYPE String = ilu.CString;

TYPE Identifier = INTEGER;

(* exceptions *)

EXCEPTION NonUniqueNameError : Identifier
    "The given name has already been registered.";

EXCEPTION BadIdentifierError : String
    "The given identifier is invalid.";

(* classes *)

TYPE ContentRegistrar = OBJECT
    METHODS
        RegisterProvider(name: String) : Identifier
            RAISES NonUniqueNameError END
            "Register a provider name.

            The return value is an opaque identifier used in
            subsequent calls.  Raise NonUniqueNameError if the name
            has already been registered.",

        UnRegisterProvider(ident: Identifier)
            RAISES BadIdentifierError END
            "Unregister the specified provider",

        RegisterLiveContent(ident: Identifier,
                            provider: ContentProviderAPI.ContentProvider)
            RAISES BadIdentifierError END
            "Register a ContentProvider for live content provision.

            This method is used to register the fact that a KP will be
            providing dynamic content.  Whenever a provider controlled
            URL is accessed, if that URL does not map to static
            content and live content has been registered, then the
            ContentProvider.GetContent() method is called on the
            registered provider object.

            Raise BadIdentifierError if the given identifier is
            invalid.",

        UnRegisterLiveContent(ident: Identifier)
            RAISES BadIdentifierError END
            "Unregister live content provision.",

        UploadStaticContent(ident: Identifier, suitcaserepr: KPRepr.NamedPart)
            RAISES BadIdentifierError END
            "Upload static content.

            The static content to be uploaded is the contents of the
            Suitcase filesystem contained in the KP.  Any files that
            have already been uploaded will be overwritten.  To
            actually delete directories or files, you must first
            UnloadStaticContent() then re-upload the entire new
            content.

            Raise BadIdentifierError if the given identifier is
            invalid.",

        UnloadStaticContent(ident: Identifier)
            RAISES BadIdentifierError END
            "Unload static content."

    END;

(*
 * Local variables:
 * indent-tabs-mode: nil
 * End:
 *)
