(* Specification for a simple File-like object.
 *
 * This allows for the expected read/write interface for file-like
 * objects.  There's one catch: depending on how the file is `opened'
 * you can either read or write to it.  For normal file-like objects,
 * you do this with an open() method which returns the object.  This
 * interface specifies a slightly different paradigm.  Here, you get a
 * file-like object through some other mechanism (e.g. the namespace),
 * and then call the open() method to specify read or write semantics.
 *)

(* $Id: FileAPI.isl,v 1.3 1997/02/06 00:26:07 guido Exp $ *)
(* $Source: /projects/cvsroot/koe/src/koe/interfaces/FileAPI.isl,v $ *)

INTERFACE FileAPI;

TYPE String = ilu.CString;
TYPE Strings = SEQUENCE OF String;
TYPE OptionalString = OPTIONAL String;

(* Exceptions *)

EXCEPTION FileIOError : OptionalString
    "A file operation failed.";

(* Type definitions *)

TYPE File = OBJECT
    METHODS
        open(mode: String)
	    RAISES FileIOError END
	    "Opens the file for reading or writing.

	    `mode' can be one of the following strings:

	    r  -- for reading
	    w  -- for writing (w/ truncation)
	    a  -- for appending

	    Assuming the underlying object understands this, `mode'
	    can also be r+, w+, and a+, and may have an optional b
	    flag.",

        close()
	    RAISES FileIOError END
	    "Closes the file",

	flush()
	    RAISES FileIOError END
	    "Flushes the file",

	read(size: INTEGER) : String
	    RAISES FileIOError END
	    "Reads size number of bytes.

	    Returns the string read.  If at EOF, an empty string is
	    returned.  If size is negative, reads until EOF.",

        readline() : String
	    RAISES FileIOError END
	    "Reads a single line.

	    Returns the string read.  If at EOF, an empty string is
	    returned.",

        readlines() : Strings
	    RAISES FileIOError END
	    "Reads all lines.

	    Returns the strings read.  If at EOF, an empty list is
	    returned.",

	seek(offset: INTEGER, whence: INTEGER)
	    RAISES FileIOError END
	    "If the underlying file object supports it, seeks the file pointer.

	    `whence' can be 0 (absolute file positioning), 1 (seek
	    relative to the current position), or 2 (seek relative to
	    the file's end).",

	tell() : INTEGER
	    RAISES FileIOError END
	    "Returns the file's current position",

	write(text: String)
            RAISES FileIOError END
	    "Writes the text to the file-like object",

	writelines(lines: Strings)
	    RAISES FileIOError END
	    "Writes the list of lines to the file-like object"

    END;
