(*  KOS kernel interfaces. *)
(*  $Revision: 1.41 $ *)
(*  $Source: /projects/cvsroot/koe/src/koe/interfaces/KOSKernel.isl,v $ *)

INTERFACE KOSKernel
    IMPORTS KOS, KPRepr, Namespace, ReportingAPI END;


TYPE String = ilu.CString;
TYPE StringSequence = SEQUENCE OF String;

EXCEPTION NotFound : String
	"Requested object not found.";

EXCEPTION NotPermitted : String
	"Insufficient authority for requested operation.";

TYPE CID = String;			(* Connector ID - in ILU, and SBH *)
TYPE CnType = String;			(* Connector type name *)

TYPE ConditionID = INTEGER;

TYPE Condition = OBJECT

	DOCUMENTATION
	  "An object to put in the namespace, so that conditions
           can be shared. A program that wants to register a trigger
           gets a Condition out of the namespace and passes it to
           the Supervisor."

	METHODS
	  GetID(): ConditionID
	END;


TYPE ProcessLink =
	OBJECT

	DOCUMENTATION

	  "The ProcessLink allows the KOS kernel to communicate with
	  subprocesses. It is the dual of the ProcessEntry. The KOS
	  uses this interface to communication with the subprocess,
	  which holds the true object."
	  (* Think of it as the Supervisor. *)
	   
	METHODS

	  (* By making signal asynchronous, we risk the possibility
	     that the supervisor will fail to execute any of the
	     callbacks. This means that a Signal call would silently
	     and create the potential for deadlock.

	     Need to think about this more.
	   *)

	  ASYNCHRONOUS Signal(var: ConditionID),

	  ASYNCHRONOUS Broadcast(var: ConditionID)

	END; (* ProcessLink *)

TYPE ProcessEntry =
	OBJECT

	DOCUMENTATION
	  "The ProcessEntry links a subprocess with the KOS kernel.
	  The ProcessEntry true server is part of the kernel process,
          and the subprocess uses the ProcessEntry interface to 
          communicate with the KOS."

	METHODS
	
	  EstablishProcessLink(link: ProcessLink)
	    "Establish a link for the KOS to make calls to 
	    the subprocess",

	(* These methods provide Connection Manager services. *)

	  RegisterConnector(instance: CID,
			    "type": CnType, name: String)
	    "Create an entry for a connector instance in the KOS.",

	  UnregisterConnector(instance: CID)
	    RAISES NotFound, NotPermitted END
	    "Remove entries for connector instance owned by requester.",

	(* Interface for triggers *)

	  CreateCondition(): ConditionID
	    "Create a new condition variable",

	  Broadcast(var: ConditionID)
	    RAISES NotFound END
	    "Notify all triggers waiting on a condition variable",

	  Signal(var: ConditionID)
	    RAISES NotFound END
	    "Notify exactly one (arbitrary) trigger",

	  RegisterTrigger(var: ConditionID)
	    RAISES NotFound END
	    "Register a trigger on a condition variable",

	  UnregisterTrigger(var: ConditionID)
	    RAISES NotFound END
	    "Unregister a trigger"

	END;	(* ProcessEntry *)


TYPE PluginEntry =
	OBJECT

	SUPERTYPES ProcessEntry END

	DOCUMENTATION
          "KOS kernel communications link with a plugin subprocess.";

TYPE KPEntry =
	OBJECT

	SUPERTYPES ProcessEntry END

	DOCUMENTATION
	  "KOS kernel communications link with a KP subprocess."

	METHODS

          Hello(link: ProcessLink,
		OUT kos-name: String,
		OUT kpid: KOS.KPID,
		OUT kprepr: KPRepr.Representation,
		OUT visualizer: ReportingAPI.Visualizer,
		OUT nsd: Namespace.Descriptor)
	    "Initial handshake between supervisor and kernel.",

	  Goodbye(new-kpid: KOS.KPID,
		  new-kos-name: String)
	    "This KP has left us via migration.",

	  Terminated(reason: String,
		     details: String)
	    "This KP has terminated."

	END;	(* KPEntry *)
