(* Specifications for the intermediate representation of a Knowbot Program.
 *
 * This interface describes methods which allow clients to get, set,
 * and delete KP metadata, program code, program state, and suitcase
 * elements.  
 *
 * The original interface was based on MIME, which may explain some of
 * the names. It has evolved away from MIME however.
 *)

(* $Id: KPRepr.isl,v 1.19 1997/02/12 06:00:19 guido Exp $ *)
(* $Source: /projects/cvsroot/koe/src/koe/interfaces/KPRepr.isl,v $ *)

INTERFACE KPRepr
	IMPORTS
		Mapping
	END;

TYPE String = ilu.CString;
TYPE BinaryString = SEQUENCE OF BYTE;
TYPE Strings = SEQUENCE OF String;

EXCEPTION PartNameError : String
    "No part matches name.";

(* We expect the content header to include other metadata in future
 * releases. If we didn't, we could just pass the type names and
 * forget defining an entire record.
 *)
  
TYPE ContentHeader = RECORD
	name : String,
	"type" : String
END;

TYPE NamedPart = OBJECT
	METHODS
	  GetNames() : Strings
	    "Returns list of all names in this part.",

	  SetPart(name: String, header: ContentHeader,
		 data: BinaryString)
            "Sets the header and data of a part. Creates a new
             part if necessary.",

	  GetPart(name: String, OUT hdr: ContentHeader,
	         OUT data: BinaryString)
	    RAISES
                PartNameError
            END
	    "Gets the content header and contents of a part.",

	  DeletePart(name: String)
	    RAISES
	    	PartNameError
	    END
	    "Deletes an existing part.",

	  SetInfo(name: String, info: ContentHeader)
	    RAISES
	    	PartNameError
	    END
	    "Set the information object for the named part.",

	  SetData(name: String, buffer: BinaryString)
	    RAISES
	    	PartNameError
	    END
	    "Set the body of the named part.",

	  GetInfo(name: String) : ContentHeader
	    RAISES
	    	PartNameError
	    END
	    "Returns the information object for the named part.",

	  GetData(name: String) : BinaryString
	    RAISES
	    	PartNameError
	    END
	    "Returns the body of the named part as a string."

	END;


(* The high-level KP representation. *)

TYPE Representation = OBJECT
	METHODS
	  GetMetadata() : Mapping.StringMap
	    "Returns the metadata dictionary.",

	  GetCode() : NamedPart
	    "Returns the code dictionary.",

	  GetState() : NamedPart
	    "Returns the state dictionary.",

	  GetSuitcase() : NamedPart
	    "Returns the suitcase dictionary.",

	  FreezeDry() : BinaryString
	    "Returns a freeze dried, transportable, opaque representation.",

	(* The last two methods are Python specific. If you decide
	 * to use them out them outside of KPReprImpl, think again.
	 *)

	  Marshal() : BinaryString
	    "Return a marshalled representation (Python specific).",

	  "Pickle"() : BinaryString
	    "Return a pickled representation (Python specific)."

	END;
