(* Specification of a generalized interface to dictionary-like objects.
 *
 * Currently the 1.x ILU ISL language definition does not directly
 * support dictionary objects.  This Mapping interface is an attempt
 * at supporting this type of object.  Some semantics of dictionaries
 * are lost here, but by using this definition, it is pretty easy to
 * convert to and from Python dictionaries.
 *
 * From a StringMap() to a Python dictionary:
 *
 * d = {}
 * for [k, v] in mapobj.GetItems():
 *     d[k] = v
 *
 * From a dictionary to a StringMap():
 *
 * mapobj.SetValues(d.items())
 *)

(* $Id: Mapping.isl,v 1.9 1998/02/13 15:56:02 bwarsaw Exp $ *)
(* $Source: /projects/cvsroot/koe/src/koe/interfaces/Mapping.isl,v $ *)

INTERFACE Mapping;

TYPE String = ilu.CString;
TYPE Strings = SEQUENCE OF String;

(* Exceptions *)

EXCEPTION KeyError : String
    "No metadata associated with given key.";

EXCEPTION WriteError : String
    "Dictionary or value is read-only.";


(* Type definitions.
 *
 * Implementation note: Values are always ILU strings.  This is bogus,
 * but necessary, since ILU does not support the equivalent of CORBA's
 * `any' type.  Making Value a UNION is not an option since unions get
 * mapped to tuples in Python, making for a very ugly interface.  You
 * have to kludge up your own conversions of higher data types to
 * strings and vice versa, based on your application.
 *)

TYPE Key  = String;
TYPE Value = String;

TYPE Keys = SEQUENCE OF Key;
TYPE Values = SEQUENCE OF Value;

TYPE KeyValuePair = ARRAY OF 2 String;
TYPE KeyValuePairs = SEQUENCE OF KeyValuePair;

(* Mapping type interface. *)
TYPE StringMap = OBJECT
    COLLECTIBLE
    METHODS
        GetKeys() : Keys
	    "Return all keys.",

	GetValue(key: Key) : Value
	    RAISES KeyError END
	    "Return value associated with key.
	    Can raise KeyError if the key is not in the mapping.",

	GetItems() : KeyValuePairs
	    "Return all key/value pairs.",

	SetValue(key: Key, value: Value)
	    RAISES WriteError END
	    "Replace or add a key/value mapping.
	    Can raise WriteError if the value is read-only.",

	DeleteKey(key: Key)
	    RAISES KeyError, WriteError END
	    "Deletes a key/value mapping.
	    Can raise WriteError if the key is non-deleteable
	    (read-only).  Can raise KeyError if the key is not in the
	    mapping."

    END;
