"""Bastionification utility.

A bastion (for another object -- the 'original') is an object that has
the same methods as the original but does not give access to its
instance variables.  Bastions have a number of uses, but the most
obvious one is to provide code executing in restricted mode with a
safe interface to an object implemented in unrestricted mode.

The bastionification routine has an optional second argument which is
a filter function.  Only those methods for which the filter method
(called with the method name as argument) returns true are accessible.
The default filter method returns true unless the method name begins
with an underscore.

There are a number of possible implementations of bastions.  We use a
'lazy' approach where the bastion's __getattr__() discipline does all
the work for a particular method the first time it is used.  This is
usually fastest, especially if the user doesn't call all available
methods.  The retrieved methods are stored as instance variables of
the bastion, so the overhead is only occurred on the first use of each
method.

Detail: the bastion class has a __repr__() discipline which includes
the repr() of the original object.  This is precomputed when the
bastion is created.

"""


from types import MethodType, UnboundMethodType, InstanceType, \
     BuiltinFunctionType


class BastionClass:

    """Helper class used by the Bastion() function.

    You could subclass this and pass the subclass as the bastionclass
    argument to the Bastion() function, as long as the constructor has
    the same signature (a get() function and a name for the object).

    """

    def __init__(self, get, name):
        """Constructor.

        Arguments:

        get - a function that gets the attribute value (by name)
        name - a human-readable name for the original object
               (suggestion: use repr(object))

        """
        self._get_ = get
        self._name_ = name

    def __repr__(self):
        """Return a representation string.

        This includes the name passed in to the constructor, so that
        if you print the bastion during debugging, at least you have
        some idea of what it is.

        """
        return "<Bastion for %s>" % self._name_

    def __getattr__(self, name):
        """Get an as-yet undefined attribute value.

        This calls the get() function that was passed to the
        constructor.  The result is stored as an instance variable so
        that the next time the same attribute is requested,
        __getattr__() won't be invoked.

        If the get() function raises an exception, this is simply
        passed on -- exceptions are not cached.

        """
        attribute = self._get_(name)
        self.__dict__[name] = attribute
        return attribute

def defaultFilter(name):
    return name[:1] != '_'

def Bastion(object, filter=None, name=None, bastionclass=None):
    """Create a bastion for an object, using an optional filter.

    See the Bastion module's documentation for background.

    Arguments:

    object - the original object
    filter - a predicate that decides whether a function name is OK;
             by default all names are OK that don't start with '_'
    name - the name of the object; default repr(object)
    bastionclass - class used to create the bastion; default BastionClass

    """

    filter = filter or defaultFilter
    name = name or `object`
    bastionclass = bastionclass or BastionClass

    # Note: we define *two* ad-hoc functions here, get1 and get2.
    # Both are intended to be called in the same way: get(name).
    # It is clear that the real work (getting the attribute
    # from the object and calling the filter) is done in get1.
    # Why can't we pass get1 to the bastion?  Because the user
    # would be able to override the filter argument!  With get2,
    # overriding the default argument is no security loophole:
    # all it does is call it.
    # Also notice that we can't place the object and filter as
    # instance variables on the bastion object itself, since
    # the user has full access to all instance variables!

    def get1(name, object=object, filter=filter):
        """Internal function for Bastion().  See source comments."""
        if filter(name):
            attribute = getattr(object, name)
            if type(attribute) in (MethodType, BuiltinFunctionType):
                return attribute
##            print "filter allows, by type doesn't", name, attribute
##            print "type:", type(attribute)
        raise AttributeError, name

    def get2(name, get1=get1):
        """Internal function for Bastion().  See source comments."""
        return get1(name)

    return bastionclass(get2, name)

class innerMethodWrapper:
    """Wraps some callable object to allow inspection of return value

    Must use a callable instance to be fully general, i.e. to wrap a
    callable object that uses any combination of *args and **keyword
    args and default values.   Pass the actual callable object into
    the constructor.

    NOTE: You can't pass one of these instance into a restricted
    environment because the restricted code could get at the method
    attribute.  You *must* pass a Bastion.
    """
    
    def __init__(self, method):
        self.method = method
    def __call__(self, *args, **kw):
        try:
            ret = apply(self.method, args, kw)
        except Exception, exc:
            # XXX need to inspect exception too!
            print "exception occurred", exc
            raise
        else:
            if type(ret) == InstanceType:
                ret = StickyBastion(ret)
            return ret

def methodWrapperFilter(name):
    return name == "__call__"

def CallableMethodWrapper(method):
    """Make Bastion that allows access only to __call__"""
    inst = innerMethodWrapper(method)
    bast = Bastion(inst, methodWrapperFilter)
    return bast

class StickyBastionClass(BastionClass):
    """A variant of Bastion that wraps instance return values

    Consider this class:
    class Foo:
        ...
        def copy(self):
            return Foo(...)

    If we create a Bastion of an instance of class Foo, a user of the
    bastion will be able to retreive an un-bastion-ized copy of the
    object by calling the copy method.  In general, the default
    BastionClass does nothing to handle methods that return
    instances.  As a result, trusted code must go to great lengths to
    verify that any return value is properly wrapped in a Bastion.

    The StickyBastionClass inspects all return values and converts
    returned instances into StickyBastion instances.  This method is
    not foolproof; a method may return a sequence of instances.

    XXX Another, more restrictive Bastion will probably be implemented 
    to attack this problem.
    
    """
    def __getattr__(self, name):
        attribute = self._get_(name)
        wrappedAttribute = CallableMethodWrapper(attribute)
        self.__dict__[name] = wrappedAttribute
        return wrappedAttribute

def StickyBastion(object, filter=None, name=None):
    return Bastion(object, filter, name, StickyBastionClass)

if __name__ == "__main__":
    """Test the Bastion() function."""
    class Original:
        def __init__(self, sum=0):
            self.sum = sum
        def add(self, n):
            self._add(n)
        def _add(self, n):
            self.sum = self.sum + n
        def total(self):
            return self.sum
        def copy(self):
            return Original(self.sum)
    o = Original()
    b = Bastion(o)
    sb = StickyBastion(o)
    testcode = """if 1:
    b.add(81)
    b.add(18)
    print "b.total() =", b.total()
    try:
        print "b.sum =", b.sum,
    except:
        print "inaccessible"
    else:
        print "accessible"
    try:
        print "b._add =", b._add,
    except:
        print "inaccessible"
    else:
        print "accessible"
    try:
        print "b._get_.func_defaults =", b._get_.func_defaults,
    except:
        print "inaccessible"
    else:
        print "accessible"
    try:
        print "b.copy().sum =", b.copy().sum,
    except:
        print "inaccessible"
    else:
        print "accessible"
    print "sb=",sb
    try:
        print "sb.copy().sum =", sb.copy().sum,
    except Exception, err:
        print "inaccessible"
        print err
    else:
        print "accessible"
    try:
        print "sb.copy().total() =", sb.copy().total(),
    except Exception, err:
        print "inaccessible"
        print err
    else:
        print "accessible"
    \n"""
    exec testcode
    print '='*20, "Using rexec:", '='*20
    import rexec
    r = rexec.RExec()
    m = r.add_module('__main__')
    m.b = b
    m.sb = sb
    r.r_exec(testcode)

