"""Event manager for KP animation.
"""

import time
import KP
import App
from Constants import *

# scale is in number of steps... should be the same length as
# Constants.KP_AGE_COLORS
## AGE_SCALE = [ 2,  4,  8,  15, 25,
##              40, 60, 85, 115, 150,
##               ]
AGE_SCALE = [20] * len(KP_AGE_COLORS)



class EventManager:
    """Manages all events."""
    def __init__(self):
        self.__events_by_kp = {}
        self.__aging_kps = {}
        App.EVENTMGR = self

    def idle_p(self):
        return not len(self.__events_by_kp)

    def step(self):
        for kpname, eventobj in self.__events_by_kp.items():
            status = eventobj.step()
            if not status:
                eventobj.close()
                del self.__events_by_kp[kpname]
                # add it to the list of aging KPs
                try:
                    kp = App.KOE.get_kp_by_name(kpname)
                except KeyError:
                    # KP has been destroyed
                    pass
                else:
                    self.__aging_kps[kp] = AGE_SCALE[:]
        # age all KPs
        for kp, scale in self.__aging_kps.items():
            step = scale[0] - 1
            if step == 0:
                kp.age()
                del scale[0]
                if not scale:
                    del self.__aging_kps[kp]
            else:
                scale[0] = step


    def add_event(self, kpname, eventobj=None):
        """Add, replace, or delete an animation event object for a KP.
        With default eventobj value, remove it.
        """
        # reset the KOE scan alarm
        App.CONTROLLER.reset_scan_alarm()
##      print 'add_event()...', int(time.time())
        try:
            # speed animate the event, because we're going to replace it
            oldeventobj = self.__events_by_kp[kpname]
            while oldeventobj.step():
                App.CANVAS.update_idletasks()
        except KeyError:
            pass
        if eventobj:
            self.__events_by_kp[kpname] = eventobj
        else:
            if self.__events_by_kp.has_key(kpname):
                del self.__events_by_kp[kpname]



class NullEvent:
    def __init__(self, kpname):
        self._kpname = kpname

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self._kpname)

    def close(self):
        pass

    def step(self):
        return 0


class ArrowEvent(NullEvent):
    """Draws an arrow between a KOS and a KP.
    Base class for Birth and Migrate events.
    """
    _DIRECTION = 'last'                 # protected
    _ARROWCOLOR = 'firebrick'
    _wait_ticks = 7

    def __init__(self, kpname, source):
        NullEvent.__init__(self, kpname)
        # protected members
        self._source = source
        self._tag = 'ARROW:' + `time.time()`
        # private members
        self.__step = 0
    
    def _render_kp(self):
        """Render the KP and return a KP.KP object."""
        raise RuntimeError, 'not implemented'

    def _render_arrow(self, x0, y0, x1, y1):
        App.CANVAS.create_line(
            x0, y0, x1, y1,
            arrow=self._DIRECTION,
            fill=self._ARROWCOLOR,
            width=3,
            tags=(self._tag, 'ARROW'))

    def step(self):
        rtnval = 1
        # there are only two steps to this animation
        self.__step = self.__step + 1
        if self.__step == 1:
            # draw the KP at the next available KOS slot
            # draw the arrow from the KOS triangle to the KP
            kp = self._render_kp()
            x0, y0 = self._get_source_center(kp)
            x1, y1 = kp.get_center()
            self._render_arrow(x0, y0, x1, y1)
        elif self.__step in range(2, self._wait_ticks):
            # keep the arrow around a little longer
            pass
        elif self.__step == self._wait_ticks:
            App.CANVAS.delete(self._tag)
            rtnval = 0
        else:
            rtnval = 0
        return rtnval

    def _get_source_center(self, kp):
        return self._source.get_center()



class BirthEvent(ArrowEvent):
    """Animates migration to the same KOS."""
    def _render_kp(self):
        kp = App.KOE.get_kp_by_name(self._kpname)
        xs, ys = self._source.slot_xy(self._kpname)
        kp.render_at(xs, ys)
        self._source.mark_kp_as_owned(kp)
        return kp

##     def _get_source_center(self, kp):
##         # return a point on the left or top border at a 45 degree angle
##         p = kp.get_center()
##         if p is None:
##             return 0, 0
##         x, y = p
##         if x < y: return 0, y-x
##         else: return x-y, 0



class MagicEvent:
    _ARROWCOLOR = 'blue'
    def __init__(self):
##      print self
##     def __repr__(self):
##      return '*****WARNING: ' + self.__class__.__name__
        pass

class MagicBirthEvent(MagicEvent, BirthEvent):
    """For KP creation that never saw a KPMigrationReceived."""
    def __init__(self, kpname, source):
        MagicEvent.__init__(self)
        BirthEvent.__init__(self, kpname, source)



class MigrateNowhereEvent(ArrowEvent):
    """For when a KP migrates to the KOS it's already at."""
    _DIRECTION = 'both'
    def _render_kp(self):
        kp = App.KOE.get_kp_by_name(self._kpname)
        kp.set_age(0)
        return kp



class CloneEvent(ArrowEvent):
    """Animates a clone migration."""
    _wait_ticks = 10
    clones_by_parent = {}

    def __init__(self, kpname, parent):
        ArrowEvent.__init__(self, kpname, parent)
        CloneEvent.clones_by_parent[parent.get_name()] = self

    def _render_kp(self):
        kp = App.KOE.get_kp_by_name(self._kpname)
        kos = App.KOE.get_kos_for_kp(self._kpname)
        xs, ys = kos.slot_xy(self._kpname)
        kp.render_at(xs, ys)
        kos.mark_kp_as_owned(kp)
        # do *not* do this earlier!
        kp.set_age(0)
        return kp

    def _render_arrow(self, x0, y0, x1, y1):
        # REF: Bowyer & Woodwark, _A Programmer's Geometry_,
        # sec 1.7, pp20 - 21
        #
        # t = 0 gives center of vector from (x0,y0) to (x1,y1)
        # t > 0 gives arc to left of vector
        # t < 0 gives arc to right of vector
        t = 0.5
        xS = (x0 + x1)/2.0 - (y1 - y0) * t
        yS = (y0 + y1)/2.0 + (x1 - x0) * t
        tags = (self._tag, 'ARROW')
        App.CANVAS.create_line(x0, y0, xS, yS, x1, y1,
                               width=3,
                               fill='#E08080',
                               smooth=1,
                               arrow='last',
                               tags=tags)


class DeathEvent(NullEvent):
    def __init__(self, kpname, kos):
        NullEvent.__init__(self, kpname)
        self.__kos = kos
        self.__kp = App.KOE.get_kp_by_name(kpname)
        self.__step = 0

    def step(self):
        if self.__step == 0:
            self.__kos.unmark_kp_as_owned(self.__kp)
            self.__kp.close()
            self.__step = 1
        return 0


class MagicDeathEvent(MagicEvent, DeathEvent):
    def __init__(self, kpname, kos):
        MagicEvent.__init__(self)
        DeathEvent.__init__(self, kpname, kos)



class MigrationEvent(NullEvent):
    """Standard KPMigrated event"""
    _ARROWCOLOR = 'firebrick'

    def __init__(self, kpname, kos_a, kos_b):
        NullEvent.__init__(self, kpname)
        self.__kos_a = kos_a
        self.__kos_b = kos_b
        self.__step = 0
        self.__tag = 'ARROW:' + `time.time()`
        self.__lastkp = None

    def _render_arrow(self, x0, y0, x1, y1):
        App.CANVAS.create_line(
            x0, y0, x1, y1,
            arrow='last',
            fill=self._ARROWCOLOR,
            width=3,
            tags=(self.__tag, 'ARROW'))

    def step(self):
        rtnval = 1
        # there are 5 steps to this animation
        self.__step = self.__step + 1
        if self.__step == 1:
            # 1. KP@A is already drawn, move it to B, but hide it
            kp_b = App.KOE.get_kp_by_name(self._kpname)
            x0, y0 = kp_b.get_origin()
            xs, ys = self.__kos_b.slot_xy(kp_b.get_name())
            kp_b.move(xs, ys)
            kp_b.hide()
            # 2. Draw new KP@A (soon to be destroyed)
            kp_a = KP.KP(self._kpname, kp_b.get_age())
            kp_a.render_at(x0, y0)
            # 3. Draw line between KP@A and KP@B
            xa, ya = kp_a.get_center()
            xb, yb = kp_b.get_center()
            self._render_arrow(xa, ya, xb, yb)
            # 4. housekeep
            self.__xa = xa
            self.__ya = ya
            self.__xb = xb
            self.__yb = yb
            self.__lastkp = kp_a
            self.__kp_b = kp_b
            self.__kos_a.unmark_kp_as_owned(kp_a)
            self.__kos_b.mark_kp_as_owned(kp_b)
        elif self.__step in (2, 3, 4):
            # add a new KP 1/4 of the way to the destination and
            # remove the KP that's 1/4 of the way back
            delta = (self.__step - 1) * 0.25
            xt = self.__xa + (self.__xb - self.__xa) * delta
            yt = self.__ya + (self.__yb - self.__ya) * delta
            kp = KP.KP(self._kpname, self.__kp_b.get_age())
            kp.render_at(xt - KP_ICON_WIDTH * 0.5,
                         yt - KP_ICON_HEIGHT * 0.5)
            App.CANVAS.delete(self.__tag)
            self._render_arrow(xt, yt, self.__xb, self.__yb)
            self.__lastkp.close()
            self.__lastkp = kp
        elif self.__step == 5:
            # remove the last transitory KP, show the destination KP,
            # and remove the arrow
            self.__kp_b.show()
            self.__lastkp.close()
            self.__lastkp = None
            App.CANVAS.delete(self.__tag)
            self.__kp_b.set_age(0)
            rtnval = 0
        else:
            rtnval = 0
        return rtnval



class MagicMigrationEvent(MagicEvent, MigrationEvent):
    """For migrations that never received a KPMigrationReceived event."""
    def __init__(self, kpname, kos_a, kos_b):
        MagicEvent.__init__(self)
        MigrationEvent.__init__(self, kpname, kos_a, kos_b)
