"""This class represents the KOE.

It watches the namespace for changes in the set of reachable KOSes and
manages the tracking of KPs as they migrate among the various KOSes.

"""

import ilu
import VKOS
import KP
import Visualizer
import Events
import App
from Constants import *
from koe.common import nstools


_DEBUGGING = 0



class KOE:
    __scanning = 0

    def __init__(self):
        # primary data structure that tracks state.
        # Key: KOSname
        # Value: (VisualizerID, AdminSurr, VKOS.KOS, [KPnamelist])
        self.__koses = {}
        # this is a parallel data structure that tracks KP location
        # Key: KPname
        # Value: (KP, KOSname)
        self.__kps = {}
        App.KOE = self

    #
    # The Visualizer, which serves double duty as the
    # ReportingStation, will inform us about changes in KP state when
    # it receives KPTerminated and KPMigrationReceived messages.  We
    # need to update our internal database when these occur.
    #
    def kp_terminated(self, kpname):
        try:
            kp, kosname = self.__kps[kpname]
        except KeyError:
            print 'Lost KP was terminated:', kpname
        else:
            visId, admin, kos, kplist = self.__koses[kosname]
            kplist[kplist.index(kpname)] = None
            del self.__kps[kpname]
            kp.close()

    def kp_migration_received(self, kpname, kosname):
        try:
            kp, srckosname = self.__kps[kpname]
            visId, admin, kos, kplist = self.__koses[srckosname]
            kplist[kplist.index(kpname)] = None
        except KeyError:
            # first time we've heard of this KP
            kp = KP.KP(kpname)
            self.__kps[kpname] = (kp, kosname)
        visId, admin, kos, kplist = self.__koses[kosname]
        try:
            kplist[kplist.index(None)] = kpname
        except ValueError:
            kplist.append(kpname)
        self.__kps[kpname] = (kp, kosname)

    def get_kp_by_name(self, kpname):
        kp, kosname = self.__kps[kpname]
        return kp

    def get_kos_for_kp(self, kpname):
        kp, kosname = self.__kps[kpname]
        visId, admin, kos, kplist = self.__koses[kosname]
        return kos

    def get_kos_by_name(self, kosname):
        visId, admin, kos, kplist = self.__koses[kosname]
        return kos

    def get_kpnames_for_kos(self, kosname):
        visId, admin, kos, kplist = self.__koses[kosname]
        return kplist

    def get_koses(self):
        koses = []
        for visId, admin, kos, kplist in self.__koses.values():
            koses.append(kos)
        return koses

    #
    # periodic scanning
    #

    def __register(self, kosname, kos):
        admin = visId = None
        try:
            admin = nstools.WorldOpen('kos/%s/administrator' % kosname)
            visId = admin.RegisterVisualizer(App.RS)
        except nstools.BadPathError:
            kos.markdead()
        return admin, visId

    def scan(self):
        # only scan if the event manager is idle
        if not App.EVENTMGR.idle_p():
            return
        # prevent recursive scanning
        if self.__scanning:
            return
        self.__scanning = 1
        self.__do_scan()
        self.__scanning = 0

    def __do_scan(self):
##      import time
##      print '__do_scan()...', int(time.time())
        kostokps = {}
        # get a snapshot of the world of KOSs and KPs
        kosns = nstools.WorldOpen('kos')
        for kosname in kosns.List():
            try:
                kostokps[kosname] = []
                kpsns = nstools.Lookup(kosns, '%s/kps' % kosname).Open()
                for kpname in kpsns.List():
                    kostokps[kosname].append(kpname)
            except nstools.BadPathError:
                kostokps[kosname] = 0
        layout = 0
        # verify that all KOSs we know about are still alive and in
        # the namespace
        koslist = kostokps.keys()
        for kosname, (visId, admin, kos, kplist) in self.__koses.items():
            markdead = 0
            if kosname not in koslist:
                # KOS has recently been removed from the namespace.
                # totally clean up its tracks.
                for kpname in filter(None, kplist):
                    kp, kosname = self.__kps[kpname]
                    kp.close()
                    del self.__kps[kpname]
                kos.close()
                del self.__koses[kosname]
                layout = 1
            elif kostokps[kosname] is 0:
                # still in the namespace, but not responding
                markdead = 1
            elif not admin:
                # a KOS we aren't registered with yet
                admin, visId = self.__register(kosname, kos)
                self.__koses[kosname] = (visId, admin, kos, [])
                markdead = not admin
            else:
                # a perfectly healthy KOS
                pass
            # if the KOS is to be marked as dead, do so and update the
            # database.
            if markdead:
                self.__mark_kos_as_dead(kos, kplist)
            else:
                kos.marklive()
        # Look for KOSes that we didn't know about before
        for kosname in koslist:
            if not self.__koses.has_key(kosname):
                kos = VKOS.KOS(kosname)
                admin, visId = self.__register(kosname, kos)
                self.__koses[kosname] = (visId, admin, kos, [])
                layout = 1
        for visId, admin, kos, kpnames in self.__koses.values():
            # only scans for new plugins
            kos.scan()
        #
        # Make sure our internal view of KP ownership by KOSes is
        # consistent with the world view.  This should always be the
        # case since we should receive KPMigrationReceived or
        # KPTerminated messages for all state changes reflected in the
        # namespace.  Most common case is when we start up and there
        # are already KPs running in KOSs
        #
        kptokos = {}
        for kosname in koslist:
            kplist = kostokps[kosname]
            if kplist is 0:
                continue
            for kpname in kplist:
                kptokos[kpname] = kosname
                e = None
                if self.__kps.has_key(kpname):
                    kp, kpskosname = self.__kps[kpname]
                    if kpskosname <> kosname:
                        # the KP has `magically' migrated to a different KOS
                        e = Events.MagicMigrationEvent(
                            kpname,
                            self.get_kos_by_name(kpskosname),
                            self.get_kos_by_name(kosname))
                else:
                    # a KP that has been `magically' birthed
                    e = Events.MagicBirthEvent(
                        kpname,
                        self.get_kos_by_name(kosname))
                if e:
                    App.EVENTMGR.add_event(kpname, e)
                    App.KOE.kp_migration_received(kpname, kosname)
        # find `magically' lost KPs
        kplist = kptokos.keys()
        for kpname, (kp, kosname) in self.__kps.items():
            if kpname not in kplist:
                e = Events.MagicDeathEvent(kpname,
                                           self.get_kos_by_name(kosname))
                App.EVENTMGR.add_event(kpname, e)
                self.kp_terminated(kpname)
        #
        # Shuffle KP's in their KOS display to get rid of empty slots
        #
        pass                            # TBD
        #
        # update the display's layout and we're done
        #
        if layout:
            App.VISUALIZER.layout_koses()

    def __mark_kos_as_dead(self, kos, kplist):
        kos.markdead()
        for kpname in filter(None, kplist):
            kp, kosname = self.__kps[kpname]
            kp.close()
            del self.__kps[kpname]
        self.__koses[kos.get_name()] = (None, None, kos, [])

    def close(self):
        for kosname, (visId, admin, group, kplist) in self.__koses.items():
            try:
                admin.UnregisterVisualizer(visId)
            except (AttributeError, ilu.IluGeneralError):
                pass
