"""Visualizer application object."""

from Tkinter import *
from types import *
import math
import string
import time
from koe.common import tktools

# We don't need ilu_tk if we're an applet
try:
    import ilu_tk
except ImportError:
    ilu_tk = None
            
import ilu
from koe.common import nstools
from koe.common import access

import Events
import KP
import App
from Constants import *
from Alarm import Alarm

from koe.interfaces.stubs import ReportingAPI, ReportingAPI__skel



class Visualizer(ReportingAPI__skel.Visualizer):
    # okay to share these since Visualizer is a singleton
    #
    __alarm = None

    INTERVAL = 5.0

    HORIZADJUST = 80
    VERTADJUST = 30
    RADIUSADJUST = 100
    R_HORIZADJUST = 60
    R_VERTADJUST = VERTADJUST

    def __init__(self, width=500, height=500, master=None):
        if master:
            self.__master = master
            self.__in_applet = 1
        else:
            self.__master = Tk(className='Visualizer')
            self.__master.protocol('WM_DELETE_WINDOW', self.stop)
            self.__master.title('KOE Visualizer')
            self.__master.iconname('Visualizer')
            self.__in_applet = 0
        #
        # create the menubar
        #
        menubar = Frame(self.__master, relief=RAISED, borderwidth=2)
        menubar.pack(fill=X)
        #
        # file menu
        #
        filebtn = Menubutton(menubar, text='File')
        if not self.__in_applet: # Don't need an exit button as an applet
            filebtn.pack(side=LEFT)
        filemenu = Menu(filebtn, tearoff=0)
        filebtn['menu'] = filemenu
        filemenu.add_command(label='Quit',
                             command=self.stop,
                             accelerator='Alt-Q')
        self.__master.bind('<Alt-q>', self.stop)
        self.__master.bind('<Alt-Q>', self.stop)
        #
        # options menu
        #
        optbtn = Menubutton(menubar, text='Layout')
        optbtn.pack(side=LEFT)
        optmenu = Menu(optbtn)
        optbtn['menu'] = optmenu
        optmenu.add_command(label='Lock all',
                            command=self.__lockallcmd,
                            accelerator='Alt-L')
        self.__master.bind('<Alt-l>', self.__lockallcmd)
        self.__master.bind('<Alt-L>', self.__lockallcmd)
        optmenu.add_command(label='Unlock all',
                            command=self.__unlockallcmd,
                            accelerator='Alt-U')
        self.__master.bind('<Alt-u>', self.__unlockallcmd)
        self.__master.bind('<Alt-U>', self.__unlockallcmd)
        optmenu.add_separator()
        optmenu.add_command(label='Refresh Layout',
                            command=self.__resize_event,
                            accelerator='Alt-R')
        self.__master.bind('<Alt-r>', self.__resize_event)
        self.__master.bind('<Alt-R>', self.__resize_event)
        #
        # create the message bar
        #
        self.__msg = Label(self.__master, anchor=W)
        self.__msg.pack(fill=BOTH)
        #
        # create the canvas
        #
        hbar, vbar, frame = tktools.make_scrollbars(
            self.__master, hbar=1, vbar=1)
        self.__frame = Frame(frame)
        self.__frame.pack(expand=1, fill=BOTH, side=LEFT)
        self.__canvas = Canvas(self.__frame, width=width, height=height,
                               relief=SUNKEN, borderwidth=2)
        self.__canvas.pack(expand=1, fill=BOTH, side=LEFT)
        self.__canvas.bind('<Configure>', self.__resize_event)
        self.__canvas['scrollregion'] = (0, 0, width, height)
        tktools.set_scroll_commands(self.__canvas, hbar, vbar)
        #
        # set global attributes.  this are singleton objects
        #
        App.CANVAS = self.__canvas
        App.VISUALIZER = self
        App.RS = self

    def __resize_event(self, event=None):
        self.layout_koses()

    def __lockallcmd(self, event=None):
        for kos in App.KOE.get_koses():
            kos.lock()

    def __unlockallcmd(self, event=None):
        for kos in App.KOE.get_koses():
            kos.unlock()

    #
    # public interface
    #

    def start(self):
        if not self.__in_applet:
            ilu_tk.RunMainLoop()

    def stop(self, event=None):
        if not self.__in_applet:
            ilu_tk.ExitMainLoop()

    def __clear_msg(self):
        self.__msg.config(text='')
        self.__alarm.close()

    def message(self, msg):
        self.__msg.config(text=msg)
        if self.__alarm:
            self.__alarm.close()
        else:
            self.__alarm = Alarm(self.INTERVAL)
            self.__alarm.register(self.__clear_msg)
        self.__alarm.arm()

    def layout_koses(self):
        """Layout algorithm."""
        koses = []
        for kos in App.KOE.get_koses():
            if not kos.get_lock():
                koses.append(kos)
        koscnt = len(koses)
        #
        # no koses... nothing to do!
        #
        if koscnt == 0:
            return
        koses.sort()
        #
        # get geometry of canvas.  force window sizes to be true
        #
        self.__master.update()
        canvas_width = self.__frame.winfo_width()
        canvas_height = self.__frame.winfo_height()
        #
        # For one KOS, center it in the canvas
        #
        if koscnt == 1:
            kos = koses[0]
            x, y = kos.get_centroid()
            if not kos.get_lock():
                kos.move((canvas_width-x)*0.5, (canvas_height-y)*0.5)
        #
        # for two KOSs, use this algorithm: find the longest axis,
        # divide it into three parts 25%, 50%, 25%.  Center kos 1
        # along the first division, kos 2 along the second.
        #
        elif koscnt == 2:
            if canvas_width > canvas_height:
                # horizontal alignment
                cx1 = canvas_width * 0.25 - self.HORIZADJUST
                cx2 = canvas_width * 0.75 - self.HORIZADJUST
                cy = canvas_height * 0.5 - self.VERTADJUST
                #
                if not koses[0].get_lock():
                    x, y = koses[0].get_centroid()
                    koses[0].move(cx1-(x*0.5), cy-(y*0.5))
                if not koses[1].get_lock():
                    x, y = koses[1].get_centroid()
                    koses[1].move(cx2-(x*0.5), cy-(y*0.5))
            else:
                # vertical alignment
                cy1 = canvas_height * 0.25 - self.VERTADJUST
                cy2 = canvas_height * 0.75 - self.VERTADJUST
                cx = canvas_width * 0.5 - self.HORIZADJUST
                #
                if not koses[0].get_lock():
                    x, y = koses[0].get_centroid()
                    koses[0].move(cx-(x*0.5), cy1-(y*0.5))
                if not koses[1].get_lock():
                    x, y = koses[1].get_centroid()
                    koses[1].move(cx-(x*0.5), cy2-(y*0.5))
        else:
            # for all other possibilities, arrange in a rough circle
            # around the outer edge of the canvas.  we need to
            # subtract a bit of room off the right side for KPs which
            # tend to be fairly long.  also subtract some room off the
            # bottom to account for KOSs that have lots of KPs.  this
            # is all magic.
            #
            cx = canvas_width * 0.5
            cy = canvas_height * 0.5
            radius = min(cx, cy) - self.RADIUSADJUST
            angle = -90.0
            delta = 360.0 / koscnt
            for kos in koses:
                rads = angle * math.pi / 180.0
                x, y = kos.get_centroid()
                tx = radius * math.cos(rads)
                ty = radius * math.sin(rads)
                if not kos.get_lock():
                    kos.move(tx + cx - x - self.R_HORIZADJUST,
                             ty + cy - y - self.R_VERTADJUST)
                angle = angle + delta


    # interface for KOS

    #
    #  Reporting Station API
    #

    # TBD and XXX: This is flawed.  We should not be assigning any
    # semantics to the reportingID, nor to the name the KP is bound to
    # in the namespace.  Instead, the KP should have an interface
    # whereby it can be queried for its globally unique pure name

    def KPTerminated(self, reportingID, reason, details, kprepr):
##      self.message('KPTerminated: %s %s %s %s' %
##                   (reportingID, reason, details, kprepr))
        kpname = reportingID
        try:
            kos = App.KOE.get_kos_for_kp(kpname)
        except KeyError:
            self.message('KPTerminated for phantom KP: ' + reportingID)
        else:
            e = Events.DeathEvent(kpname, kos)
            App.EVENTMGR.add_event(kpname, e)
            App.KOE.kp_terminated(kpname)

    def KPMigrationReceived(self, reportingID, idTo, migratingTo):
        # we ignore the idTo since it will be different for each KOS
        # and we can't string match on that part of the name.  TBD:
        # seems slightly bogus that the id'd name of the KP is in the
        # namespace.
##      self.message('KPMigrationReceived: %s %s %s' %
##                   (reportingID, idTo, migratingTo))
        kpname = reportingID
        kosname = migratingTo
        e = self.__getevent(kpname, kosname)
        App.EVENTMGR.add_event(kpname, e)
        App.KOE.kp_migration_received(kpname, kosname)

    def __getevent(self, kpname, kosname):
        targetkos = App.KOE.get_kos_by_name(kosname)
        try:
            sourcekos = App.KOE.get_kos_for_kp(kpname)
        except KeyError:
            # KP is newly birthed since no KOS is collecting it yet.
            # see if it's a clone or a traditional birth
            index = string.rfind(kpname, '.')
            if index >= 0:
                parentname = kpname[:index]
                try:
                    parentkp = App.KOE.get_kp_by_name(parentname)
                except KeyError:
                    # traditional birth
                    e = Events.BirthEvent(kpname, targetkos)
                else:
                    e = Events.CloneEvent(kpname, parentkp)
        else:
            # this is a migration
            if sourcekos == targetkos:
                # must be migrating to the KOS it is already at
                e = Events.MigrateNowhereEvent(kpname, targetkos)
            else:
                # migration to a different KOS
                e = Events.MigrationEvent(kpname, sourcekos, targetkos)
        return e

    def Shutdown(self):
        # Insure requestor is authorized
        try:
            if access.ilu_allow_access('Visushutdown'):
                if not self.__in_applet:
                    ilu_tk.ExitMainLoop()
        except access.ResourceError:
            print 'Error obtaining access information!'

