"""Classes for SGML/XML parsers that build outlines of input documents."""

__version__ = '$Revision: 1.2 $'

import sgmllib
import string
import xmllib


ALL_CONTENTS = "ALL"


class OutlineError(Exception):
    def __init__(self, msg):
        self.msg = msg
        Exception.__init__(self, (msg,))


def outline(fp, ctype, contentsof=()):
    """Return outline for the input stream according to its content-type.

    The content of selected elements may be stored by passing in the list of
    general identifiers as `contentsof'.  If instead the constant ALL_CONTENTS
    is passed, the contents of all elements will be stored in the outline
    structure.
    """
    #
    # This can't work for text/sgml or application/sgml since
    # the set of empty tags isn't known by these simple parsers.
    #
    if ctype == "text/html":
        parser = HTMLOutliner(contentsof=contentsof)
    elif ctype in ("text/xml", "application/xml"):
        parser = XMLOutliner(contentsof=contentsof)
    else:
        raise OutlineError("unsupported MIME type")
    parser.feed(fp.read())
    parser.close()
    return parser.get_outline()



class Outliner:
    """Mixin class to support outlining functionality.  Works with the
    standard SGML and XML parser classes.

    When inheriting from this class, it must preceed the standard parser
    base classes to allow it to override some of the standard methods.

    The combining class must provide the method _make_attrs_dict(), which
    should accept an attributes object as provided by the parser class
    and convert it to a dictionary.  It is an error to not define this
    method.
    """

    __closed = 0
    __discard = 0
    __keeping = 0
    __tag = None
    __flowing = 0

    def unknown_starttag(self, tag, attrs):
        self.__push(tag, attrs)
        if self.__empty_tag(tag):
            self.__pop()

    def unknown_endtag(self, tag):
        if self.close_until([tag]):
            self.__pop()
        else:
            print "unexpected closing tag: </%s>" % tag

    def handle_data(self, data):
        if self.__keeping:
            if self.__context and isinstance(self.__context[-1], type('')):
                self.__context[-1] = self.__context[-1] + data
            else:
                self.__context.append(data)
            if self.__flowing:
                data = self.__context[-1]
                prespace = (data[0] in string.whitespace) and ' ' or ''
                words = string.split(data)
                postspace = (words and data[-1] in string.whitespace) \
                            and ' ' or ''
                data = "%s%s%s" % (prespace, string.join(words), postspace)
                self.__context[-1] = data

    # outliner-specific methods for clients:

    def get_outline(self):
        return self.__outline

    # outliner-specific methods for subclasses:

    def init_outliner(self, emptytags=(), contentsof=()):
        self.__empty_tag = self.__make_set_predicate(emptytags)
        self.__keep_contents = self.__make_set_predicate(contentsof)
        self.__context = self.__outline = []
        self.__stack = []

    def close_outliner(self):
        while self.__stack:
            self.__pop()

    def close_until(self, tags):
        if self.__tag in tags:
            return 1
        if self.__stack and self.__stack[-1][0] in tags:
            self.finish_endtag(self.__tag)
            return 1
        indexes = range(len(self.__stack))
        indexes.reverse()
        for i in indexes:
            if self.__stack[i][0] in tags:
                self.finish_endtag(self.__stack[i + 1][0])
                return 1
        return 0

    def get_current_context(self):
        return self.__context

    def get_context(self):
        context = []
        elem = self.__outline[-1]
        while 1:
            context = elem[0]
            try:
                context = elem[-1][-1]
            except IndexError:
                return filter(None, context)

    def set_flowing(self, flowing):
        self.__flowing = flowing

    def discard(self, tag=None):
        if tag in (None, self.__tag):
            self.__discard = 1
            return 1
        indexes = range(len(self.__stack))
        indexes.reverse()
        for i in indexes:
            if self.__stack[i][0] == tag:
                self.__stack[i][2] = 1
                return 1
        return 0

    # internal methods:

    def __push(self, tag, attrs):
        context = []
        self.__context.append((tag, self._make_attrs_dict(attrs), context))
        self.__stack.append([self.__tag, self.__keeping, self.__discard,
                             self.__flowing, self.__context])
        self.__tag = tag
        self.__keeping = self.__keeping or self.__keep_contents(tag)
        self.__context = context
        self.__discard = 0

    def __pop(self):
        discard = self.__discard
        stuff = self.__stack[-1]
        self.__tag, self.__keeping, self.__discard, \
                    self.__flowing, self.__context = stuff
        del self.__stack[-1]
        if discard:
            del self.__context[-1]

    def __make_set_predicate(self, set):
        d = {}
        if set == ALL_CONTENTS:
            self.__keeping = 1
        else:
            for o in set:
                d[o] = o
        return d.has_key

    def _make_attrs_dict(self, attrs):
        raise OutlineError(
            "Outliner subclass failed to implement _make_attrs_dict().")



class SGMLOutliner(Outliner, sgmllib.SGMLParser):
    """Outlining parser for SGML data."""

    def __init__(self, emptytags=(), contentsof=()):
        self.init_outliner(emptytags, contentsof)
        sgmllib.SGMLParser.__init__(self)

    def close(self):
        sgmllib.SGMLParser.close(self)
        self.close_outliner()

    def _make_attrs_dict(self, attrlist):
        d = {}
        for k, v in attrlist:
            d[k] = v
        return d


class HTMLOutliner(SGMLOutliner):
    """Outlining parser for HTML data."""

    from htmlentitydefs import entitydefs

    def __init__(self, contentsof=()):
        # empty tags taken from HTML 4.0
        emptytags = ('area', 'base', 'basefont', 'br', 'col', 'frame', 'hr',
                     'img', 'input', 'isindex', 'link', 'meta', 'param')
        SGMLOutliner.__init__(self, emptytags=emptytags, contentsof=contentsof)
        self.set_flowing(1)

    def start_li(self, attrs):
        self.close_until(['ol', 'ul'])
        self.unknown_starttag('li', attrs)

    def start_dd(self, attrs):
        self.close_until(['dl'])
        self.unknown_stardtag('dd', attrs)

    def start_dt(self, attrs):
        self.close_until(['dl'])
        self.unknown_starttag('dt', attrs)

    def start_pre(self, attrs):
        self.unknown_starttag('pre', attrs)
        self.set_flowing(0)

    def start_listing(self, attrs):
        self.unknown_starttag('listing', attrs)
        self.set_flowing(0)

    def start_plaintext(self, attrs):
        self.unknown_starttag('plaintext', attrs)
        self.set_flowing(0)


class XMLOutliner(Outliner, xmllib.XMLParser):
    """Outlining parser for XML data."""

    def __init__(self, contentsof=()):
        self.init_outliner((), contentsof)
        xmllib.XMLParser.__init__(self)

    def close(self):
        xmllib.XMLParser.close(self)
        self.close_outliner()

    def _make_attrs_dict(self, attrlist):
        return attrlist



def _test(files):
    import mimetypes
    import pprint
    #
    if type(files) is type(''):
        files = [files]
    for filename in files:
        ctype, encoding = mimetypes.guess_type(filename)
        print "====", filename, "[%s]" % ctype
        tree = outline(open(filename), ctype)
        pprint.pprint(tree)


if __name__ == "__main__":
    import sys
    _test(sys.argv[1:])
