"""Simple inverted index -- supports boolean queries across multiple fields"""

import os
import gdbm
import string
from array import array
import types
import tempfile

CACHESIZE = 1000

class Cache:

    def __init__(self, miss, evict, size=None, writethru=0):
        if size is None:
            size = CACHESIZE
        self.cache = [None] * size
        self.size = size
        self.hits = 0
        self.misses = 0
        self.evicts = 0
        self.miss = miss
        self.evict = evict
        if writethru:
            # In writethru mode, we assume that we will seldom write
            # to the database.  In this case, we may generate spurious 
            # writes because an entry is read and not modified.
            self.set = self.set_writethru
            self.sync = self.sync_writethru

    def get(self, key):
        i = hash(key) % self.size
        entry = self.cache[i]
        if entry:
            if entry[0] == key:
                self.hits = self.hits + 1
                return entry[1]
            if entry[2]:  # if dirty
                self.evicts = self.evicts + 1
                self.evict(entry[0], entry[1])
        self.misses = self.misses + 1
        value = self.miss(key)
        self.cache[i] = (key, value, 0)
        return value

    def set(self, key, val):
        i = hash(key) % self.size
        entry = self.cache[i]
        if entry:
            if entry[0] != key and entry[2]:
                self.evicts = self.evicts + 1
                self.evict(entry[0], entry[1])
        self.cache[i] = (key, val, 1)

    def set_writethru(self, key, val):
        i = hash(key) % self.size
        t = (key, val, 0)
        self.cache[i] = t
        self.evict(key, val)
        self.evicts = self.evicts + 1

    def sync(self):
        for key, val, dirty in filter(None, self.cache):
            if dirty:
                self.evict(key, val)

    def sync_writethru(self):
        pass

    def flush(self):
        self.sync()
        self.cache = [None] * self.size

    def printStats(self):
        empty = len(filter(lambda x:x is None, self.cache))
        print "size=%d inuse=%d empty=%d%%" % (
            self.size,
            self.size-empty,
            int(100*float(empty)/self.size)),

        print "hits=%d misses=%d evicts=%d" % (
            self.hits, self.misses, self.evicts)


class GDBMWrapper:
    """Make a GDBM more convenient and faster.

    Automatically marshal objects as they are written in and read out
    of the DBM file.  Also cache recents objects.
    """

    def __init__(self, db, cachesize=CACHESIZE, writethru=0):
        self.db = db
        self.cache = Cache(self.__load, self.__store, cachesize, writethru) 

    def __getattr__(self, attr):
        return getattr(self.db, attr)

    # might want a fairly large cache so that multiple
    # writes of the same data don't involve marshal + disk

    def sync(self):
        self.cache.sync()
        self.db.sync()

    def __load(self, item):
        try:
            s = self.db[item]
        except KeyError:
            s = ""
        a = array('i', s)
        return a

    def __getitem__(self, item):
        return self.cache.get(item)

    def __store(self, item, value):
        self.db[item] = value.tostring()

    def __setitem__(self, item, value):
        return self.cache.set(item, value)

    def printStats(self):
        self.cache.printStats()

def union(lists):
    dict = {}
    for list in lists:
        for elt in list:
            dict[elt] = elt
    return dict.keys()

def intersection(lists):
    if not lists:
        return []
    dict = None
    for list in lists:
        dict2 = {}
        for elt in list:
            dict2[elt] = elt
        if dict is None:
            dict = dict2
        else:
            dict2_has_key = dict2.has_key
            for elt in dict.keys():
                if not dict2_has_key(elt):
                    del dict[elt]
        if not dict:
            break
    return dict.keys()

def difference(lists):
    dict = {}
    try:
        for elt in lists[0]:
            dict[elt] = elt
    except IndexError:
        pass
    else:
        for list in lists[1:]:
            if not dict:
                break
            for elt in list:
                if dict.has_key(elt):
                    del dict[elt]
    return dict.keys()

class Index:
    """Interface to multiple underlying DBM files.

    The index treats each entry as a dictionary, where the keys are
    field names and the values are a list of tokens.  The field named
    'id' is given special status; it is the ID of the document.
    """

    def __init__(self, name, mode, fields=None, writethru=0):
        self.name = name
        self.dbRecords = gdbm.open(name + '.RECORDS', mode)
        self.lastid = self.maxid()
        self.dbFields = {}
        if fields is None:
            fields = []
            # search thru the damn directory
            dir, file = os.path.split(name)
            root, ext = os.path.splitext(file)
            root = root + '.'
            l = len(root)
            for file in os.listdir(dir):
                if file[:l] == root:
                    if file[l:] != 'RECORDS':
                        fields.append(file[l:])

        self.fields = fields
        for field in self.fields:
            db = GDBMWrapper(gdbm.open(name + '.' + field, mode),
                             writethru=writethru)
            self.dbFields[field] = db

        # controlling the sync behavior
        self.writeable = ('w' in mode or 'c' in mode or 'n' in mode)
        self.dirty = 0

    def nextid(self):
        self.lastid = self.lastid + 1
        return self.lastid

    def maxid(self):
        keys = self.dbRecords.keys()
        if not keys:
            return 0
        return max(map(int, keys))

    def close(self):
        if self.writeable and self.dirty:
            self.sync()

    def sync(self):
        self.dbRecords.sync()
        for field, db in self.dbFields.items():
            db.sync()
        self.dirty = 0

    def updateField(self, field, token, ids=[]):
        if type(ids) == types.ListType:
            ids = array('i', ids)
        try:
            db = self.dbFields[field]
        except KeyError:
            raise KeyError, "no field: %s" % field
        try:
            list = db[token]
        except KeyError:
            list = array('i')
        list = list + ids
        db[token] = list

    def addRecord(self, id, record):
        self.dbRecords[str(id)] = record
        
    def addEntry(self, id, edict, orig):
        """Add entry to database, where entry is dict of field->tokens."""

        for k, v in edict.items():
            try:
                db = self.dbFields[k]
            except KeyError:
                # ignore field we aren't interested in
                continue
            for tok in v:
                try:
                    list = db[tok]
                except KeyError:
                    list = array('i')
                list.append(id)
                db[tok] = list
        self.dbRecords[str(id)] = str(orig)

    def search1(self, field, term):
        """Search for one term in one field.  Return an array of ids."""
        try:
            db = self.dbFields[field]
        except KeyError:
            raise ValueError, "invalid field name: %s" % `field`
        return db[term]

    def search(self, query, bool=None):
        """Query format: {field: (bool, tokens)}

        bool is one of ['and', 'or', 'andnot'] and is used when there are
        multiple search tokens for a field.  tokens is a sequence of search
        terms.

        The optional bool argument is applied to the search result
        sets for each field.  If the bool is 'and', then only those
        documents that appeared in every field are returned.
        """

        if len(query) > 1 and bool is None:
            raise ValueError, "must specify combiner for multi-field queries"
        resultSets = {}
        for field, (b, tokens) in query.items():
            assert type(tokens) != types.StringType, \
                   "bad list of terms:" + str(tokens)
            try:
                db = self.dbFields[field]
            except KeyError:
                raise ValueError, "invalid field name: %s" % `field`
            results = []

            for term in tokens:
                results.append(db[term])

            if len(results) == 1:
                resultSets[field] = results[0]
            else:
                if b == 'and':
                    resultSets[field] = intersection(results)
                elif b == 'or':
                    resultSets[field] = union(results)
                else:
                    resultSets[field] = difference(results)
        if len(resultSets) == 1:
            return resultSets.values()[0]
        if bool == 'and':
            return intersection(resultSets.values())
        elif bool == 'or':
            return union(resultSets.values())
        else:
            return difference(resultSets.values())

    def lookup(self, id):
        """Return the original text version of a record."""
        return self.dbRecords[str(id)]

class BulkIndex(Index):

    def __init__(self, name, mode, fields):
        Index.__init__(self, name, mode, fields, writethru=1)
        self.bulkfilename = tempfile.mktemp()
        self.bulk = None

    def addEntry(self, id, edict, orig):
        if not self.bulk:
            self.bulk = open(self.bulkfilename, "w")
        self.addRecord(id, str(orig))
        for field, words in edict.items():
            for word in words:
                line = "%s %s %d\n" % (field, word, id)
                self.bulk.write(line)

    def sync(self):
        if not self.bulk:
            return
        self.bulk.close()
        self.bulk = None
        os.system("sort -d -o %s %s" % (self.bulkfilename, self.bulkfilename))
        f = open(self.bulkfilename, "r")
        lastfield = lastword = lastids = None
        while 1:
            block = f.readlines(10000)
            if not block:
                break
            for line in block:
                # If this fails, someone @$%!& stomped on our temp file!
                field, word, id = string.split(line)
                id = int(id)
                if field != lastfield or word != lastword:
                    if lastids:
                        self.updateField(lastfield, lastword, lastids)
                    lastfield = field
                    lastword = word
                    lastids = array('i')
                lastids.append(id)
        f.close()
        os.unlink(self.bulkfilename)
        if lastids:
            self.updateField(lastfield, lastword, lastids)
        Index.sync(self)
