#! /usr/bin/env python

"""A very simple plugin of an image repository.

Implements the RAPK.isl interface, with terms and conditions
negotiation.  See class Cryptor for details.

"""

import os, sys, string
import getopt
import marshal
import pickle
import rotor
import random
import md5
import traceback
from koe.common import pathhack
from koe.common import pluginlib
from koe.common import kptools, RS, Suitcase
from koe.common import MappingImpl
from koe.interfaces.stubs import RAPK, RAPK__skel
from koe.interfaces.stubs import Mapping, Mapping__skel

_DEMODIR = os.path.join(pathhack.kos_root(),'demos/imagerepo')

# serves as the reporting station for all decrypt.py KPs launched by the
# Cryptor.launch_decryptor() method
_RS = RS.DebuggingRS()



def main():
    opts, args = getopt.getopt(sys.argv[1:], 'n:')
    if not opts:
        print 'KOS name is required for the ImageRepo'
        sys.exit(1)
    else:
        kosname = opts[0][1]
    kos, ns = pluginlib.register('BassImageRepository', kosname)
    #
    # Hardcode homebase to be the same as the KOS name.
    # XXX This used to list the namespace, but that doesn't work any more.
    home = kosname
    sid = kos.bind_service('BassImages', Repository(home), 'RAPK.Repository')
    kos.run()
    kos.unbind_service(sid)
    sys.exit(0)



# not currently used in the ImageRepo, but provided here as a placeholder.
class DemoManager(RAPK__skel.DemoManager):
    def NewObjects(self, name, newpath):
        pass



class Repository:
    def __init__(self, home):
        self.__home = home
        self.__iterators = {}
        self.__nextiterator = 0
        #
        # Dictionary of instantiated digital objects.  Key is the
        # `handle' used to look up metadata and Ts&Cs.  Value is the
        # digital object, instantiated as needed.
        #
        self.__dos = {}
        #
        # crack the index file and read the index.  each line of the
        # index file has the following format:
        #
        # imagename|thumbnailfile|thumbnailtype|hiresfile|hirestype|price
        #
        # pound sign at start of line is comment
        #
        self.__images = {}
        file = os.path.join(_DEMODIR, 'image.idx')
        idxfp = open(file, 'r')
        # the fastest way of doing this...
        blob = idxfp.read()
        idxfp.close()
        for line in map(string.strip, string.splitfields(blob, '\n')):
            if not line or line[0] == '#':
                continue
            img = ImageHelper(line)
            self.__images[img.name()] = img
        #
        # The Repository's metadata
        #
        self.__mapping = MappingImpl.StringMap({
            'RespositoryName': 'Bass Image Repository',
            'ImageCount'     : `len(self.__images)`,
            })

    #
    # RAPK interface
    #

    def GetRepoMetadata(self):
        # return the underlying StringMap object
        return self.__mapping

    def Scan(self):
        return self.__images.keys()

    def __getdo(self, handle):
        try:
            return self.__dos[handle]
        except KeyError:
            try:
                imgobj = self.__images[handle]
                self.__dos[handle] = DigitalObject(self.__home, handle, imgobj)
                return self.__dos[handle]
            except KeyError:
                raise RAPK.NotFound, handle

    def GetMetadata(self, handle):
        return self.__getdo(handle).GetMeta()

    def GetDO(self, handle):
        return self.__getdo(handle)

    def GetTermsAndConditions(self, handle):
        return self.__getdo(handle).GetTermsAndConditions()

    def CreateIterator(self):
        next = self.__nextiterator
        self.__nextiterator = self.__nextiterator + 1
        images = self.__images.keys()[:]
        images.sort()
        self.__iterators[next] = images
        return next

    def Iterate(self, iterator):
        try:
            images = self.__iterators[iterator]
            if not images:
                raise RAPK.LastDocument, 'for iterator %d' % iterator
            else:
                handle = images[0]
                del images[0]
                return handle
        except KeyError:
            raise RAPK.BadIterator, `iterator`



class DigitalObject(RAPK__skel.DigitalObject):
    def __init__(self, home, handle, imgobj):
        self.__home = home
        self.__imgobj = imgobj
        self.__metadata = MappingImpl.StringMap({
            'Handle'        : handle,
            'ThumbnailType' : imgobj.ttype(),
            'ThumbnailSize' : str(imgobj.tsize()),
            'HiresType'     : imgobj.htype(),
            'HiresSize'     : str(imgobj.hsize()),
            'HiresPrice'    : imgobj.price(),
            'Disseminations': string.join(['thumbnail', 'hires']),
            })
        # we don't currently implement Ts&Cs on digital objects
        self.__terms = MappingImpl.StringMap()

    def GetMetadata(self):
        return self.__metadata

    def GetTermsAndConditional(self):
        return self.__terms

    def GetDissemination(self, dtype):
        # return a new dissemination object for each call
        if dtype == 'thumbnail':
            file = self.__imgobj.tfile()
            cryptor = None
        elif dtype == 'hires':
            file = self.__imgobj.hfile()
            cryptor = Cryptor()
        else:
            raise RAPK.TypeError, dtype
        return Dissemination(self.__home, self.__imgobj, file, cryptor)



class ImageHelper:
    def __init__(self, descriptor):
        [self.__imgname,
         tfile, self.__ttype,
         hfile, self.__htype,
         self.__price] = string.splitfields(descriptor, '|')
        # stat the thumbnail file to get its size
        self.__tfile = os.path.join(_DEMODIR, tfile)
        m, i, d, n, u, g, self.__tsize, a, m, c = os.stat(self.__tfile)
        # stat the hires file to get its size
        self.__hfile = os.path.join(_DEMODIR, hfile)
        m, i, d, n, u, g, self.__hsize, a, m, c = os.stat(self.__hfile)

    def name(self):  return self.__imgname
    def tfile(self): return self.__tfile
    def ttype(self): return self.__ttype
    def tsize(self): return self.__tsize
    def hfile(self): return self.__hfile
    def htype(self): return self.__htype
    def hsize(self): return self.__hsize
    def price(self): return self.__price



class MagicMapping(MappingImpl.StringMap):
    """Encapsulates all knowledge about KP decryption interaction model."""
    def __init__(self, dissemination):
        MappingImpl.StringMap.__init__(self, {
            'RequiredTerms': string.join(['authentity',
                                          'ticket',
                                          'receipt',
                                          'destination',
                                          ])
            })
        #
        self.__dissem = dissemination
        #
        # state flags
        #
        # ticket is the ticket for verification but only once.  Once
        # it's been accessed, it is permanently destroyed (by setting
        # its value to the emtpy string).  These flags are set only
        # after the 'authentity' condition is met.
        #
        # key and publickey are similar to {public,}ticket and are set
        # only after the 'ticket', 'receipt', and 'destination'
        # conditions have been met.
        #
        self.__ticketfrozen_p = None
        self.__keyfrozen_p = None

    def __setvalue(self, key, value):
        dict = self.get_dict()
        if not self.__ticketfrozen_p and key == 'authentity':
            self.__dissem.prepare_data()
            dict['ticket'] = self.__dissem.get_ticket(value)
            self.__ticketfrozen_p = 1
        elif not self.__keyfrozen_p and \
             key in ('ticket', 'receipt', 'destination'):
            # make sure all three have been set
            try:
                ticket      = dict['ticket']
                receipt     = dict['receipt']
                destination = dict['destination']
                dict['key'] = self.__dissem.set_receipt(
                    ticket, receipt, destination)
                self.__keyfrozen_p = 1
            except KeyError:
                pass
        else:
            raise Mapping.WriteError, 'Key is read-only: %s' % key

    def __getvalue(self, key):
        # these values are read-once
        if key in ('ticket', 'key'):
            self.get_dict()[key] = ''

    #
    # API overrides
    #
    def GetValue(self, key):
        rtnval = MappingImpl.StringMap.GetValue(self, key)
        self.__getvalue(key)
        return rtnval

    def GetItems(self):
        keyvals = []
        for k, v in self._d.items():
            keyvals.append(k)
            keyvals.append(v)
            self.__getvalue(k)
        return keyvals

    def SetValue(self, key, value):
        MappingImpl.StringMap.SetValue(self, key, value)
        self.__setvalue(key, value)

    def DeleteKey(self, key):
        raise Mapping.WriteError, 'Key cannot be deleted: %s' % key



class Cryptor:
    """Class that implements encryption scheme.

    The current scheme breaks the decryption process into three parts.
    Each part is contained in a separate KP and it is the rendezvous
    of these KPs that allows decryption.  This decreases the
    likelihood that the rights can be compromised.

    Data values:

    ticket = md5digest(authentity + imagename + str(random))

        The ticket is a read-once data element that is presented to a
        KP once it has authenticated it's identity (i.e. its
        `authentity').  This ticket must then be presented to future
        negotiations.

    rotorsn = random[7, 17]
    updates = random[3, 8]

        Each dissemination has a different random value for these to
        attributes.  These control how the decryption/encryption rotor
        is initialized.

    key = md5digest(ticket * updatesn)

        The key is a read-once data element that is used to initialize
        the decryption/encryption rotor.

    rotor = rotor(key, rotorsn)

        The decryption/encryption rotor (i.e. the algorithm for decryption).

    edata = rotor.encrypt(data)
    data = rotor.decrypt(edata)

        The encrypted or clear-text data.

    The interaction model proceeds like so:

    1. A KP shows up at the image repository and request a
       dissemination of a digital object.

    2. If the dissemination is protected, the KP must present its
       `authentity', otherwise all reading of data will fail.

    3. When the KP presents its authentity, the encrypted data can be
       read.  The one-time ticket is also made available via the Ts
       and Cs object.  This ticket must be presented for each future
       negotiation.

    4. If the ticket, a receipt of payment, and the destination KOS is
       presented, the decryption key is made available.  In addition,
       a new KP is launched to the destination.  This new KP contains
       the decryption algorithm.  At the remote site, the three pieces
       must be combined to render the image: the encrypted data, the
       decryption key, and the algorithm.

       Actually, it is the algorithm KP that performs the decryption
       and uses the PlayerAPI interface to render the image.  The
       algorithm KP can future control playing the image, if desired.

    """
    def __init__(self):
        self.__rotorsn = random.choice(range(10))+7
        self.__updatesn = random.choice(range(5))+3
        self.__ticket = None
        self.__price = None
        self.__key = None
        self.__rotor = None

    def get_ticket(self, imgname, price, authentity):
        if not self.__ticket:
            self.__price = price
            m = md5.new(authentity)
            m.update(imgname)
            m.update(`random.random()`)
            self.__ticket = m.digest()
            self.__key = md5.new(self.__ticket * self.__updatesn).digest()
        return self.__ticket

    def launch_decryptor(self, ticket, receipt, home, destination):
        if None in (self.__key, self.__ticket) or self.__ticket <> ticket:
            # Tbd: error handling in this function isn't correct, but
            # then the interface isn't robust enough for proper error
            # handling.  When things go wrong, we just return an empty
            # key and no algorithm KP will be launched, so decryption
            # will always fail. We probably want to improve the ISL
            # interface to support authorization exceptions when using
            # the Mapping interface to negotiate terms and conditions.
            # I'm just not sure how to do this yet.
            return ''
        # TBD: this is a hack for the demo
        if receipt <> self.__price:
            return ''
        #
        # launch decrypt KP
        #
        kprepr = kptools.Representation(main='decrypt', path=[_DEMODIR])
        sfs = kprepr.get_suitcase_fs()
        for fn, val in (
            ('rotorsn' ,    self.__rotorsn),
            ('destination', destination),
            ):
            fp = sfs.open(fn, 'w')
            fp.write(str(val))
            fp.close()
        kprepr.submit(kosname=home, reporter=_RS)
        #
        # return the real decryption key
        #
        return self.__key

    def encrypt(self, data):
        if not self.__rotor:
            if self.__key is None:
                return ''
            self.__rotor = rotor.newrotor(self.__key, self.__rotorsn)
        return self.__rotor.encrypt(data)



class Dissemination(RAPK__skel.Dissemination):
    def __init__(self, home, imgobj, filename, cryptor=None):
        self.__home = home
        self.__imgobj = imgobj
        self.__filename = filename
        self.__dataready_p = not cryptor
        self.__fp = None
        self.__cryptor = cryptor
        self.__terms = cryptor and \
                       MagicMapping(self) or \
                       MappingImpl.StringMap({'RequiredTerms': ''})

    #
    # callbacks for MagicMapping
    #

    def prepare_data(self):
        self.__dataready_p = 1

    def get_ticket(self, authentity):
        return self.__cryptor.get_ticket(self.__imgobj.name(),
                                         self.__imgobj.price(),
                                         authentity)

    def set_receipt(self, ticket, receipt, destination):
        return self.__cryptor.launch_decryptor(
            ticket, receipt, self.__home, destination)

    #
    # API
    #

    def GetTermsAndConditions(self):
        return self.__terms

    def Read(self, size):
        if not self.__fp:
            if not self.__dataready_p:
                raise RAPK.TermsAndConditionsFailure, 'no authorization'
            try:
                self.__fp = open(self.__filename, 'r')
            except IOError, error:
                raise RAPK.ReadError, `error`
        try:
            data = self.__fp.read(size)
            if self.__cryptor:
                edata = self.__cryptor.encrypt(data)
                if not edata:
                    raise RAPK.ReadError, 'no authorization'
                return edata
            return data
        except IOError, error:
            raise RAPK.ReadError, `error`

    def Close(self):
        if self.__fp:
            self.__fp.close()
            self.__fp = None


if __name__ == '__main__':
    main()
