"""A very simple trader service that sends KPs to look for services.

This plugin manages a list of known services at other KOSes. It
implements an interface that lets you ask for all the known KOSes the
have a service of a particular type. If it does not know of any, it
will raise an UnknownInterface exception and send a KP to look for
that interface.

See testcases/ifinder.py for a client of this plugin.

This interface isn't useful as anything other than a toy for many
reasons. Among them are:
  *  Services are transient. But once it has asked about an interface,
  it never asks again.
  *  The client doesn't know how long to wait before asking again
  after an UnknownInterface exception.
"""

from koe.common import pluginlib
from koe.common import nstools
import ilu
import time
from koe.common import kpapp
from koe.common import RS
import pickle
from koe.interfaces.stubs import InterfaceAPI__skel, InterfaceAPI

from ifinderkp import InterfaceFinderKP

def main():
    kos, ns = pluginlib.register('IFinder')
    alarm = ilu.CreateAlarm()
    manager = KPManager(kos)
    finder = InterfaceFinder(manager, kos, ns)
    kos.bind_service("InterfaceFinder", finder,
                     "InterfaceAPI.InterfaceFinder")
    # use an alarm to do some work after the mainloop starts
    alarm.set(1, finder.send_initial_kps, ())
    kos.run()

class KPManager(RS.NullRS):

    def __init__(self, kos):
        RS.NullRS.__init__(self)
        self.kos = kos
        self.recv = {}
        self.kps = {}

    def send(self, instance, modules, kos_name, callback):
        kp = kpapp.KPConstructor(instance, modules)
        recv = self.get_receiver(kos_name)
        id = recv.submit(kp, self)
        # We might need to worry about timing here:
        # Since submit involves ILU, it is possible for KPTerminated
        # to get called before we store the id. Hopefully this is unlikely.
        # Could fix this by having a seperate routine that waits for
        # an id to be stored before referencing its callback
        self.kps[id] = callback

    def KPTerminated(self, id, reason, details, kprepr):
        kp = kpapp.KPConstructor()
        kp.set_representation(kprepr)

        # this interface might belong back in kpapp
        # or I might just use the suitcase
        repr = kp.get_representation()
        pickled_kp = repr.GetState().GetData('state')
        instance = pickle.loads(pickled_kp)
        id = repr.GetMetadata().GetValue('reporting-id')
        # end of really ugly stuff
        if reason:
            state = kprepr.GetState()
            try:
                trace = state.GetData("traceback")
            except KPRepr.PartNameError:
                trace = None
            self.kp_fail(id, instance)
        else:
            self.kp_return(id, instance)

    def kp_return(self, id, kp):
        callback = self.kps[id]
        del self.kps[id]
        callback(kp)

    def kp_fail(self, id, kp):
        callback = self.kps[id]
        del self.kps[id]
        callback(kp, fail=1)

    def get_receiver(self, kos):
        if self.recv.has_key(kos):
            return self.recv[kos]
        recv = kpapp.Receiver(kos)
        self.recv[kos] = recv
        return recv


class InterfaceFinder(InterfaceAPI__skel.InterfaceFinder):

    """Service for listing services at other KOSes.

    Simple implementation, does list_services once and never checks
    results fo freshness."""
    
    def __init__(self, kpmgr, kos, ns):
        self.kos = kos
        self.kpmgr = kpmgr
        self.interfaces = {}
        self.inprogress = []
        self.initial = ["Seller1.Seller", "RAPK.Repository"]
        koslayer = nstools.Lookup(ns, "world/kos").Open("Namespace.Context")
        self.koses = koslayer.List()

        # a little status report
        print "Starting InterfaceFinder"
        print "Known KOSes: ", self.koses
        
    def send_initial_kps(self):
        for interface in self.initial:
            self.search_for_interface(interface)

    def Find(self, interface):
        if self.interfaces.has_key(interface):
            ifaces, age = self.interfaces[interface]
            return ifaces
        self.search_for_interface(interface)
        raise InterfaceAPI.UnknownInterface, interface

    def search_for_interface(self, interface):
        if interface in self.inprogress:
            return
        ifkp = InterfaceFinderKP(interface, self.koses[1:])
        self.inprogress.append(interface)
        self.kpmgr.send(ifkp, ["ifinderkp"],
                        self.koses[0],  # name of KOS
                        self.receive_kp # callback
                        )

    def receive_kp(self, ifkp, fail=0):
        if fail == 0:
            found = []
            for kos in ifkp.found.keys():
                if ifkp.found[kos]:
                    found.append(kos)
            self.interfaces[ifkp.interface] = (found, time.time())
        self.inprogress.remove(ifkp.interface)

if __name__ == "__main__":
    main()

