"""Restricted execution for KP programs.

Three special methods of providing modules are supported:
    1.  A KP can provide additional code modules in their transportable
        object.
    2.  The restricted environment can provide methods of the name
        makemodule_<name> to build the module <name>.
    3.  The KOS administrator can provide special 'proxy' implementations of
        modules stored on the filesystem which operate in unrestricted space
        but are mirrored in the restricted environment.

The first option allows KPs to carry additional modules along with them; these
modules are only loaded into the restricted environment.

The method option of providing modules allows specialized modules to be
provided which must be tightly integrated with the KOS implementation.  The
methods which actually build modules are required to create a module of the
appropriate name in the restricted environment.  The best way to create the
restricted module is to call the add_module(<name>) method of the restricted
environment.  The various copy_{none,except,...}() methods of the rexec object
will return a module created using add_module(), so may be used as well.
Attributes of the module may be filled in as needed.

The proxy implementation can be used to allow a KOS administrator to add
modules to the restricted environment which require non-restricted
implementations.  Source files should be placed along a special 'proxy path'
which is searched in a manner similar to sys.path, but the path should not
overlap the standard path.  Modules are loaded from files named <module>.py;
no caching of .pyc files is performed.

The proxy path is set by calling the set_proxy_path method of KPRExec.
The path must by a sequence of strings, where each string is a path to 
be searched.  The current implementation of KPRExec looks for a
proxies directory in the supervisor directory, which is the directory
that contains KPRExec.py, and places it on the proxy path if it exist.

Packages are not supported using any of these approaches.
"""


import os
import sys
import os
import types
import pickle
import string
import StringIO
import linecache
import ihooks
import rexec
import ilu
from koe.common import pathhack
from koe.common import nstools
from koe.common import Suitcase
from koe.common import shadow
from koe.interfaces.stubs import KOS
from koe.interfaces.stubs import KPRepr
from koe.interfaces.stubs import Mapping
from koe.interfaces.stubs import FileAPI


#  Magic name on path indicates to check the KP document for the module.
MAGIC_DIRECTORY = '<kp-local>'

REXEC_METHOD = 42
REXEC_FILE = 43


class KPRHooks(rexec.RHooks):

    # XXX This allows the KP to mess with the path.  Do we care?
    def default_path(self):
        return [MAGIC_DIRECTORY] + rexec.RHooks.default_path(self)



#  rexec.RModuleLoader inherits from ihooks.FancyModuleLoader
#
class KPRModuleLoader(rexec.RModuleLoader):
    """Enhanced loader to provide access to code modules packed with the KP.
    """

    __kpsources = None

    def add_kprepr(self, kprepr):
        """Complete initialization by adding the KP representation."""
        self.__kpsources        = kprepr.GetCode()

    def get_code_module(self, name):
        """Retrieve the source code for a module as a (pseudo-)file.

        Return a tuple (open_file, filename) on success, or (None, None)
        if the module can't be found.

        Side effect: enter the source code's lines in the line cache.

        """
        if not self.__kpsources:
            return None, None
        try:
            src = self.__kpsources.GetData(name)
        except KPRepr.PartNameError:
            return None, None
        filename = name + '.py'

        # don't actually put the code in the linecache here
        # just tell the line call how to get it... yes?

        lines = string.splitfields(src, '\n')
        # linecache used for tracebacks and profiling
        linecache.cache[filename] = (len(lines), 0, lines, filename)
        return src, filename

    __rexec = None
    def set_environment(self, rexec):
        self.__rexec = rexec

    __msource = None
    def set_methodsource(self, msource):
        self.__msource = msource

    def find_module_in_dir(self, name, dir, allow_packages=None):
        """Extension to locate a module packed with the KP document.

        If the name of the directory matches the magic name, attempt to
        load the module from the code portion of the kprepr.  If found,
        the module can be loaded as a normal Python source module in the
        restricted environment.

        If the directory is None, search the method source for a method
        makemodule_<name> and then search the proxy path for the file
        <name>.py.  In either of these cases, special handlers must be used
        to load the module in load_module().
        """
        if dir == MAGIC_DIRECTORY:
            buf, filename = self.get_code_module(name)
            if buf:
                f = StringIO.StringIO(buf)
            else:
                f = buf
            return f and (f, filename, ('', '', ihooks.PY_SOURCE))
        elif dir is None:
            methodname = "makemodule_" + name
            if self.__msource and hasattr(self.__msource, methodname):
                maker = getattr(self.__msource, methodname)
                return maker, str(maker), ('', '', REXEC_METHOD)
            basename = name + ".py"
            for p in self.__proxypath:
                filename = os.path.join(p, basename)
                try:
                    fp = open(filename)
                except IOError:
                    pass
                else:
                    return fp, filename, ('', '', REXEC_FILE)
        if allow_packages is None:
            return rexec.RModuleLoader.find_module_in_dir(self, name, dir)
        else:
            return rexec.RModuleLoader.find_module_in_dir(self, name, dir,
                                                          allow_packages)

    def load_module(self, name, stuff):
        """Handle the special method and proxy file techniques for defining
        modules.

        If the method technique is used, call the method and extract the
        module from the restricted environment.  If the proxy file technique
        is used, load the file into an unrestricted environment and build a
        restricted module that reflects the public members of the unrestricted
        module.
        """
        file, filename, (suff, mode, type) = stuff
        if type == REXEC_METHOD:
            if not self.__rexec:
                raise ImportError, \
                      "no environment to retrieve method-based module from"
            maker = file
            maker()
            return self.__rexec.modules.get(name)
        if type == REXEC_FILE:
            try:
                dict = {"__name__": name}
                code = compile(file.read(), filename, 'exec')
                exec code in dict
                # create the restricted side
                m = self.__rexec.add_module(name)
                try:
                    exports = dict['__export__']
                except KeyError:
                    exports = dict.keys()
                    for aname in exports[:]:
                        if aname[0] == '_':
                            exports.remove(aname)
                for aname in exports:
                    if dict.has_key(aname):
                        setattr(m, aname, dict[aname])
                # this is needed for Python 1.5.1 and earlier
                m.__doc__ = dict.get('__doc__', None)
                return m
            finally:
                file.close()
        return rexec.RModuleLoader.load_module(self, name, stuff)

    __proxypath = ()
    def set_proxy_path(self, path):
        self.__proxypath = tuple(path)

    def get_proxy_path(self):
        return self.__proxypath


class KPRExec(rexec.RExec):

    """Restricted execution environment for KP user code."""

    preload_modules = ("types", "string", "pickle", "StringIO", 
                       "urlparse", "regex", "marshal")

    def __init__(self, hooks=None, verbose=0):
        """Constructor.
        
        Construct base class  and set our loader and importer object.

        """
        # The only thing we want on KPs path is the standard Python libraries
        # path, and 'common'.  The only way to get back to the standard
        # Python library path is to delete everything that has a kosroot root.
        kosroot = pathhack.kos_root()
        common = os.path.dirname(nstools.__file__)
        newpath = [common]
        for dir in sys.path:
            if os.path.commonprefix([dir, kosroot]) <> kosroot:
                newpath.append(dir)
        self.ok_path = tuple(newpath)
        rexec.RExec.__init__(self, hooks or KPRHooks(verbose), verbose)
        # Override a few things...
        self.loader = KPRModuleLoader(self.hooks, verbose)
        self.importer = rexec.RModuleImporter(self.loader, verbose)
        self.ok_builtin_modules = self.ok_builtin_modules
        self.loader.set_environment(self)       # circular reference
        self.loader.set_methodsource(self)      # and another...
        #
        # Should this really be automatic?  Probably not, but this is
        # fine for preliminary testing.
        #
        proxydir = os.path.join(kosroot, 'supervisor', 'proxies')
        if os.path.isdir(proxydir):
            self.set_proxy_path([proxydir])

    def set_proxy_path(self, path):
        self.loader.set_proxy_path(path)

    def get_proxy_path(self):
        return self.loader.get_proxy_path()

    def copy_none(self, src):
        # XXX: You don't want KPs to have to worry about packages, so put all
        # their modules at the top level
        name = string.split(src.__name__, '.')[-1]
        m = self.add_module(name)
        m.__doc__ = src.__doc__
        return m

    def make_sys(self):
        """Use a shadow of the sys module to maintain current exception info.

        Overrides base rexec.make_sys()."""
        m = shadow.shadow(sys, ['exc_type', 'exc_value',
                                'last_type', 'last_value'])
        m.modules = self.modules
        m.__name__ = 'sys'
        self.modules['sys'] = m
        m.path = map(None, self.ok_path)

        for i in self.ok_sys_names:
            if hasattr(sys, i):
                setattr(m, i, getattr(sys, i))
        m.argv = ['RESTRICTED']
        m.builtin_module_names = (self.modules.keys()
                                  + list(self.ok_builtin_modules))
        m.builtin_module_names.sort()
        m.builtin_module_names = tuple(m.builtin_module_names)

    def preload(self):
        """Preload a bunch of commonly used modules."""
        modules = (tuple(self.ok_builtin_modules) +
                   tuple(self.ok_dynamic_modules) +
                   tuple(self.preload_modules))
        # XXX: this is kind of disgusting, but the idea here is that in a
        # vanilla Python configuration, some of the ok_builtin_modules aren't
        # even built.  They will (rightly so) generate an ImportError, but we
        # don't need to be warned about them.  GvR still wants to be warned
        # about other import failures because otherwise the user's KP will
        # silently run much slower.
        quietmodules = ('audioop', 'imageop')
        for module in modules:
            try:
                self.r_import(module)
            except ImportError, msg:
                if module not in quietmodules:
                    print "Can't preload module", module, ":", msg
        # XXX: We have to patch the pickle.Unpickle class in Python 1.5
        # because standard unpickling is broken in restricted execution
        # environments.  The Python 1.5 versions attempt to assign to the
        # instance's __class__ or __dict__, both of which are prohibited.
        # These implementations work around that, and will be part of Python
        # 1.5.1 -BAW
        self._patch_pickle()

    def _patch_pickle(self):
        m = self.modules['pickle']
        c = m.__dict__['Unpickler']
        # copied and modified from the Python 1.5 pickle.py file
        def load_inst(self):
            from types import ClassType
            class _EmptyClass:
                pass
            k = self.marker()
            args = tuple(self.stack[k+1:])
            del self.stack[k:]
            module = self.readline()[:-1]
            name = self.readline()[:-1]
            klass = self.find_class(module, name)
            instantiated = 0
            if (not args and type(klass) is ClassType and
                not hasattr(klass, "__getinitargs__")):
                try:
                    value = _EmptyClass()
                    value.__class__ = klass
                    instantiated = 1
                except RuntimeError:
                    # In restricted execution, assignment to inst.__class__ is
                    # prohibited
                    pass
            if not instantiated:
                value = apply(klass, args)
            self.append(value)
        # patch the class
        c.load_inst = load_inst
        c.dispatch[m.INST] = load_inst

        def load_obj(self):
            from types import ClassType
            class _EmptyClass:
                pass
            stack = self.stack
            k = self.marker()
            klass = stack[k + 1]
            del stack[k + 1]
            args = tuple(stack[k + 1:]) 
            del stack[k:]
            instantiated = 0
            if (not args and type(klass) is ClassType and
                not hasattr(klass, "__getinitargs__")):
                try:
                    value = _EmptyClass()
                    value.__class__ = klass
                    instantiated = 1
                except RuntimeError:
                    # In restricted execution, assignment to inst.__class__ is
                    # prohibited
                    pass
            if not instantiated:
                value = apply(klass, args)
            self.append(value)
        # patch the class
        c.load_obj = load_obj
        c.dispatch[m.OBJ] = load_obj

        def load_build(self):
            stack = self.stack
            value = stack[-1]
            del stack[-1]
            inst = stack[-1]
            try:
                setstate = inst.__setstate__
            except AttributeError:
                try:
                    inst.__dict__.update(value)
                except RuntimeError:
                    # In restricted execution, the instance's __dict__ is not
                    # accessible.  Use the old way of unpickling the instance
                    # variables.  This is a semantic different when unpickling
                    # in restricted vs. unrestricted modes.
                    for k, v in value.items():
                        setattr(inst, k, v)
            else:
                setstate(value)
        # patch the class
        c.load_build = load_build
        c.dispatch[m.BUILD] = load_build


    def add_kprepr(self, kprepr):
        """Add the KP representation, completing the initialization."""
        self.loader.add_kprepr(kprepr)

    def add_stdio(self, metadata):
        """Prepare std{in,out,err} for the KP."""

        kpstdout = self.figure_file(metadata, 'stdout', 'a')
        kpstderr = self.figure_file(metadata, 'stderr', 'a')
        kpstdin = self.figure_file(metadata, 'stdin', 'r')

        s = self.modules['sys']

        self.set_files()

        s.stdout = kpstdout
        s.stderr = kpstderr
        s.stdin = kpstdin

        self.reset_files()

    def figure_file(self, metadata, name, mode):
        """Figure out which std file to use, based on the metadata."""
        key = name + '-server'
        try:
            sbh = metadata.GetValue(key)
        except Mapping.KeyError:
            sbh = None
        if sbh == 'null':
            from koe.common import iotools
            f = iotools.null_io_mechanism()
        elif sbh:
            f = ilu.ObjectOfSBH(FileAPI.File, sbh)
            # Don't ping here
            if not f:
                raise ValueError, "designated stdout-server seems dead"
        else:
            f = getattr(sys, name)
        return f

    def makemodule_types(self):
        """Create the types module.

        The standard types module has several problems that make
        just importing it in restricted mode a problem.

        """
        m = self.copy_except(types, ('__builtins__',))

    def makemodule_nstools(self):
        """Create the nstools module.

        This contains only the Lookup() and WorldOpen() functions and
        the BadPathError exception.

        """
        m = self.copy_only(nstools, ('Lookup', 'WorldOpen', 'BadPathError'))

    def makemodule_ilu(self):
        """Create the ilu module.

        This module only contains the exceptions.

        """
        m = self.copy_only(ilu, ('IluGeneralError', 'IluProtocolError',
                                 'IluUnimplementedMethodError'))

    def makemodule_Suitcase(self):
        import koe.common.Suitcase
        m = self.copy_only(koe.common.Suitcase,
                           ('SuitcaseFilesystem', 'SuitcaseFilesystemPath'))

    def makemodule_KOS(self):
        """Create a specially crafted KOS module.

        This is a subset of the real KOS module, containing only a
        selection of objects that are deemed safe.

        At the moment, only string objects whose name does not begin
        with an underscore are deemed safe -- this is sufficient to
        be able to use KOS exceptions in user code.

        """
        m = self.add_module('KOS')
        for name in dir(KOS):
            if name[0] != '_':
                value = getattr(KOS, name)
                if type(value) is types.StringType:
                    setattr(m, name, value)

    def makemodule_SURR(self):
        """Create a specially crafted SURR module.

        This is a totally handcrafted module which provides a way to
        instantiate ILU surrogate objects that have been pickled by
        the Supervisor (see ExtendedPickler.save_inst() in
        Supervisor.py

        ### Security warning!!! TBD XXX

        Surrogates are not unpicklable in restricted execution mode.
        TBD: this can probably be spoofed so we need to provide a more
        robust mechanism for migrating surrogates...

        ...but maybe not (???), because the kernel must have the class
        already.
                                                                            
        """
        m = self.add_module('SURR')
        def __birth__(modname, classname, sbh):
            # TBD: This is all hackery.  It works but GvR should check it!
            # this fails if the module has not yet been imported.
            # which I think is fine, although we don't want the
            # unpickling to fail.  in that case build a standin for
            # this instantiate, but realize that it should probably be
            # bastionized in the KP user space:
            try:
                # XXX: kludge to deal with packages
                components = string.split(modname, '.')
                if len(components) > 1:
                    pkg = __import__(string.join(components[:-1], '.'),
                                     globals(), locals(),
                                     [components[-1]])
                    module = getattr(pkg, components[-1])
                else:
                    module = __import__(modname, globals(), locals())
                class_ = getattr(module, classname)
                return ilu.ObjectOfSBH(class_, sbh)
            # TBD: if you change the finding of `module' above to
            # check in the existing sys.modules list, change the
            # ImportError below to a KeyError.  To be honest I'm not
            # sure which is the best thing to do.  From an
            # implementation standpoint, the import seems best, but
            # from a security standpoint???
            except (ImportError, ilu.IluProtocolError, ilu.IluGeneralError):
                # build a standin which will try to get instantiated
                # at the next migration.  TBD: this is a complete hack!
                exec STANDIN % (classname, sbh)
                class_ = eval('%s' % classname)
                pickle.classmap[class_] = modname
                # need to do this hackery so the object gets pickled correctly
                return class_()
        m.__birth__ = __birth__

    def load_main(self, name):
        """Load the main module.

        This is not quite an import, since the main module is always
        executed in the special module __main__; its source is retrieved
        from the module whose name is given.  (This may appear strange
        but is standard Python behavior -- a script is always executed
        in module __main__.)

        """
        buf, filename = self.loader.get_code_module(name)
        code = compile(buf, filename, 'exec')
        self.r_exec(code)

    def add_kos_bastion(self, bastion):
        """Add a function returning the KOS bastion to the KOS module.

        This is called after the bastion has been created.  Since the
        bastion depends on the delegate and the delegate will likely
        depend on the restricted execution environment, this has to be
        patched in later.

        The function __KOS_Bastion__() is needed by the unpickling
        code for an incoming KP.  The pickling code for an outgoing KP
        will replace all occurrences of its KOS bastion with
        references to KOS.__KOS_Bastion__(), so they will be replaced
        with the KOS bastion at the receiving end.

        """
        def __KOS_Bastion__(bastion=bastion):
            return bastion
        m = self.r_import('KOS')
        m.__KOS_Bastion__ = __KOS_Bastion__

    # Override some stuff in rexec.py that's too liberal.

    def r_open(self, file, mode='r', buf=-1):
        raise IOError, "can't open files"

    ok_posix_names = ('error', 'times', 'getpid')


STANDIN = """
class %s:
    def IluSBH(self):
        return '%s'
    def _IluClass(self):
        pass
"""
