"""Supervisor class -- trusted code for running a Knowbot Program."""

import os
import sys
import traceback
import StringIO
import ilu
import Bastion
import KPRExec
import ExtendedPickler
from koe.common import CnBroker
from koe.common import nstools
from koe.common import Suitcase
from koe.common import pathhack
from koe.common import qsplit
from koe.interfaces.stubs import KPRepr
from koe.interfaces.stubs import Mapping
from koe.interfaces.stubs import KOS
from koe.interfaces.stubs import KOSKernel, KOSKernel__skel
from koe.interfaces.stubs import ReportingAPI

# Spend some time preloading a restricted execution environment
preloaded_rexec = KPRExec.KPRExec()
preloaded_rexec.preload()

KPMigration = 'KPMigration'   # exception used for profiling support

# XXX: This is a horrible kludge.  See the comment in kernel/kernel.py for a
# detailed explanation.
ilu.SetFDBudget(2000)

class Supervisor(CnBroker.ConnectionBroker, KOSKernel__skel.ProcessLink):

    def __init__(self, sbh):
        """Everything happens during this constructor call this constructor."""

        self._kpentry = ilu.ObjectOfSBH(KOSKernel.KPEntry, sbh)
        CnBroker.ConnectionBroker.__init__(self, self._kpentry)

        (self._kos_name,
         self._local_id,
         self._kprepr,
         self._visualizer,
         rawdescriptor) = self._kpentry.Hello(self)

        self._metadata = self._kprepr.GetMetadata()     # Just a cache

        nsdescriptor = nstools.Descriptor(reference=rawdescriptor.reference,
                                          typename=rawdescriptor.typename)
        ns = nsdescriptor.Open(typeobj='Namespace.Context')
        self.set_ns(ns)         # A ConnectionBroker method

        self._rexec = preloaded_rexec   # Preloaded before the fork
        self._rexec.add_kos_bastion(self.get_bastion())
        self._rexec.add_kprepr(self._kprepr)
        self._rexec.add_stdio(self._metadata)

        self.__migrated = 0

        try:
            self._rexec.set_files()
            reason, details = self.run_user_code()
        finally:
            if self.__migrated:
                return
            self._rexec.reset_files()

        self.handle_cnbroker_termination()

        if self._suitcase_fs:
            self._suitcase_fs._flush()
        self._kpentry.Terminated(reason, details)

        id = self._metadata.GetValue('reporting-id')
        self._visualizer.KPTerminated(id, reason, details, self._kprepr)
        rs = self.reporting_station()
        if rs:
            rs.KPTerminated(id, reason, details, self._kprepr)
    
    def run_user_code(self):
        """Run all user code, catching user exceptions.
        
        Assume stdio streams have been connected to the user's IO server.
        Return (reason, details) for termination.
        This won't return if the user code successfully migrates.

        """
        main_module_name = self._metadata.GetValue('main-module')
        try:
            kp_class_name = self._metadata.GetValue('knowbot-class')
        except Mapping.KeyError:
            kp_class_name = "KP"
        kos = self.get_bastion()
        state = self._kprepr.GetState()
        try:
            pickle = state.GetData("state")
        except KPRepr.PartNameError:
            pickle = None

        try:
            argstrs = self._metadata.GetValue('program-arguments')
            args = qsplit.qsplit_nq(argstrs)
        except Mapping.KeyError:
            args = []

        try:
            reason = ""
            try:
                details = self.user_code(main_module_name,
                                         kp_class_name, pickle, kos, args)
            except SystemExit, value:
                # SystemExit is actually a valid way for the KP to
                # signal it is quitting to the supervisor.
                details = str(value or "0") or "0"
        except:
            t, v, tb = sys.exc_info()
            try:
                self.store_traceback(tb)
            finally:
                # eliminates the circular reference created above
                tb = None
            ## XXX store the pickle as state['corpse']?
            # It is conceivable that the user code raises an exception
            # in the conversion to string of the exception and its value.
            # In this case, we make something up and don't give the
            # traceback -- it's either a stupid mistake or a trap.
            try:
                reason = str(t)
                details = str(v)
            except:
                reason = "RuntimeError"
                details = "conversion to string of exception failed"
        
        return reason, details

    def user_code(self, main_module_name, kp_class_name, pickle, kos, args):
        """Actual user code.
        
        Any exception here is a user error, except SystemExit.
        Return exit status as a string.
        
        """
        rexec = self._rexec     # Assume this can't fail :-)
        # set up sys.argv, overriding the default ['RESTRICTED'].
        sysmod = rexec.modules['sys']
        sysmod.argv = [main_module_name] + args
        sysmod.kos = kos
        # get all the information we need to run the user code
        rexec.load_main(main_module_name)
        if pickle:
            pickle_module = rexec.r_import('pickle')
            kp = pickle_module.loads(pickle)
        else:
            main_module = rexec.r_import('__main__')
            kp_class = getattr(main_module, kp_class_name)
            kp = kp_class()
        self._kp = kp
        # Run the user code.
        rc = kp.__main__(kos)
        return str(rc or "0") or "0"

    _bastion = None

    def get_bastion(self):
        #  This creates a circular reference:
        if not self._bastion:
            self._bastion = Bastion.Bastion(self,
                                            lambda n,l=self.PUBLICS: n in l)
            # poke extra non-method attributes onto the bastion
            self._bastion.MigrationError = KOS.CommunicationError
        return self._bastion

    _rs = None

    def reporting_station(self):
        """Retrieve connection to the reporting station for this KP.
        
        Return None if the reporting station is dead.

        """
        if not self._rs:
            self._rs = ilu.ObjectOfSBH(ReportingAPI.ReportingStation,
                        self._metadata.GetValue('reporting-station'))
        return self._rs

    # List of methods a KP can access.  Other attributes are added in
    # get_bastion()
    PUBLICS = CnBroker.ConnectionBroker.PUBLICS + (
        'get_suitcase',
        'get_kos_name',
        'get_namespace',
        'get_kphandle',
        'checkpoint',
        'migrate',
        'clone',
        )

    _suitcase_fs = None

    def get_suitcase(self):
        #  Need to make this a bastion for the suitcase filesystem
        if not self._suitcase_fs:
            suitcase = self._kprepr.GetSuitcase()
            self._suitcase_fs = Suitcase.SuitcaseFilesystem(suitcase)
        return self._suitcase_fs

    def get_kos_name(self):
        return self._kos_name

    def get_namespace(self):
        # TBD: This, and all descendents must be bastionized!
        return self.get_ns()

    def get_kphandle(self):
        return '%s-%s' % (self._kos_name, self._local_id)

    # XXX Exception handling in migrate() and clone() is underwhelming

    def checkpoint(self):
        "Save the current state with the kernel."
        self.save_state()

    def migrate(self, destination):
        receiver = self.locate_receiver(destination)
        self.save_state()
        new_kpid = receiver.SubmitKP(KOS.AlphaFormat, self._kprepr.Marshal()) 
        ## If we get this far, and Goodbye raises an exception, we
        ## could have two copies of the same KP running.
        self._kpentry.Goodbye(new_kpid, destination)
        ### There are two branches here:
        ### (1) Call os._exit. Causes profiling to fail.
        ### (2) Raise KPMigration. Means the user can catch an
        ###     exception on a successful migration.
        ### For now, we'll use branch (1)
        os._exit(0)
        #self.__migrated = 1
        #raise KPMigration

    def clone(self, destination=None):
        destination = str(destination or self.get_kos_name())
        receiver = self.locate_receiver(destination)
        self.save_state()

        try:
            rs = self.reporting_station()
            if not rs: 
                raise KOS.CommunicationError, \
                      "no cloning -- could not contact reporting station"
            id = rs.AllocateID(self._metadata.GetValue('reporting-id'))
        except (ilu.IluProtocolError, ilu.IluGeneralError):
            raise KOS.CommuncationError, \
                  "no cloning -- could not contact reporting station"

        # XXX Looks like we should catch ILU exceptions in Submit too
        changed_metadata = self.swap_metadata({'reporting-id': id})
        try:
            if destination:
                try:
                    return receiver.SubmitKP(KOS.AlphaFormat,
                                             self._kprepr.Marshal())
                except (ilu.IluProtocolError, ilu.IluGeneralError):
                    raise KOS.CommuncationError, \
                          "Error while submitting KP"
            else:
                # this is quicker because there is no IPC to copy repr
                return receiver.SubmitRepr(self._kprepr)
        finally:
            self.swap_metadata(changed_metadata)

    #  These methods are used to implement some of the user-level methods,
    #  but are hidden by the bastionization filter.

    def save_state(self):
        """Save the current state to the KPRepr.

        Flushes suitcase metadata updates and saves 'pickle' of run state.
        Exceptions raised here are the user's responsibility.

        """
        if self._suitcase_fs:
            self._suitcase_fs._flush()

        pf = StringIO.StringIO()
        pickler = ExtendedPickler.ExtendedPickler(pf,
                                self.get_bastion(), self._rexec)
        pickler.dump(self._kp)
        state = self._kprepr.GetState()
        state.SetPart(
            'state',    
            KPRepr.ContentHeader('state', 'application/python-pickle'),
            pf.getvalue())

    def store_traceback(self, tb):
        """Helper for run_user_code() to store the trackback."""
        state = self._kprepr.GetState()
        f = StringIO.StringIO()
        traceback.print_tb(tb, file = f)
        trace = f.getvalue()
        f.close()
        state.SetPart(
            "traceback",
            KPRepr.ContentHeader("traceback", 'text/python-traceback'),
            trace)

    def swap_metadata(self, dict):
        """Change some metadata, returning the old values."""
        prev = {}
        metadata = self._metadata
        for k, v in dict.items():
            try:
                prev[k] = metadata.GetValue(k)
            except Mapping.KeyError:
                prev[k] = None
            if dict[k] is None:
                metadata.DeleteKey(k)
            else:
                metadata.SetValue(k, v)
        return prev

    def locate_receiver(self, destination):
        """Return the receiver object.

        Raise KOS.CommuncationError if the receiver is not found.

        """
        destpath = "kos/%s/receiver" % destination
        try:
            receiver = nstools.WorldOpen(destpath, "KOS.Receiver")
        except nstools.BadPathError:
            raise KOS.CommunicationError, \
                  "Could not contact receiver: " + destination
        return receiver
