"""KP starter process -- helper to fork off KP supervisor processes.

Since we've observed that much of the time it takes to start a new KP
goes into importing the support modules that comprise the KP
supervisor, we apply a hack: the KOS kernel forks a "KP starter"
process which imports all the support modules and then forks a child
process whenever the kernel needs a KP supervisor process.

The starter is constrained by one important restriction: it cannot use
ILU.  It can import the ILU support modules, but if it were to use any
of them, this would initialize ILU and the internal ILU state would be
totally confusing for the forked child processes.  Therefore, a simple
protocol based on pipes is used for communication between the KOS
kernel and the KP starter.

The interface between the KOS kernel and the KP starter is simple:
The kernel writes the KP entry's SBH on a pipe to the starter,
followed by a newline.  The starter writes the pid of the child
process to another pipe, also followed by a newline.  When the child
dies and the starter finds out about this, it writes the pid and the
exit status to yet another pipe, separated by a colon and followed by
a newline.  All numbers are written in decimal.

"""

# path hackery, no longer requires $PYTHONPATH to be set properly
import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../"))
import string
import select
import Supervisor
import time
import getopt

TIMEOUT = 1.0                           # Select timeout
BUFSIZE = 1024                          # Read buffer size

# XXX These three file descriptor numbers must match those in the kernel
SBHFD = 3                               # Fd to read sbh from
PIDFD = 4                               # Fd to write pid to
STSFD = 5                               # Fd to write status to

FORK_TRIES = 2


def guaranteed_fork():
    while 1:
        try:
            pid = os.fork()
        except os.error:
            # XXX: The fork failed, probably due to a resource issue.  In this
            # case, we'll just sleep it off and try to fork again in a little
            # while, hopefully giving other processes time to complete.  This
            # may not always work, e.g. if they're waiting on the new KP
            print 'we are forked'
            time.sleep(2)
        else:
            return pid


def limited_fork(tries=FORK_TRIES):
    while tries:
        tries = tries - 1
        try:
            pid = os.fork()
        except os.error:
            print 'we are forked'
            time.sleep(2)
        else:
            return pid


def main(prof_name, dir):
    # arguments are used for profiling
    fd = SBHFD
    buffer = ""
    try:
        while 1:
            r, w, x = select.select([fd], [], [], TIMEOUT)
            mourn()
            if r:
                data = os.read(fd, BUFSIZE)
                if not data:
                    break               # Kernel has closed the pipe
                buffer = buffer + data
                lines = string.split(buffer, '\n')
                lines, buffer = lines[:-1], lines[-1]
                for sbh in lines:
                    pid = limited_fork()
                    if pid is None:
                        os.write(PIDFD, "ResourceError fork failed\n")
                    elif pid:
                        # Parent
                        os.write(PIDFD, "%d\n" % pid)
                    else:
                        # Child -- everything happens in the constructor.
                        # If this fails, standard exception handling happens.
                        if prof_name is None:
                            from koe.common.ilutools import IluErrorTuple
                            try:
                                supervisor = Supervisor.Supervisor(sbh)
                            except IluErrorTuple:
                                # any ILU RPC call fail for any reasons, but
                                # we can ignore this because we're going to
                                # eventually exit below anyway.
                                pass
                        else:
                            # XXX: I suspect this profiling code is busted if
                            # the Supervisor object created is not reference
                            # counted by the profiler.  See the non-profiled
                            # version above.
                            import profile
                            filename = '%s.%s.prof' % (prof_name,
                                                       str(time.time()))
                            prof_path = os.path.join(dir, filename)
                            profile.run('Supervisor.Supervisor(\'%s\')' % sbh,
                                        prof_path)
                        sys.exit()
    except KeyboardInterrupt:
        print "\r%s:" % os.path.basename(sys.argv[0]),
        print "received KeyboardInterrupt"

def mourn():
    """Process children that have recently died, without blocking."""
    while 1:
        try:
            pid, sts = os.waitpid(-1, os.WNOHANG)
        except os.error:
            break
        if not pid:
            break
        os.write(STSFD, "%d:%d\n" % (pid, sts))

if __name__ == '__main__':
    prof_name = None
    dir = '/tmp'
    try:
        optlist, args = getopt.getopt(sys.argv[1:], 'p:', ['profile-dir='])
    except getopt.error, msg:
        print msg
        optlist = []

    for k, v in optlist:
        if k == '-p':
            prof_name = v
        if k == '--profile-dir':
            dir = v
    main(prof_name, dir)
