"""Exercise connectors registration - test:

  - connector cleanup on irregular termination of a KP, and
  - recognition of connector collision, when one KP tries to register the
    same service as another.

We use a three-phase process, involving two clonings:

 Phase 1 - original KP:
   - establish a trivial service - using Seller1, since it's easy.
   - Clone to produce the phase 2 KP.
   - And coredump, before phase 2 has time to get going.

 Phase 2 - first clone, no service collision (if proper cleanup):
   - Give some time for phase 1 clone to crash.
   - Register the same service - it should not collide, if the original
     service was properly cleaned out despite unclean phase 1 exit.
   - Also register a uniquely named service, so phase 3 can signal us
     to conclude.  (Our commonly-named connector would be clobbered by
     then.)
   - Launch the phase 3 KP.
   - Exit when phase 3 uses our uniquely named connector.

 Phase 3 - second clone, service collision:
   - Try to register service under common name - it should collide.
   - Signal phase 2 using special service, so it exits.
   - Exit, wrapping it all up."""

import sys
import os
import time
import string
import tempfile

SERVICE_NAME = "testee"
SERVICE_TYPE = "Seller1.Seller"

PHASE2_DELAY = 10       # Time for coredump and process_died processing.
PHASE2_SERVICE_PREFIX = "phase2"
got = 0

been_called = 0

class KP:
    """Three phase exercise of connector cleanup and collision.

    We should produces a core dump and some complaints from the supervisor
    (including "ProcessFailed"), but should not, if successful, produce any
    uncaught exceptions.

    See module docstring for details."""
    def __main__(self, kos):
        self.kos = kos
        try:
            self.phase = self.phase + 1
            self.prior_pid = self.pid
            self.pid = os.getpid()
        except AttributeError:
            self.phase = 1
            self.prior_pid = 0
            self.pid = os.getpid()

        if self.phase == 1:
            self.phase1()
        if self.phase == 2:
            self.phase2()
        if self.phase == 3:
            self.phase3()

    def phase1(self):
        self.note("Creating service.")
        self.create_service()
        self.clone_next()
        self.note("Coredumping!  (Supervisor complains, " +
                  "and core left in kernel dir...)")
        import coredump
        # Should never get here.

    def phase2(self):
        self.note("Delaying %d secs, so phase 1 %s",
                  PHASE2_DELAY,
                  "has time to fully crash.")
        time.sleep(PHASE2_DELAY)
        self.note("Creating service - should NOT conflict.")
        self.create_service()
        # A second service so the next clone can signal conclusion:
        self.create_service(prefix=PHASE2_SERVICE_PREFIX)
        self.clone_next()
        self.kos.run()
        self.note("Concluding.")
        sys.exit()

    def phase3(self):
        self.note("Creating service.")
        try:
            self.create_service()
            self.note("Whoops - %s %s",
                      "service should have conflicted",
                      "with that of phase 2")
        except self.kos.exception("Namespace.BindError"):
            self.note("Connector conflict appropriately detected.")
        self.note("Signalling phase 2 to exit.")
        self.get_service(prefix=PHASE2_SERVICE_PREFIX).price()
        self.note("All done.")
        sys.exit()

    def create_service(self, prefix=""):
        svcid = self.kos.bind_service(prefix + SERVICE_NAME,
                                        Seller1(self.phase, self.kos),
                                        SERVICE_TYPE)

    def get_service(self, prefix=""):
        return self.kos.lookup_service(SERVICE_TYPE,
                                         prefix + SERVICE_NAME).Open()

    def clone_next(self):
        self.note("Cloning phase %d KP.", self.phase + 1)
        self.kos.clone()

    def note(self, msg, *args):
        print "Phase %d:" % self.phase, msg % args

class Seller1:
    """Trivial service, for exercising contention as well as for
    inter-phase communication."""

    def __init__(self, phase, kos):
        self.phase = phase
        self.kos = kos

    def price(self):
        """'You can be sincere and still be stupid.' Charles F. Kettering"""
        try:
            self.kos.stop()
        except RuntimeError:
            print "Phase %d failed kos.stop()" % self.phase
        return self.phase
