"""Module of common functions for the koeboot.py and koehalt.py scripts.

By default, the boot file directory is $KOSROOT/tools/boot, and within
this directory, the parse order is `boot', `boot.cluster', and
`boot.USER', where USER is your $LOGNAME or $USER.  You can add
another boot file onto the load order by using the -f option.

The general layout of the boot files is similar to the kernel setup files.
See Python's ConfigParser.py module for details.  There are currently four
special sections:

DEFAULT    -- same as for plugins

Worldroot  -- to start the worldroot servers

Visualizer -- to start the visualizer

KERNELS
    contains keys which are hostnames (either FQDN or short names).
    Values are the list of kernels to (perhaps) start at boot on those
    hosts.  Only perhaps because, each kernel's `enable' option is
    checked before it is started up.  The special key `-' means start
    on any host for which a host-specific section was not found.

    Preference is taken for a FQDN matching the current host's
    aliases.  If no match is found, then even the non-FQDN are
    matched.

Following these four sections are sections describing how to start
each kernel named in the KERNELS section above.

The following options are used to control how the subprocesses are invoked:

    enable        -- a boolean (0 or 1).  If true, the process is started.

    args          -- a string.  The arguments to pass to the process on
                     the command line.

    exec-file     -- a string.  The absolute path to the file to exec()

    ns-path       -- a string.  Path in the namespace to check for
                     process' existence 

    ping-for-life -- a boolean.  If true, ns-path existence in the
                     namespace is not enough to guarantee object
                     liveness.  In this case, an IluPing is sent to
                     the object as well.

    wait-for      -- a boolean.  If true, after the process is forked,
                     the parent waits for ns-path to appear in the
                     namespace before continuing (or aborts if it
                     doesn't appear after a timeout).

    x-display     -- a string.  If provided, sets the DISPLAY
                     environment variable for the new subprocess.
"""

import sys
import os
import time
from koe.pythonlib import ConfigParser            # until Python 1.5.2
import getopt
import string
import socket
from koe.common import qsplit
from koe.common import pathhack
from koe.common import nstools


BOOTDIR = 'config'
BOOTFN = 'boot'
BOOTCLUSTERFN = 'boot.cluster'


def write(msg, add_nl=1):
    if msg[-1] <> '\n' and add_nl:
        msg = msg + '\n'
    sys.stdout.write(msg)
    sys.stdout.flush()


def get(parser, section, option):
    try:
        return parser.get(section, option)
    except ConfigParser.NoOptionError:
        return None


def waitfor(path, pid):
    timeout = 10                        # x 2 = 20secs
    while timeout >= 0:
        try:
            nstools.WorldOpen(path)
            return 0
        except (nstools.BadPathError, nstools.NoWorldrootError):
            timeout = timeout - 1
            cpid, status = os.waitpid(pid, os.WNOHANG)
            if cpid <> 0:
                # the process exited somehow
                return 1
            time.sleep(2)
    return 1


def unwaitfor(path):
    timeout = 10                        # x 2 = 20secs
    while timeout >= 0:
        try:
            nstools.WorldOpen(path)
            timeout = timeout - 1
            time.sleep(2)
        except nstools.BadPathError:
            return 0
    return 1


def host_kernels(parser):
    hosts = {}
    hname = socket.gethostname()
    hostname, aliases, ips = socket.gethostbyaddr(socket.gethostbyname(hname))
    for h in aliases + [hname, hostname]:
        hosts[h] = h
    hostnames = hosts.keys()
    opts = parser.options('KERNELS')
    # first match on FQDN if present
    for kh in opts:
        if string.find(kh, '.') and kh in hostnames:
            return qsplit.qsplit_nq(get(parser, 'KERNELS', kh))
    # now match against all hosts
    for kh in opts:
        if kh in hostnames:
            return qsplit.qsplit_nq(get(parser, 'KERNELS', kh))
    return []


class ArgParser:
    def __init__(self, argstr, longarglist, leftover_ok=1):
        self._argstr = argstr
        self._longarglist = longarglist
        self._leftover_ok = leftover_ok
        self.args = []

    def usage(self, status=0, errmsg=None):
        raise RuntimeError, 'unimplemented'

    def unhandled_option(self, opt, val):
        raise RuntimeError, 'unimplemented'

    def parse(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]
        try:
            opts, args = getopt.getopt(argv, self._argstr, self._longarglist)
        except getopt.error, msg:
            self.usage(status=1, errmsg=msg)

        if not self._leftover_ok and args:
            self.usage(status=1)
        else:
            self.args = args

        for opt, val in opts:
            # first character must be a `-'.  if second one is also,
            # its a long option
            if opt[1] == '-':
                methodnm = 'longopt_' + opt[2:]
            else:
                methodnm = 'shortopt_' + opt[1:]

            if hasattr(self, methodnm):
                method = getattr(self, methodnm)
                method(opt, val)
            else:
                self.unhandled_option(opt, val)


def make_configparser(extrafiles):
    kosroot = pathhack.kos_root()
    bootdir = os.path.join(kosroot, BOOTDIR)
    files = [os.path.join(bootdir, BOOTFN),
             os.path.join(bootdir, BOOTCLUSTERFN),
             ]

    user = None
    if os.environ.has_key('LOGNAME'):
        user = os.environ['LOGNAME']
    elif os.environ.has_key('USER'):
        user = os.environ['USER']

    if user:
        files.append(os.path.join(bootdir, BOOTFN + '.' + user))
    for file in extrafiles:
        files.append(file)

    defaults = {'kosroot': kosroot,
                'kosname': "",
                'hostname': pathhack.get_short_hostname(),
                }
    p = ConfigParser.ConfigParser(defaults)
    p.read(files)
    return p
