#! /usr/bin/env python

"""Knowbot Operating Environment bootstrapper.

Boots the entire KOE, including Worldroots, KOSs, and the Visualizer,
based on configuration boot files.  See the file boothalt.py for
details on the configuration file format.

Usage:

    %(program)s [-f file] [-x] [-q] [-h] [koses]

Where:

    --file file
    -f file         -- read boot configuration from the named file.  More
                       than one -f option can be used.

    --exclude
    -x              -- reverse the meaning of koses named on the command
                       line to mean: start all kernels defined in the
                       configuration files *except* those named on the
                       command line.

    --quiet
    -q              -- be quiet during startup


    --help
    -h              -- print this help message


    koses           -- optional kernels to start up or exclude.  If no
                       kernels are specified, only those found in the
                       configuration files are started.  If the -x
                       option is given, all kernels specified in the
                       configuration files except those listed here
                       are started up.  Without -x, only these kernels
                       will be started, and the visualizer and worldroot will
                       not be started.

"""

# path hackery, no longer requires $PYTHONPATH to be set properly
import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../../"))
import string
from koe.common import nstools
# yes bogus, I know
from boothalt import *


program = sys.argv[0]
quiet = None

def start_program(parser, tool, is_worldroot=0, missing_ok=0):
    section = tool
    if missing_ok and section not in parser.sections():
        section = 'DEFAULT'

    if not parser.getboolean(section, 'enable'):
        return

    nspath = get(parser, section, 'ns-path')
    if nspath:
        try:
            obj = nstools.WorldOpen(nspath)
            if obj:
                ping = parser.getboolean(section, 'ping-for-life')
                if not ping or obj.IluPing():
                    if not quiet:
                        write('%s is already running' % tool)
                    return
        except nstools.BadPathError:
            pass
        # hack alert
        except nstools.NoWorldrootError, msg:
            if not is_worldroot:
                raise

    pid = os.fork()
    if pid:
        # parent, possibly wait for the child to start up
        if not quiet:
            write('Starting %s... ' % tool, add_nl=0)
        started = 1
        if parser.getboolean(section, 'wait-for') and nspath:
            started = not waitfor(nspath, pid)
        if started:
            if not quiet:
                write('PID = %d' % pid)
        else:
            if not quiet:
                write('NOT STARTED')
            # hack alert
            if is_worldroot:
                raise nstools.NoWorldrootError('')
    else:
        # child
        argstr = get(parser, section, 'args') or ''
        program = get(parser, section, 'exec-file')
        args = [program] + qsplit.qsplit_nq(argstr)
        display = get(parser, section, 'x-display')
        if display:
            os.environ['DISPLAY'] = display
        os.execv(program, args)



class BootArgs(ArgParser):
    def __init__(self):
        ArgParser.__init__(self, 'f:xqh',
                           ['file=', 'exclude', 'quiet', 'help'])
        self.files = []
        self.exclude = None
        self.quiet = None
        self._help = None

    def usage(self, status=0, errmsg=None):
        if errmsg is not None:
            print errmsg
        print __doc__ % globals()
        sys.exit(status)

    def unhandled_option(self, opt, val):
        print 'Invalid option ignored:', opt

    def shortopt_f(self, opt, val):
        self.files.append(val)
    longopt_file = shortopt_f

    def shortopt_x(self, opt, val):
        self.exclude = 1
    longopt_exclude = shortopt_x

    def shortopt_q(self, opt, val):
        self.quiet = 1
    longopt_quiet = shortopt_q

    def shortopt_h(self, opt, val):
        self._help = 1
    longopt_help = shortopt_h

    def parse(self, argv=None):
        ArgParser.parse(self, argv)
        if self._help:
            self.usage()



def main():
    global quiet
    args = BootArgs()
    args.parse()
    quiet = args.quiet

    p = make_configparser(args.files)

    # Only start up the Worldroot and Visualizer if there no are kernels
    # specified on the command line, or if the -x option was used.
    # E.g. specifying kernels to start up disables starting the Worldroot and
    # Visualizer.
    if not args.args or args.exclude:
        try:
            start_program(p, 'Worldroot', is_worldroot=1)
        except nstools.NoWorldrootError:
            print 'No worldroot.  Boot process terminating.'
            sys.exit(1)
        start_program(p, 'Visualizer')

    # Which kernels to start?
    includes = {}
    excludes = {}

    for k in args.args:
        if args.exclude:
            excludes[k] = k
        else:
            includes[k] = k

    if args.exclude or not args.args:
        # kernels specified in the - option of the KERNELS section
        for k in string.split(get(p, 'KERNELS', '-')):
            includes[k] = k
        # kernels specified in the host part of the KERNELS section
        for k in host_kernels(p):
            includes[k] = k

    for k in excludes.keys():
        if includes.has_key(k):
            del includes[k]

    for k in includes.keys():
        defaults = p.defaults()
        defaults['kosname'] = k
        start_program(p, k, missing_ok=1)



main()
