#! /usr/bin/env python

"""Halt the Knowbot Operating Environment.

Halts all or part of the KOE based components found in the Namespace.
This is different than the koeboot.py program.

Usage:

    %(program)s [-f file] [-x] [-w] [-z] [-a] [-q] [-h] [koses]

Where:

    --file file
    -f file         -- read boot configuration from the named file.  More
                       than one -f option can be used.

    --worldroot
    -w              -- ordinarily the Worldroot server is not shutdown
                       by this script, unless this argument is
                       supplied

    --visualizer
    -z              -- ordinarily the Visualizer is not shutdown
                       by this script, unless this argument is
                       supplied

    --all
    -a              -- shuts down both the Visualizer and the Worldroot

    --exclude
    -x              -- exclude the kernels named on the command line from
                       those shut down.

    --quiet
    -q              -- be quiet during startup

    --help
    -h              -- print this help message

    koses           -- optional kernels to halt.  If specified, only
                       these kernels will be shutdown.  If not specified, all
                       kernels specified in the config files are halted.  If
                       exactly and only the kernel name `*' is used, all
                       kernels found in the namespace are halted.

                       If the -x option is used, all kernels found in the
                       namespace *except* those listed here are halted.

"""

# path hackery, no longer requires $PYTHONPATH to be set properly
import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../../"))
import string
from koe.common import pathhack
from koe.common import nstools
# yes bogus, I know
from boothalt import *

BOOTDIR = 'config'
BOOTFN = 'boot'
BOOTCLUSTERFN = 'boot.cluster'

program = sys.argv[0]
quiet = None

DidNotDieError = 'DidNotDieError'


def do_shutdowns(p, koslist, visualizer, worldroot):
    for k in koslist:
        try:
            kernel = nstools.WorldOpen('kos/%s' % k)
        except nstools.BadPathError:
            if not quiet:
                print 'No kernel:', k
                continue
        try:
            admin = nstools.WorldOpen('kos/%s/administrator' % k)
        except nstools.BadPathError:
            if not quiet:
                print 'KOS not running:', k
                continue
        if not quiet:
            write('Shutting down KOS "%s"... ' % k, add_nl=0)
        admin.Shutdown()
        status = unwaitfor('kos/%s/administrator' % k)
        if not quiet:
            if status:
                print 'DID NOT DIE'
            else:
                print 'Done'

    # now shutdown the visualizer
    if visualizer:
        nspath = get(p, 'Visualizer', 'ns-path')
        if not nspath:
            nspath = 'tools/visu.' + pathhack.get_short_hostname()
        try:
            visu = nstools.WorldOpen(nspath)
        except nstools.BadPathError:
            if not quiet:
                print 'No Visualizer'
        else:
            if not quiet:
                write('Shutting down Visualizer... ', add_nl=0)
            visu.Shutdown()
            status = unwaitfor(nspath)
            if not quiet:
                if status:
                    print 'DID NOT DIE'
                else:
                    print 'Done'

    # now shutdown the worldroot
    if worldroot:
        try:
            repl = nstools.WorldOpen('local-worldroot')
        except (nstools.NoWorldrootError, nstools.BadPathError):
            pass
        else:
            if not quiet:
                print 'Shutting down worldroot server'
            repl.Shutdown()


class HaltArgs(ArgParser):
    def __init__(self):
        ArgParser.__init__(
            self, 'fwzaxqh',
            ['file=', 'worldroot', 'visualizer', 'all',
             'exclude', 'quiet', 'help'])
        self.exclude = None
        self.quiet = None
        self.files = []
        self._help = None
        self.kill_visu = None
        self.kill_wr = None

    def usage(self, status=0, errmsg=None):
        if errmsg is not None:
            print errmsg
        print __doc__ % globals()
        sys.exit(status)

    def unhandled_option(self, opt, val):
        print 'Invalid option ignored:', opt

    def shortopt_f(self, opt, val):
        self.files.append(val)
    longopt_file = shortopt_f

    def shortopt_x(self, opt, val):
        self.exclude = 1
    longopt_exclude = shortopt_x

    def shortopt_w(self, opt, val):
        self.kill_wr = 1
    longopt_worldroot = shortopt_w

    def shortopt_z(self, opt, val):
        self.kill_visu = 1
    longopt_visualizer = shortopt_z

    def shortopt_a(self, opt, val):
        self.kill_visu = 1
        self.kill_wr = 1
    longopt_all = shortopt_a

    def shortopt_q(self, opt, val):
        self.quiet = 1
    longopt_quiet = shortopt_q

    def shortopt_h(self, opt, val):
        self._help = 1
    longopt_help = shortopt_h

    def parse(self, argv=None):
        ArgParser.parse(self, argv)
        if self._help:
            self.usage()


def main():
    global quiet
    args = HaltArgs()
    args.parse()
    quiet = args.quiet

    p = make_configparser(args.files)

    includes = {}
    excludes = []
    if args.args:
        if args.exclude:
            excludes = args.args
        else:
            for k in args.args:
                includes[k] = k

    # see if the special kernel name `*' was specified
    if len(includes) == 1 and includes.has_key('*'):
        # discover the list of kernels to shutdown from the namespace
        ns = nstools.WorldOpen('kos')
        for k in ns.List():
            includes[k] = k
        del includes['*']
    elif not includes:
        # kernels specified in the - option of the KERNELS section
        for k in string.split(get(p, 'KERNELS', '-')):
            includes[k] = k
        # kernels specified in the host part of the KERNELS section
        for k in host_kernels(p):
            includes[k] = k

    for k in excludes:
        if includes.has_key(k):
            del includes[k]

    # args now contains a list of kos to kill
    try:
        do_shutdowns(p, includes.keys(), args.kill_visu, args.kill_wr)
    except nstools.NoWorldrootError:
        if not quiet:
            print 'No Worldroot!'

main()
