import marshal
import types
import sys

class CallStats:
    "Calculates timing states for a call described by a list of times."

    def __init__(self, kind, callinfo, times):
        self.times = times
        self.callinfo = callinfo
        self.memo = {}       # memoize stats
        if kind == 0:        # skeleton
            self.stats = { 'total' : ('InputObjectID', ('FinishReply',
                                                        'NoReply')),
                           'request' : ('InputObjectID', 'RequestRead'),
                           'reply' : (('BeginReply', 'NoReply'),
                                      ('FinishReply', 'NoReply')),
                           }
        else:                # stub
            self.stats = { 'total' : ('BeginCall', 'FinishCall'),
                           'request' : ('BeginRequest', 'FinishRequest'),
                           'reply' : ('GetReply', 'ReplyRead'),
                           }
        # begin==1, end==2

    def __len__(self):
        return self.stats

    def __getitem__(self, key):
        if self.memo.has_key(key):
            return self.memo[key]
        try:
            return self.__compute(self.stats[key])
        except KeyError:
            raise KeyError, "unknown statistic"
        except ValueError:
            raise KeyError, "missing timing info needed for statistic"

    def __match(self, direction, pattern, event):
        if type(pattern) != types.TupleType:
            names = (pattern,)
        else:
            names = pattern
        if event[1] == direction:
            for name in names:
                if event[0] == name:
                    return event[2]
        return None

    def __compute(self, how):
        begin, end = how
        tbegin = tend = None
        for event in self.times:
            if tbegin is None:
                tbegin = self.__match(1, begin, event)
            if tend is None:
                tend = self.__match(2, end, event)
                if tend:
                    break
        try:
            return tend - tbegin
        except TypeError:
            # didn't find a tbegin or tend
            raise ValueError, "missing timing info"
        begin, end = how
        tbegin = tend = None
        for event in self.times:
            if tbegin is None:
                tbegin = self.__match(1, begin, event)
            if tend is None:
                tend = self.__match(2, end, event)
                if tend:
                    break
        try:
            return tend - tbegin
        except TypeError:
            # didn't find a tbegin or tend
            raise ValueError, "missing timing info"

def main(path):
    f = open(path)
    try:
        # first line is information about process
        sysinfo = marshal.load(f)
        print sysinfo
        while 1:
            kind, call, times = marshal.load(f)
            if not call:
                continue
            stat = CallStats(kind, call, times)
            methname = "%s.%s" % (call[0], call[1])
            if len(methname) > 30:
                print methname, kind, ":\n", " " * 33,
            else:
                print methname, kind, ":", " " * (30-len(methname)),
            try:
                print "total=%-10.4f" % (stat['total']),
            except KeyError:
                print "total=n/a",
            try:
                print "request=%-10.4f" % (stat['request']),
            except KeyError:
                print "request=n/a",
            try:
                print "reply=%-10.4f" % (stat['reply'])
            except KeyError:
                print "reply=n/a"
    except EOFError:
        # we're done
        pass
        

if __name__ == "__main__":
    main(sys.argv[1])
