#! /usr/bin/env python

"""A ps-like utility for the KOS."""

import sys, os
sys.path.insert(0, os.path.normpath(os.path.join(sys.path[0], "../../")))
import string
import regex
from koe.common import nstools
from koe.interfaces.stubs.AdministratorAPI import KOSProcessType

# Three functions for running ps and parsing its output.

CMD_NAMES = ['CMD', 'COMMAND', 'COMD']
LEFT = ['TTY', 'TT'] + CMD_NAMES

# start by parsing the 
def get_fields(header):
    alpha = regex.compile('[A-Z%]')
    fields = []
    start = 0
    end = 0
    offset = 0
    try:
        while 1:
            start = string.index(header, ' ', offset)
            title = alpha.search(header, offset)
            end = string.index(header, ' ', title)
            fields.append((header[title:end], start, end))
            offset = end
    except ValueError:
        fields.append(header[title:], start, len(header))
    return fields

def parse_ps_line(line, fields):
    global LEFT
    entry = {}
    for title, start, end in fields:
        if not title in LEFT:
            entry[title] = string.strip(line[start:end])
        else:
            # have to special case some entries, because they
            # are left justified
            if title in CMD_NAMES:
                entry[title] = line[start+1:]
            else:
                # not sure if anything gets lost here
                poss = line[start+1:]
                end = string.find(poss, ' ')
                entry[title] = poss[:end]
    return entry

SOLARIS_PS_COMMAND = \
 "/usr/bin/ps -e -o user,pid,ppid,pgid,vsz=VMEM_KB -o stime,etime,time,args"
IRIX_PS_COMMAND = "/bin/ps -e -f"
DEFAULT_PS_COMMAND = "ps -e -f"

if sys.platform == "irix5":
    ps_command = IRIX_PS_COMMAND
    cmd_key = "COMD"
elif sys.platform == "irix6":
    ps_command = IRIX_PS_COMMAND
    cmd_key = "CMD"
elif sys.platform == "sunos5":
    ps_command = SOLARIS_PS_COMMAND
    cmd_key = "COMMAND"
else:
    ps_command = DEFAULT_PS_COMMAND
    cmd_key = "COMMAND"

def get_ps():
    # using only names defined in the POSIX locale
    # probably doesn't buy us platform independence anyway
    ps = os.popen(ps_command)
    lines = string.splitfields(ps.read(), '\n')
    ps.close()
    header = lines[0]
    fields = get_fields(header)
    procs = {}
    for line in lines[1:]:
        entry = parse_ps_line(line, fields)
        procs[entry['PID']] = entry
    return procs, fields

def search_ps(procs, field, rx):
    "Search ps output for value matching rx in column field."
    match = []
    for pid, entry in procs.items():
        try:
            val = entry[field]
        except KeyError:
            pass
        else:
            if rx.search(val) != -1:
                match.append(pid)
    return match

def guess_local_kernels(procs):
    "Return a list of kernel names running on local host."
    # This is an ugly hack. We look for all the kernel.py processes,
    # and guess their names based on either the -n arg or the nodename.
    # ask the kernels whats up
    # merge the two lists
    kospids = search_ps(procs, cmd_key, regex.compile('kernel\.py'))
    kosnames = []
    kname = regex.compile('-n[\' \']+\([^\' \']+\)')
    for pid in kospids:
        cmd = procs[pid][cmd_key]
        if kname.search(cmd) != -1:
            kosnames.append(kname.group(1))
        else:
            hostname = os.uname()[1]
            i = string.find(hostname, '.')
            if i != -1:
                hostname = hostname[:i]
            kosnames.append(hostname)
    return kosnames

def get_kernel_admin(kos_name):
    "Get administrator interface for kernel named kos_name."
    try:
        return nstools.WorldOpen('kos/%s/administrator' % kos_name)
    except nstools.BadPathError, msg:
        print "Error: could not reach administrator interface for", kos_name
        print "\t", msg
        return None

class Formatter:

    "Produce tables based on a format description."

    def __init__(self):
        self.fields = []
        self.labels = ()
        self.fmt = ''

    def define(self, *fields):
        self.fmt = ''
        labels = []
        for field, label, len, align in fields:
            self.fields.append(field)
            labels.append(label)
            if align == 'left':
                fmt = '%-'
            else:
                fmt = '%'
            fmt = fmt + '%d.%d' % (len, len) + 's '
            self.fmt = self.fmt + fmt
        if self.fmt[-1:] == ' ':
            self.fmt = self.fmt[:-1]
        self.labels = tuple(labels)

    def header(self):
        return self.fmt % self.labels

    def format(self, *proc):
        vals = []
        for field in self.fields:
                for p in proc:
                    miss = 1
                    try:
                        val = p[field]
                    except KeyError:
                        pass
                    else:
                        vals.append(val)
                        miss = 0
                        break
                if miss:
                    vals.append('')
        return self.fmt % tuple(vals)

# Define the default format
default = Formatter()
ALEFT = 'left'
ARIGHT = 'right'
default.define(('type', 'TYPE', 4, ARIGHT),
               ('kernel', 'KERNEL', 8, ARIGHT),
               ('name', 'NAME', 10, ARIGHT),
               ('kid', 'KPID', 4, ARIGHT),
               ('PID', 'PID', 5, ARIGHT),
               ('STIME', 'STIME', 8, ARIGHT),
               ('TIME', 'TIME', 8, ARIGHT),
               ('VMEM_KB', 'MEM', 6, ARIGHT),
               ('COMMAND', 'COMMAND', 18, ALEFT))

type_map = { KOSProcessType.kos        : 'KOS',
             KOSProcessType.kp         : 'KP',
             KOSProcessType.extension  : 'EXT',
             KOSProcessType.kp_starter : 'sta'
             }

def output(kos, info, proc, fmt=default):
    "Output an entry describing a KOS process."
    global type_map
    # create custom output entries
    custom = {'kernel': kos,
              'type': type_map[info['type']],
              'COMMAND': proc[cmd_key],
              }
    if info['type'] == KOSProcessType.kp:
        i = string.find(info['name'], '-')
        if i != -1:
            custom['name'] = info['name'][i+1:]
    print fmt.format(custom, info, proc)

def main():
    # get all of the PS data
    procs, fields = get_ps()
    kos_names = guess_local_kernels(procs)
    kos_procs = {}
    for kos in kos_names:
        admin = get_kernel_admin(kos)
        if admin:
            list = admin.GetProcessInfo()
            kos_procs[kos] = list

    print default.header()
    for kos in kos_procs.keys():
        for info in kos_procs[kos]:
            pid = str(info['pid'])
            if procs.has_key(pid):
                output(kos, info, procs[pid])

if __name__ == "__main__":
    main()
