#! /usr/bin/env python

# This documentation string also serves as a usage message
"""Namespace listing program.

 This program dumps out a namespace, starting from whatever Worldroot
 is reachable.  It prints a tree of all names, following subcontexts
 (subdirectories), and properly avoids loops in the namespace.  For
 each name it finds it displays the type of object the name is bound
 to (unless the type is 'Namespace.Context').

 The program will also display a message if a loop was encountered, or
 if the bound object was un-responsive.  The latter can occur if the
 true server object has disappeared but the name was never unbound.

 By using the -c (--clean) option, you can clean out the names of
 unresponsive objects.  Use this feature with care!

Usage:
    %(program)s [-c | -q | -h] [root [root ...]]

Where:

    --clean
    -c            -- clean out unresponsive names

    --quiet
    -q            -- do not dump the namespace (useful with --clean)

    --help
    -h            -- display this help message and exit.

    Optional `root' arguments can be used to dump sub-contexts rooted
    at the supplied paths.  Without no root arguments, the entire tree
    rooted at the Worldroot is dumped.

"""

# path hackery, no longer requires $PYTHONPATH to be set properly
import sys, os; sys.path.append(os.path.join(sys.path[0], "../../"))
import string
import getopt
from koe.common import nstools
from koe.interfaces.stubs import Namespace

# keep track of object references we've seen so we can break loops
refs = {}
cleaned = {}


def dump(context, tabs=0, cleaning=0):
    tab = '    ' * tabs
    contexts = context.List()
    contexts.sort()
    for name in contexts:
        print '%s%s' % (tab, name),
        try:
            desc = context.Lookup1(name)
        except nstools.BadPathError:
            print '<deleted>'
            continue
        if refs.has_key(desc.reference()):
            if cleaning and cleaned.has_key(desc.reference()):
                print '<loop detected>', '<already cleaned>',
                try:
                    context.Unbind(name)
                    print  '<cleaned out!>',
                except Namespace.BindError, msg:
                    print "<can't clean: %s>" % str(msg)
                print
            else:
                print '<loop detected>'
            continue
        refs[desc.reference()] = 1
        #
        # always open it, looking for non-responders
        obj = desc.Open()
        if not obj:
            print '<cannot open>'
            continue
        if desc.typename() <> 'Namespace.Context':
            print '<%s>' % desc.typename(),
            ok = obj.IluPing()
        else:
            try:
                obj.List()
            except:
                ok = 0
            else:
                ok = 1
        if not ok:
            print '<not responding>',
            if cleaning:
                try:
                    context.Unbind(name)
                    print '<cleaned out!>',
                    cleaned[desc.reference()] = 1
                except Namespace.BindError, msg:
                    print "<can't clean: %s>" % str(msg),
            print
            continue
        if desc.typename() != 'Namespace.Context':
            print
        elif name == 'world':
            print '<another world>'
        else:
            print
            dump(obj, tabs+1, cleaning and name != 'replicators')


def reset():
    global cleaned, refs
    cleaned = {}
    refs = {}



program = sys.argv[0]

def usage(status):
    print __doc__ % globals()
    sys.exit(status)

def main():
    # default options
    clean = 0
    quiet = 0

    try:
        opts, args = getopt.getopt(sys.argv[1:], 'chq',
                                   ['clean', 'quiet', 'help'])
    except getopt.error, msg:
        print msg
        usage(1)

    roots = []
    if args:
        roots = args

    for opt, val in opts:
        if opt in ('-h', '--help'):
            usage(0)
        elif opt in ('-c', '--clean'):
            clean = 1
        elif opt in ('-q', '--quiet'):
            quiet = 1

    # try to get the worldroot namespace context
    try:
        worldroot = nstools.GetWorldroot()
    except nstools.NoWorldrootError, msg:
        print "No worldroot:", msg
        sys.exit(1)

    if not roots:
        roots = ['']

    dumpables = []
    for tree in roots:
        if tree == '':
            refs[worldroot.GetDescriptor().reference()] = 1
            dumpables.append((worldroot, None))
        else:
            ns = worldroot
            for path in string.split(tree, '/'):
                try:
                    desc = ns.Lookup1(path)
                except nstools.BadPathError:
                    if not quiet:
                        print tree, 'is inaccessible'
                    break
                ns = desc.Open('Namespace.Context')
                if not ns:
                    if not quiet:
                        print tree, 'is not a sub-tree'
                    break
                refs[ns.GetDescriptor().reference()] = 1
            else:
                dumpables.append((ns, tree))

    oldstdout = sys.stdout
    try:
        # TBD: blech. disgusting hack so I don't have to propagate and
        # check the quiet flag everywhere!
        if quiet:
            try:
                fp = open('/dev/null', 'w')
            except IOError:
                import StringIO
                fp = StringIO.StringIO()
            sys.stdout = fp
        for ns, path in dumpables:
            if path:
                print path
                dump(ns, 1, clean)
            else:
                dump(ns, 0, clean)
    finally:
        sys.stdout = oldstdout


if __name__ == '__main__':
    main()
