"""General Tk Canvas based output forms."""


import Tkinter; Tk = Tkinter
import copy
import string
from koe.common import physfonts
import time

from types import TupleType

#  Exception:
FormWindowError = "FormWindowError"


#  Fonts:
LIGHT_FONT = physfonts.BODY_R_M
BOLD_FONT = physfonts.BODY_B_M
LIGHT_ITALIC_FONT = physfonts.BODY_I_M
BOLD_ITALIX_FONT = physfonts.BODY_BI_M
DEFAULT_FONT = LIGHT_FONT

#  Field types:
STRING = 'string'
DATETIME = 'datetime'
LABEL = 'label'
INT = 'int'
REAL = 'real'

LABEL_DEFAULTS = {
    "anchor": Tk.SE, "justify": Tk.RIGHT, "fill": "midnightBlue"
    }
STRING_DEFAULTS = {
    "anchor": Tk.SW, "justify": Tk.LEFT
    }
DATETIME_DEFAULTS = copy.copy(STRING_DEFAULTS)
INT_DEFAULTS = copy.copy(STRING_DEFAULTS)
REAL_DEFAULTS = copy.copy(STRING_DEFAULTS)


def fmtTime(t):
    "Return time string in desired format."
    return time.ctime(t)[4:-5]


class FormWindow(Tk.Canvas):
    """Form-oriented output 

    This class uses a form-based approach with a data-structure
    providing the definitions of the fields which need to be added to
    the display.  The display itself is a canvas, so can be used for
    just about any arrangement.

    The input to the form-generator is a sequence of sequences; each
    inner sequence represents a single field of the form.  If a field
    is named, it is a variable and can be assigned to using
    `form_window["name"] = value'.  The inner sequences are formed as
    follows:

      (type, x-position, y-position, name|None, value|None [, options-dict])

     """
    _xpos = 0
    _ypos = 0

    def __init__(self, master, field_seq = (),
                 relief=Tk.SUNKEN, borderwidth=2, **kw):
        self._master = master
        kw["relief"] = relief
        kw["borderwidth"] = borderwidth
        self._window = apply(Tk.Canvas, (master,), kw)
        self._fields = {}
        xpos = ypos = 0
        for ent in field_seq:
            apply(self.add, tuple(ent))

    def __setitem__(self, name, value):
        if self._fields.has_key(name):
            self._fields[name].set(value)

    def __getitem__(self, name):
        self._fields[name].get()

    def add(self, ftype, x, y, name, value, options=None):
        try:
            method = getattr(self, 'add_' + ftype)
        except AttributeError:
            raise FormWindowError, "unknown field type: " + ftype
        if options is None:
            options = {}
        return apply(method, (x, y, name, value), options)

    def add_label(self, x, y, name, value="", **kw):
        x, y = self._position(x, y)
        kw["value"] = value
        field = apply(CanvasStringVar, (self._window, x, y),
                      _merge(kw, LABEL_DEFAULTS))
        if name:
            self._fields[name] = field
        return field

    def add_string(self, x, y, name, value="", **kw):
        x, y = self._position(x, y)
        field = apply(CanvasStringVar, (self._window, x, y),
                      _merge(kw, STRING_DEFAULTS))
        field.set(value)
        if name:
            self._fields[name] = field
        return field

    def add_int(self, x, y, name, value=None, **kw):
        if not name:
            raise RuntimeError, "name is required for int fields"
        x, y = self._position(x, y)
        self._fields[name] = apply(CanvasIntVar, (self._window, x, y),
                                   _merge(kw, INT_DEFAULTS))
        self._fields[name].set(value)
        return self._fields[name]

    def add_real(self, x, y, name, value=None, **kw):
        if not name:
            raise RuntimeError, "name is required for real fields"
        x, y = self._position(x, y)
        self._fields[name] = apply(CanvasRealVar, (self._window, x, y),
                                   _merge(kw, REAL_DEFAULTS))
        self._fields[name].set(value)
        return self._fields[name]

    def add_datetime(self, x, y, name, value=None, **kw):
        if not name:
            raise RuntimeError, "name is required for datetime fields"
        x, y = self._position(x, y)
        self._fields[name] = apply(CanvasDateTimeVar, (self._window, x, y),
                                   _merge(kw, DATETIME_DEFAULTS))
        self._fields[name].set(value)
        return self._fields[name]

    def clear(self):
        for field in self._fields.values():
            field.set(None)

    def destroy(self):
        for name, field in self._fields.items():
            field.destroy()
            del self._fields[name]
        self._window.destroy()
        self._window = None

    def pack(self, **kw):
        return apply(self._window.pack, (), kw)

    def get_window(self):
        return self._window

    def _position(self, x, y):
        if type(x) is TupleType:
            if x:
                x = self._xpos + x[0]
            else:
                x = self._xpos
        if type(y) is TupleType:
            if y:
                y = self._ypos + y[0]
            else:
                y = self._ypos
        self._xpos, self._ypos = x, y
        return x, y


class CanvasStringVar:
    def __init__(self, canvas, x, y,
                 value=None, justify=Tk.LEFT, anchor=Tk.SW,
                 font=DEFAULT_FONT, **kw):
        self._canvas = canvas
        kw["anchor"] = anchor
        kw["justify"] = justify
        self._value = value
        kw["text"] = value or ""
        kw["font"] = font
        self._item = apply(canvas.create_text, (x, y), kw)

    def set(self, value):
        self._value = value
        self._canvas.itemconfig(self._item, text=(value or ""))

    def get(self):
        return self._value

    def destroy(self):
        self._canvas = None


class CanvasReprVar(CanvasStringVar):
    def __init__(self, canvas, x, y, value=None, **kw):
        if value is not None:
            kw["value"] = `value`
        apply(CanvasStringVar.__init__, (self, canvas, x, y), kw)
        self._value = value

    def set(self, value):
        if value is None:
            v = None
        else:
            v = `value`
        CanvasStringVar.set(self, v)
        self._value = value


class CanvasIntVar(CanvasReprVar):
    pass

class CanvasRealVar(CanvasReprVar):
    pass


class CanvasDateTimeVar(CanvasStringVar):
    def __init__(self, canvas, x, y, value=None, **kw):
        kw["value"] = None
        apply(CanvasStringVar.__init__, (self, canvas, x, y), kw)
        self._value = value
        if value is not None:
            self.set(value)

    def set(self, value):
        v = value and fmtTime(value) or None
        CanvasStringVar.set(self, v)
        self._value = value


def _merge(primary, secondary):
    dict = {}
    for k, v in primary.items():
        dict[k] = v
    for k, v in secondary.items():
        if not dict.has_key(k):
            dict[k] = v
    return dict
