"""Interactive list allowing selection of a single KP in the current 'mode.'
"""

import string
import Tkinter
from koe.common import tktools


ACTIVE = 'active'
COMPLETED = 'completed'
CURRENT = 'current'
HISTORIC = 'historic'

class KPList:
    def __init__(self, master=None, width=30, clear_handler=None, **kw):
        self.__kpdicts = {ACTIVE: {},
                          COMPLETED: {},
                          CURRENT: {},
                          HISTORIC: {}}
        self.__which = Tkinter.StringVar()
        self.__clear_handler = clear_handler
        self._create_widgets(master, width, kw)
        self.select_list(CURRENT)

    def _create_widgets(self, master, width, kw):
        frame = Tkinter.Frame(master, class_="KnowbotList")
        frame.pack(expand=1, fill=Tkinter.BOTH)
        # do the options next
        fr = Tkinter.Frame(frame)
        fr.pack(anchor=Tkinter.W, side=Tkinter.BOTTOM)
        self.__prune_button = Tkinter.Button(
            fr, text="Prune", command=self.__prune_event)
        self.__prune_button.pack(side=Tkinter.LEFT)
        self.__left_radio = Tkinter.Radiobutton(
            fr, variable=self.__which, command=self.__rebuild_list_event)
        self.__left_radio.pack(side=Tkinter.LEFT)
        self.__right_radio = Tkinter.Radiobutton(
            fr, variable=self.__which, command=self.__rebuild_list_event)
        self.__right_radio.pack(side=Tkinter.LEFT)
        self.config_radios(CURRENT, HISTORIC)
        # the listbox is last, 'cause of the packer
        listbox, f = tktools.make_list_box(frame, width=width, takefocus=0)
        self.__listbox, self.__frame = listbox, frame
        listbox["exportselection"] = 0
        listbox["takefocus"] = 0

    def __getitem__(self, option):
        return self.__listbox[option]

    def __setitem__(self, option, value):
        self.__listbox[option] = value

    def pack(self, *args, **kw):
        return apply(self.__frame.pack, args, kw)

    def bind(self, *args, **kw):
        return apply(self.__listbox.bind, args, kw)

    def config_radios(self, left, right):
        self.__left_radio.config(value=left, text=string.capwords(left))
        self.__right_radio.config(value=right, text=string.capwords(right))

    def select_list(self, which):
        old_which = self.__which.get()
        if which != old_which:
            old_selected = self.get_reporting_id()
            self.unselect_kp()
            self.__which.set(which)
            self.__rebuild_list()
            self.select_kp(old_selected)

    def get_ids(self, which):
        return self.__get_kpdict(which).keys()

    def get_kpinfos(self, which):
        return self.__get_kpdict(which).values()

    def get_kpinfo(self, repid=None):
        if not repid:
            repid = self.get_reporting_id()
        # get a specific kpinfo, or None
        for which, kpdict in self.__kpdicts.items():
            if kpdict.has_key(repid):
                return kpdict[repid]
        return None

    def get_reporting_id(self, index=None):
        # returns the reporting-id of the specified index, defaulting to the
        # the selected kp.
        if index:
            return self.get(index)
        indexes = self.__listbox.curselection()
        if indexes:
            ids = self.__get_kpdict().keys()
            if ids:
                ids.sort()
                return ids[string.atoi(indexes[0])]
        return None

    def get(self, index):
        # essentially just a listbox.get(), but only returns the reporting-id
        if index is None:
            index = Tkinter.ACTIVE
        try:
            index = self.__listbox.index(index)
        except IndexError:
            return None
        ids = self.__get_kpdict().keys()
        if ids:
            ids.sort()
            return ids[index]
        return None

    def insert_kp(self, kpinfo):
        repid = kpinfo.get_reporting_id()
        self.__get_kpdict(ACTIVE)[repid] = kpinfo
        self.__get_kpdict(CURRENT)[repid] = kpinfo
        if self.__which.get() in (ACTIVE, CURRENT):
            self.__rebuild_list()

    def select_kp(self, kp):
        if not kp:
            return
        if type(kp) is type(self):      # if really a kpinfo, get reporting-id
            kp = kp.get_reporting_id()
        repid = kp
        kpdict = self.__get_kpdict()
        self.__listbox.select_clear(0, Tkinter.END)
        if kpdict.has_key(repid):
            ids = kpdict.keys()
            ids.sort()
            index = ids.index(repid)
            self.__listbox.select_set(index)
            self.__listbox.see(index)
            kpdict[repid].enable_display()
        elif self.__clear_handler:
            self.__clear_handler()

    def unselect_kp(self):
        self.__listbox.select_clear(0, Tkinter.END)
        if self.__clear_handler:
            self.__clear_handler()

    def update_kp(self, kpinfo):
        repid = kpinfo.get_reporting_id()
        which = self.__which.get()
        kpdict = self.__get_kpdict(which)
        redraw = kpdict.has_key(repid)
        active = self.__get_kpdict(ACTIVE)
        if kpinfo.get_completion_time() and active.has_key(repid):
            # kp terminated for some reason.
            if which == ACTIVE:
                self.unselect_kp()
            del active[repid]
            self.__get_kpdict(COMPLETED)[repid] = kpinfo
        if redraw or kpdict.has_key(repid):
            self.__rebuild_list()

    def prune(self, which):
        self.__listbox.select_clear(0, Tkinter.END)
        if which == CURRENT:
            self.__prune_currents()
        elif which in (HISTORIC, COMPLETED):
            self.__kpdicts[which] = {}
            self.__rebuild_list()

    def __prune_event(self, event=None):
        self.prune(self.__which.get())

    def __prune_currents(self):
        currents = self.__get_kpdict(CURRENT)
        historic = self.__get_kpdict(HISTORIC)
        changed = 0
        for repid, kpinfo in currents.items():
            if kpinfo.get_completion_time():
                historic[repid] = kpinfo
                del currents[repid]
                changed = 1
        if changed:
            self.__rebuild_list()

    def __get_kpdict(self, which=None):
        if not which:
            which = self.__which.get()
        return self.__kpdicts[which]

    def __rebuild_list_event(self, event=None):
        self.unselect_kp()
        self.__rebuild_list()

    def __rebuild_list(self):
        old_selected = self.get_reporting_id()
        old_index = self.__listbox.index(Tkinter.At(1,1))
        self.__listbox.delete(0, Tkinter.END)
        kps = self.__get_kpdict().items()
        kps.sort()
        for repid, kpinfo in kps:
            s = self.__format_kpinfo(kpinfo)
            self.__listbox.insert(Tkinter.END, s)
        self.__listbox.yview(old_index)
        which = self.__which.get()
        if which == ACTIVE:
            self.__prune_button['state'] = Tkinter.DISABLED
        else:
            self.__prune_button['state'] = Tkinter.NORMAL
        self.select_kp(old_selected)
        if self.__clear_handler and not self.get_reporting_id():
            self.__clear_handler()

    def __format_kpinfo(self, kpinfo):
        comment = kpinfo.get_completion_time() and "died" or "now"
        args = kpinfo.get_reporting_id(), comment
        format = "%s %s at %s (%s)"
        s = format % (args + kpinfo.get_location())
        return s
