"""Parse a KP template file.

The syntax is essentially a bunch of RFC822 message headers, but blank
lines and lines that don't conform to the 'key: value' format are
ignored rather than aborting the parsing.  Continuation lines
\(starting with whitespace) are honored properly.  Key names (but not
values) are converted to lowercase before they are used.  Values are
stripped of leading and trailing whitespace; continuations are
represented by an embedded newline plus a space; otherwise, internal
whitespace is left unchanged.

The only argument is an open file, which will be read until EOF.

The `parse()' return value is a triple of dictionaries.

See KOE_ROOT/../docs/specs/tmpl_files.html for information on the KP
template file format.

"""

import string
import regex


DESCRIPTOR_START = "kp-template--"

validpat = "^\([-a-z0-9_]*\):\(.*\)$"
valid = regex.compile(validpat, regex.casefold)

debug = 0

def parse(fp):
    """Parse a KP template file.  See module docstring.

    Returns a triple of metadata headers, template headers, and tool-specific
    extension headers.  The leading strings for the template and tool-specific
    headers are stripped.  The tool-specific return value is a dictionary of
    tool ID --> dictionary mappings.
    """
    mheaders = {}                       # metadata headers
    theaders = {}                       # template headers
    xheaders = {}                       # tool-specific headers
    name = None                         # used for continuation line
    lineno = 0
    desc_start_len = len(DESCRIPTOR_START)
    while 1:
        line = fp.readline()
        if not line:
            break
        lineno = lineno + 1
        if line[0] == '#' or line == "\n":
            name = None
            continue
        if line[0] in string.whitespace:
            # It looks line a continuation line.
            if name:
                # Continue the previous line
                value = string.strip(line)
                if value:
                    if headers[name]:
                        headers[name] = headers[name] + "\n\t" + value
                    else:
                        headers[name] = value
            else:
                # It's a bad line, ignore it.
                if debug:
                    print "Error at", lineno, ":", `line`
        elif valid.match(line) > 0:
            # It's a header line.
            name, value = valid.group(1, 2)
            name = string.lower(name)
            value = string.strip(value)
            # separate the three groups of headers: metadata, template, tool
            if len(name) >= desc_start_len \
               and name[:desc_start_len] == DESCRIPTOR_START:
                headers = theaders
                name = name[desc_start_len:]
                pos = string.find(name, "--")
                if pos >= 0:
                    ext, name = name[:pos], name[pos + 2:]
                    if not xheaders.has_key(ext):
                        xheaders[ext] = {}
                    headers = xheaders[ext]
            else:
                headers = mheaders
            headers[name] = value # XXX Override a previous value
        elif string.strip(line) != "":
            # It's a bad line.  Ignore it.
            if debug:
                print "Error at", lineno, ":", `line`

    return mheaders, theaders, xheaders


from koe.common import kptools


class FilenameResolver:
    def __init__(self, base, alt):
        self.__basedir = base
        self.__altdir = alt

    def __call__(self, relfn):
        fn = os.path.join(self.__basedir, relfn)
        if os.path.exists(fn):
            return fn
        fn = os.path.join(self.__altdir, relfn)
        if os.path.exists(fn):
            return fn
        return None


class TemplateRepresentation(kptools.Representation):
    def __init__(self, basedir=None, headers=None):
        kptools.Representation.__init__(self)
        if basedir and headers:
            self.load_from_descriptor(basedir, headers)

    def load_from_descriptor(self, basedir, (mheaders, theaders, xheaders)):
        import os
        meta = self.GetMetadata()
        code = self.GetCode()
        altdir = theaders.has_key("modules-base") and theaders["modules-base"]
        resolve = FilenameResolver(basedir, altdir)
        meta.SetValue("knowbot-class", "KP")
        for k, v in mheaders.items():
            meta.SetValue(k, v)
        if theaders.has_key("modules"):
            modules = string.split(theaders["modules"])
        else:
            modules = ()
        for modname in modules:
            try:
                modfile = theaders["module-" + modname]
            except KeyError:
                raise kptools.LoadingError, \
                      "missing filename for %s module" % modname
            modfile = resolve(modfile)
            if not modfile:
                raise kptools.LoadingError, \
                      "could not locate " + modfile
            self.add_module(modname, open(modfile).read())
        self.load_suitcase((mheaders, theaders, xheaders))

    def load_suitcase(self, (mheaders, theaders, xheaders)):
        """Load the suitcase from a specification dictionary."""
        if not xheaders \
           or xheaders.has_key("suitcase") \
           or xheaders["suitcase"]:
            return
        sfs = self.get_suitecase()
        # this ensures we hit directories before files...
        items = xheaders["suitcase"].items()
        items.sort()
        for path, source in items:
            if not source:
                continue
            if source[0] == "=":
                # literal data
                data = source[1:]
                fp = sfs.open(path, "wb")
                fp.write(data)
                if data and data[-1] != "\n":
                    fp.write("\n")
                fp.close()
                continue
            if os.path.isfile(source):
                pass
            else:
                pass


def test():
    """Test program for parse().

    This takes a filename as command line argument;
    if no filename is given, it parses ../grail-defaults.
    It also times how long it takes.

    """
    import os
    import sys
    import time
    global debug
    debug = 1
    if sys.argv[1:]:
        fn = sys.argv[1]
    else:
        fn = os.path.join(os.environ["HOME"], ".grail/grail-preferences")
    fp = open(fn)
    t0 = time.time()
    mheaders, theaders, xheaders = parse(fp)
    t1 = time.time()
    fp.close()
    print "Parsing time", round(t1-t0, 3)
    # print metadata headers
    if mheaders:
        print "\n# Metadata headers:"
    names = mheaders.keys()
    names.sort()
    for name in names:
        value = mheaders[name]
        print "%s: %s" % (name, value)
    # print template headers
    if theaders:
        print "\n# Standard template headers:"
    names = theaders.keys()
    names.sort()
    for name in names:
        value = theaders[name]
        print "%s%s: %s" % (DESCRIPTOR_START, name, value)
    # print tool-specific headers
    if xheaders:
        print "\n# Tool-specific headers:"
    dicts = xheaders.items()
    dicts.sort()
    for tool, dict in dicts:
        items = dict.items()
        items.sort()
        print "# [%s]" % tool
        for name, value in items:
            print "%s%s--%s: %s" % (DESCRIPTOR_START, tool, name, value)


if __name__ == '__main__':
    test()
