"""GUI KPLauncher for the graphical 'reporting station' application.
"""

import FileDialog
import fnmatch
import Tkinter ; Tk = Tkinter
import string
import os, ilu, sys
from koe.common import tktools

from koe.common import kptools
from koe.common import nstools
from koe.common import pathhack
from koe.common import physfonts
from koe.interfaces.stubs import FileAPI__skel
import GuiFileImpl
from koe.interfaces.stubs import KOS
from koe.common import RS


STDOUT_COLOR = 'grey40'
STDERR_COLOR = 'black'

NO_KERNEL_FOUND = "<none found>"

GROUP_LABEL_FONT = physfonts.BODY_R_S


def make_group(master, name, label=None, fill=Tk.Y, side=None, expand=None):
    return tktools.make_group_frame(master, name=name, label=label,
                                    font=GROUP_LABEL_FONT,
                                    fill=fill, side=side, expand=expand)


def receiver_names():
    """Retrieves list of all currently advertised receivers."""
    try:
        ns_kos = nstools.WorldOpen("kos", "Namespace.Context")
    except nstools.BadPathError:
        return ()
    except nstools.NoWorldrootError:
        return None
    result = []
    for name in ns_kos.List():
        try:
            desc = nstools.Lookup(ns_kos, name + "/receiver")
        except nstools.BadPathError:
            pass
        else:
            if desc and desc.Open("KOS.Receiver"):
                result.append(name)
    result.sort()
    return result



EXPOSE_ON_OUTPUT = 'on output'
EXPOSE_ON_ERROR = 'on error'
EXPOSE_ON_REQUEST = 'on request'

DEFAULT_EXPOSURE = EXPOSE_ON_OUTPUT


class KPConsole(GuiFileImpl.Console):
    def __init__(self, master=None, exposure=DEFAULT_EXPOSURE, *args, **kw):
        self.__kps = {}
        apply(GuiFileImpl.Console.__init__, (self, master) + args, kw)
        # specialized controls:
        frame = self.get_control_frame()
        self.__exposure = v = Tkinter.StringVar()
        self.__lbl = Tkinter.Label(
            frame, text="Expose on")
        self.__lbl.pack(side=Tkinter.LEFT)
        self.__rb1 = Tkinter.Radiobutton(
            frame, text="all output", command=self.__expose,
            variable=v, value=EXPOSE_ON_OUTPUT)
        self.__rb1.pack(side=Tkinter.LEFT)
        self.__rb2 = Tkinter.Radiobutton(
            frame, text="errors", command=self.__expose,
            variable=v, value=EXPOSE_ON_ERROR)
        self.__rb2.pack(side=Tkinter.LEFT)
        self.__rb3 = Tkinter.Radiobutton(
            frame, text="request", command=self.__expose,
            variable=v, value=EXPOSE_ON_REQUEST)
        self.__rb3.pack(side=Tkinter.LEFT)
        self.tag_configure('stdout', foreground=STDOUT_COLOR)
        self.tag_configure('stderr', foreground=STDERR_COLOR)
        self.set_exposure(exposure)
        if master:
            tktools.set_transient(self, master, expose=0)

    __app = None
    def set_app(self, app):
        self.__app = app

    def add_kp(self, repid):
        if not self.__kps.has_key(repid):
            self.__kps[repid] = 1

    def remove_kp(self, repid):
        if not self.__kps.has_key(repid):
            return
        if not self.__kps[repid]:
            return
        self.__kps[repid] = 0
        if not filter(None, self.__kps.values()):
            self.__rb1.forget()
            self.__rb2.forget()
            self.__rb3.forget()
            self.__lbl.config(text="No more active KPs.", foreground="red")
            frame = self.get_control_frame()
            self.add_button(Tk.Button(
                frame, text="Discard", width=7, command=self.__discard))

    def set_exposure(self, exposure):
        self.__exposure.set(exposure)
        self.__expose()

    def __discard(self, event=None):
        if self.__app:
            for repid in self.__kps.keys():
                kpinfo = self.__app.fetch_kpinfo(repid, create=0)
                if kpinfo:
                    kpinfo.set_console(None)
                    self.__app.update_kp(kpinfo)
            self.__app = None
        self.destroy()

    def __expose(self, event=None):
        when = self.__exposure.get()
        if when == EXPOSE_ON_REQUEST:
            self.expose_on_tag('stdout', 0)
            self.expose_on_tag('stderr', 0)
        elif when == EXPOSE_ON_ERROR:
            self.expose_on_tag('stdout', 0)
            self.expose_on_tag('stderr', 1)
        else:
            self.expose_on_tag('stdout', 1)
            self.expose_on_tag('stderr', 1)


class IOServerWriter(GuiFileImpl.TaggedWriter, FileAPI__skel.File):
    pass


class FileDialogMixin:
    def __init__(self):
        self.__control_frames = []
        # repack the left scrollbar so it isn't quite as heinous
        self.dirs.pack(before=self.dirsbar)
        self.botframe.pack(fill=Tk.X, side=Tk.BOTTOM, padx=2, pady=2)
        self.selection.pack(side=Tk.TOP, after=self.midframe)

    mask_hidden = 1
    def filter_command(self, event=None):
        # slightly changed version from FileDialog.FileDialog: removes
        # 'dot' files if pattern starts with "*" and self.mask_hidden is true
        dir, pat = self.get_filter()
        try:
            names = os.listdir(dir)
        except os.error:
            self.master.bell()
            return
        self.directory = dir
        self.set_filter(dir, pat)
        names.sort()
        subdirs = [os.pardir]
        matchingfiles = []
        mask_hidden = self.mask_hidden
        for name in names:
            fullname = os.path.join(dir, name)
            if os.path.isdir(fullname):
                if fullname != ".." and fullname[0] == "." and mask_hidden:
                    continue
                subdirs.append(name)
            elif fnmatch.fnmatch(name, pat):
                if mask_hidden and pat and pat[0] == "*" and name[0] == ".":
                    continue
                matchingfiles.append(name)
        self.dirs.delete(0, Tk.END)
        for name in subdirs:
            self.dirs.insert(Tk.END, name)
        self.files.delete(0, Tk.END)
        for name in matchingfiles:
            self.files.insert(Tk.END, name)
        head, tail = os.path.split(self.get_selection())
        if tail == os.curdir: tail = ''
        self.set_selection(tail)

    __control_group = None
    def get_control_group(self):
        if self.__control_group is None:
            f = Tk.Frame(self.top, borderwidth=2, name="controls")
            f.pack(fill=Tk.BOTH)
            self.__control_group = f
        return self.__control_group

    def get_control_frame(self, name="options"):
        """Retrieve a named control frame.

        If the named frame doesn't exist, it will be created.  The default
        name is 'options'.

        """
        for group in self.__control_frames:
            if group.group_name == name:
                return group
        f = self.get_control_group()
        if self.__control_frames:
            # add a spacer between groups
            fr = Tk.Frame(f, height=2)
            fr.pack(fill=Tk.BOTH)
        group = make_group(f, name, fill=Tk.X)
        group.group_name = name
        self.__control_frames.append(group)
        return group

    def add_filetype_selector(self, typelist, default):
        self.__filetype = var = Tk.StringVar(self.top)
        var.set("*." + default)
        f = self.get_control_frame("filetypes")
        radios = []
        cmd = self.__filetype_select
        for name, type in typelist:
            mask = "*." + type
            rb = Tk.Radiobutton(f, text=name, value=mask, variable=var,
                                command=cmd)
            rb.pack(expand=1, fill=Tk.X, side=Tk.LEFT)
            radios.append(rb)
        apply(tktools.unify_button_widths, tuple(radios))

    def __filetype_select(self, event=None):
        mask = self.__filetype.get()
        dir, pat = self.get_filter()
        self.set_filter(dir, mask)
        basename, ext = os.path.splitext(self.selection.get())
        self.selection.delete(0, Tk.END)
        self.selection.insert(0, basename + "." + mask[2:])
        self.filter_command()


class RSLoadFileDialog(FileDialogMixin, FileDialog.LoadFileDialog):
    def __init__(self, *args, **kw):
        apply(FileDialog.LoadFileDialog.__init__, (self,) + args, kw)
        FileDialogMixin.__init__(self)

class RSSaveFileDialog(FileDialogMixin, FileDialog.SaveFileDialog):
    def __init__(self, *args, **kw):
        apply(FileDialog.SaveFileDialog.__init__, (self,) + args, kw)
        FileDialogMixin.__init__(self)


class LaunchDialog(RSLoadFileDialog, RS.NullRS):
    """Knowbot Program Launcher Dialog."""
    __exposed = 1
    title = "KP Launch Window"

    def __init__(self, app, master=None, title=None):
        RS.NullRS.__init__(self)
        self.__app = app
        self.__master = master
        # locate a reachable kernel, if possible:
        names = receiver_names()
        default_kernel = kptools.default_kernel()
        if names and default_kernel in names:
            self.__kernel = default_kernel
        elif names:
            self.__kernel = names[0]
        else:
            self.__kernel = NO_KERNEL_FOUND

        #  Reconfigure the base:
        RSLoadFileDialog.__init__(self, master, title=title)
        self.top.iconname('Launcher')
        self.top.protocol("WM_DELETE_WINDOW", self.cancel_command)
        self.ok_button['text'] = 'Launch'
        self.top.bind("<Alt-C>", self.build_command)
        self.top.bind("<Alt-c>", self.build_command)
        build_button = Tk.Button(self.botframe, text="Compose",
                                 command=self.build_command)
        build_button.pack(after=self.ok_button, side=Tk.LEFT, expand=1)
        tktools.unify_button_widths(self.ok_button, self.filter_button,
                                    self.cancel_button, build_button)

        #  Add material specific to the launcher:
        self.create_launcher_widgets()

        #  Set up directory & file filter:
        directory = app.get_directory()
        self.set_filter(os.path.normpath(directory), "*.py")
        self.filter_command()

        if names is None:
            app.missing_worldroot()

    # make the base ID generator available by another name:
    generate_id = RS.NullRS.AllocateID

    ######################################################################
    #
    #  Reporting station (external) methods.
    #
    ######################################################################

    def AllocateID(self, prefix):
        if not self.__app.cloning_p():
            raise KOS.AuthorizationError, "Cloning disabled."
        id = self.generate_id(prefix)
        kpinfo = self.__app.fetch_kpinfo(prefix, create=0)
        if kpinfo:
            if not kpinfo.cloning_p():
                raise KOS.AuthorizationError, "Cloning disabled for " + prefix
            io = kpinfo.get_console() or None
            new_kpinfo = self.__app.fetch_kpinfo(id)
            new_kpinfo.set_console(io)
            new_kpinfo.set_ancestors(parent=prefix, root=kpinfo.get_root())
            kpinfo.add_child(id)
        return id

    def KPTerminated(self, id, reason, details, kprepr):
        self.__app.kp_terminated(id, reason, details, kprepr)

    def KPMigrationReceived(self, id, idTo, migratingTo):
        self.__app.kp_migrated(id, idTo, migratingTo)

    ######################################################################
    #
    #  Overridden FileDialog methods.
    #
    ######################################################################

    def ok_command(self):
        """Launch a KP based on the selected source file."""
        filename = self.selection.get()
        self.__app.set_directory(os.path.dirname(filename))
        repr = self.load_representation(filename)
        if not repr:
            self.__app.set_status("Could not open source: "
                                  + os.path.basename(filename))
            self.__app.warning("File Error",
                               "Could not open source: "
                               + os.path.basename(filename))
            return
        #
        basename, ext = os.path.splitext(os.path.basename(filename))
        kosname = self.__kernel
        try:
            repid, kpid = self.launch_kp(repr, kosname, basename)
        except KOS.EncodingError, msg:
            self.__app.set_status('Encoding error: ' + msg)
        except KOS.ResourceError, msg:
            self.__app.set_status('Resource error: ' + msg)
        except nstools.BadPathError, msg:
            self.__app.set_status("Could not locate KOS '%s'" % kosname)
        except ilu.IluGeneralError, msg:
            self.__app.set_status('ILU Error: ' + msg)
        else:
            self.__app.set_status("Submitted KP %s from %s%s to %s"
                                  % (repid, basename, ext, kosname))
        #
        self.__maybe_close()

    def cancel_command(self, event=None):
        """Hide the window after recording the selected directory."""
        dir = self.selection.get()
        if dir:
            self.__app.set_directory(os.path.dirname(dir))
        self.withdraw()

    ######################################################################
    #
    #  Application-specific command methods.
    #
    ######################################################################

    def build_command(self, event=None):
        filename = self.selection.get()
        if not os.path.isfile(filename):
            filename = ''
            directory, pat = self.get_filter()
            self.__app.set_directory(directory)
        self.__maybe_close()
        import Builder
        Builder.BuilderDialog(Tk._default_root, self.__app, self,
                              filename, self.__kernel)

    ######################################################################
    #
    #  Application-specific (internal) methods.
    #
    ######################################################################

    def create_launcher_widgets(self):
        # Insert selector for kernel:
        midframe = self.get_control_frame()
        kframe = Tk.Frame(midframe)
        Tk.Label(kframe, text = 'KOS:').pack(side = Tk.LEFT)
        KernelChooser(kframe, self.select_receiver, self.__kernel, self.__app,
                      ).pack(side=Tk.LEFT)
        kframe.pack(expand=1, side=Tk.LEFT, fill=Tk.BOTH)
        # Options:
        self.__ioserver = Tk.BooleanVar(self.top)
        self.__ioserver.set(1)
        self.__keep_open = Tk.BooleanVar(self.top)
        rframe = Tk.Frame(midframe)
        rframe.pack(side=Tk.RIGHT, expand=1, fill=Tk.X)
        Tk.Checkbutton(rframe, text="I/O Services", anchor=Tk.W,
                       variable=self.__ioserver).pack(fill=Tk.X)
        Tk.Checkbutton(rframe, text="Keep launcher open", anchor=Tk.W,
                       variable=self.__keep_open).pack(fill=Tk.X)
        #
        self.add_filetype_selector(
            [('Python', 'py'), ('Template', 'tmpl'), ('MIME', 'mime')], "py")

    def expose(self):
        if not self.__exposed:
            self.filter_command()
            self.__exposed = 1
        self.top.deiconify()
        self.top.tkraise()

    def withdraw(self):
        if self.__exposed:
            self.__exposed = 0
            self.top.withdraw()

    def launch_kp(self, repr, kosname, kpname):
        repid = self.generate_id(kpname)
        if self.__ioserver.get():
            # add i/o server functionality:
            cinfo = self.make_kp_console(repid, 0, 1, 1)
            io, stdin, stdout, stderr = cinfo
            meta = repr.GetMetadata()
            if stdin:
                meta.SetValue('stdin-server', stdin)
            if stdout:
                meta.SetValue('stdout-server', stdout)
            if stderr:
                meta.SetValue('stderr-server', stderr)
        else:
            io = None
        _repid, kpid = repr.submit(kosname=kosname, repid=repid,
                                   reporter=self)  
        # _repid should be the same as repid
        kpinfo = self.__app.fetch_kpinfo(repid)
        kpinfo.set_ancestors(root=repid)
        if io: kpinfo.set_console(io)
        self.__app.select_kp(repid)
        return repid, kpid

    def make_kp_console(self, repid, stdin, stdout, stderr):
        """Create a console object for a KP.

        repid
            Reporting-id for the KP.  This is used to give the resulting
            window a usable title and icon name.

        stdin
            True if a standard input server is required.  Currently
            false will be assumed.

        stdout
            True if a standard output server is required.

        stderr
            True if a standard error server is required.

        This method returns four values: the console object and the binding
        handles for the standard input, standard output, and standard error
        streams.  Streams which are not provided are be represented by
        false values.

        """
        io = KPConsole(self.__master, title="Console for " + repid,
                       exposure=self.__app.get_exposure())
        io.set_app(self.__app)
        io.iconname(repid)
        stdin = 0
        if stdout:
            stdout = IOServerWriter(io, ('stdout',)).IluSBH()
        if stderr:
            stderr = IOServerWriter(io, ('stderr',)).IluSBH()
        return io, stdin, stdout, stderr

    def load_representation(self, filename):
        dirname, basename = os.path.split(filename)
        basename, ext = os.path.splitext(basename)
        if ext == ".mime":
            try: repr = self.load_mime_file(filename)
            except IOError: repr = None
        elif ext == ".tmpl":
            try: repr = self.load_kpdata_file(filename)
            except IOError: repr = None
        else:
            try: repr = self.load_source_file(filename)
            except kptools.LoadingError: repr = None
        return repr

    def load_kpdata_file(self, filename):
        from KPTemplate import TemplateRepresentation, parse
        return TemplateRepresentation(os.path.dirname(filename),
                                      parse(open(filename)))

    def load_mime_file(self, filename):
        from koe.common import KPReprImpl
        return KPReprImpl.Representation(open(filename))

    def load_source_file(self, filename):
        modname = os.path.splitext(os.path.basename(filename))[0]
        return kptools.Representation((), main=filename)

    def select_receiver(self, name):
        """Select a particular kernel for additional operations.

        name
            The name of the KOS kernel which should be used for subsequent
            submissions.

        """
        self.__kernel = name

    def __maybe_close(self):
        # Close the launcher window only if the "keep open" box is not checked.
        if not self.__keep_open.get():
            self.withdraw()


class KernelChooser(Tk.OptionMenu):

    def __init__(self, master, select_cb, default, app):
        self.__app = app
        self.__select_cb = select_cb
        var = self.__var = Tk.StringVar(master)
        Tk.OptionMenu.__init__(self, master, var, default)
        var.set(default)
        self['width'] = max(len(NO_KERNEL_FOUND), len(default))
        self['anchor'] = Tk.W
        self['menu']['postcommand'] = self.__post_kernels_menu

    __required = ()
    def require(self, name):
        if name not in self.__required:
            self.__required = list(self.__required) + [name]

    __names = ()
    def __post_kernels_menu(self, event = None):
        # Update list of available receivers.
        menu = self['menu']
        names = receiver_names()
        if names is None and not self.__names:
            self.__app.missing_worldroot()
            return
        names = names or ()
        if names == self.__names:       # avoid rebuilding if easy to catch!
            return
        self.__names = names or ()
        kosnames = names[:]
        for name in self.__required:
            if name not in kosnames:
                names.append(name)
        names.sort()
        menu.delete(0, Tk.END)
        if names:
            for rn in names:
                if rn in kosnames:
                    cmd = Caller(self.__pick_kernel, rn)
                    state = Tk.NORMAL
                else:
                    cmd = None
                    state = Tk.DISABLED
                menu.add_command(label=rn, command=cmd, state=state)
        else:
            menu.add_command(label='None known.', state=Tk.DISABLED)
            self.__select_cb(NO_KERNEL_FOUND)
            self.__var.set(NO_KERNEL_FOUND)

    def __pick_kernel(self, name):
        self.__var.set(name)
        self.__select_cb(name)


class Caller:
    """Lightweight class to act as a command from the GuiRSWatcher class."""
    def __init__(self, command, *args, **kw):
        self._command = command
        self._args = args
        self._kw = kw

    def __call__(self, *args, **kw):
        """Execute the command registered at construction."""
        apply(self._command, self._args, self._kw)
