"""Text-widget based listbox.

Allows each item in the list to be multi-line.  This tries to be as similar
to a Tkinter.Listbox as possible, so use the Tk documentation for the listbox
to use this class.
"""


import copy
import os
import string
import sys
import Tkinter ; Tk = Tkinter
from types import IntType, StringType

SELECTED = "sel"                        # tag for selected elements
SELECTED_FIRST = "sel.first"
SELECTED_LAST = "sel.last"

CURRENT = "CURRENT"                     # tag for 'current' entry
CURRENT_FIRST = "CURRENT.first"
CURRENT_LAST = "CURRENT.last"

ATTRIBUTES_FOR_TEXT = ("background", "borderwidth", "cursor",
                       "exportselection",
                       "font", "foreground", "geometry", "height",
                       "highlightcolor", "highlightthickness",
                       "relief", "selectbackground", "selectforeground",
                       "selectborderwidth",
                       "setgrid", "takefocus", "width"
                       )

ORIGIN = Tk.At(0, 0)

_tlb_from_text = {}
_tlb_inited = 0

def _get_tlb(event):
    return _tlb_from_text[id(event.widget)]


class ListboxItem:
    """Representation of a single, multi-line item in a listbox.

    All information pertinant to an item is maintained using this class.
    Objects of this class publish three instance variables which may be
    read directly in order to maintain reasonable performance.

    text
        The text of the item plus a trailing newline.  The `get()' method of
        the TextListbox implementation is required to remove the trailing
        newline since it is not part of the client's data.  Get/set access
        without the newline is provided by using item['text'].

    lines
        The number of screen lines required to display the item.

    tag
        The tag used in the text widget to identify this item.  This does
        not change over the lifetime of the item, and will not be reused
        for other items for any ListboxItem in the same process.

    """
    __tag_count = 0
    __inited = 0
    text = "\n"
    lines = 1
    def __init__(self, tlb, **kw):
        """Constructor for the ListboxItem class.

        tlb
            The TextListbox instance to which we belong.  This is required
            to support the configuration machinery.

        """
        self.__tlb = tlb
        self.tag = "ENTRY_%d" % ListboxItem.__tag_count
        ListboxItem.__tag_count = ListboxItem.__tag_count + 1
        self.__options = {}
        if kw:
            apply(self.configure, (), kw)
        self.__inited = 1

    def __getitem__(self, opt):
        opt = normalize_option(opt)
        if opt == 'text':
            return self.text[:-1]
        val = None
        if self.__options.has_key(opt):
            val = self.__options[opt]
        if val in (None, ''):
            val = self.__tlb._text.tag_cget(self.tag, '-' + opt)
        return val
    cget = __getitem__

    def __setitem__(self, key, value):
        """Set a configuration option for the item.

        Setting an option to `None' causes the default appearance for that
        attribute to be restored.

        """
        if value is None:
            value = ''
        key = normalize_option(key)
        if key == 'text':
            # self.__inited will always be true here:
            self._reset_text(value)
        else:
            self.__options[key] = value
            apply(self.__tlb._text.tag_config, (self.tag,), {key: value})

    def configure(self, **kw):
        opts = {}
        for key, value in kw.items():
            key = normalize_option(key)
            if value is None:
                value = ''
            if key == 'text':
                if self.__inited:
                    self._reset_text(value)
                else:
                    self.text = value + "\n"
                    self.lines = map(None, self.text).count("\n")
            else:
                opts[key] = value
                self.__options[key] = value
        return apply(self.__tlb._text.tag_config, (self.tag,), opts)
    config = configure

    def destroy(self):
        self.__tlb = None               # discard reference

    def _reset_text(self, ntext):
        if not self.__tlb:
            return
        ntext = ntext + '\n'
        nlines = map(None, ntext).count("\n")
        index = self.__tlb._text.index(self.tag + ".first")
        self.__tlb._text['state'] = Tk.NORMAL
        self.__tlb._text.delete(self.tag + ".first", self.tag + ".last")
        self.__tlb._text.insert(index, ntext, self.tag)
        self.__tlb._text['state'] = Tk.DISABLED
        self.text = ntext
        changed = self.lines != nlines
        self.lines = nlines
        if changed:
            self.__tlb._recalc_lines()


def normalize_option(key):
    if key[0] == '-':                   # mostly for w.cget()
        key = key[1:]
    if key == 'bg':
        return 'background'
    if key == 'fg':
        return 'foreground'
    return key


class TextListbox(Tk.Widget):
    """List box with multi-line entries.
    """
    _mode = "browse"
    _anchor = None
    _active = None
    _totlines = 0
    _xscrollcmd = None
    _yscrollcmd = None
    _start_line = None
    _end_line = None

    klass = "TextListbox"

    def __init__(self, master = None, **cnf):
        global _tlb_inited
        under_current = 0
        if cnf.has_key('class'):
            self.klass = cnf['class']
            del cnf['class']
        elif cnf.has_key('class_'):
            self.klass = cnf['class_']
            del cnf['class_']
        extra = ('-class', self.klass)
        opts = {}
        if cnf.has_key('name'):
            opts['name'] = cnf['name']
            del cnf['name']
        Tk.Widget.__init__(self, master, 'frame', opts, {}, extra)

        self._entries = []              # each entry is a ListboxItem instance

        textopts = {
            'width': _extract("width", cnf, 12),
            'height': _extract("height", cnf, 10),
            'setgrid': _extract("setgrid", cnf, 0) and 1 or 0,
            'insertwidth': 0,
            'state': Tk.DISABLED,
            'wrap': Tk.NONE,
            }
        for k, v in cnf.items():
            if k in ATTRIBUTES_FOR_TEXT:
                textopts[k] = v
                del cnf[k]
        self._text = apply(Tk.Text, (self,), textopts)

        # rebind class bindings to self.klass
        bindtags = self._text.bindtags()
        self._text.bindtags((bindtags[0], self.klass,
                             bindtags[2], bindtags[3]))
        _tlb_from_text[id(self._text)] = self
        self._selection = {}

        if not _tlb_inited and self.klass == "TextListbox":
            for sequence, command in [
                ("<Button-1>", self._key_button1),
                ("<ButtonRelease-1>", self._key_buttonrelease1),
                ("<B1-Motion>", self._key_b1_motion),
                ("<Shift-Button-1>", self._key_shift_button1),
                ("<Button-2>", self._key_button2),
                ("<B2-Motion>", self._key_b2_motion),
                ("<Key-Up>", self._key_up),
                ("<Key-Down>", self._key_down),
                ("<Shift-Up>", self._key_shift_up),
                ("<Shift-Down>", self._key_shift_down),
                ("<Control-Button-1>", self._key_control_button1),
                ("<Control-B1-Motion>", self._key_control_b1_motion),
                ("<Control-Home>", self._key_control_home),
                ("<Control-End>", self._key_control_end),
                ("<Control-Shift-Home>", self._key_control_shift_home),
                ("<Control-Shift-End>", self._key_control_shift_end),
                ("<space>", self._key_space),
                ("<Select>", self._key_space),
                ("<Control-slash>", self._key_control_slash),
                ("<Control-backslash>", self._key_control_backslash),
                ("<Prior>", self._key_prior),
                ("<Next>", self._key_next),
                ("<Home>", self._key_home),
                ("<End>", self._key_end),
                ("<Left>", self._key_left),
                ("<Right>", self._key_right),
                ("<Control-Left>", self._key_control_left),
                ("<Control-Right>", self._key_control_right),
                ("<Control-Prior>", self._key_control_left),
                ("<Control-Next>", self._key_control_right),
                ("<Configure>", self._resize_event)]:
                self._text.bind_class("TextListbox", sequence, command)
            _tlb_inited = 1

        self._text.tag_config(SELECTED, borderwidth = 1, relief = Tk.RAISED,
                              background = self._text["selectbackground"])
        self._text.pack(side = Tk.LEFT, expand = 1, fill = Tk.BOTH)
        if cnf:
            apply(self.config, (), cnf)
        self._reset_scrollers()

    def __getitem__(self, name):
        return self.cget(name)

    def __setitem__(self, name, value):
        apply(self.config, (), {name: value})

    def activate(self, index):
        self._reset_active(self.index(index))

    def bind(self, *args, **kw):
        apply(self._text.bind, args, kw)

    def cget(self, name):
        name = normalize_option(name)
        if name in ATTRIBUTES_FOR_TEXT:
            return self._text[name]
        if name == "underlinecurrent":
            return self._text.tag_cget(CURRENT, "underline")
        if name == "xscrollcommand":
            return self._xscrollcmd
        if name == "yscrollcommand":
            return self._yscrollcmd
        if name == "selectmode":
            return self._mode
        raise Tk.TclError, "unknown option: " + name

    def config(self, **kw):
        text_cnf = {}
        _rescroll = 0
        for name, value in kw.items():
            name = normalize_option(name)
            if name in ATTRIBUTES_FOR_TEXT:
                text_cnf[name] = value
            elif name == "selectmode":
                if value in ("browse", "single", "extended", "multiple"):
                    self._mode = value
                    if self._selection and (value in ("browse", "single")):
                        if self.select_includes(self._anchor):
                            self.select_clear(0, Tk.END)
                            self.select_set(self._anchor)
                        else:
                            self.select_clear(0, Tk.END)
                    return
                raise Tk.TclError, "illegal select mode"
            elif name == "xscrollcommand":
                self._xscrollcmd = value
                _rescroll = 1
            elif name == "yscrollcommand":
                self._yscrollcmd = value
                _rescroll = 1
            elif name == "underlinecurrent":
                self._text.tag_config(CURRENT, underline=value)
            else:
                raise Tk.TclError, "unknown option: " + name
        if text_cnf:
            apply(self._text.config, (), text_cnf)
        if _rescroll:
            self._reset_scrollers()
    configure = config

    def itemconfig(self, index, **kw):
        return apply(self.get_item(index).configure, (), kw)
    itemconfigure = itemconfig

    def itemcget(self, index, opt):
        return self.get_item(index).cget(opt)

    def get_item(self, index):
        index = self.index(index)
        if index >= len(self._entries):
            raise ValueError, "index out of range"
        return self._entries[index]

    def curselection(self):
        result = self._selection.keys()
        result.sort()
        return tuple(result)

    def get(self, index, last=None):
        try:
            index = self.index(index)
        except ValueError:
            index = self.index(Tk.END)
        if index is None:
            raise Tk.TclError, "illegal index"
        if last is not None:
            try:
                last = self.index(last)
            except ValueError:
                last = self.index(Tk.END)
            if last is None:
                raise Tk.TclError, "illegal index"
        else:
            last = index + 1
        try:
            result = self._entries[index:last]
        except IndexError, msg:
            raise Tk.TclError, msg, sys.exc_traceback
        res = []
        for entry in result:
            res.append(entry.text[:-1])
        return res

    def delete(self, first, last=None):
        first, last = self._calc_range(first, last)
        old_size = self.size()
        self._text['state'] = Tk.NORMAL
        try:
            if first < len(self._entries):
                stag = self._entries[first].tag + ".first"
                etag = self._entries[-1].tag + ".last"
                self._text.delete(stag, etag)
            for entry in self._entries[first:last]:
                entry.destroy()
            del self._entries[first:last]

            # update indexes in the selection dictionary
            dict = self._selection
            for i in range(first, last):
                if dict.has_key(i):
                    del dict[i]
            for i in range(last, old_size):
                if dict.has_key(i):
                    dict[i - (last - first)] = 1
                    del dict[i]

            # update "active" index
            current = None
            if self._active is not None and self._entries:
                current = self._active
                if current > last:
                    self._reset_active(max(0, current - (last - first)))
                elif current >= first:
                    current = first
                    if len(self._entries) <= current:
                        current = None

            self._regen_from(first)
            self._reset_active(current)
            self._recalc_lines()
        finally:
            self._text['state'] = Tk.DISABLED

    def destroy(self):
        del _tlb_from_text[id(self._text)]
        for e in self._entries:
            e.destroy()
        self._yscrollcmd = self._xscrollcmd = self._entries = None
        Tk.Widget.destroy(self)

    def focus_set(self):
        self._text.focus_set()

    def index(self, index, end_offset=1):
        if index == 'active':
            return self._active
        if index == 'anchor':
            return self._anchor
        if index == Tk.END:
            if self._entries:
                return len(self._entries) - end_offset
            return 0
        if type(index) is IntType:
            if index == 0 or (0 < index < len(self._entries)):
                return index
            return len(self._entries)
        if type(index) is StringType and index and index[0] == '@':
            names = self._text.tag_names(index)
            if not names and self._entries:
                return len(self._entries) - 1
            entry = None
            for tag in names:
                if len(tag) > 6 and tag[:6] == "ENTRY_":
                    for i in range(len(self._entries)):
                        if self._entries[i].tag == tag:
                            return i
        raise Tk.TclError, "illegal index specification: " + `index`

    def insert(self, index, *elements):
        if not elements:
            return
        if type(index) is StringType and index and index[0] == '@':
            index = self.index(index)
            if index is None:
                index = len(self._entries)
        else:
            index = self.index(index, 0)
        # convert element tuple into a list of ListboxItem instances
        elements = map(self._get_listitem, elements)
        if index < len(self._entries):
            tup = self._text.tag_nextrange(self._entries[index].tag, '1.0')
            if tup:
                start, x = tup
                self._text['state'] = Tk.NORMAL
                self._text.delete(start, Tk.END + " - 1 char")
        self._insert_in_selection(index, len(elements))
        self._entries[index:index] = elements
        self._recalc_lines()
        self._regen_from(index)
        if self._active is not None and self._active >= index:
            self._reset_active(self._active + len(elements))

    def _get_listitem(self, e):
        # could inline for speed, but it would be ugly
        if type(e) is StringType:
            return ListboxItem(self, text=e)
        return e

    def nearest(self, y):
        names = self._text.tag_names(Tk.At(0, y))
        if not names and self._entries:
            return len(self._entries) - 1
        for tag in names:
            if len(tag) > 6 and tag[:6] == "ENTRY_":
                for i in range(len(self._entries)):
                    if tag == self._entries[i].tag:
                        return i
        return None

    def scan_mark(self, x, y):
        self._text.scan_mark(x, y)

    def scan_dragto(self, x, y):
        self._text.scan_dragto(x, y)

    def see(self, index):
        index = self.index(index)
        if index is None:
            if not self._entries:
                raise Tk.TclError, "illegal index"
            index = len(self._entries) - 1
        start, end = self._find_text_lines(index)
        initpos = self._text.index(ORIGIN)
        initpos = string.atoi(string.splitfields(initpos, '.')[1])
        self._text.see("%d.%d" % (end, initpos))
        self._text.see("%d.%d" % (start, initpos))
        self._reset_scrollers()

    def select_adjust(self, index):
        index = self.index(index)
        # ...

    selection_adjust = select_adjust

    def select_anchor(self, index):
        index = self.index(index)
        if index is not None:
            self._anchor = index

    selection_anchor = select_anchor

    def select_clear(self, start, end=None):
        if not self._entries:
            self._selection = {}
            self._redraw_selection()
            return
        start = self.index(start)
        if end is not None:
            end = self.index(end) 
            if end is None:
                raise Tk.TclError, "illegal index"
            begin = min(start, end)
            end = max(start, end) + 1
            start = begin
        else:
            end = start + 1
        for i in range(start, end):
            if self._selection.has_key(i):
                del self._selection[i]
        self._redraw_selection()

    selection_clear = select_clear

    def select_includes(self, index):
        return self._selection.has_key(self.index(index))

    selection_includes = select_includes

    def select_set(self, start, end=None):
        if not self._entries:
            if self._selection:
                self._selection = {}
            self._reset_active(None)
            self._redraw_selection()
            return
        start = self.index(start)
        if start is None:
            start = end = len(self._entries) - 1
        if end is not None and self._mode in ("extended", "multiple"):
            end = self.index(end)
            if end is None:
                raise Tk.TclError, "illegal index"
        elif self._mode in ("browse", "single"):
            self.select_clear(0, len(self._entries) - 1)
            end = start
        elif end is None:
            end = start
        for i in range(start, end + 1):
            self._selection[i] = 1
        self._redraw_selection()

    selection_set = select_set

    def size(self):
        return len(self._entries)

    def yview(self, *args):
        if not args:
            if not self._entries:
                return 0.0, 1.0
            pos = self._text.index(ORIGIN)
            pos = string.atoi(string.splitfields(pos, '.')[0])
            pos = max(pos - 1, 0)
            offset = 1.0 * pos / self._totlines
            height = self._get_height()
            if pos + height - 1 > self._totlines:
                span = 1.0
            else:
                span = 1.0 * (height + pos) / self._totlines
            return offset, min(span, 1.0)
        index = args[0]
        if index == "scroll":
            return self.yview_scroll(string.atoi(args[1]), args[2])
        if index == "moveto":
            return self.yview_moveto(string.atof(args[1]))
        start, end = self._find_text_lines(self.index(index))
        if start is not None:
            self._text.yview('%d.0' % start)
            self._reset_scrollers()

    def yview_moveto(self, fraction):
        index = self._text.index(ORIGIN)
        if not self._entries:
            return
        above = int(fraction * self._totlines)
        if above < 1:
            above = 1
        self._text.yview("%d.0" % above)
        self._reset_scrollers()

    def yview_scroll(self, num, what):
        index = self._text.index(ORIGIN)
        if not self._entries:
            return
        if what == "pages":
            num = num * self._get_height()
        pos = string.atoi(string.splitfields(index, '.')[0])
        new_pos = pos + num
        if new_pos < 1:
            new_pos = 1
        elif new_pos > self._totlines:
            new_pos = self._totlines
        self._text.yview("%d.0" % new_pos)
        self._reset_scrollers()

    # internal utility methods

    def _calc_range(self, first, last, end_offset=1):
        first = self.index(first)
        if last is not None:
            last = self.index(last, end_offset)
            if last is None:
                last = first
                #raise Tk.TclError, "illegal index"
            last = last + 1
            if last <= first:
                raise Tk.TclError, "illegal range"
        else:
            last = first + 1
        return first, last

    def _get_height(self):
        lineheight = self._text.dlineinfo(ORIGIN)[3]
        return self._text.winfo_height() / lineheight

    def _insert_in_selection(self, index, length):
        # insert new elements into the selection
        # (use for elements that are being added to self._entries)
        newsel = {}
        for i in self._selection.keys():
            if i >= index:
                i = i + length
            newsel[i] = 1
        self._selection = newsel

    def _regen_from(self, index):
        insert_args = [Tk.END]
        append = insert_args.append
        opts = []
        for entry in self._entries[index:]:
            append(entry.text)
            append(entry.tag)
        if len(insert_args) > 1:
            self._text['state'] = Tk.NORMAL
            apply(self._text.insert, tuple(insert_args))
        self._redraw_selection()
        self._text['state'] = Tk.DISABLED
        self._reset_scrollers()

    def _recalc_lines(self):
        total = 0
        for entry in self._entries:
            total = total + entry.lines
        self._totlines = total

    def _redraw_selection(self):
        self._text['state'] = Tk.NORMAL
        try: self._text.tag_remove(SELECTED, '1.0', Tk.END)
        except Tk.TclError: pass
        for [start, end] in _dict_to_ranges(self._selection):
            if start is None:
                return None, None
            start_line = 1
            for entry in self._entries[:start]:
                start_line = start_line + entry.lines
            end_line = start_line
            for entry in self._entries[start:end+1]:
                end_line = end_line + entry.lines
            self._text.tag_add(SELECTED, '%d.0' % start_line,
                               '%d.0' % end_line)

    def _reset_active(self, index):
        if self._active == index:
            return
        self._active = index
        try: self._text.tag_remove(CURRENT, '1.0', Tk.END)
        except Tk.TclError: pass
        if index is not None:
            start, end = self._find_text_lines(index)
            self._text.tag_add(CURRENT, "%d.0" % start, "%d.0" % (end + 1))

    def _reset_scrollers(self):
        if self._yscrollcmd:
            first, last = self.yview()
            self._yscrollcmd(first, last)

    def _resize_event(self, event=None):
        self._reset_scrollers()
        if len(self._selection) == 1 \
           and self._selection.has_key(0) \
           and self._selection[0][0] == self._selection[0][1]:
            self.see(self._selection[0][0])

    def _find_text_lines(self, index):
        # returns line index of first line and last line; not the first line
        # and first line of index+1!
        tag = self._entries[index].tag
        try:
            start, end = self._text.tag_nextrange(tag, '1.0')
        except Tk.TclError:
            return None, None
        # debugging:
##      except ValueError, err:
##          print "in _find_text_lines(%s):" % index
##          print self._text.tag_nextrange(tag, '1.0')
##          raise ValueError, err, sys.exc_traceback
        start = string.atoi(string.splitfields(start, '.')[0])
        end = start + self._entries[index].lines - 1
        return start, end

    #  The rest of these are the actual key bindings.
    #  Since these are *class* bindings defined for each new instance, the
    #  self parameter may not refer to the actual TextListbox instance

    def _key_button1(self, event):
        self = _get_tlb(event)
        self.select_clear(0, Tk.END)
        if not self._entries:
            return
        index = self.index(Tk.At(event.x, event.y))
        if index is None and event.y >= 0:
            index = self.size() - 1
        self._reset_active(index)
        self._anchor = self._active
        self.select_set(self._active)
        self.see(self._active)
        self.focus_set()

    def _key_buttonrelease1(self, event):
        self = _get_tlb(event)
        if self._mode != "browse":
            if self._entries:
                index = self.index(Tk.At(event.x, event.y))
                if index is None:
                    index = self.size() - 1
                self._reset_active(index)
            else:
                self._reset_active(None)

    def _key_b1_motion(self, event):
        self = _get_tlb(event)
        if not self._entries:
            return
        if self._mode == "browse":
            index = self.index(Tk.At(event.x, event.y))
            if index is not None:
                self.select_clear(0, Tk.END)
                self.select_set(index)
                self.see(index)
            elif event.y >= self._text.winfo_height():
                self.select_clear(0, Tk.END)
                self.select_set(self.size() - 1)
                self.see(self._active)
        elif self._mode == "extended":
            index = self.index(Tk.At(event.x, event.y))
            if index is not None and index != self._anchor:
                start = min(index, self._anchor)
                end = max(index, self._anchor)
                self.select_set(start, end)

    def _key_control_b1_motion(self, event):
        self = _get_tlb(event)
        if self._mode != "extended" or not self._entries:
            return
        index = self.index(Tk.At(event.x, event.y))
        if index is not None and index != self._anchor:
            start = min(index, self._anchor)
            end = max(index, self._anchor)
            if self.select_includes(self._anchor):
                self.select_set(start, end)
            else:
                self.select_clear(start, end)

    def _key_shift_button1(self, event):
        self = _get_tlb(event)
        index = self.index(Tk.At(event.x, event.y))
        if index is None or not self._entries:
            return
        self._reset_active(index)
        if self._mode == "extended" and self._anchor != index:
            start = min(index, self._anchor)
            end = max(index, self._anchor)
            self.select_set(start, end)

    def _key_control_button1(self, event):
        self = _get_tlb(event)
        if self._mode != "extended" or not self._entries:
            return
        index = self.index(Tk.At(event.x, event.y))
        if index is None:
            index = len(self._entries) - 1
        if self.select_includes(index):
            self.select_clear(index)
        else:
            self.select_set(index)
        self._anchor = index

    def _key_button2(self, event):
        self = _get_tlb(event)
        self.scan_mark(event.x, event.y)

    def _key_b2_motion(self, event):
        self = _get_tlb(event)
        self.scan_dragto(event.x, event.y)

    def _key_up(self, event):
        self = _get_tlb(event)
        if self._mode == "single":
            if self._entries and self._active:
                self._reset_active(self._active - 1)
                self.see(self._active)
            return
        if self._active:
            self._reset_active(self._active - 1)
            self.select_clear(0, Tk.END)
            self.select_set(self._active)
            self.see(self._active)

    def _key_shift_up(self, event):
        self = _get_tlb(event)
        if self._mode == "multiple" and self._active:
            self._reset_active(self._active - 1)
            self.see(self._active)
        elif self._mode == "extended" and self._active:
            self._reset_active(self._active - 1)
            self.select_set(self._active)
            self.see(self._active)

    def _key_shift_down(self, event):
        self = _get_tlb(event)
        if self._mode in ("extended", "multiple"):
            if self._active is not None \
               and self._active + 1 < len(self._entries):
                self._reset_active(self._active + 1)
                if self._mode == "extended":
                    self.select_set(self._active)
                self.see(self._active)

    def _key_down(self, event):
        self = _get_tlb(event)
        if self._mode == "single":
            if self._entries and self._active is not None \
               and self._active + 1 < self.size():
                self._reset_active(self._active + 1)
                self.see(self._active)
            return
        if self._active is not None \
           and self._active + 1 < len(self._entries):
            self.select_clear(0, Tk.END)
            self._reset_active(self._active + 1)
            self.select_set(self._active)
            self.see(self._active)

    def _key_control_home(self, event):
        self = _get_tlb(event)
        if self._entries:
            self.select_clear(0, Tk.END)
            self.select_set(0)
            self._reset_active(0)
            self.see(0)

    def _key_control_shift_home(self, event):
        self = _get_tlb(event)
        if self._mode == "extended" and self._entries:
            if not self._selection:
                self._reset_active(0)
                self.select_set(0)
            elif self._anchor != 0:
                self.select_set(0, self._anchor)
        elif self._mode == "multiple":
            if self._entries:
                self._reset_active(0)
            else:
                self._reset_active(None)

    def _key_control_end(self, event):
        self = _get_tlb(event)
        if self._entries:
            self._reset_active(len(self._entries) - 1)
            self.select_clear(0, Tk.END)
            self.select_set(self._active)
            self.see(self._active)

    def _key_control_shift_end(self, event):
        self = _get_tlb(event)
        if self._mode == "extended" and self._entries:
            if self._selection:
                sefl.select_set(Tk.ANCHOR, Tk.END)
                self.see(len(self._entries) - 1)
            elif self._active is not None:
                self.select_clear(0, Tk.END)
                self.select_set(self._active, len(self._entries) - 1)
            else:
                self.select_clear(0, Tk.END)
                self._reset_active(len(self._entries) - 1)
                self.select_set(self._active)
        elif self._mode == "multiple" and self._entries:
            self.activate(Tk.END)
            self.see(self._active)

    def _key_control_slash(self, event):
        self = _get_tlb(event)
        if self._mode in ("extended", "multiple"):
            if self._entries:
                self.select_set(0, Tk.END)
            else:
                self.select_clear(0)
        elif self._active is not None:
            self.select_clear(0, Tk.END)
            self.select_set(self._active)
            self.see(self._active)

    def _key_space(self, event):
        self = _get_tlb(event)
        if self._active is not None:
            self.select_clear(0, Tk.END)
            self.select_set(self._active)

    def _key_prior(self, event):
        self = _get_tlb(event)
        self.yview_scroll(-1, "pages")

    def _key_next(self, event):
        self = _get_tlb(event)
        self.yview_scroll(1, "pages")

    def _key_home(self, event):
        self = _get_tlb(event)
        self._text.xview("moveto", 0.0)

    def _key_end(self, event):
        self = _get_tlb(event)
        self._text.xview("moveto", 1.0)

    def _key_left(self, event):
        self = _get_tlb(event)
        self._text.xview("scroll", -1, "units")

    def _key_right(self, event):
        self = _get_tlb(event)
        self._text.xview("scroll", 1, "units")

    def _key_control_left(self, event):
        self = _get_tlb(event)
        self._text.xview("scroll", -1, "pages")

    def _key_control_right(self, event):
        self = _get_tlb(event)
        self._text.xview("scroll", 1, "pages")

    def _key_control_backslash(self, event):
        self = _get_tlb(event)
        if self._mode != "browse":
            self.select_clear(0, Tk.END)


#  Utilities to keep the code from getting uglier....

def _dict_to_ranges(dict):
    curr = None
    selection = []
    elements = dict.keys()
    elements.sort()
    for entry in elements:
        if curr:
            if curr[1] == entry - 1:
                curr[1] = entry
            else:
                selection.append(curr)
                curr = [entry, entry]
        else:
            curr = [entry, entry]
    if curr:
        selection.append(curr)
    return selection


def _extract(attr, dict, value=None):
    if dict.has_key(attr):
        value = dict[attr]
        del dict[attr]
    return value
