"""Information object pertaining to a KP.

This is fairly simple, and intended primarily to record information about
the whereabouts of a KP and some information about how it died.

"""

import os
import string
import time
from koe.common import RS

# raise by get_lastmove() if the KP has not been received; value is (msg, id)
LocationError = "TrackingInfo.LocationError"

# raised on attempt to re-use a reporting-id; value is (msg, id)
UniquenessError = "TrackingInfo.UniquenessError"


_corepath = '.'


class TrackingInfo:
    """Information maintained to track the current disposition of a KP."""
    __repid_stash = {}

    _status = None
    __core = None
    __ltime = None                      # launch time
    __ttime = None                      # termination time
    __clonable = 1                      # is cloning enabled?
    __root = None                       # progenetor KP repid
    __parent = None                     # immediate parent KP repid

    def __init__(self, repid):
        if self.__repid_stash.has_key(repid):
            raise UniquenessError, ('KPRepr IDs must be unique.', repid)
        self.__repid_stash[repid] = repid
        self.__repid = repid
        self.__lochst = []
        self.__children = []
        self.__exception(None, None)

    def set_ancestors(self, root=None, parent=None):
        self.__parent = parent
        self.__root = root

    def add_child(self, child):
        self.__children.append(child)

    def get_parent(self):
        return self.__parent

    def get_root(self):
        return self.__root

    def get_children(self):
        return tuple(self.__children)

    def set_corefile(self, fileName):
        """Set the URL of the kpcore image."""
        self.__core = fileName

    def get_corefile(self):
        """Retrieves filename of the corpsicle for this KP."""
        return self.__core

    def cloning_p(self):
        return self.__clonable

    def set_cloning(self, flag):
        self.__clonable = flag and 1 or 0

    def get_completion_time(self):
        return self.__ttime

    def get_hopcount(self):
        if self.__lochst:
            return len(self.__lochst) - 1
        return 0                        # still need starting location

    def get_launchtime(self):
        """Retrieves the launch time for this KP, or None."""
        if self.__lochst:
            return self.__lochst[-1][2]
        return None

    def get_lastmove(self):
        if self.__lochst:
            return self.__lochst[-1]
        raise LocationError, ("KP has not been started.", self.__repid)

    def get_location(self):
        """Retrieves the current location and KOS KPID for the KP."""
        if self.__lochst:
            return self.__lochst[-1][:2]
        return None, None

    def get_reporting_id(self):
        """Returns the KP's unique ID for the reporting station."""
        return self.__repid

    def relocate(self, kosname, kpid):
        """Update the location information for this KP."""
        tup = (kosname, kpid, time.time())
        if self.__lochst:
            if self.get_location() != (kosname, kpid):
                self.__lochst.append(tup)
        else:
            self.__lochst.append(tup)

    def terminate(self, reason, details, kprepr):
        self.__terminate(time.time(),
                        reason and ("Exception: " + reason)
                        or ("Exit: " + str(details)))
        if reason:
            from koe.interfaces.stubs import KPRepr
            state = kprepr.GetState()
            try: 
                traceback = state.GetData('traceback')
            except KPRepr.PartNameError:
                traceback = None
            self.__exception(reason, details, traceback)

    def __terminate(self, time, status):
        # overridden in subclass
        self.__ttime = time
        self._status = status

    def __exception(self, t, v, tb=None):
        self.__exc_type = t
        self.__exc_value = v
        self.__exc_traceback = tb

    def get_exception(self):
        return (self.__exc_type,
                self.__exc_value,
                self.__exc_traceback)


def set_core_path(path):
    global _corepath
    _corepath = path
