#! /usr/bin/env python

# This documentation string also serves as a usage message
"""Submit Knowbot Program(s) and wait for completion.

Usage: submit.py [option] ... [argument] ... -- [kpoptions]

Options:

--kernel name, -k name
    kernel name (default: your hostname)
--monitor name, -m name
    external reporting station name (default: internal reporting station)
--nogo, -n
    do everything but don't submit
--noioserver, -i
    inhibit redirection of KP's I/O (use kernel's console)
--showmime, -x
    just before submission, write the submitted KP as MIME to stdout
--single, -1
    only submit a single KP; arguments are modules
--suitcase spec, -s spec
    suitcase specification (see below)
--nocvs, -C
    when using the path:directory suitcase spec, don't copy CVS directories or
    files (.cvsignore) into the suitcase.
--read-suitcase spec, -r spec
    copy a file from the suitcase to the local filesystem when the KP
    finishes. (uses the same specification format as --suitcase)
--verbose, -v
    verbose mode -- be chatty about KP status changes etc.

--
    delineates start of options to be passed to the KP

Arguments:

Each non-option argument is either a MIME-encoded KP ("-" or a file
ending in ".mime") or a comma-separated list of Python modules.  If
omitted or when an argument is "-", a single MIME-encoded KP is read
from standard input.

If --single is given, a single KP is submitted and each argument
specifies a module that becomes part of a list; thus, "submit --single
a b c" is equivalent to "submit a,b,c".  This may be handy if you rely
on filename completion in your shell a lot.  In this case, '-' loses
its special meaning and no mime files can be submitted.

Suitcase specifications:

One or more --suitcase options control the contents of the suitcase
sent along with the KP.  The suitcase is only added to the first KP
sent (should there be a way to specify suitcases for each of them?).
The suitcase option can have the following forms:

--suitcase path:file      : insert a file in the suitcase
--suitcase path:directory : insert an entire directory tree in the suitcase
--suitcase path:=data     : insert literal data in the suitcase

"""


import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../../"))
import time
import string
import getopt
import signal
import traceback
import ilu
from koe.common import RS
from koe.common import nstools
from koe.common import ilutools
from koe.common import kptools
from koe.common import qsplit
from koe.interfaces.stubs import KOS
from koe.interfaces.stubs import KPRepr
from koe.interfaces.stubs import FileAPI__skel


def main():
    """Main program."""

    global verbose

    t0 = time.time()

    # Defaults
    kernelname = kptools.default_kernel()
    suitcasespecs = []
    suitcaseresults = []
    nogo = 0
    showit = 0
    ioserver = 1
    oneonly = 0
    verbose = 0
    external_rs = 0
    freezedir = None
    rsinfo = None

    # find the start of the KP options
    submitargs = sys.argv[1:]
    kpargs = []
    for i, arg in map(None, range(len(submitargs)), submitargs):
        if arg == '--':
            kpargs = submitargs[i+1:]
            submitargs = submitargs[:i]
            break
    program_arguments = qsplit.qjoin(kpargs)

    # Parse command line arguments
    try:
        opts, args = getopt.getopt(
            submitargs,
            'k:m:nix1s:vr:Cf:',
            [
                'kernel=',              # -k
                'monitor=',             # -m
                'nogo',                 # -n
                'noioserver',           # -i
                'showmime',             # -x
                'single',               # -1
                'suitcase=',            # -s
                'read-suitcase',        # -r
                'verbose',              # -v
                'freeze-dry=',          # -f
                'nocvs',                # -C
            ])
    except getopt.error, msg:
        usage(msg)
    if not args:
        args = ['-']
    includecvs = 1
    for o, a in opts:
        if o in ('-k', '--kernel'):
            kernelname = a
        if o in ('-s', '--suitcase'):
            i = string.find(a, ':')
            if i < 0:
                usage("missing ':' in --suitcase specification")
            name, what = a[:i], a[i+1:]
            if what[:1] == '=':
                how = '='
                what = what[1:]
            else:
                if os.path.isdir(what):
                    how = 'd'
                else:
                    how = 'f'
                    if not os.path.exists(what):
                        error("File or directory %s not found", `what`)
            suitcasespecs.append((name, how, what))
        if o in ('-r', '--read-suitcase'):
            i = string.find(a, ':')
            if i < 0:
                usage("missing ':' in --read-suitcase specification")
            suitcaseresults.append(a[:i], a[i], a[i+1:])
        if o in ('-n', '--nogo'):
            nogo = 1
        if o in ('-x', '--showmime'):
            showit = 1
        if o in ('-i', '--noio'):
            ioserver = 0
        if o in ('-1', '--single'):
            oneonly = 1
        if o in ('-v', '--verbose'):
            verbose = verbose+1
        if o in ('-m', '--monitor'):
            external_rs = a
        if o in ('-f', '--freeze-dry'):
            freezedir = a
        if o in ('-C', '--nocvs'):
            includecvs = 0

    # Get external monitor object:
    if external_rs:
        try:
            external_rs = nstools.WorldOpen("tools/"+external_rs,
                                            "MonitorAPI.MonitorServer")
        except nstools.BadPathError:
            error("Could not locate monitor server %s." % external_rs)
        rs = None
    else:
        # Create an I/O server
        if ioserver:
            ioserver = kptools.IOServer()

        # Create reporting station and local id allocator
        rs = MyRS(freezedir=freezedir)

    # Submit each argument, generating reporting ids
    if oneonly:
        args = [args]
    errors = 0
    for arg in args:
        localid = None
        try:
            if not oneonly and (arg == '-' or arg[-5:] == '.mime'):
                # Argument is a file containing MIME data
                try:
                    if arg == '-':
                        fp = seekable(sys.stdin)
                        arg = '<stdin>'
                    else:
                        fp = open(arg)
                    repr = kptools.Representation()
                    try:
                        repr.loadfromfile(fp)
                    except KPRepr.KPMimeError, msg:
                        error("MIME parsing error for %s: %s", `arg`, str(msg))
                    if fp is not sys.stdin:
                        fp.close()
                except IOError, msg:
                    error("Can't open or read %s: %s", `arg`, str(msg))
                if arg == '-':
                    main = ""
                else:
                    main = arg
            else:
                if oneonly:
                    # Argument is already a list
                    modules = arg
                else:
                    # Argument is a comma-separated list of modules
                    modules = string.split(arg, ',')
                main = modules[0]
                try:
                    repr = kptools.Representation(modules[1:], main=main)
                    metadata = repr.GetMetadata()
                except (IOError, kptools.LoadingError), msg:
                    error("Can't load %s: %s", `arg`, str(msg))
                except SyntaxError, value:
                    list = traceback.format_exception_only(SyntaxError, value)
                    details = string.join(list, "  ")
                    error("Syntax error in source module:\n  %s" % details)

            # Generate a local (reporting) id for the KP
            if main:
                main = os.path.basename(main)
                if main[-3:] == '.py':
                    main = main[:-3]
                elif main[-5:] == '.mime':
                    main = main[:-5]
            if external_rs:
                try:
                    rsinfo = external_rs.RequestReporter(main, ioserver)
                except KOS.AuthorizationError, msg:
                    error("Could not allocate monitor: " + msg)
                localid = rsinfo.reporting_id
                repsbh = rsinfo.binding_handle
            else:
                localid = rs.AllocateID(main)
                repsbh = rs.IluSBH()
                if suitcaseresults:
                    try:
                        rs.saveid(localid, suitcaseresults)
                    finally:
                        suitcaseresults = []

            # Add the IO server to the metadata
            if ioserver:
                if external_rs:
                    metadata.SetValue('stdout-server', rsinfo.stdout_server)
                    metadata.SetValue('stderr-server', rsinfo.stderr_server)
                else:
                    ioserver.linktokp(repr)

            # Update the suitcase according to the suitcase specification
            # Only the first KP gets a suitcase
            if suitcasespecs:
                try:
                    update_suitcase(repr, suitcasespecs, includecvs)
                finally:
                    suitcasespecs = []

            # Add program arguments meta data
            if program_arguments:
                metadata.SetValue('program-arguments', program_arguments)

            # Submit the KP to the kernel
            if showit:
                mimedata = repr.FreezeDry()
                sys.stdout.write(mimedata)
            if nogo:
                sys.exit(0)     # go to 'except SystemExit' below
            try:
                repid, kpid = repr.submit(kosname=kernelname, repid=localid,
                                          repsbh=repsbh)
            except KOS.EncodingError, msg:
                error("Encoding error submitting %s: %s", `arg`, str(msg))
            except KOS.MetadataError, msg:
                error("Metadata error submitting %s: %s", `arg`, str(msg))
            except KOS.AuthorizationError, msg:
                error("Authorization error submitting %s: %s", `arg`, str(msg))
            except nstools.BadPathError, msg:
                error("Error looking up kernel %s: %s", kernelname, str(msg))
            # XXX What about other errors?
            else:
                message("KP %s submitted from %s, KP id %s",
                        `localid`, `arg`, `kpid`)

        except SystemExit:
            errors = errors + 1
            if rs and localid:
                rs.delid(localid)

    if not rs or rs.empty():
        if errors:
            sys.exit(1)
        message("No reports requested.  Exit.")
    else:
        # Start the ILU main loop
        t1 = time.time()
        message("Elapsed time for submission: %.3f secs.", t1-t0)
        message("Waiting for report(s)...")
        t0 = time.time()
        rs.start_signal_handling()
        ilutools.RunMainLoop()
        t1 = time.time()
        message("\nElapsed time for execution: %.3f secs.\n", t1-t0)


def usage(message):
    """Write command line syntax error and usage message, and exit."""
    sys.stderr.write("Usage error: %s\n\n%s" % (str(message), __doc__))
    sys.exit(2)


def message(format, *args):
    """Write message to stderr, if verbose."""
    if verbose:
        t = "%.4f " % (time.time())
        sys.stderr.write(t + "submit: %s\n" % format % args)


def warning(format, *args):
    """Write warning message to stderr."""
    sys.stderr.write("submit: %s\n" % format % args)


def error(format, *args):
    """Write error message, and exit."""
    sys.stderr.write("submit: %s\n" % format % args)
    sys.exit(1)


def seekable(fp):
    """Return a seekable version of fp, possibly a StringIO copy."""
    try:
        fp.seek(fp.tell(), 0)
    except IOError:
        import StringIO
        data = fp.read()
        return StringIO.StringIO(data)
    else:
        return fp

def save_suitcase(repr, suitcaseresults):
    from koe.common import Suitcase
    sfs = Suitcase.SuitcaseFilesystem(repr.GetSuitcase())
    for src, op, dst in suitcaseresults:
        print "copying %s from suitcase to %s" % (src, dst)
        # ignore the op for now. it is always ':'
        try:
            sf = sfs.open(src)
        except IOError, msg:
            warning("Could not find suitcase file '%s': %s", src, msg[1])
            continue
        try:
            f = open(dst, 'w')
        except IOError, msg:
            warning("Could not create destination file '%s': %s", dst, msg[1])
        else:
            f.write(sf.read())
        sf.close()

def update_suitcase(repr, suitcasespecs, includecvs):
    """Update the suitcase according to the specs (see usage message)."""
    from koe.common import Suitcase
    sfs = Suitcase.SuitcaseFilesystem(repr.GetSuitcase())
    for name, how, what in suitcasespecs:
        head, tail = sfs.path.split(name)
        make_suitcase_dir(sfs, head)
        if how == '=':
            # Immediate data
            f = sfs.open(name, 'w')
            f.write(what)
            if what and what[-1:] != '\n':
                f.write('\n')           # Ensure last line is terminated
            f.close()
        elif how == 'f':
            # Data from file
            load_suitcase_file(sfs, name, what)
        elif how == 'd':
            # Data from directory
            load_suitcase_dir(sfs, name, what, includecvs)
        else:
            error("Internal error: how = %s", `how`)
    sfs._flush()

def load_suitcase_file(sfs, name, what):
    try:
        fp = open(what)
        data = fp.read()
        fp.close()
    except IOError, msg:
        error("Can't open or read %s: %s", `what`, str(msg))
    f = sfs.open(name, 'w')
    f.write(data)
    f.close()

def load_suitcase_dir(sfs, name, what, includecvs):
    make_suitcase_dir(sfs, name)
    try:
        names = os.listdir(what)
    except IOError, msg:
        error("Can't list directory %s: %s", `what`, str(msg))
    for fname in names:
        if not includecvs and fname in ('CVS', '.cvsignore'):
            continue
        fullname = os.path.join(what, fname)
        sname = sfs.path.join(name, fname)
        if os.path.isdir(fullname):
            load_suitcase_dir(sfs, sname, fullname, includecvs)
        else:
            load_suitcase_file(sfs, sname, fullname)

def make_suitcase_dir(sfs, name):
    if not name or sfs.path.isdir(name): return
    head, tail = sfs.path.split(name)
    make_suitcase_dir(sfs, head)
    sfs.mkdir(name)


class MyRS(RS.NullRS):
    """Class supporting the interface of the ReportingStation's log.

    This interface involves providing just one method:

    append - append an entry to the log

    Other methods are provided for keeping track of ids of submitted
    KPs:

    addid - add a local id to the table of KPs to watch out for
    empty - test whether there are any ids left in the table
    saveid - save the suitcase of the KP with this id

    """
    
    stop_the_clones = 0

    def __init__(self, freezedir=None):
        RS.NullRS.__init__(self)
        self.freezedir = freezedir
        self.ids = {}
        self.suitcase = {}
        self.locations = {}

    def KPTerminated(self, id, reason, details, kprepr):
        message("KPTerminated message received for KP %s." % `id`)
        if reason:
            state = kprepr.GetState()
            try:
                trace = state.GetData("traceback")
            except KPRepr.PartNameError:
                trace = None
            self.exception(id, reason, details, trace)
            msg = "terminated by %s exception" % reason
        else:
            msg = "exited with status %s" % details
        if self.suitcase.has_key(id):
            save_suitcase(kprepr, self.suitcase[id])
        else:
            message("Ignoring suitcase of KP %s." % `id`)
        if self.freezedir:
            if not os.path.isdir(self.freezedir):
                os.mkdir(self.freezedir)
            fp = open(os.path.join(self.freezedir, id + ".mime"), "w")
            fp.write(kprepr.FreezeDry())
            fp.close()
        self.kpdone(id, msg)

    def KPMigrationReceived(self, id, kpid, kernel):
        message("KP %s migrated to %s; KP id = %s.", `id`, `kernel`, `kpid`)
        self.locations[id] = (kpid, kernel)

    def AllocateID(self, prefix):
        if self.stop_the_clones:
            raise KOS.AuthorizationError, "no cloning any more!"
        id = RS.NullRS.AllocateID(self, prefix)
        self.addid(id)
        return id

    # New methods (not in base class)...

    def saveid(self, id, spec):
        self.suitcase[id] = spec

    def addid(self, id):
        self.ids[id] = id

    def delid(self, id):
        del self.ids[id]
        if self.locations.has_key(id):
            del self.locations[id]

    def empty(self):
        return not self.ids

    def kpdone(self, id, msg):
        if self.ids.has_key(id):
            self.delid(id)
        else:
            warning("Oops!  Never heard of KP %s." % `id`)
        message("KP %s %s.", `id`, str(msg))
        if self.empty():
            message("No more KPs left to report on.  Exit.")
            ilutools.ExitMainLoop()

    def exception(self, id, reason, details, trace=None):
        message("Exception in KP %s:", `id`)
        if trace:
            sys.stderr.write("Traceback (innermost last):\n")
            sys.stderr.write(trace)
            if trace[-1] != '\n':
                sys.stderr.write('\n')
        sys.stderr.write("%s: %s\n" % (reason, details))
    
    def start_signal_handling(self):
        signal.signal(signal.SIGINT, self.kill_callback)
        self.set_watchdog()
    
    need_to_kill = 0

    def kill_callback(self, sig=None, frame=None, *extras):
        message("Kill callback invoked by signal %s...", str(sig))
        self.need_to_kill = 1
        self.stop_the_clones = 1

    def killemall(self):
        # XXX Use our SBH as a password
        password = self.IluSBH()
        sig = signal.SIGTERM
        for id, (kpid, kernel) in self.locations.items():
            message("Killing KP %s at %s, KP id %s, with signal %d",
                    `id`, `kernel`, `kpid`, sig)
            r = nstools.WorldOpen("kos/%s/receiver"%kernel)
            try:
                r.KillKP(kpid, sig, password)
            except KOS.ResourceError, msg:
                print "Can't kill", `id`, "at", `kernel`, ":", msg
                if self.ids.has_key(kpid):
                    self.delid(kpid)
            except KOS.AuthorizationError, msg:
                print "Not allowed to kill", `id`, "at", `kernel`, ":", msg
                if self.ids.has_key(kpid):
                    self.delid(kpid)
            except:
                print "Unexpected kill error:"
                traceback.print_exc()
                if self.ids.has_key(kpid):
                    self.delid(kpid)
        message("Done killing.")
        if self.empty():
            message("No more KPs left to report on.  Exit.")
            ilutools.ExitMainLoop()

    watchdog_alarm = None

    def set_watchdog(self):
        if not self.watchdog_alarm:
            self.watchdog_alarm = ilu.CreateAlarm()
        when = ilu.FineTime_Now() + ilu.FineTime(1)
        self.watchdog_alarm.set(when, self.watchdog_callback, ())

    def watchdog_callback(self):
        if verbose > 1: message("Watchdog...")
        if self.need_to_kill:
            self.need_to_kill = 0
            self.killemall()
        self.set_watchdog()


# Standard script boilerplate -- run main() if we are the main program
if __name__ == '__main__':
    main()
