#! /usr/bin/env python

# This documentation string also serves as a usage message
"""Replicated Worldroot server.

Usage: worldroot.py [-f] [-h] [-n name] [-p port] [-q] [-s] [-v]

Where:

    -f -- force overwrite of var/worldroot.sbh
    -h -- print a help message
    -n name -- this worldroot's name (default: full hostname)
    -p port -- use a different TCP port
    -q -- quiet (print less debugging output)
    -s -- run a sweep before entering server main loop
    -v -- verbose (print more debugging output)

The following files in the <KOSROOT>/var directory are used:

worldroot.sbh     -- the SBH of our root context as seen by the KOS
wr-replicator.sbh -- the SBH of our worldroot replicator object
wr-peers.tab      -- table of peer replicators
wr-currentdb*     -- database containing the current events

All these files are written; the wr-peers.tab file is also read (at
startup), and the worldroot.sbh file is read implicitly to test for
the existence of another live local worldroot server.  The database
of current events is read at startup time and updated as events come
in.  The peers table is updated each time a new peer connects.

The format of the peers table file is as follows: each line contains
two whitespace-delimited fields, the first is the name of a peer, the
second is the last known SBH of its worldroot replicator.  Blank lines
and lines whose first nonblank character is a hash sign are ignored.
Lines that do not contain exactly 2 files are flagged as errors and
ignored.

If no peers table is found, the worldroot server starts off in
stand-alone mode (although if another worldroot server connects, it
will be added to the peers table).

The worldroot.sbh file is removed on exit, to signal e.g. starting KOS
kernels that there is no running worldroot server.  The other files
are not removed; the peers table and current events database are
persistent state that are used to recover after a worldroot restart,
and the wr-replicator.sbh file is only written so that you can
manually transfer it to another worldroot replicator's peers table.

"""

# path hackery
import sys, os; sys.path.insert(0, os.path.join(sys.path[0], "../../"))
from koe.common import pathhack

import string
import getopt
import shelve, anydbm
import os
import time
import socket

from koe.interfaces.stubs import Namespace
from koe.common import NamespaceImpl
from koe.common import nstools
from koe.common import ilutools
import ilu
from koe.common import access

from koe.interfaces.stubs import WorldAPI, WorldAPI__skel

# Filenames

WR_SBH = "worldroot.sbh"
REP_SBH = "wr-replicator.sbh"
PEERS_TAB = "wr-peers.tab"
WR_CURDB = "wr-currentdb"

# Exceptions
AuthorizationError = "Insufficient permissions"
ResourceError = "Insufficient resources"


ilu.SetFDBudget(2 * ilu.GetFDBudget())


# XXX The definition order in this module is roughly bottom-up, except
# that in the WorldrootImpl class, the methods are defined roughly
# top-down.  Sorry for the inconsistency!


def varfn(fn):
    """Return a filename in the <KOSROOT>/var subdirectory."""
    root = pathhack.kos_root()
    vardir = os.path.join(root, "var")
    return os.path.join(vardir, fn)


def varopen(fn, mode='r', bufsize=-1):
    """Open a file in the <KOSROOT>/var subdirectory."""
    return open(varfn(fn), mode, bufsize)


def sbhsave(fn, sbh):
    """Save an SBH to a file in the <KOSROOT>/var subdirectory."""
    f = varopen(fn, 'w')
    f.write("%s\n" % sbh)
    f.close()


def sbhload(fn):
    """Load an SBH from a file in the <KOSROOT>/var subdirectory."""
    f = varopen(fn, 'r')
    sbh = string.strip(f.readline())
    f.close()
    return sbh


def fullhostname():
    """Return our full hostname.
    
    (This is complicated by the fact that on some systems
    socket.gethostname() returns a nickname only; in that case, we
    return the first ``alternative'' name that contains a period.)

    """
    host = socket.gethostname()
    # Make sure to find the fully-qualified domain name
    if '.' not in host:
        name, nicknames, addr = socket.gethostbyaddr(host)
        for name in [name] + nicknames:
            if '.' in name:
                host = name
                break
    return host


def shortevent(event):
    """Return a condensed string representation of an event record."""
    return "%s(%s%s)" % (event.action, `event.dbname + '/' + event.key`,
                            event.value and ", ..." or "")


def dumpevent(event, verbose=0):
    """Print the complete contents of an event record."""
    print "    originator : %s" % `event.originator`
    print "    uniqueid   : %s" % `event.uniqueid`
    print "    generation : %s" % event.generation
    print "    timestamp  : %s" % time.ctime(event.timestamp)
    print "    dbname     : %s" % `event.dbname`
    print "    action     : %s(%s%s)" % (event.action, `event.key`,
                                         event.value and ", value" or "")
    if event.value and verbose:
        print "    value.typename  : %s" % `event.value.typename`
        print "    value.reference : %s" % `event.value.reference`


def synchack(shelf):
    # Hack to add a sync method to a shelf when using pure Python 1.4.
    # This creates a circular reference, but that's OK.
    if hasattr(shelf, 'sync'):
        return
    def sync(self=shelf):
        if hasattr(self.dict, 'sync'):
            self.dict.sync()
    shelf.sync = sync


class ConnectionImpl(WorldAPI__skel.Connection):

    """Peer-to-peer connection.

    Connections are lightweight objects whose purpose is to be able to
    be able to tell which peer sent a request -- there is one
    connection per peer.

    If the peer reconnects, it will be given the same connection
    object.

    If our server is reincarnated, it will create new connections and
    the old ones will be invalid (connection objects use the default
    server).

    """

    def __init__(self, worldroot, peername):
        self.worldroot = worldroot
        self.peername = peername

    def DispatchEvent(self, event):
        self.worldroot.dispatch_event(self.peername, event)


class WorldrootImpl(WorldAPI__skel.Worldroot):

    """Worldroot replicator.

    Events are propagated to all known live peers except the
    originator and the peer from which the event was received (since
    no path information for the event is known).

    XXX This comment is out of date.

    At startup time, all known live peers are asked for a complete
    list of their events, and all such events are processed locally.

    Clearly, this does not scale very well -- when we find that we
    the number of worldroot replicators is becoming a bottleneck, we
    can improve the algorithms without significant changes to the
    basic approach.

    Also note that currently the worldroot state is *not* persistent.
    However, as long as at least one worldroot server is up and
    current, other worldroot servers will be able to recover their
    state from that server.

    """

    generation = 0
    lastsweep = 0

    def __init__(self, server, whoami, verbose=0, quiet=1):
        self.IluServer = server
        self.whoami = whoami
        self.verbose = verbose
        self.quiet = quiet
        self.peers = {}
        self.peerconnections = {}
        self.connections = {}
        try:
            self.currentevents = shelve.open(varfn(WR_CURDB))
        except anydbm.error:
            print "wr-current.db is corrupt. ignoring it."
            try: os.unlink(varfn(WR_CURDB))
            except os.error: pass
            try: os.unlink(varfn(WR_CURDB) + ".pag")
            except os.error: pass
            try: os.unlink(varfn(WR_CURDB) + ".dir")
            except os.error: pass
            self.currentevents = shelve.open(varfn(WR_CURDB))
        synchack(self.currentevents)
        self.eventsseen = {}
        self.databases = {}

    def recover(self):
        if self.verbose:
            print "recovering from locally saved state..."
        for key in self.currentevents.keys():
            event = self.currentevents[key]
            if self.verbose:
                print "recover event", shortevent(event)
                if self.verbose > 1: dumpevent(event, self.verbose > 2)
            self.dispatch_event_locally(event)
        if self.verbose:
            print "done recovering from locally saved state"

    def closedb(self):
        self.currentevents.close()

    def add_database(self, dbname, db):
        self.databases[dbname] = db

    def load_peers(self):
        try:
            fp = varopen(PEERS_TAB, 'r')
        except IOError:
            print "No peers file -- assume we're alone."
            return
        nblist = fp.readlines()
        fp.close()
        lineno = 0
        for line in nblist:
            lineno = lineno + 1
            words = string.split(line)
            if not words or words[0][0] == '#':
                continue
            if len(words) != 2:
                print "%s(%d): bad line" % (PEERS_TAB, lineno)
                print ' ', `line`
                continue
            [name, sbh] = words
            if name != self.whoami:
                self.add_peer(name, sbh)

    def save_peers(self):
        fp = varopen(PEERS_TAB, 'w')
        fp.write("%s %s\n" % (self.whoami, self.IluSBH()))
        for (peername, peer) in self.peers.items():
            fp.write("%s %s\n" % (peername, peer.IluSBH()))
        fp.close()

    def add_peer(self, peername, sbh):
        
        if self.verbose:
            print "add peer", peername

        # Insure that this peer is still a host we allow access *to*
        port = access.ilu_port_from_sbh(sbh)
        if not port:
            if not self.quiet or self.verbose:
                print "Error obtaining port for Peer: %s", peername
            return
        
        if self.verbose:
            print 'checking access permissions to: %s, port: %s ' \
                  % (peername, port),

        perm = access.LimitParser()
        if not perm.is_allowed_to('worldrootpeer', peername, port=port):
            if self.verbose:
                print 'denied.'
            elif not self.quiet:
                print "Permission denied to Peer: %s, Port: %s" \
                      % (peername, port)
            return
        if self.verbose:
            print 'allowed.'
            
        if not self.quiet:
            print "Contacting Peer: %s, Port: %s " % (peername, port)
            
        peer = ilu.ObjectOfSBH(WorldAPI.Worldroot, sbh)
        self.peers[peername] = peer
        c = self.connection_for(peername)
        try:
            pc = peer.Connect(self.whoami, self, c)
        except ilutools.IluErrorTuple, msg:
            if self.verbose:
                print "add peer", peername, "failed:", msg
        else:
            self.peerconnections[peername] = pc
            try:
                events = peer.GetEvents(self.lastsweep+1, 0)
            except ilutools.IluErrorTuple, msg:
                if self.verbose:
                    print "failed to get events from peer", peername
            else:
                if self.verbose:
                    print "recovering events from peer", peername
                for event in events:
                    if self.verbose:
                        print "dispatching recovery event", shortevent(event)
                    self.dispatch_event_locally(event)
                if self.verbose:
                    print "done recovering events from peer", peername

    def Connect(self, peername, peer, conn):
        
        # Insure we allow connections *from* peername... but only on
        # the initial contact for efficiency
        sbh = peer.IluSBH()
        port = access.ilu_port_from_sbh(sbh)
        if not port:
            raise ResourceError, "Can't get port for: %s" % peername
        if self.verbose:
            print 'checking access from: %s, port: %s ' % (peername, port),
        perm = access.LimitParser()
        if not perm.is_allowed_from('worldrootpeer', peername, port=port):
            if self.verbose:
                print "denied."
            raise AuthorizationError, \
                  "\nConnections from: %s, Port: %s not authorized." \
                  % (peername, port)

        if self.verbose:
            print "allowed."

        if not self.quiet:
            print "Connection request from: %s, Port: %s" \
                  % (peername, port) 
        self.peers[peername] = peer
        self.save_peers()
        self.peerconnections[peername] = conn
        return self.connection_for(peername)

    def GetEvents(self, lower, upper):

        if upper == 0:
            upper = self.generation
        elif upper > self.generation:
            self.generation = upper
        events = []
        for key in self.currentevents.keys():
            event = self.currentevents[key]
            if lower <= event.generation <= upper:
                events.append(event)
        return events

    def PutEvents(self, generation, events):

        if generation > self.generation:
            self.generation = generation
        for event in events:
            if self.verbose:
                print "dispatching sweep event", shortevent(event)
            self.dispatch_event_locally(event)

    def run_sweep(self, lower=0):

        # Collect events from all known replicators, and then push
        # them back to all of them.  This will take a while, but of
        # course we will be handling incoming requests during the
        # sweep.

        upper = self.generation
        if not lower:
            try:
                sweepevent = self.currentevents["./."]
            except KeyError:
                lower = 0
            else:
                lower = sweepevent.generation
        rdir = nstools.WorldOpen("replicators")
        rnames = rdir.List()
        rnames.sort()
        if self.verbose:
            print "Starting sweep, lower=%d, upper=%d" % (lower, upper)
            if self.verbose > 1:
                print "Replicator names:", rnames
        allevents = {}
        for rname in rnames:
            if self.verbose:
                print "Sweeping", rname, "..."
            r = rdir.Lookup1(rname).Open()
            try:
                events = r.GetEvents(lower, upper)
            except ilutools.IluErrorTuple, msg:
                print "While sweeping", rname, ":"
                print "ILU general error:", msg
                print "Sweep aborted."
                return
            for event in events:
                allevents[event.uniqueid] = event
        if self.verbose:
            print "Sweeping completed, now pushing results."
        events = allevents.values()
        sweepevent = self.make_sweep_event(upper)
        events.append(sweepevent)
        for event in events:
            if self.verbose:
                print "dispatching sweep event locally", shortevent(event)
            self.dispatch_event_locally(event)
        for rname in rnames:
            if self.verbose:
                print "Pushing", rname, "..."
            r = rdir.Lookup1(rname).Open()
            try:
                r.PutEvents(upper, events)
            except ilutools.IluErrorTuple, msg:
                print "While pushing", rname, ":"
                print "ILU general error:", msg
                print "Push continues."
        if self.verbose: print "Push done."

    def make_sweep_event(self, upper):
        uniqueid = self.generate_uid()
        timestamp = time.time()
        dbname = "."
        action = "Sweep"
        key = "."
        value = None
        event = WorldAPI.Event(self.whoami,
                               uniqueid, upper, timestamp,
                               dbname, action, key, value)
        return event

    def connection_for(self, peername):
        try:
            c = self.connections[peername]
        except KeyError:
            c = self.connections[peername] = ConnectionImpl(self, peername)
        return c

    def dispatch_event(self, peername, event):

        if self.verbose:
            print "dispatching event from", peername, shortevent(event)
        if self.dispatch_event_locally(event):
            self.propagate_event(event, peername)

    def dispatch_event_locally(self, event):
        uniqueid = event.uniqueid
        if self.eventsseen.has_key(uniqueid):
            # Already processed this one
            if self.verbose:
                print "discarding previously seen event", shortevent(event)
            return 0
        self.eventsseen[event.uniqueid] = event
        originator = event.originator
        generation = event.generation
        timestamp = event.timestamp
        dbname = event.dbname
        action = event.action
        key = event.key
        value = event.value
        dbkey = dbname+'/'+key
        try:
            curevent = self.currentevents[dbkey]
        except KeyError:
            pass
        else:
            if ((curevent.generation, curevent.uniqueid) >
                (event.generation, event.uniqueid)):
                # We already have an event pertaining to the same key
                # with a newer generation number -- or one with the
                # same generation and a winning uniqueid (to break a
                # tie for conflicting updates)
                if self.verbose:
                    print "discarding overruled event", shortevent(event)
                return 0
        self.currentevents[dbkey] = event
        self.currentevents.sync()
        if generation > self.generation:
            self.generation = generation
        try:
            db = self.databases[dbname]
        except KeyError:
            # We don't have this database -- still propagate the event
            if dbkey == "./." and action == "Sweep":
                # Internal event to set lastsweep
                if self.lastsweep < event.generation:
                    if self.verbose:
                        print "Bumping lastsweep from", self.lastsweep,
                        print "to", event.generation
                    self.lastsweep = event.generation
                elif self.verbose:
                    print "Ignore sweep event with generation",
                    print event.generation, "; lastsweep =", self.lastsweep
            elif self.verbose:
                print "no local action for event", shortevent(event)
        else:
            if self.verbose:
                print "update local database for event", shortevent(event)
            # Update the local database
            if action == 'Bind':
                db.local_rebind(key, value)
            elif action == 'Unbind':
                try:
                    db.local_unbind(key)
                except Namespace.BindError:
                    pass
            else:
                print "Bad event.action:", `action`
                dumpevent(event, 1)
        return 1

    def create_event(self, dbname, action, key, value):
        self.generation = self.generation + 1
        uniqueid = self.generate_uid()
        timestamp = time.time()
        event = WorldAPI.Event(self.whoami,
                               uniqueid, self.generation, timestamp,
                               dbname, action, key, value)
        self.currentevents[dbname+'/'+key] = self.eventsseen[uniqueid] = event
        self.currentevents.sync()
        self.propagate_event(event)

    def propagate_event(self, event, source=None):
        if self.verbose:
            if source:
                print "propagating event from", source, shortevent(event)
            else:
                print "propagating local event", shortevent(event)
            if self.verbose > 1: dumpevent(event, self.verbose > 2)
        for (peername, pc) in self.peerconnections.items():
            if peername not in (source, event.originator):
                if self.verbose:
                    print "propagate to", peername
                try:
                    pc.DispatchEvent(event)
                except ilutools.IluErrorTuple, msg:
                    if self.verbose:
                        print "propagate failed:", msg
        if self.verbose:
            print "done propagating", shortevent(event)
            if self.verbose > 1:
                print

    uid_counter = 0
    uid_host = None
    uid_pid = 0

    def generate_uid(self):
        self.uid_counter = self.uid_counter + 1
        if not self.uid_host:
            self.uid_host = fullhostname()
        if not self.uid_pid:
            self.uid_pid = os.getpid()
        return "%s-%d-%.3f-%d" % (self.uid_host, self.uid_pid,
                                  time.time(), self.uid_counter)

    def Shutdown(self):
        try:
            if access.ilu_allow_access('Worldrootshutdown'):
                ilutools.ExitMainLoop()
        except access.ResourceError:
            print 'Error obtaining access information!'



class ReplicatedContext(NamespaceImpl.Context):

    """Replicated context.

    This extends the Bind/Rebind/Unbind methods to invoke the
    replicator after updating the local state.  It also defines
    methods local_bind/local_rebind/local_unbind for use by the
    replicator when dispatching events received from other
    replicators.

    Note that for performance reasons, it is important that the
    replicator's DispatchEvent is declared as an ASYNCHRONOUS method
    in the ISL file.

    """

    def __init__(self, server, dbname, replicator):
        self.IluServer = server
        NamespaceImpl.Context.__init__(self)
        self.__dbname = dbname
        self.__replicator = replicator
        replicator.add_database(dbname, self)

    # Internal names for base class methods
    __bind = NamespaceImpl.Context.Bind
    __rebind = NamespaceImpl.Context.Rebind
    __unbind = NamespaceImpl.Context.Unbind

    # Non-replicating methods, called by replicator
    local_bind = __bind
    local_rebind = __rebind
    local_unbind = __unbind

    # Published methods, originating new events

    def Bind(self, key, value):
        self.__bind(key, value)
        self.__replicate('Bind', key, value)

    def Rebind(self, key, value):
        try:
            oldvalue = self.Lookup1(key)
        except Namespace.ResolutionError:
            pass
        else:
            oldr = oldvalue['reference']
            oldt = oldvalue['typename']
            newr = value['reference']
            newt = value['typename']
            if oldr == newr and oldt == newt:
                # No change -- forget it
                return
        self.__rebind(key, value)
        self.__replicate('Bind', key, value)

    def Unbind(self, key):
        self.__unbind(key)
        self.__replicate('Unbind', key, None)

    # Internal shorthand to call the replicator

    def __replicate(self, action, key, value):
        self.__replicator.create_event(self.__dbname, action, key, value)


class WorldrootContext(NamespaceImpl.Context):

    """Worldroot top level namespace."""

    def __init__(self, server, replicator, verbose=0):
        self.IluServer = server
        NamespaceImpl.Context.__init__(self)
        for name in "kos", "types", "tools", "replicators":
            ctx = ReplicatedContext(server, name, replicator)
            NamespaceImpl.Context.Bind(self, name, ctx.GetDescriptor())
        rdesc = Namespace.Descriptor("WorldAPI.Worldroot", replicator.IluSBH())
        NamespaceImpl.Context.Bind(self, "local-worldroot", rdesc)
        replicator.recover()

    def Bind(self, key, value):
        raise BindError, "worldroot name space is frozen"

    def Rebind(self, key, value):
        raise BindError, "worldroot name space is frozen"

    def Unbind(self, key):
        raise BindError, "worldroot name space is frozen"


def usage(msg=None, sts=2):
    """Write an error message and a usage message to stderr, and exit."""
    sys.stdout = sys.stderr
    if msg:
        print msg
        print
    print __doc__,
    sys.exit(sts)


def main():

    """Main program."""

    # Default options

    port = 0
    whoami = None
    verbose = 0
    quiet = 0
    force = 0
    sweep = 0

    # Command line parsing

    args = None
    try:
        optlist, args = getopt.getopt(sys.argv[1:], "fhn:p:qsv")
    except getopt.error, msg:
        usage(msg)

    if args:
        usage("No non-option arguments expected")

    for opt, val in optlist:
        if opt == "-f":
            force = 1
        if opt == "-h":
            usage(sts=0)
        if opt == "-n":
            whoami = val
        if opt == "-p":
            try:
                port = string.atoi(val)
            except ValueError, msg:
                usage(msg)
        if opt == "-q":
            verbose = 0
            quiet = 1
        if opt == "-s":
            sweep = 1
        if opt == "-v":
            verbose = verbose + 1
            quiet = 0

    # Fill in the real default options

    if not whoami:
        whoami = fullhostname()

    if not port:
        port = 7438

    # Check that we don't overwrite an existing var/worldroot.sbh file

    try:
        fp = varopen(WR_SBH)
    except IOError:
        pass
    else:
        fp.close()
        if force:
            if not quiet: print "Will overwrite existing", WR_SBH
        else:
            sys.stdout = sys.stderr
            print "Won't overwrite existing", WR_SBH, "(use -f to overwrite)"
            sys.exit(1)

    # Create the default server, using a random port etc.
    # (The first server you create becomes the default server.)

    if verbose: print "Creating default server"
    default_server = ilu.CreateServer()

    # Create a specific server, for objects whose SBH must be
    # persistent.  (See <ILUSRC>/examples/reconnect/server.py for an
    # example of how to use ilu.CreateServer with a given port.)

    if verbose: print "Creating specific server"
    try:
        specific_server = ilu.CreateServer(whoami,
                                           ["sunrpcrm", "tcp_0_%d" % port])
    except ilutools.IluErrorTuple, msg:
        sys.stdout = sys.stderr
        print "ILU general error:", msg
        print "Whoops -- worldroot port conflict!"
        sys.exit(1)

    # Create the WorldAPI replicator

    replicator = WorldrootImpl(specific_server, whoami, verbose, quiet)
    # save the wr-replicator.sbh file in a format that makes it very easy to
    # cut and past into a wr-peers.tab file
    sbhsave(REP_SBH, fullhostname() + ' ' + replicator.IluSBH())

    # Create the worldroot namespace server

    wrcontext = WorldrootContext(specific_server, replicator, verbose)

    # Save the worldroot SBH

    worldrootSBH = wrcontext.GetDescriptor()['reference']
    sbhsave(WR_SBH, worldrootSBH)

    # Connect the replicator with its peers and load the database

    replicator.load_peers()

    # Save the peers file (adding our own SBH)

    replicator.save_peers()

    # Bind the replicator in the /replicators directory
    
    ctx = nstools.WorldOpen("replicators")
    ctx.Rebind(whoami,
               Namespace.Descriptor("WorldAPI.Worldroot", replicator.IluSBH()))

    # Run a sweep, if requested

    if sweep:
        if not quiet: print "Running sweep..."
        replicator.run_sweep()

    # Crank it up

    try:
        if not quiet: print "Starting Worldroot."
        ilutools.RunInterruptableMainLoop()
        if not quiet: print "Worldroot stopped."
    finally:
        replicator.closedb()
        sbhfile = varfn(WR_SBH)
        try:
            os.unlink(sbhfile)
        except os.error, msg:
            if not quiet: print "Failed to unlink", sbhfile, ":", msg


# Run the main program

if __name__ == '__main__':
    main()
